
!--------------------------------------
!
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!		Compact Modelling Framework	
!			CMF2.0
!
! authors: 
!	2013.08.07 / Kalmykov, Vladimir, IO RAS, Russia, vvk88@mail.ru
!
! reference:
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Vychisl. Metody Programm., 2013, v. 14, iss. 3, 88-95. 
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Certificate of state registration of the computer progam number 2013619320. 
!		Date of state registration in the Registry of the computer programs October 1, 2013
!--------------------------------------


	module c_map_module
	
	use shared_module
	use shared_utils_module
	use c_module
	use c_rw_module
	use netcdf
	use c_gd_module

	implicit none

	! Main structure for dynamic-size vectors
	type dynamic_node
		integer, allocatable :: src(:)				! addresses for source array
		integer, allocatable :: dst(:)				! addresses for destination array
		real (kind = 8), allocatable :: wts(:)			! mapping weights

		integer, allocatable :: s_type_addr_in_un(:)		! table for decoding real addresses from unique array
		integer, allocatable :: s_type_for_send_un(:)		! unique addresses, wchich i should send to other workers

		integer, allocatable :: d_type_sum_owner(:)		! addresses of destination cells, who own sums
		integer, allocatable :: d_type_for_send_all(:)		! all required source cells, which used for partial sums composition
		integer, allocatable :: d_type_for_send_num_sums(:)	! number of member in each sum for decoding 'd_type_for_send_all'
		real (kind = 8), allocatable :: d_type_for_send_wts(:)	! weights, corresponding to source cells 

	end type dynamic_node

	! Vector with dynamic_node structure
	type dynamic_array
		type (dynamic_node), allocatable :: dyn_arr(:)
	end type dynamic_array

	! Main table for all combinations of src-dst components	
	type (dynamic_array), allocatable :: c_map_data(:,:)

	! Desstination array - result of mapping(analogue of stripe_2D)
	real(kind = 4), allocatable :: c_map_dst_array(:,:)

	! Send-recv vectors for runtime exchange with mapping fields
	real(kind = 4), allocatable :: c_map_send_buf(:,:)
	real(kind = 4), allocatable :: c_map_recv_buf(:,:)

	! MPI requests for MPI_INIT procedures
	integer, allocatable :: c_map_send_request(:,:,:)
	integer, allocatable :: c_map_recv_request(:,:,:)
		
	CONTAINS
	
!===============================================================================================================================================================================================================

!==========================================================================
! This subroutine build required data for 2 cases of mapping: source and
! destination decompositions. Source method is faster, when source grid has
! fever points(ATM => OCN) and destination in another case(OCN => ATM).
!==========================================================================

	subroutine c_map_ini
	
	! Work variables
	integer src_counter, dst_counter, w, counter, cplCounter 
	integer max_2D

	! Allocate mapping table for information about numbers of cells for each coupler neighbour
	allocate(c_map_data(NUM_OF_COMP, NUM_OF_COMP))
	
	! Allocate result array
	max_2D = 0
	do counter = 1, NUM_OF_COMP	
		if(max_2D <  MAXVAL(c_inf(counter) % wk(:) % stripe_count)*c_inf(counter) % il) max_2D =  MAXVAL(c_inf(counter) % wk(:) % stripe_count)*c_inf(counter) % il
	end do
	allocate (c_map_dst_array (max_2D, MAXVAL(c_map_info(:,:,1,MAP_NUM_FUNC)))); c_map_dst_array = 0
	
	!===============================================================================================
	! Make whole initialization sequential or partly parallel depends on ' run_ini_np_parallel' flag
	!===============================================================================================
	if(rank_local == 0) write(*,'(A$)') 'Coupler: file reading map_ini phase...'	
	do cplCounter = 0, run_ini_np_parallel-1
		
		if(MOD(rank_local,run_ini_np_parallel) == cplCounter) then
		!========================================================================================
		! Analize decompositions for src and dst components. Consider only turned-on components.
		! Allocate 'dynamic_array' structure for every coupler's neighbour and every src-dst pair
		!========================================================================================
		do src_counter = 1, NUM_OF_COMP
		do dst_counter = 1, NUM_OF_COMP  

			if(c_inf(src_counter) % proc_count /= 0 .AND. c_inf(dst_counter) % proc_count /= 0 .AND. src_counter /= dst_counter &
			.AND. c_map_info(src_counter, dst_counter, 1 ,MAP_PERIOD_MIN) /= 0) then

				call c_map_ini_analize_decomp(src_counter, dst_counter)
				allocate(c_map_data(src_counter,dst_counter) % dyn_arr(size_local))

				do w = 1, size_local
					allocate(c_map_data(src_counter,dst_counter) % dyn_arr(w) % src(c_map_info(src_counter, dst_counter, w, MAP_NUM_LINKS)))
					allocate(c_map_data(src_counter,dst_counter) % dyn_arr(w) % dst(c_map_info(src_counter, dst_counter, w, MAP_NUM_LINKS)))
					allocate(c_map_data(src_counter,dst_counter) % dyn_arr(w) % wts(c_map_info(src_counter, dst_counter, w, MAP_NUM_LINKS)))
					c_map_data(src_counter,dst_counter) % dyn_arr(w) % src = 0
					c_map_data(src_counter,dst_counter) % dyn_arr(w) % dst = 0
					c_map_data(src_counter,dst_counter) % dyn_arr(w) % wts = 0
					
				end do
			end if

		end do
		end do

		!====================
		! Fill mapping arrays
		!====================
		do src_counter = 1, NUM_OF_COMP
		do dst_counter = 1, NUM_OF_COMP   
			if(c_inf(src_counter) % proc_count /= 0 .AND. c_inf(dst_counter) % proc_count /= 0 .AND. src_counter /= dst_counter &
			.AND. c_map_info(src_counter, dst_counter, 1 ,MAP_PERIOD_MIN) /= 0) call c_map_ini_make_arrays(src_counter, dst_counter)
		end do
		end do
		
		end if
	
		call MPI_BARRIER(comm_local, ierr)
	end do
	if(rank_local == 0) write(*,'(A, f6.1)') "done. Time from start: ", MPI_WTIME()-wall_time_start_ini
	!========================================
	! End of sequential initialization region
	!========================================

	!=============================================================================================
	! Exchange src-arrays with all required neighbours for both interpolation schemes: src and dst
	!=============================================================================================
	if(rank_local == 0) write(*,'(A$)') 'Coupler: src-dst map_ini phase...'		
	do src_counter = 1, NUM_OF_COMP
 	do dst_counter = 1, NUM_OF_COMP

		if(c_inf(src_counter) % proc_count /= 0 .AND. c_inf(dst_counter) % proc_count /= 0 .AND. src_counter /= dst_counter &
			.AND. c_map_info(src_counter, dst_counter, 1 ,MAP_PERIOD_MIN) /= 0) then
			if(c_map_info(src_counter, dst_counter, 1, MAP_TYPE) == MAP_SRC_TYPE) call c_map_ini_src_type(src_counter, dst_counter)
			if(c_map_info(src_counter, dst_counter, 1, MAP_TYPE) == MAP_DST_TYPE) call c_map_ini_dst_type(src_counter, dst_counter)
		end if
	end do
	end do
	call MPI_BARRIER(comm_local, ierr)
	if(rank_local == 0) write(*,'(A, f6.1)') "done. Time from start: ", MPI_WTIME()-wall_time_start_ini
! 	do w = 0, size_local-1
! 		if(rank_local == w) write(*,*) 'Process', rank_local, 'finished all personal ini.'
! 		call MPI_BARRIER(comm_local, ierr)
! 	end do
! 	
! 	if(rank_local == 0) call shared_memory_usage("CPL","before alloc sendrecv map")
	
	!====================================================================
	! Allocate send-recv buffers as MAXVAL per all components and workers
	!====================================================================
	allocate(c_map_send_buf( MAX(1,MAXVAL(c_map_info(:,:,:, MAP_NUM_LINKS_FOR_SEND_UN_SRC))*MAXVAL(c_map_info(:, :, 1, MAP_NUM_FUNC)),			&
					MAXVAL(c_map_info(:,:,:, MAP_NUM_LINKS_FOR_SEND_PART_DST))*MAXVAL(c_map_info(:, :, 1, MAP_NUM_FUNC))), size_local)	)
	c_map_send_buf = 0
	allocate(c_map_recv_buf( MAX(1,MAXVAL(c_map_info(:,:,:, MAP_NUM_LINKS_UN_SRC))*MAXVAL(c_map_info(:, :, 1, MAP_NUM_FUNC)),				&
					MAXVAL(c_map_info(:,:,:, MAP_NUM_LINKS_PART_DST))*MAXVAL(c_map_info(:, :, 1, MAP_NUM_FUNC))), size_local)		)
	c_map_recv_buf = 0
	
	allocate(c_map_send_request(NUM_OF_COMP, NUM_OF_COMP, size_local))
	allocate(c_map_recv_request(NUM_OF_COMP, NUM_OF_COMP, size_local))
	c_map_send_request = MPI_REQUEST_NULL
	c_map_recv_request = MPI_REQUEST_NULL

	!============================================
	! Initiate send-recv operations for all cases
	!============================================
	do src_counter = 1, NUM_OF_COMP
 	do dst_counter = 1, NUM_OF_COMP
		if(c_inf(src_counter) % proc_count /= 0 .AND. c_inf(dst_counter) % proc_count /= 0 .AND. src_counter /= dst_counter &		
			.AND. c_map_info(src_counter, dst_counter, 1 ,MAP_PERIOD_MIN) /= 0) then
		
			do w = 1, size_local
				call c_map_ini_sendrecv_init(src_counter, dst_counter, w)
			end do
		
		end if
	end do
	end do
	
	CONTAINS
		
!===============================================================================================================================================================================================================
	
!=====================================================
! Analize numer of local entires for memory allocating
!=====================================================

	subroutine c_map_ini_analize_decomp(src_comp, dst_comp)

	! Work variables
	integer, allocatable :: src_grid_addr(:), dst_grid_addr(:)
	
	integer num_links
	integer nc_id, num_links_id, src_id, dst_id
	integer src_comp, src_first_cell, src_last_cell
	integer dst_comp, dst_first_cell, dst_last_cell
	integer w
	
	! Open file
	call nc_rc(NF90_OPEN(TRIM(c_map_build_interp_filename(src_comp, dst_comp)), NF90_NOWRITE, nc_id))

	! Read 'num_links' dimension
	call nc_rc(NF90_INQ_DIMID(nc_id, "num_links", num_links_id))	
	call nc_rc(NF90_INQUIRE_DIMENSION(nc_id, num_links_id, len=num_links))

	! Read 'addr' variables
	allocate(src_grid_addr(num_links)); src_grid_addr = 0
	allocate(dst_grid_addr(num_links)); dst_grid_addr = 0
	call nc_rc(NF90_INQ_VARID(nc_id, "src_address", src_id))
	call nc_rc(NF90_INQ_VARID(nc_id, "dst_address", dst_id))
	
!	write(*,*) 'Rank', rank_local, 'before get_var', TRIM(c_map_build_interp_filename(src_comp, dst_comp))
	call nc_rc( NF90_GET_VAR(nc_id, src_id, src_grid_addr))
	call nc_rc( NF90_GET_VAR(nc_id, dst_id, dst_grid_addr))
!	write(*,*) 'Rank', rank_local, 'after get_var', TRIM(c_map_build_interp_filename(src_comp, dst_comp))

	! Close file
	call nc_rc( NF90_CLOSE(nc_id))
	
	! For each coupler, calculate number of my points in another worker's domains
	dst_first_cell = (c_inf(dst_comp) % wk(me) % stripe_start-1)*c_inf(dst_comp) % il + 1
	dst_last_cell = dst_first_cell + c_inf(dst_comp) % il*c_inf(dst_comp) % wk(me) % stripe_count - 1

	do w = 1, run_np(COUPLER)
	
		src_first_cell = (c_inf(src_comp) % wk(w) % stripe_start-1)*c_inf(src_comp) % il + 1
		src_last_cell = src_first_cell + c_inf(src_comp) % il*c_inf(src_comp) % wk(w) % stripe_count - 1

		c_map_info(src_comp, dst_comp, w, MAP_NUM_LINKS) = COUNT(dst_grid_addr >= dst_first_cell .AND. dst_grid_addr <= dst_last_cell 	&
									      .AND. src_grid_addr >= src_first_cell .AND. src_grid_addr <= src_last_cell)
	end do

	! Free memory
	deallocate(src_grid_addr, dst_grid_addr)

	end subroutine c_map_ini_analize_decomp

!=======================================================================================================================================================================================================================================

!=====================================
! Fill all work arrays and close files
!=====================================

	subroutine c_map_ini_make_arrays(src_comp, dst_comp)

	! Work variables
	integer, allocatable :: src_grid_addr(:), dst_grid_addr(:)
	real (kind = 8), allocatable::	wts_map(:,:)
	
	integer num_links, num_wts	
	integer nc_id, num_links_id, num_wts_id, src_id, dst_id, wts_id
	integer src_comp, src_first_cell, src_last_cell
	integer dst_comp, dst_first_cell, dst_last_cell
	integer n, w
	integer :: counter_in_coupler(run_np(COUPLER))

	! Open file
	call nc_rc(NF90_OPEN(TRIM(c_map_build_interp_filename(src_comp, dst_comp)), NF90_NOWRITE, nc_id))

	! Read dimensions
	call nc_rc(NF90_INQ_DIMID(nc_id, "num_links", num_links_id))
	call nc_rc(NF90_INQ_DIMID(nc_id, "num_wgts", num_wts_id))
	call nc_rc( NF90_INQUIRE_DIMENSION(nc_id, num_links_id, len=num_links))
	call nc_rc( NF90_INQUIRE_DIMENSION(nc_id, num_wts_id, len=num_wts))
			
	! Read map data
	allocate(src_grid_addr(num_links)); src_grid_addr = 0
	allocate(dst_grid_addr(num_links)); dst_grid_addr = 0
	allocate(wts_map(num_wts, num_links)); wts_map = 0
	call nc_rc(NF90_INQ_VARID(nc_id, "src_address", src_id))
	call nc_rc(NF90_INQ_VARID(nc_id, "dst_address", dst_id))
	call nc_rc(NF90_INQ_VARID(nc_id, "remap_matrix", wts_id))
	call nc_rc( NF90_GET_VAR(nc_id, src_id, src_grid_addr))
	call nc_rc( NF90_GET_VAR(nc_id, dst_id, dst_grid_addr))
	call nc_rc( NF90_GET_VAR(nc_id, wts_id, wts_map))

	! Close files
	call nc_rc( NF90_CLOSE(nc_id))	

	! Calculate boundaries for dst and src decompositions
	dst_first_cell = (c_inf(dst_comp) % wk(me) % stripe_start-1)*c_inf(dst_comp) % il + 1
	dst_last_cell = dst_first_cell + c_inf(dst_comp) % il*c_inf(dst_comp) % wk(me) % stripe_count - 1

	! Fill arrays for every worker
	counter_in_coupler(:) = 1

	do n = 1, num_links
		if (dst_grid_addr(n) >= dst_first_cell .AND. dst_grid_addr(n) <= dst_last_cell) then

			do w = 1, run_np(COUPLER)

				src_first_cell = (c_inf(src_comp) % wk(w) % stripe_start-1)*c_inf(src_comp) % il + 1
				src_last_cell = src_first_cell + c_inf(src_comp) % il*c_inf(src_comp) % wk(w) % stripe_count - 1

				if(src_grid_addr(n) >= src_first_cell .AND. src_grid_addr(n) <= src_last_cell) then
					c_map_data(src_comp, dst_comp) % dyn_arr(w) % src(counter_in_coupler(w)) = src_grid_addr(n)
					c_map_data(src_comp, dst_comp) % dyn_arr(w) % dst(counter_in_coupler(w)) = dst_grid_addr(n)
					c_map_data(src_comp, dst_comp) % dyn_arr(w) % wts(counter_in_coupler(w)) = wts_map(1,n)
			  		counter_in_coupler(w) = counter_in_coupler(w) + 1
				end if
			end do
		end if
	end do

	! Deallocate memory
	deallocate(src_grid_addr, dst_grid_addr, wts_map)

	end subroutine c_map_ini_make_arrays

!===============================================================================================================================================================================================================

!======================================================================
! Initiate send-recv operations for all cases of components and workers
!======================================================================

	subroutine c_map_ini_sendrecv_init(src_comp, dst_comp, worker)
	
	! Input variables
	integer src_comp, dst_comp, worker
	
	! Work variables
	integer num_cells_send, num_cells_recv
	integer rank_send, rank_recv

	! Calculate mesage size
	if(c_map_info(src_comp, dst_comp, 1, MAP_TYPE) == MAP_SRC_TYPE) then
		num_cells_send = c_map_info(src_comp, dst_comp, 1, MAP_NUM_FUNC)*c_map_info(src_comp, dst_comp, worker, MAP_NUM_LINKS_FOR_SEND_UN_SRC)
		num_cells_recv = c_map_info(src_comp, dst_comp, 1, MAP_NUM_FUNC)*c_map_info(src_comp, dst_comp, worker, MAP_NUM_LINKS_UN_SRC)
	else
		num_cells_send = c_map_info(src_comp, dst_comp, 1, MAP_NUM_FUNC)*c_map_info(src_comp, dst_comp, worker, MAP_NUM_LINKS_FOR_SEND_PART_DST)
		num_cells_recv = c_map_info(src_comp, dst_comp, 1, MAP_NUM_FUNC)*c_map_info(src_comp, dst_comp, worker, MAP_NUM_LINKS_PART_DST)
	end if
	
	! Set 'MPI_PROC_NULL' ranks for non-exist interactions
	if(num_cells_send /= 0) then
		rank_send = worker-1
	else
		rank_send = MPI_PROC_NULL
	end if

	if(num_cells_recv /= 0) then
		rank_recv = worker-1
	else
		rank_recv = MPI_PROC_NULL
	end if
 
	call MPI_SEND_INIT(c_map_send_buf(1,worker),				& 
			   num_cells_send,					&
			   MPI_REAL, 						&
			   rank_send, 						&
			   MAP_NULL,						&
			   comm_local, 						&
			   c_map_send_request(src_comp, dst_comp, worker),	&
			   ierr							)
 			
	call MPI_RECV_INIT(c_map_recv_buf(1,worker),				& 
			   num_cells_recv,					&
			   MPI_REAL, 						&
			   rank_recv, 						&
			   MAP_NULL,						&
			   comm_local, 						&
			   c_map_recv_request(src_comp, dst_comp, worker),	&
			   ierr							)
	
	end subroutine c_map_ini_sendrecv_init

!===============================================================================================================================================================================================================

!==============================
! Build interpolation file name
!==============================

	character(NAME_LEN) function c_map_build_interp_filename(src_comp, dst_comp)

	integer src_comp, dst_comp
	logical file_exists
	
	c_map_build_interp_filename = 	TRIM(run_data_folder)		//	&
				TRIM(c_inf(src_comp) % name)		//	&
				'_'					//	&
				TRIM(I2S(c_inf(src_comp) % il))	//	&
				'x'					//	&
				TRIM(I2S(c_inf(src_comp) % jl))	//	&
				'_to_'					//	&
				TRIM(c_inf(dst_comp) % name)		//	&
				'_'					//	&
				TRIM(I2S(c_inf(dst_comp) % il))	//	&
				'x'					//	&
				TRIM(I2S(c_inf(dst_comp) % jl))	//	&
				".nc"
				
	! Open netCDF file
	INQUIRE(file = c_map_build_interp_filename, exist = file_exists)
	if(.NOT. file_exists) call shared_error("Cant open interpolation weights file - "//TRIM(c_map_build_interp_filename))
		
	end function c_map_build_interp_filename
	
!===============================================================================================================================================================================================================

	end subroutine c_map_ini
	
	end module c_map_module
	
	