
!--------------------------------------
!
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!		Compact Modelling Framework	
!			CMF2.0
!
! authors: 
!	2013.08.07 / Kalmykov, Vladimir, IO RAS, Russia, vvk88@mail.ru
!	2019.01.10 / Ushakov Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!
! reference:
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Vychisl. Metody Programm., 2013, v. 14, iss. 3, 88-95. 
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Certificate of state registration of the computer progam number 2013619320. 
!		Date of state registration in the Registry of the computer programs October 1, 2013
!--------------------------------------



	module c_module

	use shared_const_module
	use shared_utils_module
	use shared_module
	
	implicit none

	PRIVATE

!=======================================================================================================================================================================================
!============
! Public data
!============

	! Number of current event
	PUBLIC :: c_ev_num, c_ev_count
	
	! Coupler information tables
	PUBLIC :: c_inf
	PUBLIC :: c_events
	PUBLIC :: c_map_info

	! Public interfaces
	PUBLIC :: c_ini_register, c_ini_make_events
	
!=======================================================================================================================================================================================

	TYPE(ComponentInfo) :: c_inf(0:NUM_OF_COMP)
	
	integer, allocatable :: c_map_info(:,:,:,:)

	! Wall time variables
	real(kind=8) w_time_start_gd, w_time_start_map, w_time_map_with_gd, w_time_map
	
	! Work variables
	integer :: a, t, c, i, j, k, m, n, s, sr_counter, func
	
	! Events information
	TYPE EventType
		integer :: src, dst
		integer :: m
		integer :: act
		integer :: num_4D, num_3D, num_2D, num_2D_kind8
	END TYPE EventType
	
	type(EventType), allocatable :: c_events(:)

	integer :: c_ev_num, c_ev_count
	
	CONTAINS
	
!=======================================================================================================================================================================================
	
!==================================================================================================
! Receive main information, mapping and RW-tables from components and construct 'c_comp_info' table
!==================================================================================================

	subroutine c_ini_register
	
	! Receive arrays
	integer mapping_table(NUM_OF_COMP, NUM_OF_COMP, MAP_NUM_FIELDS)

	! Work variables
	integer st(MPI_STATUS_SIZE)
	integer w, arr_count
	
!=================================================
! Allocate and initiate coupler information tables
!=================================================
	! Basic coupler initialization
	c_inf(COUPLER) % name = 'COUPLER'
	c_inf(COUPLER) % proc_count = run_np(COUPLER)
	
	do c = 1, NUM_OF_COMP
		c_inf(c) % proc_count = run_np(c)
		
		allocate(c_inf(c) % wk(run_np(COUPLER)))
		c_inf(c) % wk(me) % slave_count = c_inf(c) % proc_count/run_np(COUPLER)
		
		do w = 1, run_np(COUPLER)
			allocate(c_inf(c) % wk(w) % sl(c_inf(c) % wk(me) % slave_count))
		end do
		
		c_inf(c) % wk(me) % slave_start = SUM(c_inf(0:c-1) % proc_count) + rank_local*c_inf(c) % wk(me) % slave_count	
	end do
	
	! Allocate
 	allocate(c_map_info(NUM_OF_COMP, NUM_OF_COMP, run_np(COUPLER), MAP_NUM_FIELDS)); c_map_info = 0	
	
!=============================================
! Initial interaction with not-null components
!=============================================
	do c = 1, NUM_OF_COMP

		if(c_inf(c) % proc_count == 0) CYCLE
		
		! Collect slave personal information
		do s = 1, c_inf(c) % wk(me) % slave_count
		
			call MPI_RECV(c_inf(c) % wk(me) % sl(s) % west, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start+s-1, c, MPI_COMM_WORLD, st, ierr)
			call MPI_RECV(c_inf(c) % wk(me) % sl(s) % east, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start+s-1, c, MPI_COMM_WORLD, st, ierr)
			call MPI_RECV(c_inf(c) % wk(me) % sl(s) % south, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start+s-1, c, MPI_COMM_WORLD, st, ierr)
			call MPI_RECV(c_inf(c) % wk(me) % sl(s) % north, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start+s-1, c, MPI_COMM_WORLD, st, ierr)
			c_inf(c) % wk(me) % sl(s) % il_belt = c_inf(c) % wk(me) % sl(s) % east - c_inf(c) % wk(me) % sl(s) % west + 1
			c_inf(c) % wk(me) % sl(s) % jl_belt = c_inf(c) % wk(me) % sl(s) % north - c_inf(c) % wk(me) % sl(s) % south + 1
		end do

		! Calculate and exhange with decomposition data between coupler workers
		c_inf(c) % wk(me) % stripe_start = c_inf(c) % wk(me) % sl(1) % south
		c_inf(c) % wk(me) % stripe_count = c_inf(c) % wk(me) % sl(c_inf(c) % wk(me) % slave_count) % north - c_inf(c) % wk(me) % stripe_start + 1
		call MPI_ALLGATHER(c_inf(c) % wk(me) % stripe_start, 1, MPI_INTEGER, c_inf(c) % wk(:) % stripe_start, 1, MPI_INTEGER, comm_local, ierr)
		call MPI_ALLGATHER(c_inf(c) % wk(me) % stripe_count, 1, MPI_INTEGER, c_inf(c) % wk(:) % stripe_count, 1, MPI_INTEGER, comm_local, ierr)
		
		! Collect common information
		call MPI_RECV(c_inf(c) % il, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		call MPI_RECV(c_inf(c) % jl, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		call MPI_RECV(c_inf(c) % kl, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		call MPI_RECV(c_inf(c) % ml, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		call MPI_RECV(c_inf(c) % inp, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		call MPI_RECV(c_inf(c) % jnp, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		call MPI_RECV(c_inf(c) % read_cp, 1, MPI_LOGICAL, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
			
		call MPI_RECV(c_inf(c) % read_fd, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		call MPI_RECV(c_inf(c) % save_dg, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		call MPI_RECV(c_inf(c) % save_it, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		call MPI_RECV(c_inf(c) % sync_it, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
			
		call MPI_RECV(c_inf(c) % name, NAME_LEN, MPI_CHARACTER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
		
		! Receive map and rw-data from components
		do a = 1, ACTION_NUM_FIELDS
		do t = 1, ARRAY_TYPE_NUM_FIELDS
			call MPI_RECV(c_inf(c) % rw(a,t) % ar_size, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)

			do n = 1,c_inf(c) % rw(a,t) % ar_size
				call MPI_RECV(c_inf(c) % rw(a,t) % ar(n) % name, NAME_LEN, MPI_CHARACTER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
				call MPI_RECV(c_inf(c) % rw(a,t) % ar(n) % lname, NAME_LEN, MPI_CHARACTER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
				call MPI_RECV(c_inf(c) % rw(a,t) % ar(n) % units, NAME_LEN, MPI_CHARACTER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)
				call MPI_RECV(c_inf(c) % rw(a,t) % ar(n) % file, NAME_LEN, MPI_CHARACTER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)

				call MPI_RECV(c_inf(c) % rw(a,t) % ar(n) % period, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)				
			end do
		end do
		end do
		
		call MPI_RECV(mapping_table(:,:,MAP_NUM_FUNC), NUM_OF_COMP**2,MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)	
		call MPI_RECV(mapping_table(:,:,MAP_TYPE), NUM_OF_COMP**2,MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)	
		call MPI_RECV(mapping_table(:,:,MAP_PERIOD_MIN), NUM_OF_COMP**2,MPI_INTEGER, c_inf(c) % wk(me) % slave_start, c, MPI_COMM_WORLD, st, ierr)	
				
		! Check errors in mapping synchronization and set coupler mapping table
		do n = 1, NUM_OF_COMP
			if(run_np(n) == 0 .OR. c == n) CYCLE
			
			if(c < n) then
				c_map_info(c, n, 1, MAP_NUM_FUNC) = mapping_table(c, n, MAP_NUM_FUNC)
				c_map_info(n, c, 1, MAP_NUM_FUNC) = mapping_table(n, c, MAP_NUM_FUNC)
				c_map_info(c, n, 1, MAP_PERIOD_MIN) = mapping_table(c, n, MAP_PERIOD_MIN)
				c_map_info(n, c, 1, MAP_PERIOD_MIN) = mapping_table(n, c, MAP_PERIOD_MIN)
			else 
				if(c_map_info(n, c, 1, MAP_NUM_FUNC) /= mapping_table(n, c, MAP_NUM_FUNC) .OR. &
				c_map_info(c, n, 1, MAP_NUM_FUNC) /= mapping_table(c, n, MAP_NUM_FUNC))		&
				call shared_error("Number of mapping functions between "//TRIM(c_inf(c) % name)//" and "//TRIM(c_inf(n) % name)//" are not synchronized: "//i2s(c_map_info(n, c, 1, MAP_NUM_FUNC))//i2s(mapping_table(n, c, MAP_NUM_FUNC))//i2s(c_map_info(c, n, 1, MAP_NUM_FUNC))//i2s(mapping_table(c, n, MAP_NUM_FUNC)))
				if(c_map_info(n, c, 1, MAP_PERIOD_MIN) /= mapping_table(n, c, MAP_PERIOD_MIN) .OR. &
				c_map_info(c, n, 1, MAP_PERIOD_MIN) /= mapping_table(c, n, MAP_PERIOD_MIN))		&
				call shared_error("Mapping periods between "//TRIM(c_inf(c) % name)//" and "//TRIM(c_inf(n) % name)//" are not synchronized.")
			end if
			
			c_map_info(c, n, 1, MAP_TYPE) = mapping_table(c, n, MAP_TYPE)
		end do
		
		! Send some data to components
		call c_rw_get_run_info(c)	
		do s = 1, c_inf(c) % wk(me) % slave_count
			call MPI_SEND(time_start_min, 1, MPI_INTEGER, c_inf(c) % wk(me) % slave_start + s - 1, c, MPI_COMM_WORLD, ierr)
		end do
		
	end do
	
	
	end subroutine c_ini_register

	
!=======================================================================================================================================================================================
	
!====================================================
! Make table of events for coupler system. Every step
! of coupler is controlled by this event queue.
!
! Possible events: 
!
! - read control point
! - read real initial conditions
! - read file data
! - write control point
! - write diagnostics
! - mapping from one component to another
!====================================================

	subroutine c_ini_make_events

	! Work variables
	integer ev_ind, ev_counter, ev_ini_counter, mm
	type(EventType), allocatable :: ev_arr(:), ev_arr_stencil(:), ev_tmp(:)
	
!=======================================
! Allocate main table and work variables
!=======================================
	allocate (ev_arr_stencil (NUM_OF_COMP*(NUM_OF_COMP-1) +  BUFFER_LEN*NUM_OF_COMP))
	allocate (ev_arr (NUM_OF_COMP*(NUM_OF_COMP-1) +  BUFFER_LEN*NUM_OF_COMP))		! Explicit assume of BUFFER_LEN possible event types
	allocate (c_events(BUFFER_LEN**2)) ! Main events table
	
!========================================
! Create stencil with all types of events 
!========================================
	ev_ini_counter = 0

	! Initialize read file data events
	do c = 1, NUM_OF_COMP
		if(c_inf(c) % proc_count /= 0) then

			call c_ini_add_event(ev_arr_stencil, ev_ini_counter,											&
				ACTION_READ_FD, COUPLER, c, c_inf(c) % read_fd,						&
				c_inf(c) % rw(ACTION_READ_FD,ARRAY_TYPE_4D) % ar_size, c_inf(c) % rw(ACTION_READ_FD,ARRAY_TYPE_3D) % ar_size, 	&
				c_inf(c) % rw(ACTION_READ_FD,ARRAY_TYPE_2D) % ar_size, c_inf(c) % rw(ACTION_READ_FD,ARRAY_TYPE_2D_KIND8) % ar_size	)
		end if
	end do	 
	
	! Initialize mapping events
	do c = 1, NUM_OF_COMP
	do n = 1, NUM_OF_COMP
		if(c_inf(c) % proc_count /= 0 .AND. c_inf(n) % proc_count /= 0 .AND. n /= c .AND.			&
		   c_map_info(c, n, 1, MAP_PERIOD_MIN) /= 0 .AND. c_map_info(c, n, 1, MAP_NUM_FUNC) /= 0) then

			call c_ini_add_event(ev_arr_stencil, ev_ini_counter,											&
				ACTION_MAPPING, c, n, c_map_info(c, n, 1, MAP_PERIOD_MIN),						&
				0, 0, 																&
				c_map_info(c, n, 1, MAP_NUM_FUNC), 0										)
		end if
	end do
	end do
       
	! Initialize save control points events
	do c = 1, NUM_OF_COMP
		if(c_inf(c) % proc_count /= 0 .AND. time_save_cp /= 0) then

			call c_ini_add_event(ev_arr_stencil, ev_ini_counter,											&
				ACTION_SAVE_CP, c, COUPLER, time_save_cp,										&
				c_inf(c) % rw(ACTION_SAVE_CP,ARRAY_TYPE_4D) % ar_size, c_inf(c) % rw(ACTION_SAVE_CP,ARRAY_TYPE_3D) % ar_size, 	&
				c_inf(c) % rw(ACTION_SAVE_CP,ARRAY_TYPE_2D) % ar_size, c_inf(c) % rw(ACTION_SAVE_CP,ARRAY_TYPE_2D_KIND8) % ar_size	)

		end if
	end do

	! Initialize save diagnostics events	
	do c = 1, NUM_OF_COMP
		if(c_inf(c) % proc_count /= 0) then

			call c_ini_add_event(ev_arr_stencil, ev_ini_counter,											&
				ACTION_SAVE_DG, c, COUPLER, c_inf(c) % save_dg,						&
				0, c_inf(c) % rw(ACTION_SAVE_DG,ARRAY_TYPE_3D) % ar_size, c_inf(c) % rw(ACTION_SAVE_DG,ARRAY_TYPE_2D) % ar_size, 0	)

		end if
	end do

	! Initialize save integral events	
	do c = 1, NUM_OF_COMP
		if(c_inf(c) % proc_count /= 0 .AND. c_inf(c) % save_it /= 0) then

			call c_ini_add_event(ev_arr_stencil, ev_ini_counter,											&
				ACTION_SAVE_IT, c, COUPLER, c_inf(c) % sync_it*c_inf(c) % save_it,						&
				0, 0, 0, 0						)

		end if
	end do
	
	
	
!==================================
! Make event table from our stencil
!==================================
	ev_arr = ev_arr_stencil
	
	! Define first events depends of 'time_start_min'
	do i = 1, ev_ini_counter
		!This is the original code of V. Kalmykov
			!do while(ev_arr(i) % m <= time_start_min)
			!	ev_arr(i) % m = ev_arr(i) % m + ev_arr_stencil(i) % m 
			!end do
		
		!This is the #modification of K. Ushakov
			!Set the first raising moment for each event: it must be in our experiment time bounds,
			!and be equal to a periodic or to a specified aperiodic moment
			do while(ev_arr(i) % m <= time_start_min 		.OR. &
				(mod(ev_arr(i) % m, ev_arr_stencil(i) % m).ne.0 .AND. .not. is_aperiodic_moment(ev_arr(i) % m, ev_arr(i) % act) )	)
				ev_arr(i) % m = ev_arr(i) % m + 1 
			end do
		
	
	
	end do
	
!===========================================
! Firstly, null-step actions (if necessary):
! - read control point
! - read real initial conditions
! - read file data
! - initial mapping send
!===========================================
	ev_counter = 0

	! Read real initial conditions
 	do c = 1, NUM_OF_COMP
 	      if(c_inf(c) % proc_count /= 0) then

			call c_ini_add_event(c_events, ev_counter,												&
				ACTION_READ_IC, COUPLER, c, time_start_min,							&
				c_inf(c) % rw(ACTION_READ_IC,ARRAY_TYPE_4D) % ar_size, c_inf(c) % rw(ACTION_READ_IC,ARRAY_TYPE_3D) % ar_size, 	&
				c_inf(c) % rw(ACTION_READ_IC,ARRAY_TYPE_2D) % ar_size, c_inf(c) % rw(ACTION_READ_IC,ARRAY_TYPE_2D_KIND8) % ar_size	)
 	      end if
 	end do

	! Read control point
 	do c = 1, NUM_OF_COMP
	      if(c_inf(c) % proc_count /= 0 .AND. c_inf(c) % read_cp) then

			call c_ini_add_event(c_events, ev_counter,												&
				ACTION_READ_CP, COUPLER, c, time_start_min,							&
				c_inf(c) % rw(ACTION_READ_CP,ARRAY_TYPE_4D) % ar_size, c_inf(c) % rw(ACTION_READ_CP,ARRAY_TYPE_3D) % ar_size, 	&
				c_inf(c) % rw(ACTION_READ_CP,ARRAY_TYPE_2D) % ar_size, c_inf(c) % rw(ACTION_READ_CP,ARRAY_TYPE_2D_KIND8) % ar_size	)
 	      end if
 	end do

	! Initial file data read
	do m = 1, ev_ini_counter
 		if(ev_arr_stencil(m) % act == ACTION_READ_FD) then
 			ev_counter = ev_counter + 1
 			c_events(ev_counter) = ev_arr_stencil(m)
 			c_events(ev_counter) % m = time_start_min
 		end if
 	end do
 	
	! Initial mapping send
	do m = 1, ev_ini_counter
		if(ev_arr_stencil(m) % act == ACTION_MAPPING) then
			ev_counter = ev_counter + 1
			c_events(ev_counter) = ev_arr_stencil(m)
			c_events(ev_counter) % m = time_start_min
		end if
	end do
	
!==================
! Main cycle events
!==================
	do while(MINVAL(ev_arr(1:ev_ini_counter) % m) <= time_start_min + 24*60*time_days_in_run)

		ev_counter = ev_counter + 1

 		! Dinamically increase event table if necessary
 		if(ev_counter == SIZE(c_events)) then
 			allocate(ev_tmp(ev_counter))
			ev_tmp = c_events
			deallocate(c_events)
			allocate(c_events(ev_counter+BUFFER_LEN*10))
			c_events(1:SIZE(ev_tmp)) = ev_tmp
			deallocate(ev_tmp)
		end if

		ev_ind = 1
		do while(ev_arr(ev_ind) % m /= MINVAL(ev_arr(1:ev_ini_counter) % m))
			ev_ind = ev_ind + 1
		end do

		c_events(ev_counter) = ev_arr(ev_ind)
		
		!This is the original code of V. Kalmykov
			!ev_arr(ev_ind) % m = ev_arr(ev_ind) % m + ev_arr_stencil(ev_ind) % m
		
		!This is the modification of K. Ushakov
			!Scanning the interval from ev_arr(ev_ind)%m till the next periodic moment.
			!Step of scanning is 1 hour
			!Are there any aperiodic moments to raise the same event?
			mm=ev_arr(ev_ind) % m + 1
			do while (MOD(mm,ev_arr_stencil(ev_ind) % m)/=0 .AND. .NOT. is_aperiodic_moment(mm, ev_arr(ev_ind) % act))
				mm=mm+1
			end do
			
			ev_arr(ev_ind) % m = mm
	end do

	! Total number of events
	c_ev_count = ev_counter

	deallocate(ev_arr, ev_arr_stencil)
		
	CONTAINS
	
!===============================================================================================================================================================================================================

!==========================
! Add event to events array
!==========================

	subroutine c_ini_add_event(ev,pos,typ,src,dst,minute,d4_4,d3_4,d2_4,d2_8)

	type(EventType) :: ev(:)
	integer pos,typ,src,dst,minute,d4_4,d3_4,d2_4,d2_8

	if(d4_4 == 0 .AND. d3_4 == 0 .AND. d2_4 == 0 .AND. d2_8 == 0 .AND. typ /= ACTION_SAVE_IT) return
	pos = pos + 1
	
	ev(pos) % act = typ
 	ev(pos) % src = src
 	ev(pos) % dst = dst
 	ev(pos) % m = minute
	ev(pos) % num_4D = d4_4
	ev(pos) % num_3D = d3_4
	ev(pos) % num_2D = d2_4
	ev(pos) % num_2D_kind8 = d2_8

	end subroutine c_ini_add_event
	
!===============================================================================================================================================================================================================

	end subroutine c_ini_make_events
	
	end module c_module

	