
!--------------------------------------
!
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!		Compact Modelling Framework	
!			CMF2.0
!
! authors: 
!	2013.08.07 / Kalmykov, Vladimir, IO RAS, Russia, vvk88@mail.ru
!
! reference:
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Vychisl. Metody Programm., 2013, v. 14, iss. 3, 88-95. 
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Certificate of state registration of the computer progam number 2013619320. 
!		Date of state registration in the Registry of the computer programs October 1, 2013
!--------------------------------------


	module comp_module

 	use shared_const_module
 	use shared_module
 	use shared_utils_module

	implicit none

!=============================================================================
! Open interface of 'comp' module. Contains of common procedures and variables 
! (wchich user could use) and absract interfaces (which user should define)
!=============================================================================

	TYPE, ABSTRACT, PUBLIC :: comp
		
	CONTAINS
		
		! Common comp procedures
		PROCEDURE :: reg_model => comp_reg_model
		PROCEDURE :: reg_array => comp_reg_array
		PROCEDURE :: send_data => comp_send_data
		PROCEDURE :: send_intg => comp_send_intg
		PROCEDURE :: recv_data => comp_recv_data	
		PROCEDURE :: reg_sendrecv => comp_reg_sendrecv
		PROCEDURE :: reduce_itg => comp_reduce_itg
		
		! Abstract interfaces for components
		PROCEDURE(I_ini_params), DEFERRED :: ini_params
		PROCEDURE(I_ini_reg_comp), DEFERRED :: ini_reg_comp
		PROCEDURE(I_ini_reg_data), DEFERRED :: ini_reg_data
		PROCEDURE(I_ini_allocate), DEFERRED :: ini_allocate
		PROCEDURE(I_ini_main), DEFERRED :: ini_main
		PROCEDURE(I_main_step), DEFERRED :: main_step
	
	END TYPE comp
		
	TYPE :: ModelInfo
		character(LEN = NAME_LEN) :: name = ''
		character(LEN = NAME_LEN) :: names(0:NUM_OF_COMP) = 'none'
		integer :: proc_count
		integer :: me
		integer :: master_rank, major_slave_rank
		integer :: i_size, j_size, k_size, m_size
		integer :: i_loc_size, j_loc_size
		integer :: inp, jnp
		character(2) :: decomp_type
		integer :: west, east, south, north
	
		type(VectorType) :: inf(0:NUM_OF_COMP, 0:NUM_OF_COMP, ACTION_NUM_FIELDS + 1, ARRAY_TYPE_NUM_FIELDS)
	END TYPE ModelInfo
	
	type(ModelInfo) :: cmp
	
	class(comp), pointer :: comp_p	

	integer :: comp_westrank, comp_eastrank, comp_southrank, comp_northrank
	integer :: comp_serank, comp_swrank, comp_nerank, comp_nwrank
	logical comp_read_cp
	logical comp_i_cycle, comp_bsc_grid
	integer comp_halo_stencil, comp_halo_wide, comp_halo_max_width
	
!================================================================================================	

!======================
! Send-recv information
!======================

	! Send and receive data buffers
	real(kind=4), allocatable :: comp_s_pack_4D(:), comp_r_pack_4D(:)
	real(kind=4), allocatable :: comp_s_pack_3D(:), comp_r_pack_3D(:)
	real(kind=4), allocatable :: comp_s_pack_2D(:), comp_r_pack_2D(:)
	real(kind=8), allocatable :: comp_s_pack_2D_kind8(:), comp_r_pack_2D_kind8(:)
	
	! MPI-requests
	integer, allocatable :: comp_request_sendrecv(:,:,:,:)
	integer comp_cur_request_send(ARRAY_TYPE_NUM_FIELDS)

	! Work variables
	integer, private :: i,j,k,m,c,a,t,d,n,s,f
	
	! Integral accumulation buffer
	real(kind=8), allocatable :: comp_intgrl_values(:,:)
	character(NAME_LEN) :: comp_ingtrl_names(BUFFER_LEN) = ''
	real(kind=8) :: comp_intgrl_area
	integer :: comp_intgrl_num = 0
	integer :: comp_ingtrl_c = 1
	
	ABSTRACT INTERFACE
		subroutine I_ini_params(this)
			import comp
			class(comp) :: this
		end subroutine I_ini_params
		
		subroutine I_ini_reg_comp(this)
			import comp
			class(comp) :: this
		end subroutine	I_ini_reg_comp	

		subroutine I_ini_reg_data(this)
			import comp
			class(comp) :: this
		end subroutine	I_ini_reg_data
		
		subroutine I_ini_allocate(this)
			import comp
			class(comp) :: this
		end subroutine I_ini_allocate
		
		subroutine I_ini_main(this)
			import comp
			class(comp) :: this
		end subroutine	I_ini_main	
		
		subroutine I_main_step(this)
			import comp
			class(comp) :: this
		end subroutine	I_main_step	
		
	END INTERFACE
	
	
	CONTAINS

!===============================================================================================================================================================	
!==============================================
! Registrate randogm comp in coupled system
!==============================================

	subroutine comp_reg_model(this,						&
				reg_np,							&
				reg_i_size, reg_j_size, reg_k_size, reg_m_size, 	&
				reg_decomp_type_2D,					&
				reg_i_cycle,						&
				reg_bsc_grid,						&
				reg_time_steps_day,					&
				reg_time_save_it,					&
				reg_read_cp,						&
				reg_halo_stencil,					&
				reg_halo_wide						)

	! Parameters of register function
	class (comp) :: this	
	integer reg_np
	integer reg_i_size, reg_j_size, reg_k_size, reg_m_size
	logical reg_read_cp
	integer reg_time_steps_day
	integer reg_time_read_fd, reg_time_save_dg, reg_time_save_it
	logical reg_decomp_type_2D, reg_i_cycle, reg_bsc_grid 
	integer reg_halo_stencil, reg_halo_wide
	
	! Sending arrays and work variables
	integer size_rw_mes
	logical all_restrictions
	logical file_exists

!=================================
! Define main component parameters
!=================================
	cmp % me = proc_where()
	cmp % name = cmp % names(cmp % me)
	cmp % proc_count = reg_np
	cmp % i_size = reg_i_size
	cmp % j_size = reg_j_size
	cmp % k_size = reg_k_size
	cmp % m_size = reg_m_size
	time_steps_day = reg_time_steps_day
	time_save_it = reg_time_save_it
	comp_read_cp = reg_read_cp
	comp_i_cycle = reg_i_cycle
	comp_bsc_grid = reg_bsc_grid
	comp_halo_stencil = reg_halo_stencil
	comp_halo_wide = reg_halo_wide
	comp_halo_max_width = MAX(comp_halo_stencil, comp_halo_wide)
	if(reg_decomp_type_2D) then 
		cmp % decomp_type = "2D"
	else
		cmp % decomp_type = "1D"
	end if
	
	! Call decomposition subroutine
	call comp_decompose
	
	! Calculate other component parameters
	cmp % master_rank = rank_local/(cmp % proc_count/run_np(COUPLER))
	cmp % major_slave_rank = cmp % master_rank*(cmp % proc_count/run_np(COUPLER))
	
	time_step_min = 24*60/REAL(time_steps_day)
	if(24*60/time_step_min /= ANINT(24*60/time_step_min)) call shared_error(TRIM(cmp % name)//" - <time_step_min> cant be converted to integer number of steps in day.")
	time_run_steps = time_steps_day*time_days_in_run

	end subroutine comp_reg_model
	
!===============================================================================================================================================================
!======================================================================
! If (reg_decomp_type_2D == .FALSE.) then make 1D latitude decomposition
! or decompose domain into rectangles(squares ideally).
! Do not permit longitude decomposition(except run_np(COUPLER) = 1).
! Check for dividing of cmp % jnp on coupler_np.
! Also, cmp % inp should be even(except 1) for exchange procedures.                        
!======================================================================

	subroutine comp_decompose
	
	integer :: j_rad
	
	! Define cmp % jnp corresponding to decomposition style
	if(cmp % decomp_type == "2D") then
		cmp % jnp = MAX(1, NINT(SQRT( REAL(cmp % j_size)*REAL(cmp % proc_count)/REAL(cmp % i_size) )))
		if((cmp % name == 'ICE' .or. cmp % name == 'ICE_CICE') .and. cmp % proc_count .gt.2 ) cmp % jnp = 2
	else
		cmp % jnp = cmp % proc_count
	end if

	! Check cmp % jnp for restrictions: check all jnp from optimal in both directions
	j_rad = 1
	do while(.TRUE.)
		if(MOD(cmp % proc_count, cmp % jnp) == 0) then
			cmp % inp = cmp % proc_count/cmp % jnp
			if(MOD(cmp % jnp, run_np(COUPLER)) == 0 .AND. (MOD(cmp % inp,2) == 0 .OR. cmp % inp == 1) .AND. MOD(cmp % i_size, cmp % inp) == 0) then

				cmp % i_loc_size = cmp % i_size/cmp % inp			
				cmp % west = MOD(rank_local, cmp % inp)*cmp % i_loc_size + 1
				cmp % east = cmp % west + cmp % i_loc_size - 1
			
				if(MOD(cmp % j_size,cmp % jnp) == 0) then
					! If devided ok				
					cmp % j_loc_size = cmp % j_size/cmp % jnp
					cmp % south = (rank_local/cmp % inp)*cmp % j_loc_size + 1
					cmp % north = cmp % south + cmp % j_loc_size - 1
					EXIT
				else
					! Different size for last processor row				
					cmp % j_loc_size = cmp % j_size/(cmp % jnp-1)
					cmp % south = (rank_local/cmp % inp)*cmp % j_loc_size + 1
					cmp % north = cmp % south + cmp % j_loc_size - 1
					
					if(rank_local >= size_local - cmp % inp) then
						cmp % j_loc_size = MOD(cmp % j_size, cmp % jnp-1)
						cmp % north = cmp % j_size
					end if
					if(MOD(cmp % j_size, cmp % jnp-1) <= cmp % j_size/(cmp % jnp-1) .AND. MOD(cmp % j_size, cmp % jnp-1) > 1) EXIT
				end if

			end if
		end if
		
		! Search in increasing radius
		cmp % jnp = cmp % jnp + j_rad
		if(j_rad > 0) then 
			j_rad = -j_rad
		else
			j_rad = -j_rad + 1
		end if
		
		if(cmp % jnp < 1 .OR. cmp % jnp > cmp % proc_count) call shared_error(TRIM(cmp % name)//' cant devide domain between processors and coupler. Conditions: '//		&
								       'MOD(cmp % proc_count, cmp % jnp) == 0 .AND. MOD(cmp % jnp, run_np(COUPLER)) == 0 .AND. '//			&
								       '(MOD(cmp % inp,2) == 0 .OR. cmp % inp == 1) .AND. MOD(cmp % i_size, cmp % inp) == 0 .AND.'//			&
								       'MOD(cmp % j_size, cmp % jnp-1) <= cmp % j_size/(cmp % jnp-1) .AND. MOD(cmp % j_size, cmp % jnp-1) > 1'		)
	end do	
	
	
	end subroutine comp_decompose
		
!===============================================================================================================================================================	
	
	subroutine comp_reg_array(this, send_to, recv_from, action, 				&
							DIM4_arr, DIM3_arr, DIM2_arr, 						&
							DIM4_kind8_arr, DIM3_kind8_arr, DIM2_kind8_arr,	&
							name, long_name, units, file, period, map_type)
	
	! Input variables
	class (comp) :: this
	integer, optional :: action, send_to, recv_from
	integer, optional :: period, map_type
	character(*), optional :: name, long_name, units, file
	real (kind=4), dimension(:,:,:,:), target, optional :: DIM4_arr
	real (kind=4), dimension(:,:,:), target, optional :: DIM3_arr
	real (kind=4), dimension(:,:), optional, target :: DIM2_arr
	real (kind=8), dimension(:,:,:,:), optional, target :: DIM4_kind8_arr
	real (kind=8), dimension(:,:,:), optional, target :: DIM3_kind8_arr
	real (kind=8), dimension(:,:), optional, target :: DIM2_kind8_arr
	
	integer :: cur, s, d, a
	type(ArrInfo) :: curArr
	
	!======================
	! Check argument errors
	!======================
	if(.NOT. PRESENT(action)) call shared_error(TRIM(cmp % name)//" reg_array() subroutine - <action> argument missing.")	
	if(.NOT.PRESENT(send_to) .AND. .NOT.PRESENT(recv_from)) call shared_error(TRIM(cmp % name)//" <reg_data> - no <send_to> or <recv_from> argument.")

	! Get short names
	if(PRESENT(send_to)) then
		s = cmp % me
		d = send_to
	else
		s = recv_from
		d = cmp % me
	end if
	a = action

	if(PRESENT(send_to)) then
		if(action == ACTION_READ_CP .OR. a == ACTION_READ_FD .OR. a == ACTION_READ_IC) 					&
			call shared_error(TRIM(cmp % name)//" reg_array() subroutine - some arguments are not correspond to <send_to> parameter." )
		if(a /= ACTION_MAPPING .AND. send_to /= COUPLER) 										&
			call shared_error(TRIM(cmp % name)//" reg_array() subroutine - <action> argument not correspond to <send_to> argument." )	
	end if
		
	if(PRESENT(recv_from)) then
		if(a == ACTION_SAVE_CP .OR. a == ACTION_SAVE_DG)									&
			call shared_error(TRIM(cmp % name)//"reg_array() subroutine - some arguments are not correspond to <recv_from> parameter.") 
		if(a /= ACTION_MAPPING .AND. recv_from /= COUPLER) 										&
			call shared_error(TRIM(cmp % name)//" reg_array() subroutine - <action> argument not correspond to <recv_from> argument." )	
	end if
	
	if(.NOT. PRESENT(name)) call shared_error(TRIM(cmp % name)//" reg_array() subroutine - <name> argument missing.")
	
	if(.NOT. PRESENT(period) .AND. (a == ACTION_SAVE_DG .OR. a == ACTION_MAPPING .OR. a == ACTION_READ_FD)) 		&
		call shared_error(TRIM(cmp % name)//" reg_array() subroutine - <period> argument missing.")
	
	if(PRESENT(period)) then
		if(period <= 0) call shared_error(TRIM(cmp % name)//" reg_array() subroutine - <period> should be > 0.")
	end if

	if(a == ACTION_READ_FD .AND. .NOT. PRESENT(file)) call shared_error(TRIM(cmp % name)//" reg_array() subroutine - <file> argument missing for ACTION_READ_FD.")

	! Check inappropriate action-array type pair
	if((PRESENT(DIM3_arr) .OR. PRESENT(DIM3_kind8_arr))														&
		.AND. (a == ACTION_READ_FD .OR. a == ACTION_MAPPING))										&
			call shared_error(TRIM(cmp % name)//" reg_array() subroutine - 3D arrays are not permitted for action "//TRIM(event_type_name(a))//".")

	if((PRESENT(DIM4_arr) .OR. PRESENT(DIM4_kind8_arr)) 	&
		.AND. (a == ACTION_READ_FD .OR. a == ACTION_MAPPING .OR. a == ACTION_SAVE_DG))										&
			call shared_error(TRIM(cmp % name)//" reg_array() subroutine - 4D arrays are not permitted for action "//TRIM(event_type_name(a))//".")			

	do i = 1, ARRAY_TYPE_NUM_FIELDS
		if(ANY(cmp % inf(s,d,a,i) % ar(:) % name == name)) call shared_error(TRIM(cmp % name)//" reg_array() subroutine - two arrays have equal name <"//TRIM(name)//">.")
	end do
	
	!=======================
	! Construct array object
	!=======================
	if(PRESENT(name)) curArr % name = name
	if(PRESENT(long_name)) curArr % lname = long_name
	if(PRESENT(units)) curArr % units = units
	if(PRESENT(file)) curArr % file = file	

	
	if(PRESENT(DIM2_arr)) then
		t = ARRAY_TYPE_2D	
		curArr % p_2D => DIM2_arr
	end if
	
	if(PRESENT(DIM2_kind8_arr)) then
		t = ARRAY_TYPE_2D_KIND8
		curArr % p_2D_kind8 => DIM2_kind8_arr
		! Forbid REAL(8) save for diagnostics
		if (a == ACTION_SAVE_DG) t = ARRAY_TYPE_2D 
	end if
	
	if(PRESENT(DIM3_arr) .OR. PRESENT(DIM3_kind8_arr)) then
		t = ARRAY_TYPE_3D
		if (PRESENT(DIM3_arr)) then
			curArr % p_3D => DIM3_arr
		else
			curArr % p_3D_kind8 => DIM3_kind8_arr
		end if
	end if
	
	if(PRESENT(DIM4_arr) .OR. PRESENT(DIM4_kind8_arr)) then
		t = ARRAY_TYPE_4D
		if (PRESENT(DIM4_arr)) then
			curArr % p_4D => DIM4_arr
		else
			curArr % p_4D_kind8 => DIM4_kind8_arr
		end if
	end if

	if(PRESENT(period) .OR. a == ACTION_SAVE_CP) then
		if(a == ACTION_SAVE_CP) then
			curArr % period = time_save_cp 
		else
			curArr % period = period
		end if
	end if
	
	! Object is ready - add it to structure
	cmp % inf(s,d,a,t) % ar_size = cmp % inf(s,d,a,t) % ar_size + 1
	n = cmp % inf(s,d,a,t) % ar_size
	cmp % inf(s,d,a,t) % ar(n) = curArr
			
	!===================================
	! Check errors for all saved objects
	!===================================	
	if((a == ACTION_SAVE_DG .OR. a == ACTION_MAPPING) .AND. 											&
		MAXVAL(cmp % inf(s,d,a,t) % ar(:) % period) /= MINVAL(cmp % inf(s,d,a,t) % ar(:) % period, cmp % inf(s,d,a,t) % ar(:) % period > 0))	&
			call shared_error(TRIM(cmp % name)//" reg_array() subroutine - periods should be the same for all DG or MAP functions." )
			
	if(cmp % inf(s,d,a,t) % ar(n) % period/time_step_min /= ANINT(cmp % inf(s,d,a,t) % ar(n) % period/time_step_min)) 		&
		call shared_error(TRIM(cmp % name)//" - period for action "//TRIM(event_type_name(a))//" cant be achieved with current <time_step_min>.")
	
	if(a == ACTION_MAPPING) then
		if(PRESENT(send_to)) then
			if(.NOT. PRESENT(map_type)) call shared_error(TRIM(cmp % name)//" reg_array() subroutine - set mapping parameters: <map_type>.")
			if(map_type /= MAP_SRC_TYPE .AND. map_type /= MAP_DST_TYPE) call shared_error(TRIM(cmp % name)//" reg_array() subroutine - unknown mapping type.")
			cmp % inf(s,d,a,:) % ar(n) % map_type = map_type
		end if
	end if
		
	end subroutine comp_reg_array

!===============================================================================================================================================================	

	subroutine comp_reg_sendrecv(this)

	class (comp) :: this	
	integer status_ini_recv(MPI_STATUS_SIZE)

	!==========================================
	! Check for errors and define equal periods
	!==========================================

	! Control point
	time_l_save_cp = INT(time_save_cp/time_step_min)
	
	! Diagnostics
	time_save_dg = MAXVAL(cmp % inf(cmp % me, COUPLER, ACTION_SAVE_DG,:) % ar(1) % period)
	time_l_save_dg = INT(time_save_dg/time_step_min)

	! Swapping data
	if(cmp % inf(COUPLER, cmp % me, ACTION_READ_FD,ARRAY_TYPE_2D) % ar_size > 0) time_read_fd = MINVAL(cmp % inf(COUPLER, cmp % me, ACTION_READ_FD,ARRAY_TYPE_2D) % ar(:) % period, 	&
													cmp % inf(COUPLER, cmp % me, ACTION_READ_FD,ARRAY_TYPE_2D) % ar(:) % period > 0)
	time_l_read_fd = INT(time_read_fd/time_step_min)
			
	! Integral diagnostics
	if(time_save_it/time_step_min /= ANINT(time_save_it/time_step_min)) 															&
			call shared_error(TRIM(cmp % name)//" - <time_save_it> period cant be achieved with current <time_step_min>.")		
	time_l_save_it = INT(time_save_it/time_step_min)
	if(time_save_it /= 0 .AND. (time_save_it <= 24*60 .AND. MOD(24*60,time_save_it) /= 0 ) .OR. (time_save_it >= 24*60 .AND. MOD(time_days_in_run*24*60,time_save_it) /= 0))					&
			call shared_error(TRIM(cmp % name)//" - <time_save_it> period cannot be equally distributed in run time. Make it more round, e.g. 20 or 24*60 minutes.")
	
	! Make 'time_sync_it' explicitly greater then 1 day
	if(time_save_it < 1*24*60) then
		if(time_save_it /= 0) time_sync_it = 1*24*60/time_save_it
	else 
		time_sync_it = 1
	end if
	
	! Read control point
	if(ALL(cmp % inf(COUPLER, cmp % me, ACTION_READ_CP,:) % ar_size == 0) .AND. comp_read_cp) 												&
			call shared_error(TRIM(cmp % name)//" reg_array() subroutine - you define start from control point, but set no data to read.")
			
	!============================
	! Send information to coupler
	!============================
	! Personal information
	call MPI_SEND(cmp % west, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
	call MPI_SEND(cmp % east, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
	call MPI_SEND(cmp % south, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
	call MPI_SEND(cmp % north, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)

	! Common information
	if(rank_local == cmp % major_slave_rank) then

		call MPI_SEND(cmp % i_size, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(cmp % j_size, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(cmp % k_size, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(cmp % m_size, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(cmp % inp, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(cmp % jnp, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
				
		call MPI_SEND(comp_read_cp, 1, MPI_LOGICAL, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(time_read_fd, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(time_save_dg, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(time_save_it, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(time_sync_it, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)

		call MPI_SEND(cmp % name, NAME_LEN, MPI_CHARACTER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)

		! Send RW-table
		do a = 1, ACTION_NUM_FIELDS
		do t = 1, ARRAY_TYPE_NUM_FIELDS
		
			if(a == ACTION_READ_CP .OR. a == ACTION_READ_IC .OR. a == ACTION_READ_FD) then
				s = COUPLER
				d = cmp % me
			else
				s = cmp % me
				d = COUPLER
			end if
			
			call MPI_SEND(cmp % inf(s,d,a,t) % ar_size, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
				
			do n = 1, cmp % inf(s,d,a,t) % ar_size
				call MPI_SEND(cmp % inf(s,d,a,t) % ar(n) % name, NAME_LEN, MPI_CHARACTER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
				call MPI_SEND(cmp % inf(s,d,a,t) % ar(n) % lname, NAME_LEN, MPI_CHARACTER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
				call MPI_SEND(cmp % inf(s,d,a,t) % ar(n) % units, NAME_LEN, MPI_CHARACTER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
				call MPI_SEND(cmp % inf(s,d,a,t) % ar(n) % file, NAME_LEN, MPI_CHARACTER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
				call MPI_SEND(cmp % inf(s,d,a,t) % ar(n) % period, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
			end do
		
		end do
		end do
		
		call MPI_SEND(cmp % inf(1:NUM_OF_COMP,1:NUM_OF_COMP,ACTION_MAPPING,ARRAY_TYPE_2D) % ar_size, NUM_OF_COMP**2, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(cmp % inf(1:NUM_OF_COMP,1:NUM_OF_COMP,ACTION_MAPPING,ARRAY_TYPE_2D) % ar(1) % map_type, NUM_OF_COMP**2, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		call MPI_SEND(cmp % inf(1:NUM_OF_COMP,1:NUM_OF_COMP,ACTION_MAPPING,ARRAY_TYPE_2D) % ar(1) % period, NUM_OF_COMP**2, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, ierr)
		
	end if
	
	!============================================
	! Receive some parameters from master coupler
	!============================================
	call MPI_RECV(time_min, 1, MPI_INTEGER, cmp % master_rank, cmp % me, MPI_COMM_WORLD, status_ini_recv, ierr)

	! Note, that saved timestep is already calculated step, so we should increment it
	time_l = nint(dble(time_min)/dble(time_step_min)) + 1
	
	! Initiate send-recv operations
	call comp_sendrecv_ini
	
	end subroutine comp_reg_sendrecv
	
	
!===============================================================================================================================================================	
	
	subroutine comp_send_data(this, send_to, action)
	
	class (comp) :: this	
	integer send_to, action	
	
	integer comp_status_sr(MPI_STATUS_SIZE, ARRAY_TYPE_NUM_FIELDS)
	integer c, cur_kl, cur_ml, arrType
	
	!=================================================================
	! Sending operations are non-synchronized for continous
	! work of component model, thus we must 'Wait' untill one
	! previous 'ISend' operation complete and we can use send buffer.
	!=================================================================
	
!	write(*,*) TRIM(cmp % name)//' start send '//TRIM(event_type_name(action))//' to '//TRIM(cmp % names(send_to))//" on step "//I2S(time_l)
	call MPI_WAITALL (ARRAY_TYPE_NUM_FIELDS, comp_cur_request_send, comp_status_sr, ierr)
!	if(send_to /= 0) write(*,*) 'Comp '//TRIM(cmp % name)//' started send to '//TRIM(cmp % names(send_to))//' '//TRIM(event_type_name(action))//' on step '//I2S(time_l)
	
	!===============================================================
	! Copy arrays with appropriate action and destination to buffers
	!===============================================================
#define SEND
#define ARR_TYPE ARRAY_TYPE_4D
#define PTR p_4D
#define PTRd p_4D_kind8
#define BUF comp_s_pack_4D
#define IND k,i,j,m
#include "comp_template.f90"

#define ARR_TYPE ARRAY_TYPE_3D
#define PTR p_3D
#define PTRd p_3D_kind8
#define BUF comp_s_pack_3D
#define IND k,i,j
#include "comp_template.f90"

#define ARR_TYPE ARRAY_TYPE_2D
#define PTR p_2D
#define PTRd p_2D_kind8
#define BUF comp_s_pack_2D
#define IND i,j
#include "comp_template.f90"

#define ARR_TYPE ARRAY_TYPE_2D_KIND8
#define PTR p_2D_kind8
#define BUF comp_s_pack_2D_kind8
#define IND i,j
#include "comp_template.f90"
#undef SEND

	!=============================================
	! Start persistent requests for not null cases
	!=============================================
	comp_cur_request_send = MPI_REQUEST_NULL
	c = 0
	do i = 1, ARRAY_TYPE_NUM_FIELDS
		if(comp_request_sendrecv(INF_SR_SEND, action, i, send_to) /= MPI_REQUEST_NULL) then
			c = c + 1
			comp_cur_request_send(c) = comp_request_sendrecv(INF_SR_SEND, action, i, send_to)
		end if
	end do
	
	call MPI_STARTALL(c, comp_cur_request_send(1:c), ierr)			

	end subroutine comp_send_data

!===============================================================================================================================================================	
	
	subroutine comp_recv_data(this, recv_from, action)
	
	class (comp) :: this	
	integer recv_from, action	

	integer c_4D, c_3D, c_2D, c_2D8
	integer comp_status_sr(MPI_STATUS_SIZE, ARRAY_TYPE_NUM_FIELDS)
	integer cur_request(ARRAY_TYPE_NUM_FIELDS)
	integer c, cur_kl, cur_ml, arrType

	!=============================================
	! Start persistent requests for not null cases
	!=============================================
!	write(*,*) 'Comp '//TRIM(cmp % name)//' started recv '//TRIM(event_type_name(action))//' on step '//I2S(time_l)

	cur_request = MPI_REQUEST_NULL
	c = 0
	do i = 1, ARRAY_TYPE_NUM_FIELDS
		if(comp_request_sendrecv(INF_SR_RECV, action, i, recv_from) /= MPI_REQUEST_NULL) then
			c = c + 1
			cur_request(c) = comp_request_sendrecv(INF_SR_RECV, action, i, recv_from)
		end if
	end do

!	write(*,*) TRIM(cmp % name)//' start receive '//TRIM(event_type_name(action))//' from '//TRIM(cmp % names(recv_from))//" on step "//I2S(time_l)
	
	call MPI_STARTALL(c, cur_request(1:c), ierr)
	call MPI_WAITALL (c, cur_request(1:c), comp_status_sr, ierr)
	
!	write(*,*) TRIM(cmp % name)//' receive '//TRIM(event_type_name(action))//' from '//TRIM(cmp % names(recv_from))//" on step "//I2S(time_l)
	
	!==============================================================
	! Copy arrays with appropriate action and destination to arrays
	!==============================================================
#define RECV
#define ARR_TYPE ARRAY_TYPE_4D
#define PTR p_4D
#define PTRd p_4D_kind8
#define BUF comp_r_pack_4D
#define IND k,i,j,m
#include "comp_template.f90"

#define ARR_TYPE ARRAY_TYPE_3D
#define PTR p_3D
#define PTRd p_3D_kind8
#define BUF comp_r_pack_3D
#define IND k,i,j
#include "comp_template.f90"

#define ARR_TYPE ARRAY_TYPE_2D
#define PTR p_2D
#define PTRd p_2D_kind8
#define BUF comp_r_pack_2D
#define IND i,j
#include "comp_template.f90"

#define ARR_TYPE ARRAY_TYPE_2D_KIND8
#define PTR p_2D_kind8
#define BUF comp_r_pack_2D_kind8
#define IND i,j
#include "comp_template.f90"
#undef RECV
		
	end subroutine comp_recv_data
	
	
!===============================================================================================================================================================	
	
!=========================================
! Accumulate and send integral diagnostics
!=========================================

	subroutine comp_send_intg(this, area, ints_vector, names_vector)
	
	class (comp) :: this
	real(kind=8), optional :: area
	real(kind=8), optional :: ints_vector(:)
	character(LEN=NAME_LEN), optional :: names_vector(:)

	if(time_l_save_it == 0) return
	
	! Accumulate data at runtime and send it at the finalize stage
	if(PRESENT(area) .AND. PRESENT(ints_vector) .AND. PRESENT(names_vector)) then

		if(MOD(time_l, time_l_save_it) /= 0) return
	
		! At first call allocate buffer and send names
		if(.NOT. ALLOCATED(comp_intgrl_values)) then
			allocate(comp_intgrl_values(time_sync_it, BUFFER_LEN))
			
			if(rank_local == 0) then 

				call MPI_SEND(	SIZE(ints_vector),					& 
						1,											&
						MPI_INTEGER,	 							&
						0, 											&
						shared_mpi_tags(ACTION_SAVE_IT,COUPLER),	&
						MPI_COMM_WORLD, ierr)

				call MPI_SEND(	names_vector,						& 
						SIZE(ints_vector)*NAME_LEN,					&
						MPI_CHARACTER,	 							&
						0, 											&
						shared_mpi_tags(ACTION_SAVE_IT,COUPLER),	&
						MPI_COMM_WORLD, ierr)
						
			end if
		end if
		
		! Accumulate values
		comp_intgrl_num = SIZE(ints_vector)
		comp_intgrl_area = area
		
		comp_intgrl_values(comp_ingtrl_c, 1:comp_intgrl_num) = ints_vector(:)
		comp_ingtrl_c = comp_ingtrl_c + 1
		if(comp_ingtrl_c > time_sync_it) comp_ingtrl_c = 1
		
	else	
		! If user set nonzero period of IT, but forget pass integrals
		if(.NOT. ALLOCATED(comp_intgrl_values) .AND. rank_local == 0) call shared_error(TRIM(cmp % name)//": nonzero IT-save is set, but no integrals are passed.")
		
		! Everyone send integrals
		call MPI_SEND(	comp_intgrl_values(:,1:comp_intgrl_num),	& 
				SIZE(comp_intgrl_values(:,1:comp_intgrl_num)),		&
				MPI_DOUBLE_PRECISION, 								&
				0, 													&
				shared_mpi_tags(ACTION_SAVE_IT,COUPLER),			&
				MPI_COMM_WORLD, ierr)

		! Everyone send area
		call MPI_SEND(	comp_intgrl_area,							& 
				1,													&
				MPI_DOUBLE_PRECISION, 								&
				0, 													&
				shared_mpi_tags(ACTION_SAVE_IT,COUPLER),			&
				MPI_COMM_WORLD, ierr)		
	end if

	end subroutine comp_send_intg
	
!===============================================================================================================================================================	
!===================================================================
! Initiate MPI_INIT calls for all possible cases of send and receive
!===================================================================

	subroutine comp_sendrecv_ini
	
	integer :: rw_sizes(ACTION_NUM_FIELDS+1, ARRAY_TYPE_NUM_FIELDS)
		
	!=======================================
	! Allocate send and receive data vectors
	!=======================================
	
	! Compute sizes of arrays
	do a = 1, ACTION_NUM_FIELDS+1
		rw_sizes(a,ARRAY_TYPE_4D) = MAXVAL(cmp % inf(:,:,a,ARRAY_TYPE_4D) % ar_size)*cmp % i_loc_size*cmp % j_loc_size*cmp % k_size*cmp % m_size
		rw_sizes(a,ARRAY_TYPE_3D) = MAXVAL(cmp % inf(:,:,a,ARRAY_TYPE_3D) % ar_size)*cmp % i_loc_size*cmp % j_loc_size*cmp % k_size
		rw_sizes(a,ARRAY_TYPE_2D) = MAXVAL(cmp % inf(:,:,a,ARRAY_TYPE_2D) % ar_size)*cmp % i_loc_size*cmp % j_loc_size
		rw_sizes(a,ARRAY_TYPE_2D_KIND8) = MAXVAL(cmp % inf(:,:,a,ARRAY_TYPE_2D_KIND8) % ar_size)*cmp % i_loc_size*cmp % j_loc_size
	end do

	allocate (comp_s_pack_4D(MAXVAL(rw_sizes(ACTION_SAVE_CP:ACTION_SAVE_DG,ARRAY_TYPE_4D)))); comp_s_pack_4D = 0
	allocate (comp_s_pack_3D(MAXVAL(rw_sizes(ACTION_SAVE_CP:ACTION_SAVE_DG,ARRAY_TYPE_3D)))); comp_s_pack_3D = 0
	allocate (comp_s_pack_2D(MAXVAL(rw_sizes(:,ARRAY_TYPE_2D)))); comp_s_pack_2D = 0
	allocate (comp_s_pack_2D_kind8(MAXVAL(rw_sizes(ACTION_SAVE_CP:ACTION_SAVE_DG,ARRAY_TYPE_2D_KIND8)))); comp_s_pack_2D_kind8 = 0

	allocate (comp_r_pack_4D(MAXVAL(rw_sizes(ACTION_READ_CP:ACTION_READ_FD,ARRAY_TYPE_4D)))); comp_r_pack_4D = 0
	allocate (comp_r_pack_3D(MAXVAL(rw_sizes(ACTION_READ_CP:ACTION_READ_FD,ARRAY_TYPE_3D)))); comp_r_pack_3D = 0
	allocate (comp_r_pack_2D(MAXVAL(rw_sizes(:, ARRAY_TYPE_2D)))); comp_r_pack_2D = 0
	allocate (comp_r_pack_2D_kind8(MAXVAL(rw_sizes(ACTION_READ_CP:ACTION_READ_FD,ARRAY_TYPE_2D_KIND8)))); comp_r_pack_2D_kind8 = 0

	!=====================================================
	! Allocate and initiate request and last send variable
	!=====================================================
	allocate (comp_request_sendrecv(INF_SR_NUM_FIELDS, ACTION_NUM_FIELDS + 1, ARRAY_TYPE_NUM_FIELDS, 0:NUM_OF_COMP))
	comp_request_sendrecv = MPI_REQUEST_NULL
	comp_cur_request_send = MPI_REQUEST_NULL

	!================================================
	! Initiate all possible cases of send and receive
	!================================================

	! Save control point
	if(time_save_cp /= 0) then
		call comp_send_mpi_init(rw_sizes(ACTION_SAVE_CP, ARRAY_TYPE_4D), ACTION_SAVE_CP, ARRAY_TYPE_4D, COUPLER, buf4 = comp_s_pack_4D)
		call comp_send_mpi_init(rw_sizes(ACTION_SAVE_CP, ARRAY_TYPE_3D), ACTION_SAVE_CP, ARRAY_TYPE_3D, COUPLER, buf4 = comp_s_pack_3D)
		call comp_send_mpi_init(rw_sizes(ACTION_SAVE_CP, ARRAY_TYPE_2D), ACTION_SAVE_CP, ARRAY_TYPE_2D, COUPLER, buf4 = comp_s_pack_2D)
		call comp_send_mpi_init(rw_sizes(ACTION_SAVE_CP, ARRAY_TYPE_2D_KIND8), ACTION_SAVE_CP, ARRAY_TYPE_2D_KIND8, COUPLER, buf8 = comp_s_pack_2D_kind8)
	end if
	
	! Save diagnostics
	call comp_send_mpi_init(rw_sizes(ACTION_SAVE_DG, ARRAY_TYPE_3D), ACTION_SAVE_DG, ARRAY_TYPE_3D, COUPLER, buf4 = comp_s_pack_3D)
	call comp_send_mpi_init(rw_sizes(ACTION_SAVE_DG, ARRAY_TYPE_2D), ACTION_SAVE_DG, ARRAY_TYPE_2D, COUPLER, buf4 = comp_s_pack_2D)
	
	! Read control point
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_CP, ARRAY_TYPE_4D), ACTION_READ_CP, ARRAY_TYPE_4D, COUPLER, buf4 = comp_r_pack_4D)
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_CP, ARRAY_TYPE_3D), ACTION_READ_CP, ARRAY_TYPE_3D, COUPLER, buf4 = comp_r_pack_3D)
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_CP, ARRAY_TYPE_2D), ACTION_READ_CP, ARRAY_TYPE_2D, COUPLER, buf4 = comp_r_pack_2D)
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_CP, ARRAY_TYPE_2D_KIND8), ACTION_READ_CP, ARRAY_TYPE_2D_KIND8, COUPLER, buf8 = comp_r_pack_2D_kind8)
	
	! Read real initial conditions
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_IC, ARRAY_TYPE_4D), ACTION_READ_IC, ARRAY_TYPE_4D, COUPLER, buf4 = comp_r_pack_4D)
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_IC, ARRAY_TYPE_3D), ACTION_READ_IC, ARRAY_TYPE_3D, COUPLER, buf4 = comp_r_pack_3D)
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_IC, ARRAY_TYPE_2D), ACTION_READ_IC, ARRAY_TYPE_2D, COUPLER, buf4 = comp_r_pack_2D)
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_IC, ARRAY_TYPE_2D_KIND8), ACTION_READ_IC, ARRAY_TYPE_2D_KIND8, COUPLER, buf8 = comp_r_pack_2D_kind8)

	! Read file data
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_FD, ARRAY_TYPE_4D), ACTION_READ_FD, ARRAY_TYPE_4D, COUPLER, buf4 = comp_r_pack_4D)
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_FD, ARRAY_TYPE_3D), ACTION_READ_FD, ARRAY_TYPE_3D, COUPLER, buf4 = comp_r_pack_3D)
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_FD, ARRAY_TYPE_2D), ACTION_READ_FD, ARRAY_TYPE_2D, COUPLER, buf4 = comp_r_pack_2D)
	call comp_recv_mpi_init(rw_sizes(ACTION_READ_FD, ARRAY_TYPE_2D_KIND8), ACTION_READ_FD, ARRAY_TYPE_2D_KIND8, COUPLER, buf8 = comp_r_pack_2D_kind8)

	! Mapping send-receive. Note, that tag is defined only for one-way mapping case
	do n = 1, NUM_OF_COMP
		call comp_send_mpi_init(cmp % inf(cmp % me, n, ACTION_MAPPING, ARRAY_TYPE_2D) % ar_size*cmp % i_loc_size*cmp % j_loc_size, ACTION_MAPPING, ARRAY_TYPE_2D, n, buf4 = comp_s_pack_2D)
		call comp_recv_mpi_init(cmp % inf(n, cmp % me,ACTION_MAPPING, ARRAY_TYPE_2D) % ar_size*cmp % i_loc_size*cmp % j_loc_size, ACTION_MAPPING, ARRAY_TYPE_2D, n, buf4 = comp_r_pack_2D)
	end do

	return

	contains
	
!========================================================================================================================================================================	
!==================================
! Add MPI_SEND_INIT for common case
!==================================

	subroutine comp_send_mpi_init(mesSize, act, arrType, adress, buf4, buf8)
	
	implicit none
	
	real(kind=4), optional :: buf4(:)
	real(kind=8), optional :: buf8(:)
	integer act, arrType, dataType, adress, mesSize
	
	if(mesSize == 0) return
	
	if(PRESENT(buf4)) then	

		call MPI_SSEND_INIT(buf4,								& 
				mesSize,								&
				MPI_REAL, 								&
				cmp % master_rank, 							&
				shared_mpi_tags(act,adress),						&
				MPI_COMM_WORLD, 							&
				comp_request_sendrecv(INF_SR_SEND, act, arrType, adress), ierr	)

	else if(PRESENT(buf8)) then
	
		call MPI_SSEND_INIT(buf8,								& 
				mesSize,								&
				MPI_DOUBLE_PRECISION, 							&
				cmp % master_rank, 							&
				shared_mpi_tags(act,adress),						&
				MPI_COMM_WORLD, 							&
				comp_request_sendrecv(INF_SR_SEND, act, arrType, adress), ierr	)
	end if
	
	return
	end subroutine comp_send_mpi_init
	
	
!========================================================================================================================================================================	
!==================================
! Add MPI_RECV_INIT for common case
!==================================

	subroutine comp_recv_mpi_init(mesSize, act, arrType, adress, buf4, buf8)
	
	implicit none
	
	real(kind=4), optional :: buf4(:)
	real(kind=8), optional :: buf8(:)
	integer act, arrType, dataType, adress, mesSize

	if(mesSize == 0) return
	
	if(PRESENT(buf4)) then	

		call MPI_RECV_INIT(buf4,								& 
				mesSize,								&
				MPI_REAL, 								&
				cmp % master_rank, 							&
				shared_mpi_tags(act,adress),						&
				MPI_COMM_WORLD, 							&
				comp_request_sendrecv(INF_SR_RECV, act, arrType, adress), ierr	)

	else if(PRESENT(buf8)) then
	
		call MPI_RECV_INIT(buf8,								& 
				mesSize,								&
				MPI_DOUBLE_PRECISION, 							&
				cmp % master_rank, 							&
				shared_mpi_tags(act,adress),						&
				MPI_COMM_WORLD, 							&
				comp_request_sendrecv(INF_SR_RECV, act, arrType, adress), ierr	)
	end if
	
	end subroutine comp_recv_mpi_init
	
	end subroutine comp_sendrecv_ini	
		
!========================================================================================================================================================================		

	subroutine comp_reduce_itg(this, area, ints, gInts)
	
	class (comp) :: this
	real(kind=8) :: area, ints(:), gInts(:)
	real(kind=8) :: gArea, rbuf(SIZE(ints)+1)
		
	call MPI_ALLREDUCE([area, ints(:)], rbuf, SIZE(ints)+1, MPI_DOUBLE_PRECISION, MPI_SUM, comm_local, ierr)

	gArea = rbuf(1)
	
	gInts = 0
	do i = 1, SIZE(gInts)
		if(1+i <= SIZE(rbuf)) gInts(i) = rbuf(1+i)/gArea
	end do	
	
	end subroutine comp_reduce_itg


!========================================================================================================================================================================


	end module comp_module
