
!--------------------------------------
!
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!		Compact Modelling Framework	
!			CMF2.0
!
! authors: 
!	2013.08.07 / Kalmykov, Vladimir, IO RAS, Russia, vvk88@mail.ru
!
! reference:
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Vychisl. Metody Programm., 2013, v. 14, iss. 3, 88-95. 
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Certificate of state registration of the computer progam number 2013619320. 
!		Date of state registration in the Registry of the computer programs October 1, 2013
!--------------------------------------


	module off_module

	use shared_utils_module
	use netcdf
	use c_rw_module
	
	implicit none

	TYPE, ABSTRACT, PUBLIC :: GridConstructor
	
	character(50) name
	character(1) gType
	integer nx, ny, nz
	integer corners
	integer rank
	integer, allocatable :: dims(:)
	real (kind = 8), allocatable :: center_lat(:,:), center_lon(:,:)
	real (kind = 8), allocatable :: corner_lat(:,:,:), corner_lon(:,:,:)
	real (kind = 4), allocatable :: z(:)
	real (kind = 4), allocatable :: mask2D(:,:)
	
	CONTAINS
	
	! Common offline procedures	
	PROCEDURE :: put_var => off_put_var
	PROCEDURE :: get_var => off_get_var
	PROCEDURE :: make_CDO_file => off_make_CDO_file
	PROCEDURE :: make_SCRIP_file => off_make_SCRIP_file	
	PROCEDURE :: alloc => off_constructor

	! Abstract interfaces for components		
	PROCEDURE(I_make_grid), DEFERRED :: make_grid
	PROCEDURE(I_make_mask), DEFERRED :: make_mask
	PROCEDURE(I_make_ic), DEFERRED :: make_ic	
	
	END TYPE GridConstructor
	
	ABSTRACT INTERFACE
		subroutine I_make_grid(this)
			import GridConstructor
			class(GridConstructor) :: this
		end subroutine I_make_grid

		subroutine I_make_mask(this)
			import GridConstructor
			class(GridConstructor) :: this
		end subroutine I_make_mask
		
		subroutine I_make_ic(this)
			import GridConstructor
			class(GridConstructor) :: this
		end subroutine I_make_ic
	END INTERFACE

	! Data folder
	character(*), parameter :: data_folder = './data/'
	class(GridConstructor), pointer :: grid_p
	
	CONTAINS
	
!=====================================================================================================================================================
!===========================
! Allocate memory for object
!===========================
	subroutine off_constructor(this)
			
	class(GridConstructor) :: this
	
	! Allocate object
	this % rank = 2
	this % corners = 4
	allocate(this % dims(this % rank))
	this % dims(1) = this % nx
	this % dims(2) = this % ny	
	allocate(this % center_lat(this % nx, this % ny))
	allocate(this % center_lon(this % nx, this % ny))
	allocate(this % corner_lat(this % corners, this % nx, this % ny))
	allocate(this % corner_lon(this % corners, this % nx, this % ny))
	allocate(this % z(this % nz)); this % z = 0
	allocate(this % mask2D(this % nx, this % ny)) ; this % mask2D = 1

	end subroutine off_constructor

!=====================================================================================================================================================
!==================================
! Add array to IC-file of component
!==================================

	subroutine off_put_var(this, var, long_name, units, axis, arr_1D, arr_2D, arr_3D, arr_2D_8, external_filename, add_nx, add_ny)
	
	! Input variablesR
	class(GridConstructor) :: this
	character(*) :: var, long_name, units
	character(*), optional :: axis 
	real(kind=4), optional :: arr_1D(:), arr_2D(:,:), arr_3D(:,:,:)
	real(kind=8), optional :: arr_2D_8(:,:)

	!====> ice start <====!
	character(*), optional :: external_filename
	integer, optional :: add_nx, add_ny
	integer additional_nx, additional_ny
	!====>  ice end  <====!

	! Work variables
	character(100) :: filename	
	logical file_exists	
	integer nc_id, var_id
	integer i_dimid, j_dimid, k_dimid	
	
	!====> ice start <====!
	if(PRESENT(external_filename)) then
        filename = TRIM(external_filename)
    	else
    	!====>  ice end  <====!
	
	filename = "../data/"//TRIM(this % name)//'_'//TRIM(I2S(this % nx))//'x'//TRIM(I2S(this % ny))//'x'//TRIM(I2S(this % nz))//"_IC.nc"	

	!====> ice start <====!
	endif
	additional_nx = 0
	additional_ny = 0
	if(PRESENT(add_nx)) &
		additional_nx = add_nx
	if(PRESENT(add_ny)) &
		additional_ny = add_ny
	!====>  ice end  <====!

	write(*,'(A$)') "Add variable <"//var//"> to file '"//TRIM(filename)//"'..."

	INQUIRE(file = filename, exist = file_exists)	
	if(.NOT. file_exists) then
		! Create and prepare netCDF file
		call nc_rc(NF90_CREATE (TRIM(filename), IOR(NF90_NETCDF4,NF90_CLOBBER), nc_id))
		!call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL, 'title', filename))
		call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL, "conventions", "CF-1.7"))
        call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL,"_FillValue", missval))
        call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL,"missing_value", missval))
		!call nc_rc( NF90_DEF_DIM(nc_id, "i", this % nx, i_dimid) 
		!call nc_rc( NF90_DEF_DIM(nc_id, "j", this % ny, j_dimid) )	
		call nc_rc( NF90_DEF_DIM(nc_id, "i", this % nx + additional_nx, i_dimid) )			! add additional_nx
		call nc_rc( NF90_DEF_DIM(nc_id, "j", this % ny + additional_ny, j_dimid) )			! add additional_ny
		call nc_rc( NF90_DEF_DIM(nc_id, "k", this % nz, k_dimid) )

		call nc_rc( NF90_ENDDEF(nc_id))
	else
		! Open existing file and inquire dimensions
		call nc_rc(NF90_OPEN(TRIM(filename), NF90_WRITE, nc_id))
		call nc_rc( NF90_INQ_DIMID(nc_id, "i", i_dimid) )
		call nc_rc( NF90_INQ_DIMID(nc_id, "j", j_dimid) )
		call nc_rc( NF90_INQ_DIMID(nc_id, "k", k_dimid) )	
	end if
	
	! Define variables and attributes
	call nc_rc( NF90_REDEF(nc_id))
	if(PRESENT(arr_1D)  .AND. NF90_INQ_VARID(nc_id, var, var_id) /= NF90_NOERR) call nc_rc(NF90_DEF_VAR(nc_id, var, NF90_REAL,  (/ k_dimid /), var_id))
	if(PRESENT(arr_2D)  .AND. NF90_INQ_VARID(nc_id, var, var_id) /= NF90_NOERR) call nc_rc(NF90_DEF_VAR(nc_id, var, NF90_REAL,  (/ i_dimid, j_dimid /), var_id))
	if(PRESENT(arr_3D)  .AND. NF90_INQ_VARID(nc_id, var, var_id) /= NF90_NOERR) call nc_rc(NF90_DEF_VAR(nc_id, var, NF90_REAL,  (/ i_dimid, j_dimid , k_dimid/), var_id))
	if(PRESENT(arr_2D_8).AND. NF90_INQ_VARID(nc_id, var, var_id) /= NF90_NOERR) call nc_rc(NF90_DEF_VAR(nc_id, var, NF90_DOUBLE,(/ i_dimid, j_dimid /), var_id))

	call nc_rc( NF90_PUT_ATT(nc_id, var_id, "units", units) )
	call nc_rc( NF90_PUT_ATT(nc_id, var_id, "long_name", long_name) )
	if(PRESENT(axis)) call nc_rc( NF90_PUT_ATT(nc_id, var_id, "axis", axis) )
		
	call nc_rc( NF90_ENDDEF(nc_id))
		
	! Write variables
	if(PRESENT(arr_1D))  call nc_rc( NF90_PUT_VAR(nc_id, var_id, arr_1D) )
	if(PRESENT(arr_2D))  call nc_rc( NF90_PUT_VAR(nc_id, var_id, arr_2D) )
	if(PRESENT(arr_3D))  call nc_rc( NF90_PUT_VAR(nc_id, var_id, arr_3D) )	
	if(PRESENT(arr_2D_8))call nc_rc( NF90_PUT_VAR(nc_id, var_id, arr_2D_8) )
	
	call nc_rc( NF90_CLOSE(nc_id))
	
	write(*,'(A)') ' success.'
	
	end subroutine off_put_var	
	
!=====================================================================================================================================================
!==================================
! Read 2D variable from netCDF file
!==================================

	subroutine off_get_var(this, filename, var,arr_1D, arr_2D, arr_3D, arr_2D_8, arr_3D_8, misVal)
	
	! Intput/output variables
	class(GridConstructor) :: this
	character(*) :: filename, var
	real(kind=4), optional :: arr_1D(:), arr_2D(:,:), arr_3D(:,:,:)
	real(kind=8), optional :: arr_2D_8(:,:),arr_3D_8(:,:,:)
	real(kind=4), optional :: misVal

	! Work variables
	logical file_exists
	integer nc_id, var_id
	
	! Check errors and open file
	INQUIRE(file = filename, exist = file_exists)	
	if(.NOT. file_exists) then
		write(*,'(/A/)') "Error: try to read, but have no file: "//filename
		stop
	end if
	
	call nc_rc(NF90_OPEN(TRIM(filename), NF90_NOWRITE, nc_id))

	! Read variable
	call nc_rc( NF90_INQ_VARID(nc_id, var, var_id))
	if(PRESENT(arr_1D))  call nc_rc( NF90_GET_VAR(nc_id, var_id, arr_1D))	
	if(PRESENT(arr_2D))  call nc_rc( NF90_GET_VAR(nc_id, var_id, arr_2D))
	if(PRESENT(arr_3D))  call nc_rc( NF90_GET_VAR(nc_id, var_id, arr_3D))
	if(PRESENT(arr_2D_8))call nc_rc( NF90_GET_VAR(nc_id, var_id, arr_2D_8))
	if(PRESENT(arr_3D_8))call nc_rc( NF90_GET_VAR(nc_id, var_id, arr_3D_8))

	! Read missing value
	if(PRESENT(misVal)) then
		call nc_rc(NF90_GET_ATT(nc_id, var_id, "_FillValue", misVal))
	end if
	
	! Close file
	call nc_rc( NF90_CLOSE(nc_id))
	
	end subroutine off_get_var	
	
!=====================================================================================================================================================
!================================================================
! Make netCDF file in modified SCRIP format for CDO operators
!================================================================

	subroutine off_make_CDO_file(this,add_nx,add_ny)

	! Input grid
	class(GridConstructor) :: this	

	! netCDF variables
	integer nc_id	
	integer :: size_dimid, xsize_dimid, ysize_dimid, zsize_dimid, rank_dimid, corners_dimid
	integer :: dims_id, center_lat_id, center_lon_id, corner_lat_id, corner_lon_id, mask_id, z_id
	
	! Unrolled arrays
	real(kind = 8), dimension(this % nx * this % ny) :: vect_cent_lat, vect_cent_lon
	real(kind = 8), dimension(this % corners, this % nx * this % ny) :: vect_corn_lat, vect_corn_lon
	integer, dimension(this % nx * this % ny) :: vect_mask2D

	!====> ice start <====!
	integer, optional :: add_nx, add_ny
	integer additional_nx, additional_ny
	!====> ice end <====!
	
	! Work variables
	integer (kind=4) :: i, j, counter
	character(100) filename
	
	! Create file and set title
	filename = data_folder//TRIM(this % name)//'_'//TRIM(I2S(this % nx))//'x'//TRIM(I2S(this % ny))//'_'//this % gType//"_cdo.nc"

	write(*,'(A$)') "Add "//TRIM(this % name)//" grid to file '"//TRIM(filename)//"'..."
	
	!====> ice start <====!
	additional_nx = 0
	additional_ny = 0
	if(PRESENT(add_nx)) additional_nx = add_nx
	if(PRESENT(add_ny)) additional_ny = add_ny
	!====>  ice end  <====!
	
	
	call nc_rc(NF90_CREATE (filename, NF90_CLOBBER, nc_id))
	!call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL, 'title', this % name))
	call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL, "conventions", "CF-1.7"))
    call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL,"_FillValue", missval))
    call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL,"missing_value", missval))
	call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL, 'title', "CDO grid file"))
	! Define dimensions
	call nc_rc( NF90_DEF_DIM (nc_id, 'grid_size',  (this % nx+additional_nx) * (this % ny + additional_ny), size_dimid))
	call nc_rc( NF90_DEF_DIM (nc_id, 'grid_xsize', this % nx+additional_nx, xsize_dimid))
	call nc_rc( NF90_DEF_DIM (nc_id, 'grid_ysize', this % ny+additional_ny, ysize_dimid))
	call nc_rc( NF90_DEF_DIM (nc_id, 'grid_zsize', this % nz, zsize_dimid))
	call nc_rc( NF90_DEF_DIM (nc_id, 'grid_corners', this % corners,  corners_dimid))
	call nc_rc( NF90_DEF_DIM (nc_id, 'grid_rank', this % rank, rank_dimid))

	! Define coordinate variables and their attributes
	call nc_rc( NF90_DEF_VAR (nc_id, 'grid_dims', NF90_INT, rank_dimid, dims_id))
	call nc_rc( NF90_DEF_VAR (nc_id, 'grid_center_lat', NF90_FLOAT, (/ xsize_dimid, ysize_dimid /), center_lat_id))
	call nc_rc( NF90_DEF_VAR (nc_id, 'grid_center_lon', NF90_FLOAT, (/ xsize_dimid, ysize_dimid /), center_lon_id))
	call nc_rc( NF90_DEF_VAR (nc_id, 'grid_imask', NF90_INT, (/ xsize_dimid, ysize_dimid /), mask_id))	
	call nc_rc( NF90_DEF_VAR (nc_id, 'grid_corner_lat', NF90_FLOAT, (/ corners_dimid, xsize_dimid, ysize_dimid /), corner_lat_id))
	call nc_rc( NF90_DEF_VAR (nc_id, 'grid_corner_lon', NF90_FLOAT, (/ corners_dimid, xsize_dimid, ysize_dimid /), corner_lon_id))
	call nc_rc( NF90_DEF_VAR (nc_id, 'z', NF90_FLOAT, (/ zsize_dimid /), z_id))
	
	call nc_rc( NF90_PUT_ATT (nc_id, center_lat_id, 'units', 'degrees'))
	call nc_rc( NF90_PUT_ATT (nc_id, center_lat_id, 'bounds', 'grid_corner_lat'))
	call nc_rc( NF90_PUT_ATT (nc_id, center_lon_id, 'units', 'degrees'))
	call nc_rc( NF90_PUT_ATT (nc_id, center_lon_id, 'bounds', 'grid_corner_lon'))
	call nc_rc( NF90_PUT_ATT (nc_id, corner_lat_id, 'units', 'degrees'))
	call nc_rc( NF90_PUT_ATT (nc_id, corner_lon_id, 'units', 'degrees'))
	call nc_rc( NF90_PUT_ATT (nc_id, mask_id, 'units', 'unitless'))
	call nc_rc( NF90_PUT_ATT (nc_id, mask_id, 'coordinates', 'grid_center_lon grid_center_lat'))	
	call nc_rc( NF90_PUT_ATT (nc_id, mask_id, 'units', 'unitless'))
	call nc_rc( NF90_PUT_ATT (nc_id, z_id, 'units', 'meters'))	
	
	call nc_rc( NF90_ENDDEF(nc_id))

	! Write data
 	this % dims(1)=this % dims(1)+additional_nx		!====> ice start - end<====!
 	this % dims(2)=this % dims(2)+additional_ny		!====> ice start - end<====!
 	call nc_rc( NF90_PUT_VAR (nc_id, dims_id, this % dims))	
 	call nc_rc( NF90_PUT_VAR (nc_id, center_lat_id, REAL(this % center_lat(1-additional_nx:this%nx, 1-additional_ny:this%ny)))) 	!====> ice start - end<====!
 	call nc_rc( NF90_PUT_VAR (nc_id, center_lon_id, REAL(this % center_lon(1-additional_nx:this%nx, 1-additional_ny:this%ny))))	!====> ice start - end<====!
 	call nc_rc( NF90_PUT_VAR (nc_id, corner_lat_id, REAL(this % corner_lat(:,1-additional_nx:this%nx, 1-additional_ny:this%ny))))	!====> ice start - end<====!
 	call nc_rc( NF90_PUT_VAR (nc_id, corner_lon_id, REAL(this % corner_lon(:,1-additional_nx:this%nx, 1-additional_ny:this%ny))))	!====> ice start - end<====!
 	call nc_rc( NF90_PUT_VAR (nc_id, mask_id, NINT(this % mask2D(1-additional_nx:this%nx, 1-additional_ny:this%ny))))		!====> ice start - end<====!
 	call nc_rc( NF90_PUT_VAR (nc_id, z_id, this % z))
 		
	call nc_rc( NF90_CLOSE(nc_id))

	write(*,'(A)') ' success.'	
	
	end subroutine off_make_CDO_file	

	
!=====================================================================================================================================================
!===============================================================================
! This subroutine creates netCDF file from our meshes for SCRPIP purposes. You 
! should provide filled 'grid' structures for it - so look on 'GridType' type.
!===============================================================================

	subroutine off_make_SCRIP_file(this)

	! Input grid
	class(GridConstructor) :: this	

	! netCDF variables
	integer nc_id	
	integer :: size_dimid, rank_dimid, corners_dimid
	integer :: dims_id, center_lat_id, center_lon_id, corner_lat_id, corner_lon_id, mask_id
	
	! Unrolled arrays
	real(kind = 8), dimension(this % nx * this % ny) :: vect_cent_lat, vect_cent_lon
	real(kind = 8), dimension(this % corners, this % nx * this % ny) :: vect_corn_lat, vect_corn_lon
	integer, dimension(this % nx * this % ny) :: vect_mask2D
	
	! Work variables
	integer (kind=4) :: i, j, counter
	
	! Unroll arrays(SCRIP need it)
	counter = 1
	do j = 1, this % ny
	do i = 1, this % nx
		vect_cent_lat(counter) = this % center_lat(i,j)
		vect_cent_lon(counter) = this % center_lon(i,j)
		vect_corn_lat(:,counter) = this % corner_lat(:,i,j)
		vect_corn_lon(:,counter) = this % corner_lon(:,i,j)
		vect_mask2D(counter) = this % mask2D(i,j)
		
		counter = counter + 1
	end do
	end do

	! Create file and set title
	call nc_rc (NF90_CREATE ('SCRIP/'//TRIM(this % name)//'.nc', IOR(NF90_NETCDF4, NF90_CLOBBER), nc_id))
	call nc_rc (NF90_PUT_ATT (nc_id, NF90_GLOBAL, 'title', this % name))
	call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL, "conventions", "CF-1.7"))
    call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL,"_FillValue",missval))
    call nc_rc(NF90_PUT_ATT (nc_id, NF90_GLOBAL,"missing_value",missval))

	! Define dimensions
	call nc_rc ( NF90_DEF_DIM (nc_id, 'grid_size', this % nx * this % ny, size_dimid))
	call nc_rc ( NF90_DEF_DIM (nc_id, 'grid_corners', this % corners,  corners_dimid))
	call nc_rc ( NF90_DEF_DIM (nc_id, 'grid_rank', this % rank, rank_dimid))

	! Define coordinate variables
	call nc_rc ( NF90_DEF_VAR (nc_id, 'grid_dims', NF90_INT, rank_dimid, dims_id))
	call nc_rc ( NF90_DEF_VAR (nc_id, 'grid_center_lat', NF90_DOUBLE, size_dimid, center_lat_id))
	call nc_rc ( NF90_DEF_VAR (nc_id, 'grid_center_lon', NF90_DOUBLE, size_dimid, center_lon_id))
	call nc_rc ( NF90_DEF_VAR (nc_id, 'grid_corner_lat', NF90_DOUBLE, (/ corners_dimid, size_dimid /), corner_lat_id))
	call nc_rc ( NF90_DEF_VAR (nc_id, 'grid_corner_lon', NF90_DOUBLE, (/ corners_dimid, size_dimid /), corner_lon_id))
	
	call nc_rc ( NF90_PUT_ATT (nc_id, center_lat_id, 'units', 'degrees'))
	call nc_rc ( NF90_PUT_ATT (nc_id, center_lon_id, 'units', 'degrees'))
	call nc_rc ( NF90_PUT_ATT (nc_id, corner_lat_id, 'units', 'degrees'))
	call nc_rc ( NF90_PUT_ATT (nc_id, corner_lon_id, 'units', 'degrees'))

	! Define mask variable
	call nc_rc ( NF90_DEF_VAR (nc_id, 'grid_imask', NF90_INT, size_dimid, mask_id))
	call nc_rc ( NF90_PUT_ATT (nc_id, mask_id, 'units', 'unitless'))

	! End defining
	call nc_rc ( NF90_ENDDEF(nc_id))

	! Write data
 	call nc_rc ( NF90_PUT_VAR (nc_id, dims_id, this % dims))
 	call nc_rc ( NF90_PUT_VAR (nc_id, mask_id, vect_mask2D))
 	call nc_rc ( NF90_PUT_VAR (nc_id, center_lat_id, vect_cent_lat))
 	call nc_rc ( NF90_PUT_VAR (nc_id, center_lon_id, vect_cent_lon))
 	call nc_rc ( NF90_PUT_VAR (nc_id, corner_lat_id, vect_corn_lat))
 	call nc_rc ( NF90_PUT_VAR (nc_id, corner_lon_id, vect_corn_lon))
	
	call nc_rc ( NF90_CLOSE(nc_id))

	end subroutine off_make_SCRIP_file



!=============================================================================================================================================================================
	
	
	
	end module off_module	
	
	
	
