
!--------------------------------------
!
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!		Compact Modelling Framework	
!			CMF2.0
!
! authors: 
!	2013.08.07 / Kalmykov, Vladimir, IO RAS, Russia, vvk88@mail.ru
!
! reference:
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Vychisl. Metody Programm., 2013, v. 14, iss. 3, 88-95. 
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Certificate of state registration of the computer progam number 2013619320. 
!		Date of state registration in the Registry of the computer programs October 1, 2013
!--------------------------------------

!=====================================================
! Contains shared variables for coupler and components
!=====================================================

	module shared_module

	use shared_const_module
	use shared_utils_module
	use mpi
	
	implicit none
	
	! MPI decomposition variables
	integer size_world, rank_world
	integer size_local, rank_local, comm_local, me
	integer ierr

	! Experiment variables
	character(LEN = 3) run_experiment
	logical run_save_log
	real(kind = 8) wall_time_start_cycle, wall_time_finish, wt_with_io, wall_time_start_ini
	integer :: run_np(0:NUM_OF_COMP) = 0
	integer :: run_comps
	integer :: run_log_file_num = 0
	integer :: run_ini_np_parallel = 1
	
	! Time variables
	integer, dimension(12) :: time_month_size = (/31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31/)
	
	integer time_l
	integer time_l_in_run

	real(kind=4) :: time_step_min
	
	integer time_run_years, time_run_months, time_run_days, time_run_steps, time_days_in_run
	integer time_start_year, time_start_month, time_start_day
	integer time_save_cp
	logical time_append_cp, time_append_dg
	integer :: time_sync_fs

	integer :: time_read_fd = 0
	integer :: time_save_dg = 0
	integer :: time_save_it = 0
	integer :: time_sync_it = 0
	
	integer time_l_save_it, time_l_save_cp, time_l_save_dg, time_l_read_fd
	integer time_steps_day

	integer time_year
	integer time_l_year
	integer time_l_in_year
	integer time_day_in_year
	integer time_month
	integer time_day
	integer time_l_in_day
	integer time_hour
	integer time_min
	character(NAME_LEN) time_read_cp
	
	integer :: time_start_min
	
	character(LEN = 21) time_model_date

	! Service variables
	integer, allocatable :: shared_mpi_tags(:,:)
	character(3) run_phase
	
	TYPE SlaveInfo
		integer :: west, east, south, north
		integer :: il_belt, jl_belt
	END TYPE SlaveInfo
	
	TYPE WorkerInfo
		integer :: slave_start, slave_count		! Start and count of slave rank
		integer :: stripe_start, stripe_count		! Start and count of rows in stripe
		type(SlaveInfo), allocatable :: sl(:)
	END TYPE WorkerInfo
		
	TYPE :: ArrInfo
		character(NAME_LEN) :: name = ""
		character(NAME_LEN) :: lname = ""
		character(NAME_LEN) :: units = ""
		character(NAME_LEN) :: file = ""
		integer :: period = 0
		integer :: map_type = MYNULL
		real(kind=4), pointer :: p_2D(:,:) => NULL()
		real(kind=4), pointer :: p_3D(:,:,:) => NULL()
		real(kind=4), pointer :: p_4D(:,:,:,:) => NULL()
		real(kind=8), pointer :: p_4D_kind8(:,:,:,:) => NULL()	
		real(kind=8), pointer :: p_3D_kind8(:,:,:)	=> NULL()
		real(kind=8), pointer :: p_2D_kind8(:,:) => NULL()
	END TYPE ArrInfo
	
	TYPE :: VectorType
		type(ArrInfo) :: ar(BUFFER_LEN)
		integer :: ar_size = 0
	END TYPE VectorType
		
	TYPE ComponentInfo
		character(NAME_LEN) :: name = 'none'
		integer :: proc_count
		integer :: il, jl, kl, ml
		integer :: inp, jnp
		
		logical :: read_cp
		integer :: read_fd
		integer :: save_dg
		integer :: save_it 	! period of save IT in minutes
		integer :: sync_it 	! period of real receives and writes to file
		
		type(VectorType) :: rw(ACTION_NUM_FIELDS, ARRAY_TYPE_NUM_FIELDS)
		
		type(WorkerInfo), allocatable :: wk(:)
	END TYPE ComponentInfo
	
	
	CONTAINS

!=======================================================================================================================================================================================

!======================================================================================
! Shared inialization for all kernels - read command line parameters and call MPI_SPLIT
!======================================================================================

	subroutine shared_ini_params

	integer counter, i, j
	character(10) comp_arg, np_arg, arg
	integer np, ios
	logical file_exists
	
	namelist /RUN_INF/ time_start_year, time_start_month, time_start_day,			&
	time_save_cp, time_sync_fs, time_append_cp, time_append_dg, time_read_cp, run_save_log, run_log_file_num, run_ini_np_parallel

	!==========================================
	! Get global arguments: MPI and comman line
	!==========================================
	call MPI_INIT (ierr)
	call MPI_COMM_SIZE(MPI_COMM_WORLD, size_world, ierr)
	call MPI_COMM_RANK(MPI_COMM_WORLD, rank_world, ierr)

	! Read input namelist witn main run inforation
	INQUIRE(file = "run_list.in", exist = file_exists)
	if(file_exists) then
		open(11, file = "run_list.in")
		read(11, nml = RUN_INF)
		close(11)
	else
		call shared_error("Cant find input namelist 'run_list.in'.")
	end if
	
	if(run_log_file_num /= 6 .AND. rank_world == 0) write(*,*) "Output is redirected to '_log_file.txt'."
	
	! Read part of command line arguments (other part in c_main.f90)
	if(MOD(COMMAND_ARGUMENT_COUNT() - 4,2) /= 0) call shared_error("Usage: CPL 2 OCN 4 ATM 2 exp 0 0 1")
	run_comps =  (COMMAND_ARGUMENT_COUNT() - 4)/2	
	if(run_comps <= 1) call shared_error("At least one component should be run.")
			
	! Get experiment description
	call GET_COMMAND_ARGUMENT(run_comps*2 + 1, run_experiment)
	call GET_COMMAND_ARGUMENT(run_comps*2 + 2, arg)
	read(arg, *) time_run_years
	call GET_COMMAND_ARGUMENT(run_comps*2 + 3, arg)
	read(arg, *) time_run_months
	call GET_COMMAND_ARGUMENT(run_comps*2 + 4, arg)
	read(arg, *) time_run_days
	
	time_days_in_run = 365*time_run_years + SUM(time_month_size(1:time_run_months)) + time_run_days
	
	if(time_run_months > 12) call shared_error("'time_run_months' should be in interval [1,12].")	
	
	end subroutine shared_ini_params
	
!=======================================================================================================================================================================================
	
!===================================================================
! Decompose processor grid on colors per components. Each group will 
! have it's own internal communicator 'comm_local'.
!===================================================================

	subroutine shared_ini_split
	
	integer counter, i, j
	
	if(size_world /= SUM(run_np(0:NUM_OF_COMP))) call shared_error('<mpirun> size of communicator /= total sum of component sizes.')
	if(run_np(COUPLER) == 0) call shared_error("Coupler component is not set. Usage: CPL 2 OCN 4 ATM 2 exp 0 0 1")
	
	call MPI_COMM_SPLIT(MPI_COMM_WORLD, proc_where(), rank_world, comm_local, ierr)

	! Get local 'rank' and 'size' for components
	call MPI_COMM_SIZE(comm_local, size_local, ierr)
	call MPI_COMM_RANK(comm_local, rank_local, ierr)
	me = rank_local + 1

	! Fill system MPI_TAG table with unique numbers
	allocate(shared_mpi_tags(ACTION_NUM_FIELDS + 1, 0:NUM_OF_COMP))

	counter = 1
	do i = 1, ACTION_NUM_FIELDS + 1
	do j = 0, NUM_OF_COMP
		shared_mpi_tags(i,j) = counter
		counter = counter + 1
	end do
	end do
	
	if(run_ini_np_parallel < 0) call shared_error("'run_ini_np_parallel' should be non-negative integer.")	
	if(run_ini_np_parallel == 0) run_ini_np_parallel = size_local

	end subroutine shared_ini_split

!=======================================================================================================================================================================================
	
!=========================================================================================
! Compute real current time data from online variable:
!	-time_l 		- current global timestep			(0,1,2,..)
!
! Use external parameters:
! 	-time_start_year	- real year of modelling beginning		(0,1,2,..)
!	-time_steps_day		- number of timesteps per day
!
! Discription:
!	-time_year 		- current real year 				(1,2,3,..) (e.g., for CORE: 1948,1949, ...)
!	-time_l_year		- number of years from the starting point 	(0,1,2,..)
!	-time_l_in_year		- timestep from the beginning of the year 	(0,1,2,..)
!	-time_day_in_year	- current real day in the year			(1:365)
!	-time_month		- current real month in the year		(1:12)
!	-time_day		- current real day in the month		(1:31)
!	-time_l_in_day		- timestep from the beginning of the day 	(0,1,2,..)
!	-time_hour		- current real time in whole hours in the day	(0:23) 
!=========================================================================================
	
	subroutine shared_timer

	integer sum_day

	! Year
	time_l_year = time_l/(365*time_steps_day)
	time_year = time_l_year
	time_l_in_year = MOD(time_l, 365*time_steps_day)
	time_day_in_year = time_l_in_year/time_steps_day + 1

	! Month
	time_month = 1
	sum_day = time_month_size(time_month)

	do while(time_day_in_year > sum_day)
		time_month = time_month + 1
		sum_day = sum_day + time_month_size(time_month)
	end do

	! Day
	time_day = time_day_in_year - (sum_day - time_month_size(time_month))
	
	! Hour
	time_l_in_day = MOD(time_l_in_year, time_steps_day)
	time_hour = INT(time_l_in_day * 24./time_steps_day)
!-------	
!	add Minutes
!-------
	time_min = time_l_in_day * (24.*60)/time_steps_day - time_hour*60

	! Now add time_start_* values
	time_day = time_start_day + time_day - 1
	if(time_day > time_month_size(time_month)) then
		time_day = MOD(time_day, time_month_size(time_month))
		time_month = time_month + 1
	end if
	
	time_month = time_start_month + time_month - 1
	if(time_month > 12) then
		time_month = MOD(time_month, 12)
		time_year = time_year + 1
	end if	
		
	time_year = time_start_year + time_year
	
	! Make model date and time string
	time_model_date = TRIM(I2S(time_day,2))//"."//TRIM(I2S(time_month,2))//"."//TRIM(I2S(time_year))//" - "//TRIM(I2S(time_hour,2))//":"//"00:00"
		
	end subroutine shared_timer

!=======================================================================================================================================================================================
!========================================
! Calculate time by current global minute
!========================================
	
	subroutine shared_timer_min(cur_min)

	real cur_min
	integer sum_day

	! Year
	time_year = NINT(cur_min)/(60*24*365)
	time_day_in_year = MOD(NINT(cur_min)/(60*24), 365) + 1

	! Month
	time_month = 1
	do while(SUM(time_month_size(1:time_month)) < time_day_in_year)
		time_month = time_month + 1
	end do
	
	! Day
	time_day = time_day_in_year - (SUM(time_month_size(1:time_month)) - time_month_size(time_month))
	
	! Hour
	time_hour = MOD(NINT(cur_min)/60, 24)
	
	! Min
	time_min = MOD(NINT(cur_min), 60)
	
	! Now add time_start_* values
	time_day = time_start_day + time_day - 1
	if(time_day > time_month_size(time_month)) then
		time_day = MOD(time_day, time_month_size(time_month))
		time_month = time_month + 1
	end if
	
	time_month = time_start_month + time_month - 1
	if(time_month > 12) then
		time_month = MOD(time_month, 12)
		time_year = time_year + 1
	end if	
		
	time_year = time_start_year + time_year

	! Make model date and time string
	time_model_date = TRIM(I2S(time_day,2))//"."//TRIM(I2S(time_month,2))//"."//TRIM(I2S(time_year))//" - "//TRIM(I2S(time_hour,2))//":"//TRIM(I2S(time_min,2))//":00"
	
	end subroutine shared_timer_min
	
!=======================================================================================================================================================================================

	! Logical function for belonging to groups
	logical function proc_in(test_rank, test_comp)
	integer, intent(in) :: test_rank, test_comp

	if(test_rank >= SUM(run_np(COUPLER:test_comp))-run_np(test_comp) .AND. test_rank < SUM(run_np(COUPLER:test_comp))) then
		proc_in = .TRUE.
	else
		proc_in = .FALSE.
	end if

	end function proc_in

	! Function of defining name of group
	integer function proc_where()
	integer c

	proc_where = MYNULL
	do c = 0, NUM_OF_COMP
		if(proc_in(rank_world, c)) then
			proc_where = c
			EXIT
		end if
	end do

	end function proc_where
	
	! Rank of first proc in component
	integer function proc_first(test_comp)
	integer, intent(in) :: test_comp
	integer :: c
	
	do c = 0, size_world-1
		if(proc_in(c, test_comp)) then
			proc_first = c
			EXIT
		end if
	end do
	
	end function proc_first
	

!===================================================================================================================================================================================================

!=====================================================================
! Print error and stop program. Only 1 kernel print his error message.
! This is controlled by shared resource 'err_flag' file.
!=====================================================================

	subroutine shared_error(err_message)
	
	! Input variables
	character(LEN = *) err_message
	
	! Work variables
	integer cur_time_arr(8)
	
	if((run_phase == 'SHR' .AND. rank_world == 0) .OR. (run_phase == 'INI' .AND. rank_local == 0) .OR. (run_phase == 'RUN')) then

		call DATE_AND_TIME(VALUES = cur_time_arr)
	      
		if(run_log_file_num /= 6) open(run_log_file_num, file = '_log_file.txt')
		
		write(run_log_file_num, '(/1x,a18,i2.2,a1,i2.2,a1,i2.2)') 'Real time of run: ', cur_time_arr(5),':',cur_time_arr(6),':',cur_time_arr(7)
		write(run_log_file_num, '(1x,a18,i2.2,a1,i2.2,a1,i4.4/)') 'Real date of run: ', cur_time_arr(3),'.',cur_time_arr(2),'.',cur_time_arr(1)

		write(run_log_file_num, '(1x,A)') 'Error message: '//err_message	
		write(run_log_file_num, '(/1x,A/)') 'Terminate task...'
		
		if(run_log_file_num /= 6) close(run_log_file_num)
		
		call sleep(1)
		stop
	end if

	end subroutine shared_error
	
!====================================================================================================================================================================================================

!=========================================================
! Analize system file '/proc/self/status' for memory usage
!=========================================================

	subroutine shared_memory_usage(compName, strMessage)

	use shared_utils_module

	implicit none
	
	character(*) compName
	character(*), optional :: strMessage
	character(100) vmPeak, vmSize
	integer i

	open(111, file = '/proc/self/status')
	
	write(*, '(a/a/a/)') 'Virtual memory characteristics of '//TRIM(compName)//' kernel '//TRIM(I2S(rank_world))//' '//strMessage//': '	
	do i = 1,21
		read(111, '(a100)') vmSize
		write(*, '(a100)') vmSize
	end do
	write(*,*)
	
	close(111)
	
	
! 	vmPeak = ''
! 	do while(INDEX(vmPeak,'VmPeak') == 0)
! 		read(111,'(a100)') vmPeak
! 	end do
! 	
! 	vmSize = ''
! 	do while(INDEX(vmSize,'VmSize') == 0)
! 		read(111,'(a100)') vmSize
! 	end do
! 	
! 	close(111)
! 
! 	write(run_log_file_num, '(a/a/a/)') 'Virtual memory characteristics of '//TRIM(compName)//' kernel '//TRIM(I2S(rank_world))//' '//strMessage//': ', TRIM(vmPeak), TRIM(vmSize)

	end subroutine shared_memory_usage

!====================================================================================================================================================================================================
	
	end module shared_module

