
module array_info_module

use const_module
use utils_array_module
implicit none

type ArrayInfo
private
    character(NAME_LEN) :: name_ = ""
    character(NAME_LEN) :: lname_ = ""
    character(NAME_LEN) :: units_ = ""
    character(NAME_LEN) :: type_ = ""
    character(NAME_LEN) :: arr_dims_ = ""
    character(3) :: indx_ = ""
    
    ! Pointers to model data, unknown indexing
    real(kind=4), pointer :: p_2D4_(:,:) => NULL()
    real(kind=8), pointer :: p_2D8_(:,:) => NULL()    
    real(kind=4), pointer :: p_3D4_(:,:,:) => NULL()
    real(kind=8), pointer :: p_3D8_(:,:,:) => NULL()

    integer :: dims_(3)
 contains 
 
    ! === PUBLIC API ===
    generic :: init => init_ai_2D_real4, init_ai_2D_real8, init_ai_3D_real4, init_ai_3D_real8
    
    generic, public :: set_data => set_data_3D_real4, set_data_3D_real8, set_data_2D_real4, set_data_2D_real8
    
    procedure, public :: arr_name
    procedure, public :: arr_type           
    procedure, public :: arr_dims           
    
    procedure, public :: p_2D4
    procedure, public :: p_2D8
    procedure, public :: p_3D4
    procedure, public :: p_3D8
    ! === PUBLIC API ===
    
    procedure, private :: init_ai_2D_real4
    procedure, private :: init_ai_2D_real8
    procedure, private :: init_ai_3D_real4
    procedure, private :: init_ai_3D_real8
    
    procedure, private :: set_data_3D_real4
    procedure, private :: set_data_3D_real8
    procedure, private :: set_data_2D_real4
    procedure, private :: set_data_2D_real8
    
 end type

 CONTAINS

!======================================================================================== 

#define _INIT_2D_ARR_INFO_(_FUNC_, _TYPE_, _DIMS_, _MODEL_TYPE_, _MODEL_DIMS_, _POINTER_) \
subroutine _FUNC_(this, name, lname, units, arr, indx); \
    class(ArrayInfo) :: this; \
    character(*), intent(in) :: name, lname, units; \
    character(*), intent(in), optional :: indx; \
    _TYPE_, _DIMS_, intent(in), target :: arr; \
    this % name_ = name; \
    this % lname_ = lname; \
    this % units_ = units; \
    this % indx_ = "ij"; \
    this % type_ = _MODEL_TYPE_; \
    this % arr_dims_ = _MODEL_DIMS_; \
    this % _POINTER_ => arr; \
end subroutine

_INIT_2D_ARR_INFO_(init_ai_2D_real4,   real(4),    dimension(:,:), "real4", "2D", p_2D4_)
_INIT_2D_ARR_INFO_(init_ai_2D_real8,   real(8),    dimension(:,:), "real8", "2D", p_2D8_)

#define _INIT_3D_ARR_INFO_(_FUNC_, _TYPE_, _DIMS_, _MODEL_TYPE_, _MODEL_DIMS_, _POINTER_) \
subroutine _FUNC_(this, name, lname, units, arr, indx); \
    class(ArrayInfo) :: this; \
    character(*), intent(in) :: name, lname, units; \
    character(*), intent(in), optional :: indx; \
    _TYPE_, _DIMS_, intent(in), target :: arr; \
    this % name_ = name; \
    this % lname_ = lname; \
    this % units_ = units; \
    this % type_ = _MODEL_TYPE_; \
    this % arr_dims_ = _MODEL_DIMS_; \
    this % _POINTER_ => arr; \
    this % indx_ = MERGE(indx, "ijk", PRESENT(indx)); \
    select case (this % indx_); \
        case("ijk"); \
            this % dims_ = (/ SIZE(arr, 1), SIZE(arr, 2), SIZE(arr, 3) /); \
        case("kij"); \
            this % dims_ = (/ SIZE(arr, 2), SIZE(arr, 3), SIZE(arr, 1) /); \
        case default; \
            write(*,*) "ArrayInfo: unsupported indexing <"//indx//"> of array <"//name//">."; \
            call exit (1); \
    end select; \
end subroutine

_INIT_3D_ARR_INFO_(init_ai_3D_real4,   real(4),    dimension(:,:,:), "real4", "3D", p_3D4_)
_INIT_3D_ARR_INFO_(init_ai_3D_real8,   real(8),    dimension(:,:,:), "real8", "3D", p_3D8_)

!======================================================================================== 

#define _CONVERT_TO_MODEL_INDX_3D_(_FUNC_, _TYPE_, _DIMS_, _GOAL_) \
subroutine _FUNC_(this, adapter); \
    class(ArrayInfo) :: this; \
    _TYPE_, _DIMS_, intent(in) :: adapter; \
    if (this % indx_ == "kij") _GOAL_ = indexing_ijk_to_kij(adapter); \
    if (this % indx_ == "ijk") _GOAL_ = adapter; \
end subroutine

 _CONVERT_TO_MODEL_INDX_3D_(set_data_3D_real4, real(4), dimension(:,:,:), this % p_3D4_)
 _CONVERT_TO_MODEL_INDX_3D_(set_data_3D_real8, real(8), dimension(:,:,:), this % p_3D8_)

 
#define _CONVERT_TO_MODEL_INDX_2D_(_FUNC_, _TYPE_, _DIMS_, _GOAL_) \
subroutine _FUNC_(this, adapter); \
    class(ArrayInfo) :: this; \
    _TYPE_, _DIMS_, intent(in) :: adapter; \
    if (this % indx_ == "ij") _GOAL_ = adapter; \
end subroutine
 
 _CONVERT_TO_MODEL_INDX_2D_(set_data_2D_real4, real(4), dimension(:,:), this % p_2D4_)
 _CONVERT_TO_MODEL_INDX_2D_(set_data_2D_real8, real(8), dimension(:,:), this % p_2D8_)

!======================================================================================== 

function p_2D4(this)
    class(ArrayInfo) :: this
    real(kind=4), dimension(:,:), pointer :: p_2D4
    p_2D4 => this % p_2D4_
end function

function p_2D8(this)
    class(ArrayInfo) :: this
    real(kind=8), dimension(:,:), pointer :: p_2D8
    p_2D8 => this % p_2D8_
end function

function p_3D4(this)
    class(ArrayInfo) :: this
    real(kind=4), dimension(this % dims_(1), this % dims_(2), this % dims_(3)) :: p_3D4
    if (this % indx_ == "kij") p_3D4 = indexing_kij_to_ijk(this % p_3D4_)
    if (this % indx_ == "ijk") p_3D4 = this % p_3D4_
end function

function p_3D8(this)
    class(ArrayInfo) :: this
    real(kind=8), dimension(this % dims_(1), this % dims_(2), this % dims_(3)) :: p_3D8    
    if (this % indx_ == "kij") p_3D8 = indexing_kij_to_ijk(this % p_3D8_)
    if (this % indx_ == "ijk") p_3D8 = this % p_3D8_
end function

character(NAME_LEN) function arr_name(this)
    class(ArrayInfo) :: this
    arr_name = this % name_
end function

character(NAME_LEN) function arr_type(this)
    class(ArrayInfo) :: this
    arr_type = this % type_
end function

character(NAME_LEN) function arr_dims(this)
    class(ArrayInfo) :: this
    arr_dims = this % arr_dims_
end function

!======================================================================================== 

end module

!======================================================================================== 

#define CONCRETE_TYPE ArrayInfo
#define CONCRETE_MODULE array_info_module
#define CONCRETE_MODULE_FLIBS array_info_hash_module
#define OUTPUT_MODULE array_info_storage_module
#include "HashStorage.f90"

!======================================================================================== 



   