!
! Communicator datatype provides packing/unpacking metainfo and actual
! array data and send/recv.
!

!========================================================================================      
! Magic for enable of use hash-table of GA-ids
!
!=======================================================================================
! Define set GA interface i4 or i8 ( recomend i8 for x64 OS)
!========================================================================================   
#define GA_IS 8     ! GA Integer Size 8 or 4

module global_array_info_module

use const_module
implicit none

type GlobalArrayInfo
    integer :: id = -1    
    character(SHORT_NAME_LEN) :: arr_type = "null" 
    character(SHORT_NAME_LEN) :: arr_dimnum = "null" 
end type    
end module

#define CONCRETE_TYPE GlobalArrayInfo
#define CONCRETE_MODULE global_array_info_module
#define CONCRETE_MODULE_FLIBS global_array_info_hash_module
#define OUTPUT_MODULE global_array_info_storage_module
#include "HashStorage.f90"

!========================================================================================      

module ga_includes_module
#include "mafdecls.fh"
#include "global.fh"
end module

module communicator_ga_module

use global_array_info_module
use decomposition_module
use ga_includes_module
use global_array_info_storage_module, HashStorageGa => HashStorage
use debug_module
use component_splitter_module
implicit none

integer, parameter :: COMM_GA_STATUS = 1
integer, parameter :: COMM_GA_TIME = 2
integer, parameter :: COMM_GA_OPT = 3

integer, parameter, private :: MEMORY_STACK_SIZE = 100*(10**6), MEMORY_HEAP_SIZE = 100*(10**6)

type :: Communicator

private
    integer, allocatable :: pgroups_(:,:)
    type(HashStorageGa) :: ga_storage_                        ! internal, use to store ga_id
           
 contains
    !========================
    ! ====== PUBLIC API =====
    procedure, public :: init => init_communicator          ! (max_index) communicator constructor
    procedure, public :: destroy => destroy_communicator    ! destructor
   
    procedure, public :: init_group                         ! initialize group for 2 components
                                                            ! initialize global array for 1 or 2 components
    generic :: init_array => init_array_decomp, init_array_dims
    procedure, public :: destroy_array                      ! destroy GA array
    
                                                            ! put data to global aray
    generic, public :: put => put_ga_1D_int, put_ga_1D_int8, put_ga_1D_real4, put_ga_1D_real8, put_ga_2D_int, put_ga_2D_int8, put_ga_2D_real4, put_ga_2D_real8, put_ga_3D_real4, put_ga_3D_real8
    generic, public :: acc => acc_ga_1D_int, acc_ga_1D_int8, acc_ga_1D_real4, acc_ga_1D_real8, acc_ga_2D_int, acc_ga_2D_int8, acc_ga_2D_real4, acc_ga_2D_real8, acc_ga_3D_real4, acc_ga_3D_real8
                                                            ! get data from global array
    generic, public :: get => get_ga_1D_int, get_ga_1D_int8, get_ga_1D_real4, get_ga_1D_real8, get_ga_2D_int, get_ga_2D_int8, get_ga_2D_real4, get_ga_2D_real8, get_ga_3D_real4, get_ga_3D_real8
    procedure, public :: sum_arrays                         ! Sum 2 GA
    procedure, public :: spd_invert                         ! SPD Invert GA (n ^ 2 ops)
    procedure, public :: dgemm_arrays                       ! mult matrix * matrix
    procedure, public :: print_arr                          ! print GA  for Debug
    procedure, public :: zero_arr                           ! zero GA
    procedure, public :: sync                               ! barrier for processor group
    procedure, public :: set_group                          ! set default processor group
    procedure, public :: is_registered                      ! is GA registered?
    procedure, public :: id                                 ! id of registered array
    procedure, public :: put_info
    procedure, public :: get_info
    
    ! ====== PUBLIC API =====    
    !========================

    procedure, private :: convert_datatype
    procedure, private :: init_array_decomp
    procedure, private :: init_array_dims
    procedure, private :: init_info_array
    procedure, private :: comm_err
    
    procedure, private :: put_ga_1D_int
    procedure, private :: put_ga_1D_int8
    procedure, private :: put_ga_1D_real4
    procedure, private :: put_ga_1D_real8
    procedure, private :: put_ga_2D_int
    procedure, private :: put_ga_2D_int8
    procedure, private :: put_ga_2D_real4
    procedure, private :: put_ga_2D_real8
    procedure, private :: put_ga_3D_real4
    procedure, private :: put_ga_3D_real8
    
    procedure, private :: acc_ga_1D_int
    procedure, private :: acc_ga_1D_int8
    procedure, private :: acc_ga_1D_real4
    procedure, private :: acc_ga_1D_real8
    procedure, private :: acc_ga_2D_int
    procedure, private :: acc_ga_2D_int8
    procedure, private :: acc_ga_2D_real4
    procedure, private :: acc_ga_2D_real8
    procedure, private :: acc_ga_3D_real4
    procedure, private :: acc_ga_3D_real8

    procedure, private :: get_ga_1D_int
    procedure, private :: get_ga_1D_int8
    procedure, private :: get_ga_1D_real4
    procedure, private :: get_ga_1D_real8
    procedure, private :: get_ga_2D_int
    procedure, private :: get_ga_2D_int8
    procedure, private :: get_ga_2D_real4
    procedure, private :: get_ga_2D_real8
    procedure, private :: get_ga_3D_real4
    procedure, private :: get_ga_3D_real8  
end type
!========================================================================================      

type(Communicator) :: Communicator_GA

!========================================================================================      

 CONTAINS

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      construct object for <max_index> components
! Parameters:       
!   max_index     - number of sides to communicate
!   proc_local_count - size of local communicator, required for agile memory allocation
!----------------------------------------------------------------------------------------

subroutine init_communicator(this, max_index, proc_local_count) 
    class(Communicator) :: this
    integer, intent(in) :: max_index, proc_local_count
    integer :: status

    call this % ga_storage_ % init() 

    call GA_INITIALIZE()
    
    call CompSplitter % show_memory_usage("after GA_INIT", 0)
    
    if (.NOT. MA_INIT(  int( MT_F_REAL, GA_IS), int( MEMORY_STACK_SIZE/4/proc_local_count, GA_IS), &
                        int( MEMORY_HEAP_SIZE/4/proc_local_count, GA_IS) ) ) then
        write(*,*) "Communicator: cant init enough memory."
        call exit(1)
    end if

    
    call CompSplitter % show_memory_usage("after MA_INIT", 0)
    
    allocate(this % pgroups_(max_index, max_index)); this % pgroups_ = 0
end subroutine

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      destruct object
! Parameters:       
!----------------------------------------------------------------------------------------

subroutine destroy_communicator(this)
    class (Communicator) :: this

    call this % ga_storage_ % destroy() 
    deallocate(this % pgroups_)
    call GA_PGROUP_SET_DEFAULT( int( GA_PGROUP_GET_WORLD(), GA_IS))
    call GA_TERMINATE()          
end subroutine

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      register processor group between two sides
! Parameters:       
!   src_id, dst_id           - ids of sides
!   src_ranks, dst_ranks     - ranks of all processes of sides 
!----------------------------------------------------------------------------------------

subroutine init_group(this, src_id, src_ranks, dst_id, dst_ranks)
    class(Communicator) :: this
    integer, intent(in) :: src_id, src_ranks(:) 
    integer, intent(in), optional :: dst_id, dst_ranks(:)

!     if (CompSplitter % rank_local() == 0) then
!     
! !         call Debugger % log_msg("Make group between", CompSplitter % comp_name(src_id)//" and "//CompSplitter % comp_name(dst_id))
!         call Debugger % log_msg("src_ranks", src_ranks)
!         if (PRESENT(dst_ranks) .AND. src_id /= dst_id) then
!             call Debugger % log_msg("dst_ranks", dst_ranks)
!         else
!             call Debugger % log_msg("no dst")
!         end if
!     end if

    call GA_PGROUP_SET_DEFAULT( int( GA_PGROUP_GET_WORLD(), GA_IS) )

!     if (CompSplitter % i_am_id() /= src_id .AND. (PRESENT(dst_id) .AND. CompSplitter % i_am_id() /= dst_id)) return
    
    if (PRESENT(dst_id) .AND. src_id /= dst_id) then
        this % pgroups_(src_id, dst_id) = int( GA_PGROUP_CREATE( int( (/ src_ranks, dst_ranks /), GA_IS), &
                                                            int( SIZE(src_ranks) + SIZE(dst_ranks), GA_IS) ), GA_IS)
    else
        this % pgroups_(src_id, src_id) = int( GA_PGROUP_CREATE( int( src_ranks, GA_IS), int( SIZE(src_ranks), GA_IS) ), GA_IS)
    end if
 
    if (CompSplitter % rank_local() == 0) then
        call Debugger % log_msg("Group created")
    end if
    

end subroutine

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      initialize array based on Decomposition object, it can be accessed from <holder_id>
!                   and <subscriber_id> components, but stored on <holder_id>. If array with such name already
!                   exists - delete previous and create new.
! Parameters:       
!   arr_name        - string name of array
!   datatype        - supported datatype string: "real4", real8", "int4"
!   dimnum          - supported dimension string: "2D", "3D"
!   holder_id       - id of source component who hold array in memory
!   holder_decomp   - decomposition of holder side
!   subscriber_id   - id of subscriber component
!----------------------------------------------------------------------------------------
subroutine init_array_decomp(this, arr_name, datatype, dimnum, holder_id, holder_decomp, subscriber_id)

    class(Communicator) :: this
    character(*), intent(in) :: arr_name, datatype, dimnum
    type(Decomposition), intent(in) :: holder_decomp
    integer, intent(in) :: holder_id    
    integer, intent(in), optional :: subscriber_id
    type(GlobalArrayInfo) :: node
    integer :: ga_id, status
    integer :: subscriber_id_
        
    if (PRESENT(subscriber_id)) then
        subscriber_id_ = subscriber_id
    else
        subscriber_id_ = holder_id
    end if
    
    ! If already have such array, destroy it first
    if (this % is_registered(arr_name)) call this % destroy_array(arr_name)

    !if (CompSplitter % i_am_id() /= holder_id .AND. CompSplitter % i_am_id() /= subscriber_id_) return
    
    ! Synchronize
    call this % set_group(holder_id, subscriber_id_)
    call this % sync(holder_id, subscriber_id_)
    
    ga_id = int( GA_CREATE_HANDLE(), GA_IS)
    call GA_SET_ARRAY_NAME( int( ga_id, GA_IS), arr_name)
    call GA_SET_RESTRICTED_RANGE( int( ga_id, GA_IS), int( 0, GA_IS), int( holder_decomp % proc_count() - 1, GA_IS) )

    if (dimnum == "2D") then
        call GA_SET_DATA(   int( ga_id, GA_IS), int( 2, GA_IS) , &
                            int( (/ holder_decomp % i_size(), holder_decomp % j_size() /), GA_IS), &
                            int( this % convert_datatype(datatype), GA_IS))
        call GA_SET_IRREG_DISTR(    int( ga_id, GA_IS),                                                  &
                                    int( holder_decomp % to_cartesian_ij(), GA_IS),                     &
                                    int( (/ holder_decomp % inp(), holder_decomp % jnp() /), GA_IS) )
    else if (dimnum == "3D") then
        call GA_SET_DATA(   int( ga_id, GA_IS), int( 3, GA_IS), &
                            int( (/ holder_decomp % i_size(), holder_decomp % j_size(), holder_decomp % k_size() /), GA_IS), &
                            int( this % convert_datatype(datatype), GA_IS) )
        call GA_SET_IRREG_DISTR(    int( ga_id, GA_IS),                                                  &
                                    int( holder_decomp % to_cartesian_ijk(), GA_IS),                     &
                                    int( (/ holder_decomp % inp(), holder_decomp % jnp(), 1/), GA_IS) )
    else 
        write(*,*) "Communicator: bad dimension size:", dimnum
        call exit(1)
    end if 

    status = GA_ALLOCATE( int( ga_id, GA_IS) )
!     if(GA_NODEID() == 0) call GA_SUMMARIZE(ga_id)

    ! Store info in hash-table
    node % id = ga_id
    node % arr_dimnum = dimnum
    node % arr_type = datatype
    call this % ga_storage_ % put(arr_name, node)

    ! Mark array as free    
    call this % init_info_array(arr_name)
    if(CompSplitter % is_first_rank()) call this % put_info(arr_name, COMM_GA_STATUS, "free")
    ! Synchronize
    call this % sync(holder_id, subscriber_id_)    
    
!     call Debugger % log_msg("This is registration call of :"//TRIM(arr_name))
        
end subroutine

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      initialize array based on dimension sizes, it can be accessed from <holder_id>
!                   and <subscriber_id> components, but stored on <holder_id>. If array with such name already
!                   exists - delete previous and create new.
! Parameters:       
!   arr_name        - string name of array
!   datatype        - supported datatype string: "real4", real8", "int4"
!   dimnum          - supported dimension string: "2D", "3D"
!   dim*_len,       - size of each dimension
!   holder_id       - id of source component who hold array in memory
!   holder_size     - how many processors owns the GA
!   subscriber_id   - id of subscriber component
!----------------------------------------------------------------------------------------

subroutine init_array_dims(this, arr_name, datatype, dim1_len, dim2_len, dim3_len, holder_id, holder_size, subscriber_id)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name, datatype
    integer, intent(in), optional :: dim1_len, dim2_len, dim3_len
    integer, intent(in) :: holder_id, holder_size    
    integer, intent(in), optional :: subscriber_id
    type(GlobalArrayInfo) :: node
    integer :: ga_id, status
    integer :: subscriber_id_
    integer :: dims(3), num_dims
    
    num_dims = 0
    if (PRESENT(dim1_len)) then
        num_dims = num_dims + 1
        dims(num_dims) = dim1_len
    end if
    if (PRESENT(dim2_len)) then
        num_dims = num_dims + 1
        dims(num_dims) = dim2_len
    end if
    if (PRESENT(dim3_len)) then
        num_dims = num_dims + 1
        dims(num_dims) = dim3_len
    end if
    
    if (PRESENT(subscriber_id)) then
        subscriber_id_ = subscriber_id
    else
        subscriber_id_ = holder_id
    end if
    
    ! If already have such array, destroy it first
    if (this % is_registered(arr_name)) call this % destroy_array(arr_name)

    !if (CompSplitter % i_am_id() /= holder_id .AND. CompSplitter % i_am_id() /= subscriber_id_) return
    
    ! Synchronize    
    call this % set_group(holder_id, subscriber_id_)    
    call this % sync(holder_id, subscriber_id_)
        
    ga_id = int( GA_CREATE_HANDLE(), GA_IS)
    call GA_SET_ARRAY_NAME(         int( ga_id, GA_IS), arr_name)
    call GA_SET_RESTRICTED_RANGE(   int( ga_id, GA_IS), int( 0, GA_IS), int( holder_size-1, GA_IS))
    call GA_SET_DATA(   int( ga_id, GA_IS), int( num_dims, GA_IS), &
                        int( dims(1:num_dims), GA_IS), int( this % convert_datatype(datatype), GA_IS) )
    status = GA_ALLOCATE( int(ga_id, GA_IS) )
!     if(GA_NODEID() == 0) call GA_SUMMARIZE( int( ga_id, GA_IS))

    ! Store info in hash-table
    node % id = ga_id
    node % arr_dimnum = TRIM(n2s(num_dims))//"D"
    node % arr_type = datatype
    call this % ga_storage_ % put(arr_name, node)
        
    ! Mark array as free    
    call this % init_info_array(arr_name)
    call this % put_info(arr_name, COMM_GA_STATUS, "free")
    ! Synchronize
    call this % sync(holder_id, subscriber_id_)    
        
        
end subroutine

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      destroy global array (you should set appropriate group before this call)
! Parameters:       
!   arr_name        - string name of array
!----------------------------------------------------------------------------------------
subroutine destroy_array(this, arr_name)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name
    type(GlobalArrayInfo) :: node
    logical :: status

    node = this % ga_storage_ % get(TRIM(arr_name))    
    status = GA_DESTROY( int( node % id, GA_IS) )

end subroutine

!========================================================================================      

subroutine init_info_array(this, arr_name)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name
    type(GlobalArrayInfo) :: node    
    integer :: ga_id, status
    
    ga_id = GA_CREATE_HANDLE()
    call GA_SET_ARRAY_NAME( int( ga_id, GA_IS), TRIM(arr_name)//"_info")
    call GA_SET_DATA( int( ga_id, GA_IS), int( 2, GA_IS), int( (/ 100, 100 /), GA_IS), int( MT_F_INT, GA_IS) )
    
    status = GA_ALLOCATE( int( ga_id, GA_IS) )

    ! Store info in hash-table
    node % id = ga_id
    call this % ga_storage_ % put(TRIM(arr_name)//"_info", node)
    
!     if(GA_NODEID() == 0) call GA_SUMMARIZE(ga_id)    
end subroutine

!========================================================================================      

subroutine put_info(this, arr_name, field, value)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name, value
    integer, intent(in) :: field
    character(NAME_LEN) :: value_
    
    value_ = value
    call NGA_PUT(   int( this % id(TRIM(arr_name)//"_info"), GA_IS),&
                    int( (/ field, 1 /), GA_IS), &
                    int( (/ field, NAME_LEN/4 + 1 /), GA_IS), &
                    value_, int( (/ 1 /), GA_IS) )
 
#ifdef DEBUG
    call Debugger % log_msg("Put ga-info status of array <"//TRIM(arr_name)//"> :"//value_)
#endif 
end subroutine

!========================================================================================      

character(NAME_LEN) function get_info(this, arr_name, field)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name
    integer, intent(in) :: field    
    
    call NGA_GET(   int( this % id(TRIM(arr_name)//"_info"), GA_IS), int( (/ field, 1 /), GA_IS), &
                    int( (/ field, NAME_LEN/4 + 1 /), GA_IS), get_info, int( (/ 1 /), GA_IS) )

#ifdef DEBUG
    call Debugger % log_msg("Get ga-info status of array <"//TRIM(arr_name)//"> :"//get_info)
#endif
    
end function    

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      barrier for src_id, dst_id
! Parameters:       
!    src_id, dst_id - indexes of groups
!----------------------------------------------------------------------------------------
subroutine sync(this, src_id, dst_id)
    class(Communicator) :: this
    integer, intent(in) :: src_id
    integer, intent(in), optional :: dst_id 
    integer :: dst_id_
    
    dst_id_ = MERGE(dst_id, src_id, PRESENT(dst_id))
    
    if (this % pgroups_(src_id, dst_id_) == 0) then
        write(*,*) "Communicator GA: no group for id pair ("//TRIM(n2s(src_id))//","//TRIM(n2s(dst_id_))//")"
        call exit(1)
    end if
    
    call this % set_group(src_id, dst_id_)
    call GA_SYNC()
end subroutine

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      set default group to src_id, dst_id
! Parameters:       
!    src_id, dst_id - indexes of groups
!----------------------------------------------------------------------------------------

subroutine set_group(this, src_id, dst_id)
    class(Communicator) :: this
    integer, intent(in) :: src_id, dst_id
    
    if (this % pgroups_(src_id, dst_id) == 0) then
        write(*,*) "Communicator GA: no group for id pair ("//TRIM(n2s(src_id))//","//TRIM(n2s(dst_id))//")"
        call exit(1)
    end if
    
    call GA_PGROUP_SET_DEFAULT( int( this % pgroups_(src_id, dst_id), GA_IS) )
end subroutine
    
!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      is array name is registered in hash table?
! Parameters:       
!   arr_name        - string name of array 
!----------------------------------------------------------------------------------------
logical function is_registered(this, arr_name)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name    
    is_registered = this % id(arr_name) /= -1  
end function

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:      return ga_id of array with given name. Return -1 if no such array.
! Parameters:       
!   arr_name        - string name of array 
!----------------------------------------------------------------------------------------

integer function id(this, arr_name)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name
    type(GlobalArrayInfo) :: node

    node = this % ga_storage_ % get(TRIM(arr_name))    
    id = node % id
end function

!========================================================================================
! PRIVATE
!========================================================================================      


integer function convert_datatype(this, datatype)
    class(Communicator) :: this
    character(*), intent(in) :: datatype
        
    select case (datatype)
        case("real4")
            convert_datatype = MT_F_REAL
        case("real8")
            convert_datatype = MT_F_DBL
        case("int4")
            convert_datatype = MT_F_INT
        case("int8")
            convert_datatype = MT_F_INT
        case default
            write(*,*) "Communicator: unknown type: ", datatype
            call exit(1)
        end select
end function

!========================================================================================      

!----------------------------------------------------------------------------------------
! Description:      concrete implementations of put, acc and get operations
! Parameters:       
!   arr_name        - string name of array 
!   lo, hi          - arrays representing area (in global indexing) you want to put/get
!   arr             - your buffer for data
!---------------------------------------------------------------------------------------

#define _PUT_GA_(_FUNC_, _TYPE_, _DIMS_, _LD_)              \
subroutine _FUNC_ (this, arr_name, lo, hi, arr);            \
    class(Communicator) :: this;                            \
    character(*), intent(in) :: arr_name;                   \
    integer, intent(in) :: lo(:), hi(:);                    \
    _TYPE_, _DIMS_ :: arr;                                  \
    if (.NOT. this % is_registered(arr_name)) call this % comm_err("global array is not registered."); \
    call NGA_PUT( int( this % id(arr_name), GA_IS), int( lo, GA_IS), int( hi, GA_IS), arr, int( _LD_ , GA_IS) );  \
end subroutine

_PUT_GA_(put_ga_1D_int,    integer(4),    dimension(:),       (/ SIZE(arr, 1) /))
_PUT_GA_(put_ga_1D_int8,   integer(8), dimension(:),       (/ SIZE(arr, 1) /))
_PUT_GA_(put_ga_1D_real4,  real,       dimension(:),       (/ SIZE(arr, 1) /))
_PUT_GA_(put_ga_1D_real8,  real(8),    dimension(:),       (/ SIZE(arr, 1) /))
_PUT_GA_(put_ga_2D_int,    integer(4),    dimension(:,:),     (/ SIZE(arr, 1) /))
_PUT_GA_(put_ga_2D_int8,   integer(8), dimension(:,:),     (/ SIZE(arr, 1) /))
_PUT_GA_(put_ga_2D_real4,  real,       dimension(:,:),     (/ SIZE(arr, 1) /))
_PUT_GA_(put_ga_2D_real8,  real(8),    dimension(:,:),     (/ SIZE(arr, 1) /))
_PUT_GA_(put_ga_3D_real4,  real,       dimension(:,:,:),   (/ SIZE(arr, 1), SIZE(arr, 2) /))
_PUT_GA_(put_ga_3D_real8,  real(8),    dimension(:,:,:),   (/ SIZE(arr, 1), SIZE(arr, 2) /))

#define _ACC_GA_(_FUNC_, _TYPE_, _DIMS_, _LD_)              \
subroutine _FUNC_ (this, arr_name, lo, hi, arr, alpha);     \
    class(Communicator) :: this;                            \
    character(*), intent(in) :: arr_name;                   \
    _TYPE_, intent(in) :: alpha;                            \
    integer, intent(in) :: lo(:), hi(:);                    \
    _TYPE_, _DIMS_ :: arr;                                  \
    if (.NOT. this % is_registered(arr_name)) call this % comm_err("global array is not registered."); \
    write(*,*), "alpha = ", alpha; \
    call NGA_ACC( int( this % id(arr_name), GA_IS), int( lo, GA_IS), int( hi, GA_IS), arr, int( _LD_ , GA_IS), alpha );  \
end subroutine

_ACC_GA_(acc_ga_1D_int,    integer(4), dimension(:),       (/ SIZE(arr, 1) /))
_ACC_GA_(acc_ga_1D_int8,   integer(8), dimension(:),       (/ SIZE(arr, 1) /))
_ACC_GA_(acc_ga_1D_real4,  real,       dimension(:),       (/ SIZE(arr, 1) /))
_ACC_GA_(acc_ga_1D_real8,  real(8),    dimension(:),       (/ SIZE(arr, 1) /))
_ACC_GA_(acc_ga_2D_int,    integer(4), dimension(:,:),     (/ SIZE(arr, 1) /))
_ACC_GA_(acc_ga_2D_int8,   integer(8), dimension(:,:),     (/ SIZE(arr, 1) /))
_ACC_GA_(acc_ga_2D_real4,  real,       dimension(:,:),     (/ SIZE(arr, 1) /))
_ACC_GA_(acc_ga_2D_real8,  real(8),    dimension(:,:),     (/ SIZE(arr, 1) /))
_ACC_GA_(acc_ga_3D_real4,  real,       dimension(:,:,:),   (/ SIZE(arr, 1), SIZE(arr, 2) /))
_ACC_GA_(acc_ga_3D_real8,  real(8),    dimension(:,:,:),   (/ SIZE(arr, 1), SIZE(arr, 2) /))

#define _GET_GA_(_FUNC_, _TYPE_, _DIMS_, _LD_)              \
subroutine _FUNC_ (this, arr_name, lo, hi, arr);            \
    class(Communicator) :: this;                            \
    character(*), intent(in) :: arr_name;                   \
    integer, intent(in) :: lo(:), hi(:);                    \
    _TYPE_, _DIMS_ :: arr;                                  \
    if (.NOT. this % is_registered(arr_name)) call this % comm_err("global array is not registered: "//arr_name); \
    call NGA_GET( int( this % id(arr_name), GA_IS), int( lo, GA_IS), int( hi, GA_IS), arr, int( _LD_ , GA_IS) );  \
end subroutine

_GET_GA_(get_ga_1D_int,    integer(4),    dimension(:),       (/ SIZE(arr, 1) /))
_GET_GA_(get_ga_1D_int8,   integer(8), dimension(:),       (/ SIZE(arr, 1) /))
_GET_GA_(get_ga_1D_real4,  real,       dimension(:),       (/ SIZE(arr, 1) /))
_GET_GA_(get_ga_1D_real8,  real(8),    dimension(:),       (/ SIZE(arr, 1) /))
_GET_GA_(get_ga_2D_int,    integer(4),    dimension(:,:),     (/ SIZE(arr, 1) /))
_GET_GA_(get_ga_2D_int8,   integer(8), dimension(:,:),     (/ SIZE(arr, 1) /))
_GET_GA_(get_ga_2D_real4,  real,       dimension(:,:),     (/ SIZE(arr, 1) /))
_GET_GA_(get_ga_2D_real8,  real(8),    dimension(:,:),     (/ SIZE(arr, 1) /))
_GET_GA_(get_ga_3D_real4,  real,       dimension(:,:,:),   (/ SIZE(arr, 1), SIZE(arr, 2) /))
_GET_GA_(get_ga_3D_real8,  real(8),    dimension(:,:,:),   (/ SIZE(arr, 1), SIZE(arr, 2) /))

!========================================================================================      

subroutine sum_arrays (this, alpha, arr_name1, beta, arr_name2 , arr_name)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name, arr_name1, arr_name2
    real*4, intent(in) :: alpha, beta
    if (.NOT. this % is_registered( arr_name1) )  call this % comm_err("global array is not registered.")
    if (.NOT. this % is_registered( arr_name2) )  call this % comm_err("global array is not registered.")
    if (.NOT. this % is_registered( arr_name ) )  call this % comm_err("global array is not registered.")
    call GA_add(    alpha,  int( this % id(arr_name1), GA_IS), &
                    beta,   int( this % id(arr_name2), GA_IS), &
                            int( this % id(arr_name) , GA_IS)   )
end subroutine

!========================================================================================

subroutine dgemm_arrays (this,ta,tb, alpha, beta, arr_name1, lo1 , hi1, arr_name2 , lo2, hi2, arr_name, lo, hi )
    class(Communicator) :: this
    character(*), intent(in) :: arr_name, arr_name1, arr_name2
    integer, intent(in) :: lo1(:), hi1(:), lo2(:), hi2(:), lo(:), hi(:)
    character(1), intent(in) :: ta, tb
    real*8, intent(in) :: alpha, beta
    if (.NOT. this % is_registered( arr_name1) )  call this % comm_err("global array is not registered.")
    if (.NOT. this % is_registered( arr_name2) )  call this % comm_err("global array is not registered.")
    if (.NOT. this % is_registered( arr_name ) )  call this % comm_err("global array is not registered.")

    call nga_matmul_patch(  ta, tb, alpha, beta, &
                            int( this % id(arr_name1), GA_IS), int( lo1, GA_IS), int( hi1, GA_IS), &
                            int( this % id(arr_name2), GA_IS), int( lo2, GA_IS), int( hi2, GA_IS), &
                            int( this % id(arr_name ), GA_IS), int( lo,  GA_IS) , int( hi,  GA_IS))

end subroutine

subroutine spd_invert (this,  arr_name)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name
    integer ierr
    
    call Debugger % log_msg("spd_invert")

    if (.NOT. this % is_registered( arr_name ) )   call this % comm_err("global array is not registered.")
    ierr = ga_spd_invert( int( this % id(arr_name), GA_IS)  )
    if ( ierr /= 0 ) call this % comm_err("spd invert problem")

    
end subroutine

subroutine print_arr (this,  arr_name)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name
    if (.NOT. this % is_registered( arr_name ) )   call this % comm_err("global array is not registered.")
    call ga_print( int( this % id(arr_name), GA_IS) )

end subroutine

subroutine zero_arr (this,  arr_name, lo , hi)
    class(Communicator) :: this
    character(*), intent(in) :: arr_name
    integer, intent(in) :: lo(:), hi(:)
    if (.NOT. this % is_registered( arr_name ) )   call this % comm_err("global array is not registered.")
    call nga_zero_patch( int( this % id(arr_name), GA_IS), int( lo, GA_IS), int( hi, GA_IS) )

end subroutine


subroutine comm_err(this, msg)
    class(Communicator) :: this
    character(*), intent(in) :: msg
    write(*,*) "Communicator: "//msg
    call exit(1)

end subroutine

end module

