module component_splitter_module

use const_module
use utils_module
use mpi
implicit none

character(3), parameter :: COMPONENT_NAMES(10) = (/ "DTR", "CPL", "IOD", "OCN", "ATM", "ICE", "LND", "SEA", "AOI","NST" /)
integer, parameter :: NUMBER_OF_COMPONENTS = SIZE(COMPONENT_NAMES)
character(3), parameter :: SERVICE_COMPS(5) = (/ "DTR", "CPL", "IOD", "AOI", "NST" /)

type, private :: ComponentSplitter
private
    integer :: size_world_ = 0, rank_world_ = -1
    integer :: size_local_ = 0, rank_local_ = -1, comm_local_ = -1
    integer :: i_am_id_ = -1
    integer :: proc_num_(SIZE(COMPONENT_NAMES)) = 0
    integer :: comp_launched_ = -1
    character(3) :: run_phase_
    
 contains
 
    !=======================
    ! ===== PUBLIC API =====
    procedure, public :: init => init_splitter          ! Constructor. Read command line arguments (e.g. CPL 1 OCN 1) and split comm
    procedure, public :: destroy => destroy_splitter    ! Destructor. Finalize MPI

    procedure, public :: rank_world                     ! Rank of process in MPI_COMM_WORLD
    procedure, public :: rank_local                     ! Rank of process in comm_local
    procedure, public :: comm_world                     ! Global communicator (MPI_COMM_WORLD)
    procedure, public :: comm_world_size                ! Total number of launched processes
    procedure, public :: comm_local                     ! Local component communicator
    procedure, public :: comm_local_size                ! Number of processes per component
    procedure, public :: proc_first                     ! First process in comm_local
    procedure, public :: proc_list                      ! List of all processes in comm_local
    
    procedure, public :: i_am                           ! Component name of this process (e.g. "OCN")
    procedure, public :: i_am_id                        ! Component id of this process (e.g. OCN)
    procedure, public :: is_first_rank                  ! is this rank first in communicator
    
    procedure, public :: is_launched                    ! Is this component launched?
    procedure, public :: comp_id => get_comp_id         ! Return integer id od component by it's name
    procedure, public :: comp_name => get_comp_name     ! Name of component
    
    procedure, public :: is_model                       ! Is this component model or service?
    procedure, public :: comp_launched                  ! Number of components launched
    procedure, public :: model_launched                 ! Number of models launched
    procedure, public :: comp_defined                   ! Number of components defined
    
    procedure, public :: sync_error                     ! (msg) : Print <msg> and exit (try to syncromize)
    procedure, public :: barrier                        ! (msg) : Set MPI_BARRIER and print <msg> after it
    
    procedure, public :: set_run_phase                  ! Set run phase (used by sync_error to understand phase of program)
    
    procedure, public :: show_split
    procedure, public :: show_memory_usage
    
    ! ===== PUBLIC API =====
    !=======================
    
    procedure, private :: proc_in
    procedure, private :: read_command_line_args    
end type

!=====================================================================
! Define global object
    type(ComponentSplitter) :: CompSplitter

 CONTAINS

!=====================================================================
    
subroutine init_splitter(this, run_comps)
    class(ComponentSplitter) :: this
    integer, intent(in) :: run_comps
    integer :: ierr
        
    call this % read_command_line_args(run_comps)
        
    call MPI_INIT (ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, this % size_world_, ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD, this % rank_world_, ierr)
    
    call MPI_COMM_SPLIT(MPI_COMM_WORLD, this % i_am_id(), this % rank_world_, this % comm_local_, ierr)

    call MPI_COMM_SIZE(this % comm_local_, this % size_local_, ierr)
    call MPI_COMM_RANK(this % comm_local_, this % rank_local_, ierr)

    if(SUM(this % proc_num_(:)) /= this % size_world_) call this % sync_error("Number of mpirun processes /= sum of all component processes.", "LCL")
                
end subroutine

subroutine destroy_splitter(this)
    class(ComponentSplitter) :: this
    integer :: ierr
    call MPI_BARRIER(MPI_COMM_WORLD, ierr)
    call MPI_FINALIZE(ierr)
end subroutine

!=====================================================================

subroutine show_split(this)
    class(ComponentSplitter) :: this
    integer :: c, p
    integer, allocatable :: procs(:)
    
    write(*,*) "Parallel decomposition of MPI_COMM_WORLD:"
    do c = 1, this % comp_defined()
        if (this % is_launched(this % comp_name(c))) then
            write(*,*) this % comp_name(c), this % comm_local_size(this % comp_name(c))
            write(*,*) "Processor list:"
            
            call this % proc_list(this % comp_name(c), list = procs)
            write(*,*) procs
            deallocate(procs)
        end if
    end do
    
end subroutine


subroutine sync_error(this, msg, mode)
    
    class(ComponentSplitter) :: this
    character(*), intent(in) :: msg
    character(3), intent(in) :: mode
    integer :: cur_time_arr(8)
    
    if((mode == 'GLB' .AND. this % rank_world_ == 0) .OR. &
       (mode == 'LCL' .AND. this % rank_local_ == 0) .OR. &
       (mode == 'IND')) then

        call DATE_AND_TIME(VALUES = cur_time_arr)
                  
        write(*, '(/1x,a18,i2.2,a1,i2.2,a1,i2.2)') 'Real time of run: ', cur_time_arr(5),':',cur_time_arr(6),':',cur_time_arr(7)
        write(*, '(1x,a18,i2.2,a1,i2.2,a1,i4.4/)') 'Real date of run: ', cur_time_arr(3),'.',cur_time_arr(2),'.',cur_time_arr(1)

        write(*, '(1x,A)') char(27)//'[31mError: '//char(27)//'[0m'//msg    
        write(*, '(/1x,A/)') 'Terminate task...'

!         call sleep(1)
        call exit(1)
    end if

end subroutine

!=====================================================================

subroutine barrier(this, msg, complete, mode)
    class(ComponentSplitter) :: this
    character(*), optional :: msg
    character(80) :: msg_
    logical, optional :: complete
    character(*), intent(in), optional :: mode
    integer :: ierr   
        
    if (PRESENT(mode) .AND. mode == "GLB" .OR. .NOT. PRESENT(mode)) then
        call MPI_BARRIER(MPI_COMM_WORLD, ierr)
        if (PRESENT(msg) .AND. this % rank_world_ == 0) then
            msg_ = msg
            write(*,'(A)') msg_
        end if    
        
    else if (PRESENT(mode) .AND. mode == "LCL") then
        call MPI_BARRIER(this % comm_local_, ierr)
        if (PRESENT(msg) .AND. this % is_first_rank()) then
            msg_ = msg
            write(*,'(A)') msg_
        end if    
        
    end if  
end subroutine
!=====================================================================

integer function rank_world(this)
    class(ComponentSplitter) :: this
    rank_world = this % rank_world_
end function

integer function rank_local(this)
    class(ComponentSplitter) :: this
    rank_local = this % rank_local_
end function

integer function comm_world(this)
    class(ComponentSplitter) :: this
    comm_world = MPI_COMM_WORLD
end function

integer function comm_world_size(this)
    class(ComponentSplitter) :: this
    integer :: ierr
    call MPI_COMM_SIZE(MPI_COMM_WORLD, comm_world_size, ierr)
end function

integer function comm_local(this)
    class(ComponentSplitter) :: this
    comm_local = this % comm_local_
end function

integer function comm_local_size(this, name, id)
    class(ComponentSplitter) :: this
    character(3), intent(in), optional :: name
    integer, intent(in), optional :: id
    integer :: id_
    
    if (PRESENT(name)) then
        id_ = this % comp_id(name)
    else if (PRESENT(id)) then
        id_ = id
    else
        id_ = this % i_am_id()
    end if
    
    comm_local_size = this % proc_num_(id_)

end function

!=====================================================================

integer function proc_first(this, name, id)
    class(ComponentSplitter) :: this
    character(3), intent(in), optional :: name
    integer, intent(in), optional :: id
    integer :: c
    character(3) :: name_
    
    if (PRESENT(name)) then
        name_ = name
    else if (PRESENT(id)) then
        name_ = this % comp_name(id)
    else
        name_ = this % i_am()
    end if
      
      
    do c = 0, this % size_world_-1
        if(this % proc_in(c, name_)) then
            proc_first = c
            EXIT
        end if
    end do
end function

!=====================================================================

subroutine proc_list(this, name, id, list)
    class(ComponentSplitter) :: this
    integer, allocatable, intent(inout) :: list(:)
    character(3), intent(in), optional :: name
    integer, intent(in), optional :: id
    character(3) :: name_
    integer :: i
    
    if (PRESENT(name)) then
        name_ = name
    else if (PRESENT(id)) then
        name_ = this % comp_name(id)
    else
        name_ = this % i_am()
    end if
    
    allocate(list(this % comm_local_size(name_)))
    
    list = (/ (i, i = this % proc_first(name_), this % proc_first(name_) + this % comm_local_size(name_)-1) /)
end subroutine
    
!=====================================================================

function i_am(this)
    class(ComponentSplitter) :: this
    character(LEN=3) :: i_am
    i_am = COMPONENT_NAMES(this % i_am_id())
end function


integer function i_am_id(this)
    class(ComponentSplitter) :: this
    integer :: c
    
    if (this % i_am_id_ == -1) then
        do c = 1, this % comp_defined()
            if(this % proc_in(this % rank_world_, COMPONENT_NAMES(c))) then
                this % i_am_id_ = c
                EXIT
            end if
        end do
   end if
   
   i_am_id = this % i_am_id_     
end function

integer function get_comp_id(this, comp_name)
    class(ComponentSplitter) :: this
    character(3), intent(in) :: comp_name
    get_comp_id = search_linear(comp_name, COMPONENT_NAMES, 1, SIZE(COMPONENT_NAMES))
    if (get_comp_id == 0) call this % sync_error("ComponentSplitter: invalid component name: <"//TRIM(comp_name)//">", "IND")
end function

character(3) function get_comp_name(this, comp_id)
    class(ComponentSplitter) :: this
    integer, intent(in) :: comp_id
    get_comp_name = COMPONENT_NAMES(comp_id)
end function

!=====================================================================

logical function is_model(this, name, id)
    class(ComponentSplitter) :: this
    character(3), intent(in), optional :: name
    integer, intent(in), optional :: id
    integer :: ind
    character(3) :: name_
    
    if (PRESENT(name)) then
        name_ = name
    else if (PRESENT(id)) then
        name_ = this % comp_name(id)
    else
        name_ = this % i_am()
    end if
    
    ind = search_linear(name_, SERVICE_COMPS, 1, SIZE(SERVICE_COMPS))
    is_model = ind == 0
    
end function


logical function is_first_rank(this)
    class(ComponentSplitter) :: this
    is_first_rank = this % rank_local_ == 0
end function
    
!=====================================================================

integer function comp_defined(this)
    class(ComponentSplitter) :: this
    comp_defined = SIZE(COMPONENT_NAMES)
end function

integer function comp_launched(this)
    class(ComponentSplitter) :: this
    comp_launched = this % comp_launched_
end function

integer function model_launched(this)
    class(ComponentSplitter) :: this
    integer :: c
    
    model_launched = 0
    do c = 1, SIZE(COMPONENT_NAMES)
        if (this % is_model(COMPONENT_NAMES(c)) .AND. this % is_launched(COMPONENT_NAMES(c))) &
                    model_launched = model_launched + 1
    end do
end function

logical function is_launched(this, name, id)
    character(3), intent(in), optional :: name
    integer, intent(in), optional :: id
    class(ComponentSplitter) :: this
    is_launched = this % comm_local_size(name, id) /= 0
end function

!=====================================================================

subroutine set_run_phase(this, phase)
    class(ComponentSplitter) :: this
    character(3), intent(in) :: phase
    this % run_phase_ = phase
end subroutine

!=====================================================================

subroutine show_memory_usage(this, msg, certain_rank)
    class(ComponentSplitter) :: this
    character(*), intent(in) :: msg
    integer :: i
    integer, intent(in), optional :: certain_rank

    if (this % rank_world() == 0) write(*,*) "Memory usage "//TRIM(msg)
    
    if (PRESENT(certain_rank)) then
        if (this % rank_world() == certain_rank) &
            call memory_usage_short("rank "//n2s(certain_rank))
    else
        do i = 0, this % comm_world_size() - 1
            if (this % rank_world() == i) then
                call memory_usage_short("rank "//n2s(i))
            end if
            call this % barrier()
        end do
    end if
end subroutine


!========================================================================================
! PRIVATE
!========================================================================================
!========================================================================================

! Main function which define decomposition of processes - linear distribution now     
logical function proc_in(this, test_rank, test_comp)
    class(ComponentSplitter) :: this
    integer, intent(in) :: test_rank
    character(3), intent(in) :: test_comp
    integer :: test_comp_id, lo, hi

    test_comp_id = this % comp_id(test_comp)
    
    lo = SUM(this % proc_num_(1:test_comp_id)) - this % proc_num_(test_comp_id)
    hi = SUM(this % proc_num_(1:test_comp_id))
    
    proc_in = test_rank >= lo .AND. test_rank < hi
    
end function

!========================================================================================

subroutine read_command_line_args(this, run_comps)
    class(ComponentSplitter) :: this
    integer, intent(in) :: run_comps
    character(20) name_arg, np_arg
    integer :: np, comp
    
    this % comp_launched_ = run_comps
    
    if (run_comps == 0) call this % sync_error("ComponentSplitter: incorrect arguments.", "GLB")
    
    do comp = 1, run_comps
        call GET_COMMAND_ARGUMENT((comp-1)*2+1, name_arg)
        call GET_COMMAND_ARGUMENT((comp-1)*2+2, np_arg)
        read(np_arg, *) np

        this % proc_num_(this % comp_id(name_arg)) = np
    end do
                    
end subroutine

!========================================================================================

end module


