
module datetime_module

use utils_string_module
implicit none

public :: DateTime
                                                        ! ****** MODULE PUBLIC API *******
public :: days_in_month                                 ! Days in current month
public :: days_in_year                                  ! Days in current year
public :: is_leap_year                                  ! True if year is leap and HAS_LEAP_YEARS
public :: date2sec                                      ! Convert date to seconds since NULL_YEAR
public :: sec2date                                      ! Convert seconds from NULL_YEAR to date
public :: operator(==)                                  ! is two DateTime objects equal?
public :: epoch_start                                   ! Null date (NULL_YEAR, 1, 1, 0, 0, 0)

integer, parameter, private :: NULL_YEAR = 1900            ! Start of epoch (better 1900, but some files has "days since 0000")
logical, parameter, private :: HAS_LEAP_YEARS = .TRUE.  ! Change if dont want leap years


interface operator(==)
  module procedure eq
endinterface

!========================================================================================
type :: DateTime
    integer :: year = 0                             
    integer :: month = 0
    integer :: day = 0
    integer :: hour = 0
    integer :: minute = 0
    integer :: second = 0
    
 contains                                               ! ****** CLASS PUBLIC API ******
    procedure, public :: is_valid                       ! Check of data is valid
    procedure, public :: add_time                       ! Add time (d, h, m, s) to date
    procedure, public :: year_day                       ! Day number in this year
    procedure, public :: year_sec                       ! Second number in this year
    procedure, public :: to_string                      ! ISO representation of date
    
    procedure, private :: add_days
    procedure, private :: add_hours
    procedure, private :: add_minutes
    procedure, private :: add_seconds
        
end type 

    interface DateTime                                  
        procedure :: DateTimeFromString, DateTimeFromNumbers   ! Constructor for Date from ISO string or y,m,d,h,m,s integers
    end interface
    
 CONTAINS
!========================================================================================

function DateTimeFromNumbers(year, month, day, hour, minute, second) result(this)
    type(DateTime) :: this
    integer, intent(in) :: year, month, day, hour, minute, second
    
    this = DateTime(year, month, day, hour, minute, second)
    
    if(.NOT. this % is_valid()) then
        write(*,'(a)') "DateTime: invalid format of date string" 
        write(*,'(a)') "If you construct it from string, should be in ISO 8601 format, e.g.: 1988-11-12 12:00:00"
        write(*,'(a)') "If you construct it from numbers, all of them shoul be in the right range"
        write(*,*), year, month, day, hour
        !HAKEL NAGEL MNK
        if ( year == 0 ) then
            this = DateTime(1900, month, day, hour, minute, second)        
        else
            call assert_error("Date: date is not valid: "//this % to_string())
        endif
    end if
    
end function

function DateTimeFromString(str) result(this)
    type(DateTime) :: this
    character(*), intent(in) :: str
    
    this = DateTime(S2I(str(1:4)), S2I(str(6:7)), S2I(str(9:10)), S2I(str(12:13)), S2I(str(15:16)), S2I(str(18:19)))

end function    
    
!========================================================================================
 
character(19) function to_string(this)
    class(DateTime) :: this
    to_string = n2s(this % year, 4)    //"-"// &
                n2s(this % month, 2)   //"-"// &
                n2s(this % day, 2)     //" "// &
                n2s(this % hour, 2)    //":"// &
                n2s(this % minute, 2)  //":"// &
                n2s(this % second, 2)
                    
end function
!========================================================================================

subroutine add_time(this, d, h, m, s)

    class(DateTime), intent(inout) :: this
    integer, intent(in), optional :: d, h, m, s
        
    call this % add_days   (MERGE(d, 0, PRESENT(d)))
    call this % add_hours  (MERGE(h, 0, PRESENT(h)))
    call this % add_minutes(MERGE(m, 0, PRESENT(m)))
    call this % add_seconds(MERGE(s, 0, PRESENT(s)))
        
end subroutine

!========================================================================================

integer function year_day(this)

    class(Datetime),intent(in) :: this

    integer :: month

    year_day = 0
    do month = 1,this % month - 1
        year_day = year_day + days_in_month(month, this % year)
    enddo
    year_day = year_day + this % day

end function

!========================================================================================

integer function year_sec(this)

    class(Datetime),intent(in) :: this

    year_sec =  (this % year_day() - 1)*60*60*24 + &
                this % hour*60*60 + this % minute*60 + this % second

end function

!========================================================================================

subroutine add_days(this, d)

    class(DateTime),intent(inout) :: this
    integer, intent(in) :: d

    integer :: days_in_cur_month

    this % day = this % day + d
    
    do 
        days_in_cur_month = days_in_month(this % month, this % year)
        
        if(this % day > days_in_cur_month) then
            this % day = this % day - days_in_cur_month
            this % month = this % month + 1
            if (this % month > 12) then
                this % year = this % year + this % month/12
                this % month = MOD(this % month, 12)
            end if
        else
            EXIT
        end if
    end do

end subroutine

!========================================================================================

subroutine add_hours(this, h)

    class(Datetime),intent(inout) :: this
    integer, intent(in) :: h

    this % hour = this % hour + h

    if (this % hour >= 24) then
        call this % add_days(this % hour/24)
        this % hour = MOD(this % hour, 24)
    endif

end subroutine

!========================================================================================

subroutine add_minutes(this, m)

    class(DateTime), intent(inout) :: this
    integer, intent(in) :: m

    this % minute = this % minute + m

    if(this % minute >= 60)then
        call this % add_hours(this % minute/60)
        this % minute = MOD(this % minute, 60)
    end if
end subroutine

!========================================================================================

subroutine add_seconds(this, s)
    class(DateTime), intent(inout) :: this
    integer, intent(in) :: s

    this % second = this % second + s

    if(this % second >= 60)then
        call this % add_minutes(this % second/60)
        this % second = MOD(this % second, 60)
    endif

end subroutine

!========================================================================================
!========================================================================================

logical function is_valid(this)

    class(datetime), intent(in) :: this

    is_valid = .TRUE.

    if ((this % year < NULL_YEAR)                               .OR. &
        (this % month < 1 .OR. this % month > 12)               .OR. &
        (this % day < 1)                                        .OR. & 
        (this % day > days_in_month(this % month, this % year)) .OR. &
        (this % hour < 0 .OR. this % hour > 23)                 .OR. &
        (this % minute < 0 .OR. this % minute > 59)             .OR. &
        (this % second < 0 .OR. this % second > 59)                ) &    
            is_valid = .FALSE.
end function

!========================================================================================

integer function days_in_month(month, year)
    integer, intent(in) :: month
    integer, intent(in) :: year

    integer, parameter,dimension(12) :: MONTH_SIZE = [31,28,31,30,31,30,31,31,30,31,30,31]

    if (month < 1 .OR. month > 12) then
        days_in_month = 0
        return
    endif

    if (month == 2 .AND. is_leap_year(year)) then
        days_in_month = 29
    else
        days_in_month = MONTH_SIZE(month)
    endif

end function

!========================================================================================

integer function days_in_year(year)
    integer, intent(in) :: year

    if (is_leap_year(year)) then
        days_in_year = 366
    else
        days_in_year = 365
    endif    
end function

!========================================================================================

logical function is_leap_year(year)
    integer, intent(in) :: year
    is_leap_year = MOD(year, 4) == 0 .AND. MOD(year, 100) /= 0 .OR. MOD(year, 400) == 0
    is_leap_year = HAS_LEAP_YEARS .AND. is_leap_year
end function

!========================================================================================

integer(8) function date2sec(dt)

    type(Datetime), intent(in) :: dt
    integer :: year, days_in_years

    date2sec = 0
    days_in_years = 0
    do year = NULL_YEAR, dt % year - 1
        days_in_years = days_in_years + days_in_year(year)
    enddo
    
    date2sec = (days_in_years +  dt % year_day()-1)*24*60*60
    date2sec = date2sec + dt % hour*60*60 + dt % minute*60 + dt % second
    
end function

!========================================================================================

type(Datetime) function sec2date(sec)
    integer(8), intent(in) :: sec

    integer :: year, month, day, hour, minute, second
    integer(8) :: days, sec_left
    
    days = sec/(24*60*60) + 1
    year = NULL_YEAR
    do
        if (days <= days_in_year(year)) EXIT
        days = days - days_in_year(year)
        year = year + 1
    end do

    month = 1
    do
        if (days <= days_in_month(month, year)) EXIT
        days = days - days_in_month(month, year)
        month = month + 1
    end do
    
    day      = days
    sec_left = MOD(sec, 24*60*60) 
    hour     = sec_left/(60*60)
    minute   = sec_left/60 - hour*60
    second   = sec_left - hour*60*60 - minute*60
    
    sec2date = Datetime(year, month, day, hour, minute, second)

end function

!========================================================================================

logical function eq(d1, d2)
  type(DateTime), intent(in) :: d1,d2

  eq = d1 % year   == d2 % year   .AND. &
       d1 % month  == d2 % month  .AND. &
       d1 % day    == d2 % day    .AND. &
       d1 % hour   == d2 % hour   .AND. &
       d1 % minute == d2 % minute .AND. &
       d1 % second == d2 % second
end function

!========================================================================================

type(DateTime) function epoch_start()
    epoch_start = DateTime(NULL_YEAR, 1, 1, 0, 0, 0)
end function

!========================================================================================


end module
