!========================================================================================
! Decomposition class incapsulate logic of logically-rectangular decomposition of
! domain between processes and define many small accessors to data.
!========================================================================================

module decomposition_module

use utils_mpi_module
use utils_string_module
use file_handler_nc_module
use component_splitter_module

implicit none

!========================================================================================

type :: Decomposition
! private
    integer :: i_size_, j_size_, k_size_
    integer :: inp_, jnp_
    character(2) :: dtype_ = ''
    integer, allocatable :: west_all_(:), east_all_(:), south_all_(:), north_all_(:)
    integer :: rank_local_ = -1
    integer :: proc_count_ = -1
    integer, allocatable :: proc_decomp_(:,:)
    logical :: is_icycle_ = .false.
    logical :: is_bsc_ = .false.
            
 contains
    !========================
    ! ====== PUBLIC API =====
    final  :: destroy_decomposition
    procedure, private :: alloc
    
!     procedure, public :: get_proc_decomp
    procedure, private :: build_inp_jnp
    procedure, private :: build_borders
    procedure, public :: build_proc_decomp
    procedure, public :: rank_by_ij
    procedure, public :: ij_by_rank
    
    procedure, public :: dump_proc_decomp_to_file
    procedure, private :: read_proc_decomp_from_file
    
    procedure, public :: broadcast => broadcast_decomposition
    procedure, public :: show => show_decomposition
    procedure, public :: to_string_2D
    procedure, public :: to_string_3D
     
    procedure, public :: global_dims 
    procedure, public :: i_size
    procedure, public :: j_size
    procedure, public :: k_size

    procedure, public :: size_2D
    
    procedure, public :: proc_decomp_np
    procedure, public :: inp
    procedure, public :: jnp
    procedure, public :: proc_count
    
    !procedure, public :: local_domain
    
    procedure, public :: borders
    procedure, public :: neighbour_ranks
    procedure, public :: neighbour_coords
    
    procedure, public :: west
    procedure, public :: east
    procedure, public :: south
    procedure, public :: north
    
    procedure, public :: i_size_loc
    procedure, public :: j_size_loc
    procedure, public :: loc_sizes_2D
    procedure, public :: loc_sizes_3D
    
    procedure, public :: lower_bound_2D
    procedure, public :: lower_bound_3D
    procedure, public :: upper_bound_2D
    procedure, public :: upper_bound_3D

    procedure, public :: points_num
    procedure, public :: points_set
    
    
    
    procedure, public :: global_index
    procedure, public :: global_to_i
    procedure, public :: global_to_j
    procedure, public :: first_global_index

    procedure, public :: to_cartesian_ij    
    procedure, public :: to_cartesian_ijk
    
    procedure, public :: is_icycle
    procedure, public :: is_bsc
    
    procedure, private :: land_mask_corner_case_patch
    
    ! ====== PUBLIC API =====    
    !========================
    
end type

    interface Decomposition
        procedure :: decomposition_constructor
    end interface

 CONTAINS
  

!========================================================================================

subroutine alloc(this, proc_count)
    class(Decomposition) :: this
    integer, intent(in) :: proc_count
    if (.NOT. ALLOCATED(this % west_all_)) then
        allocate(this % west_all_(0:proc_count-1))
        allocate(this % east_all_(0:proc_count-1))
        allocate(this % south_all_(0:proc_count-1))
        allocate(this % north_all_(0:proc_count-1))
    end if
    
    if (.NOT. allocated(this % proc_decomp_)) then
        allocate(this % proc_decomp_(this % inp_, this % jnp_))
    end if
    
end subroutine


!========================================================================================
! Processor neighbour coordinates (west, east, south, north) taking 
! into account i_cycle, is_bsc (-1 if no neighbour)
!========================================================================================


function neighbour_coords(this, coords) 
    class(Decomposition) :: this
    integer, intent(in) :: coords(2)
    integer, dimension(4,2) :: neighbour_coords
    integer :: i, j
    
    i = coords(1)
    j = coords(2)
    
    ! Init answer
    neighbour_coords = -1
        
    ! West neighbour (using i_cycle)
    if (i > 1) then
        neighbour_coords(1,:) = (/ i-1, j /)
    else if (this % is_icycle_) then
        neighbour_coords(1,:) = (/ this % inp_, j /)
    end if

    ! East neighbour (using i_cycle)
    if (i < this % inp_) then
        neighbour_coords(2,:) = (/ i+1, j /)
    else if (this % is_icycle_) then
        neighbour_coords(2,:) = (/ 1, j /)
    end if

    ! South neighbour
    if (j > 1) then
        neighbour_coords(3,:) = (/ i, j-1 /)
    end if
    
    ! North neighbour (using is_bsc)
    if (j < this % jnp_) then
        neighbour_coords(4,:) = (/ i, j+1 /)
    else if (this % is_bsc_) then
        neighbour_coords(4,:) = (/ this % inp_ - i + 1, j /)
    end if

end function

!========================================================================================
! Processor ranks (west, east, south, north) taking 
! into account i_cycle, is_bsc (-1 if no neighbour)
!========================================================================================

function neighbour_ranks(this, rank_test)
    class(Decomposition) :: this
    integer, intent(in), optional :: rank_test
    integer, dimension(4) :: neighbour_ranks
    integer :: rank_test_, nc(4,2), n
        
    rank_test_ = MERGE(rank_test, this % rank_local_, PRESENT(rank_test))
    
    nc = this % neighbour_coords(this % ij_by_rank(rank_test_))
    
    neighbour_ranks = -1
    do n = 1, 4
        if (nc(n,1) /= -1) then
            neighbour_ranks(n) = this % proc_decomp_(nc(n,1), nc(n,2))
        end if
    end do
end function

!========================================================================================


function decomposition_constructor(il, jl, kl, dtype, proc_count, rank_local, proc_decomp_file, is_icycle, is_bsc) result(this)
    type(Decomposition) :: this
    integer, intent(in) :: il, jl, kl, proc_count, rank_local
    character(*), intent(in), optional :: proc_decomp_file
    logical, intent(in), optional :: is_icycle, is_bsc
    character(*), intent(in) :: dtype

    this % dtype_ = dtype
    this % i_size_ = il
    this % j_size_ = jl
    this % k_size_ = kl
    this % rank_local_ = rank_local
    this % proc_count_ = proc_count
    if (PRESENT(is_icycle)) this % is_icycle_ = is_icycle
    if (PRESENT(is_bsc)) this % is_bsc_ = is_bsc
    
    if (proc_count == 0) then
        write(*,*) 'Decomposition: Processor count is equal to zero.'
        call exit(1)
    end if            

    ! Allocate memory
    call this % alloc(proc_count)
    
    ! Build proc decomp based on (inp, jnp) or use prebuild
    if (PRESENT(proc_decomp_file) .AND. proc_decomp_file /= "") then
        call this % read_proc_decomp_from_file(proc_decomp_file)
    else
        ! Build inp and jnp
        call this % build_inp_jnp(il, jl, kl, dtype, proc_count)    
        call this % build_proc_decomp()
    end if
    
    ! Build actual border coordinates using calculated inp, jnp
    call this % build_borders(il, jl, this % inp_, this % jnp_)    
    
end function
    

subroutine destroy_decomposition(this)
    type(Decomposition) :: this
    if (ALLOCATED(this % west_all_)) &
        deallocate(this % west_all_, this % east_all_, this % south_all_, this % north_all_)
    if (allocated(this % proc_decomp_)) &
        deallocate(this % proc_decomp_)
end subroutine
    
               
!========================================================================================
! Calculate (inp jnp) pair using some restrictions (start from optimal square and grow)
!========================================================================================

subroutine build_inp_jnp(this, il, jl, kl, dtype, proc_count)
    class(Decomposition) :: this
    integer, intent(in) :: il, jl, kl, proc_count
    character(*), intent(in) :: dtype
    integer :: j_rad
    
    ! Define cmp % jnp corresponding to decomposition style
    if (dtype == "2D") then
        this % jnp_ = MAX(1, NINT(SQRT( REAL(jl)*REAL(proc_count)/REAL(il) )))
        if(CompSplitter % i_am() == 'ICE' .and. proc_count>2) this % jnp_ = 2
    else
        this % jnp_ = proc_count
        this % inp_ = 1

        return
    end if

    ! Check cmp % jnp for restrictions: check all jnp from optimal in both directions
        this % inp_ = 1
    j_rad = 1
    do while(.TRUE.)    
        if(MOD(proc_count, this % jnp_) == 0) then
            this % inp_ = proc_count/this % jnp_
            if ((MOD(this % inp_, 2) == 0 .OR. this % inp_ == 1) .AND. MOD(il, this % inp_) == 0) then
                ! Decompose in j-direction                
                if(MOD(jl, this % jnp_) == 0) then
                    EXIT
                else       
                    ! Decompose in j-direction (if not equal distribution, let last row be smaller(! not bigger - performance!!)             
                    if(MOD(jl, this % jnp_-1) <= jl/(this % jnp_-1) .AND. MOD(jl, this % jnp_-1) > 1) EXIT
                end if
            end if
        end if
        
        ! Search in increasing radius
        this % jnp_ = this % jnp_ + j_rad
        if(j_rad > 0) then 
            j_rad = -j_rad
        else
            j_rad = -j_rad + 1
        end if
        
        if(this % jnp_ < 1 .OR. this % jnp_ > proc_count)  then
            write(*,*) 'Decomposition: cant '//dtype//'-devide domain '//TRIM(this % to_string_2D())// ' between '//TRIM(n2s(proc_count))//' cores.'
            call exit(1)
        end if
    end do  
    
end subroutine

!========================================================================================
! Build processor decomposition matrix.
! It is simple sequentional filling from bottom to top without matrix
! and ignoring earth blocks in case of matrix presence.
!========================================================================================

subroutine build_proc_decomp(this, proc_mask)
    class(Decomposition) :: this
    integer, optional, intent(in) :: proc_mask(:,:)
    integer :: rank, i, j, mod_proc_mask(SIZE(proc_mask, 1), SIZE(proc_mask, 2))
    
    ! Rewrite inp, jnp using matrix
    if (PRESENT(proc_mask)) then
        this % inp_ = SIZE(proc_mask, 1)
        this % jnp_ = SIZE(proc_mask, 2)
        mod_proc_mask = proc_mask
        ! Modify land mask for special case
        call this % land_mask_corner_case_patch(mod_proc_mask)
    end if

    ! If it is allocated, replace
    if (allocated(this % proc_decomp_)) deallocate(this % proc_decomp_)
    allocate(this % proc_decomp_(this % inp_, this % jnp_))
    
    ! Build processor mapping    
    rank = 0
    do j = 1, this % jnp_
    do i = 1, this % inp_
        ! No mask - add rank
        if (.NOT. PRESENT(proc_mask)) then
            this % proc_decomp_(i,j) = rank
            rank = rank + 1
        ! Mask and not null cell -> add rank
        else if (PRESENT(proc_mask) .AND. mod_proc_mask(i,j) == 1) then
            this % proc_decomp_(i,j) = rank
            rank = rank + 1
        else
            this % proc_decomp_(i,j) = -1
        end if
    end do
    end do
    
    ! Rebuild borders
    call this % build_borders(this % i_size_, this % j_size_, this % inp_, this % jnp_)    
    
end subroutine

    
! Corner case: if case of chess-like land, diagonal cells won't be updated with 4-neighbours
! scheme, so dirty solution is to unmask this area).
! Simple idea: check diagonal neighbours and if they are non masked - unmask (dirty)
subroutine land_mask_corner_case_patch(this, land_mask)

    class(Decomposition) :: this
    integer, intent(inout) :: land_mask(:,:)
    integer :: i, j
    integer :: my_ns(4,2)
    integer :: new_mask(SIZE(land_mask,1), SIZE(land_mask,2))

    new_mask = land_mask
!    write(*,*) "land mask at start", land_mask
    
    do j = 1, this % jnp()
    do i = 1, this % inp()    
        if (land_mask(i,j) == 0) CYCLE;
        
        ! Get neighbours of this rank
        my_ns = this % neighbour_coords( (/i, j/))
!         write(*,*) 'check', (/i, j/), 'land_mask', land_mask(i,j)
        ! Get coord of south and its neighbours. If some of them unmasked, then unmask center
        call check(my_ns(3,:))
        call check(my_ns(4,:))
    end do
    end do    

    land_mask = new_mask
    
    CONTAINS
    
        subroutine check(this_coords)
            integer, intent(in) :: this_coords(2)
            integer :: ns_coords(4,2)

            ! No such neighbour
            if (this_coords(1) == -1) return
            
            ! It is unmasked already
            if (land_mask(this_coords(1), this_coords(2)) == 1) return
            
            ! Get west and east neighbours of this soutth or north neighbour
            ns_coords = this % neighbour_coords(this_coords)
            
            ! Check condition: neighbour is presented, it masked => unmask center
            if (ns_coords(1,1) /= -1) then
                if (land_mask(ns_coords(1,1), ns_coords(1,2)) == 1) then
!                     write(*,*) "west neighbour", ns_coords(1,:), " cause", this_coords
                    new_mask(this_coords(1), this_coords(2)) = 1
                end if
            end if
            if (ns_coords(2,1) /= -1) then
                if (land_mask(ns_coords(2,1), ns_coords(2,2)) == 1) then
!                     write(*,*) "east neighbour", ns_coords(2,:), " cause", this_coords
                    new_mask(this_coords(1), this_coords(2)) = 1
                end if
            end if
        end subroutine
        
end subroutine

!========================================================================================

subroutine dump_proc_decomp_to_file(this, filename)
    class(Decomposition) :: this
    character(*), intent(in) :: filename
    type(FileHandler_NC) :: fh
    write(*,*), "Dump prco decomp to", TRIM(filename) 
    call fh % open_or_create_file(filename)
    call fh % create_dim("inp", this % inp_)
    call fh % create_dim("jnp", this % jnp_)
    call fh % open_or_create_var("proc_decomp", "int", "inp", "jnp")
    call fh % put(this % proc_decomp_, (/ 1, 1 /), this % proc_decomp_np())
    call fh % close_file()
end subroutine

subroutine read_proc_decomp_from_file(this, filename)
    class(Decomposition) :: this
    character(*), intent(in) :: filename
    type(FileHandler_NC) :: fh
    
    call fh % open_file(filename)
    this % inp_ = fh % get_dim_size("inp")
    this % jnp_ = fh % get_dim_size("jnp")
    call fh % open_var("proc_decomp")    
    if (allocated(this % proc_decomp_)) then
        deallocate(this % proc_decomp_)
        allocate(this % proc_decomp_(this % inp_, this % jnp_))
    end if
    call fh % get(this % proc_decomp_, (/ 1, 1 /), this % proc_decomp_np()) 
    call fh % close_file()
    
end subroutine    

!========================================================================================

subroutine build_borders(this, il, jl, inp, jnp)
    class(Decomposition) :: this
    integer, intent(in) :: il, jl, inp, jnp
    integer :: rank, i_loc_size, j_loc_size, i, j

    if ( this % dtype_ == "1D") then
        this % west_all_(0:(jnp-1)) = 1
        this % east_all_(0:(jnp-1)) = il

        this % south_all_( 0 ) = 1
        j_loc_size =  jl / jnp
        this % north_all_( 0 ) = j_loc_size

        do rank = 1, jnp-1
            this % south_all_(rank) = this % south_all_(rank -1) + j_loc_size
            j_loc_size = (jl - this % north_all_(rank - 1) + 1) / (jnp - rank)
            this % north_all_(rank) = this % north_all_(rank-1) + j_loc_size
!write(*,'(i4,a,4(i4,x))'),this % rank_local_," j_loc_size", rank, j_loc_size, (jl - this % north_all_(rank - 1) + 1), (jnp - rank)
        enddo
        this % north_all_(jnp-1) = jl
        return
    endif

    ! Decompose in i-direction equally
    i_loc_size = il/inp
    ! Decompose in j-direction (if not equal distribution, let last row be smaller(!)
    j_loc_size = MERGE(jl/jnp, jl/(jnp - 1), MOD(jl, jnp) == 0)

    
    do j = 1, jnp
    do i = 1, inp
        if (this % proc_decomp_(i,j) /= -1) then
            rank = this % proc_decomp_(i,j)
            this % west_all_(rank) = (i-1)*i_loc_size + 1
            this % east_all_(rank) = this % west_all_(rank) + i_loc_size - 1            
            this % south_all_(rank) = (j-1)*j_loc_size + 1
            this % north_all_(rank) = MIN(this % south_all_(rank) + j_loc_size - 1, jl)            
        end if
    end do
    end do
            
end subroutine

!========================================================================================

function borders(this, rank_test)
    class(Decomposition) :: this
    integer, dimension(4) :: borders
    integer, optional, intent(in) :: rank_test
    integer :: r
    r = MERGE(rank_test, this % rank_local_, PRESENT(rank_test))
    borders = (/ this % west_all_(r),  this % east_all_(r),  this % south_all_(r),  this % north_all_(r) /)
end function

    
integer function west(this, rank_test)
    class(Decomposition) :: this
    integer, optional, intent(in) :: rank_test
    if(PRESENT(rank_test)) then
        west = this % west_all_(rank_test)
    else 
        west = this % west_all_(this % rank_local_)
    end if
end function west

integer function east(this, rank_test)
    class(Decomposition) :: this
    integer, optional, intent(in) :: rank_test
    if(PRESENT(rank_test)) then
        east = this % east_all_(rank_test)
    else 
        east = this % east_all_(this % rank_local_)
    end if
end function east

integer function south(this, rank_test)
    class(Decomposition) :: this
    integer, optional, intent(in) :: rank_test
    if(PRESENT(rank_test)) then
        south = this % south_all_(rank_test)
    else 
        south = this % south_all_(this % rank_local_)
    end if
end function south

integer function north(this, rank_test)
    class(Decomposition) :: this
    integer, optional, intent(in) :: rank_test
    if(PRESENT(rank_test)) then
        north = this % north_all_(rank_test)
    else 
        north = this % north_all_(this % rank_local_)
    end if
end function north

integer function i_size_loc(this)
    class(Decomposition) :: this
    i_size_loc = this % east_all_(this % rank_local_) - this % west_all_(this % rank_local_) + 1
end function i_size_loc

integer function j_size_loc(this)
    class(Decomposition) :: this
    j_size_loc = this % north_all_(this % rank_local_) - this % south_all_(this % rank_local_) + 1
end function j_size_loc

function loc_sizes_2D(this) 
    class(Decomposition) :: this
    integer :: loc_sizes_2D(2)
    loc_sizes_2D = (/ this % i_size_loc(), this % j_size_loc() /)
end function     

function loc_sizes_3D(this) 
    class(Decomposition) :: this
    integer :: loc_sizes_3D(3)
    loc_sizes_3D = (/ this % i_size_loc(), this % j_size_loc(), this % k_size() /)
end function     

!========================================================================================

function global_dims(this)
    class(Decomposition) :: this
    integer, dimension(3) :: global_dims
    global_dims = (/ this % i_size_, this % j_size_, this % k_size_ /)
end function


integer function i_size(this)
    class(Decomposition) :: this
    i_size = this % i_size_
end function i_size

integer function j_size(this)
    class(Decomposition) :: this
    j_size = this % j_size_
end function j_size

integer function k_size(this)
    class(Decomposition) :: this
    k_size = this % k_size_
end function k_size

!========================================================================================

function proc_decomp_np(this)
    integer, dimension(2) :: proc_decomp_np
    class(Decomposition) :: this
    proc_decomp_np = (/ this % inp_, this % jnp_ /)
end function

integer function rank_by_ij(this, i, j)
    class(Decomposition) :: this
    integer, intent(in) :: i, j
    rank_by_ij = this % proc_decomp_(i,j)
end function

function ij_by_rank(this, rank_test)
    integer :: ij_by_rank(2)
    class(Decomposition) :: this
    integer, intent(in) :: rank_test
    integer :: i, j
    do i = 1, this % inp_
    do j = 1, this % jnp_
        if (this % proc_decomp_(i,j) == rank_test) then
            ij_by_rank = (/ i, j /)
            return
        end if
    end do
    end do
    ij_by_rank = (/ -1, -1 /)
end function

logical function is_icycle(this)
    class(Decomposition) :: this
    is_icycle = this % is_icycle_
end function

logical function is_bsc(this)
    class(Decomposition) :: this
    is_bsc = this % is_bsc_
end function


integer function inp(this)
    class(Decomposition) :: this
    inp = this % inp_
end function inp

integer function jnp(this)
    class(Decomposition) :: this
    jnp = this % jnp_
end function

!========================================================================================

integer function size_2D(this)
    class(Decomposition) :: this
    size_2D = this % i_size_ * this % j_size_
end function


integer function proc_count(this)
    class(Decomposition) :: this
    proc_count = this % inp_ * this % jnp_
end function


function lower_bound_2D(this) 
    class(Decomposition) :: this
    integer :: lower_bound_2D(2)
    lower_bound_2D = (/ this % west(), this % south() /)    
end function

function lower_bound_3D(this) 
    class(Decomposition) :: this
    integer :: lower_bound_3D(3)
    lower_bound_3D = (/ this % west(), this % south(), 1 /)    
end function

    function upper_bound_2D(this) 
    class(Decomposition) :: this
    integer :: upper_bound_2D(2)
    upper_bound_2D = (/ this % east(), this % north() /)    
end function

    function upper_bound_3D(this) 
    class(Decomposition) :: this
    integer :: upper_bound_3D(3)
    upper_bound_3D = (/ this % east(), this % north(), this % k_size() /)    
end function


!========================================================================================

subroutine broadcast_decomposition(this, root, comm)
    class(Decomposition) :: this
    integer, intent(in) :: root, comm
    
    call broadcast(this % i_size_,      root, comm)
    call broadcast(this % j_size_,      root, comm)
    call broadcast(this % k_size_,      root, comm)
    call broadcast(this % inp_,         root, comm)
    call broadcast(this % jnp_,         root, comm)
    call broadcast(this % proc_count_,  root, comm)
    call broadcast(this % is_icycle_,   root, comm)
    call broadcast(this % is_bsc_,      root, comm)
    
    call this % alloc(this % proc_count_)
    call broadcast(this % west_all_,    root, comm)
    call broadcast(this % east_all_,    root, comm)
    call broadcast(this % south_all_,   root, comm)
    call broadcast(this % north_all_,   root, comm)   
    call broadcast(this % proc_decomp_, root, comm)
        
end subroutine

!========================================================================================

subroutine show_decomposition(this)
    class(Decomposition) :: this

    write(*, '(a)') 'Number of processes:          '//TRIM(n2s(this % proc_count_))
    
    write(*, '(a)') 'Mesh size[i x j x k]:         '//          &
                    '['//TRIM(this % to_string_3D())//']'
    
    write(*, '(a)') 'Processor decomp[inp x jnp]:  '//       &
                    '['//TRIM(n2s(this % inp_))//' x '//TRIM(n2s(this % jnp_))//']'
#ifdef DEBUG
    write(*, '(a,<this % proc_count_>(i3, x))') 'west_all:  ',this % west_all_(0: this % proc_count_ - 1)
    write(*, '(a,<this % proc_count_>(i3, x))') 'east_all:  ',this % east_all_(0: this % proc_count_ - 1)
    write(*, '(a,<this % proc_count_>(i3, x))') 'south_all:  ',this % south_all_(0: this % proc_count_ - 1)
    write(*, '(a,<this % proc_count_>(i3, x))') 'north_all:  ',this % north_all_(0: this % proc_count_ - 1)
    write(*, '(a,i4)') 'to_cartes_ij:  ',this % to_cartesian_ij()
    !write(*, '(a,< this % jnp_>(i3, x))') ', proc_decom[1,1:jnp] = :  ', this % proc_decomp_(1, 1 : this % jnp_)

#endif
end subroutine
    
!========================================================================================
    
character(100) function to_string_2D(this)
    class(Decomposition) :: this
    to_string_2D = TRIM(n2s(this % i_size()))//'x'//TRIM(n2s(this % j_size()))
end function

character(100) function to_string_3D(this)
    class(Decomposition) :: this
    to_string_3D = TRIM(n2s(this % i_size()))//'x'//TRIM(n2s(this % j_size()))//'x'//TRIM(n2s(this % k_size()))
end function


function to_cartesian_ijk(this)
    class(Decomposition) :: this
    integer :: to_cartesian_ijk(this % inp_ + this % jnp_ + 1)
    integer :: i, counter
    counter = 1
    do i = 1, this % inp_
        to_cartesian_ijk(counter) = this % west_all_(counter-1)
        counter = counter + 1
    end do
    do i = 1, this % jnp_
        to_cartesian_ijk(counter) = this % south_all_((i-1)*this % inp_)
        counter = counter + 1
    end do
    to_cartesian_ijk(counter) = 1    
end function

function to_cartesian_ij(this)
    class(Decomposition) :: this
    integer :: to_cartesian_ij(this % inp_ + this % jnp_)
    integer :: i, counter
    counter = 1
    do i = 1, this % inp_
        to_cartesian_ij(counter) = this % west_all_(counter-1)
        counter = counter + 1
    end do
    do i = 1, this % jnp_
        to_cartesian_ij(counter) = this % south_all_((i-1)*this % inp_)
        counter = counter + 1
    end do
end function    
        
!========================================================================================

integer function global_index(this, i_coord, j_coord)
    class(Decomposition) :: this
    integer, intent(in) :: i_coord, j_coord
    global_index = (j_coord - 1)*this % i_size() + i_coord
end function

integer function points_num(this)
    class(Decomposition) :: this
    points_num = this % i_size_loc() * this % j_size_loc()
end function
    
subroutine points_set(this, points)
    class(Decomposition) :: this
    integer, allocatable, intent(inout) :: points(:)
    integer :: i
    
    allocate(points (this % points_num()))
    points = (/ (i, i = this % global_index(this % west(), this % south()), & 
                    this % global_index(this % east(), this % north())) /)
end subroutine

integer function global_to_i(this, global_ind)
    class(Decomposition) :: this
    integer, intent(in) :: global_ind
    global_to_i = MOD(global_ind, this % i_size())
    if (global_to_i == 0) global_to_i = this % i_size()
end function    
    
integer function global_to_j(this, global_ind)
    class(Decomposition) :: this
    integer, intent(in) :: global_ind
    global_to_j = global_ind/(this % i_size() + 1) + 1
end function    

integer function first_global_index(this)
    class(Decomposition) :: this
    first_global_index = this % global_index(this % west(), this % south())
end function


! function get_proc_decomp(this)
!     integer, pointer :: get_proc_decomp(:,:)
!     class(Decomposition) :: this
!     get_proc_decomp => this % proc_decomp_
! end function

!========================================================================================

! type :: LocalDomain
!     integer, dimension(4) :: borders            ! (west, east, south, north)
!     integer, dimension(3) :: size_loc_3D        ! (i_loc_size, j_loc_size)
!     integer, dimension(3) :: lower_bound_3D     ! (west, south, 1)
!     integer, dimension(3) :: upper_bound_3D     ! (east, north, k_size)    
! end type

end module

