!========================================================================================
! Struct with all information about Event
!========================================================================================

module event_module

use const_module
use actions_module
use mpi
use datetime_module
use array_info_module

implicit none

type Event
private
    character(NAME_LEN) :: time_ = ""
    character(NAME_LEN) :: src_ = ""
    character(NAME_LEN) :: dst_ = ""    
    character(NAME_LEN) :: owner_ = ""
    character(NAME_LEN) :: action_ = ""
    character(NAME_LEN) :: arr_name_ = ""    
    character(NAME_LEN) :: arr_dims_ = ""
    character(NAME_LEN) :: arr_type_ = ""
    character(NAME_LEN) :: ga_name_ = ""
    character(NAME_LEN) :: time_rec_ = ""
    character(NAME_LEN) :: info_ = ""
    
   
 contains  
    !=======================
    ! ===== PUBLIC API =====
    procedure, public :: to_string => to_string_event                       ! print event information  
    procedure, public :: less                           ! (that) .TRUE. if this < that
    procedure, public :: set_time                       ! setter for Event time
    procedure, public :: set_time_rec                   ! setter for Event time_rec
    
    procedure, public :: time => get_time               ! DateTime of event in string representation
    procedure, public :: action => get_action           ! action of event (e.g. "SAVE_DG")
    procedure, public :: src => get_src                 ! id of source component (e.g. OCN)
    procedure, public :: dst => get_dst                 ! id of destination component (e.g. CPL or ATM)
    procedure, public :: set_dst                 ! id of destination component (e.g. CPL or ATM)
    
    procedure, public :: owner => get_owner                 ! id of destination component (e.g. CPL or ATM)    
    
    procedure, public :: arr_name => get_arr_name       ! name of array in event (e.g. "temp")
    procedure, public :: arr_type => get_array_type     ! array type (e.g. "real8")
    procedure, public :: ga_name => get_ga_name
    procedure, public :: arr_dims => get_array_dims     ! array dims (e.g. "2D")
    procedure, public :: time_rec => get_time_rec       ! array dims (e.g. "2D")
    procedure, public :: info => get_info       ! array dims (e.g. "2D")
    
    procedure, public :: make_ga_name
        
    ! ===== PUBLIC API =====
    !=======================    
end type

    interface Event
        procedure :: init_event_comp
    end interface
    
type(Event) :: EMPTY_DATA

 CONTAINS

!========================================================================================

function init_event_comp(arr_info, act, owner, src, dst, dt, dt_rec, file_prefix, info) result(this)
    type(Event) :: this
    type(ArrayInfo), intent(in) :: arr_info
    character(*), intent(in) :: act, owner
    character(*), intent(in), optional :: src, dst, file_prefix, info
    type(DateTime), intent(in), optional :: dt, dt_rec
    character(10) :: yyyymmdd
    
    this % action_ = act
    this % owner_ = owner
    this % arr_name_ = arr_info % arr_name()
    this % arr_type_ = arr_info % arr_type()
    this % arr_dims_ = arr_info % arr_dims()
    if (PRESENT(dt_rec)) this % time_rec_ = dt_rec % to_string()
    if (PRESENT(info)) this % info_ = info
    if (PRESENT(dt)) then
        this % time_ = dt % to_string()
        yyyymmdd = this % time_(1:10)
    end if
    select case (act)
            
        case("SAVE_DG","SAVE_CP")
            this % src_ = owner
            if (PRESENT(dst)) then
                this % dst_ = dst
            else if (act == "SAVE_CP") then
                this % dst_ = TRIM(file_prefix)//"_CP_"//yyyymmdd//".nc"
            else if (act == "SAVE_DG") then
                this % dst_ = TRIM(file_prefix)//"_DG_"//TRIM(this % arr_name_)//".nc"   
            end if

        case("READ_FD", "READ_IC", "READ_CP")
            this % action_ = "READ_FD"
!             this % ga_name_ = TRIM(this % arr_name_)//"_"//TRIM(this % action_)
            if (PRESENT(src)) then
                this % src_ = src
            else if (act == "READ_CP") then
                this % src_ = TRIM(file_prefix)//"_CP_"//yyyymmdd//".nc"
                this % info_ = "READ_CP"
            else if (act == "READ_IC") then
                this % src_ = TRIM(file_prefix)//"_IC"//".nc"      
            else 
                write(*,*) "Event: you need to set <src> for READ_FD action."
                call exit(1)
            end if
            this % dst_ = owner
            
        case("SEND_MP")
            this % src_ = owner
            if (PRESENT(dst)) then
                if (dst == owner) then
                    write(*,*) "Event: SEND_MP to yourself is not permitted."
                    call exit(1)
                end if
                this % dst_ = dst
            else
                write(*,*) "Event: you need to set <dst> for SEND_MP action."
                call exit(1)
            end if
            
        case("RECV_MP")
            if (PRESENT(src)) then 
                if (src == owner) then
                    write(*,*) "Event: REVC_MP from yourself is not permitted."
                    call exit(1)
                end if
            
                this % src_ = src
            else
                write(*,*) "Event: you need to set <src> for RECV_MP action."
                call exit(1)
            end if        
            this % dst_ = owner
        case("SEND_DAS")
            this % src_ = owner
            if ( TRIM(this % arr_name_ )  == TRIM("t_f") .or. &
                 TRIM(this % arr_name_ )  == TRIM("s_f") .or. &
                 TRIM(this % arr_name_ )  == TRIM("aice") .or. &
                 TRIM(this % arr_name_ )  == TRIM("ssl_f") )  then
                    !write(*,*),"DAS: GET ", TRIM(this % arr_name_ )
            else
                write(*,*) "Unknown variable for assimilation SEND", this % arr_name_
                call exit(1)
            end if
        case("RECV_AOI","RECV_ENKF")
            this % dst_ = owner
            if ( TRIM(this % arr_name_ ) == TRIM("t_f")  .or. TRIM(this % arr_name_)  == TRIM("t_dif") .or. &
                 TRIM(this % arr_name_ ) == TRIM("ssl_f").or. TRIM(this % arr_name_)  == TRIM("ssl_dif") .or. &
                 TRIM(this % arr_name_ ) == TRIM("aice").or. TRIM(this % arr_name_)  == TRIM("aice_dif") .or. &
                 TRIM(this % arr_name_ ) == TRIM("s_f")  .or. TRIM(this % arr_name_ ) == TRIM("s_dif") ) then
                 !write(*,*),"DAS RECV_ENKF", TRIM(this % arr_name_ )
            else
                write(*,*) "Unknown variable for assimilation RECV", this % arr_name_
                call exit(1)
            end if


        case("SEND_NST")
            this % src_ = owner
            write(*,*) "Event NST SEND", this % arr_name_

        case("RECV_NST")
            this % dst_ = owner
            write(*,*) "Event NST RECV", this % arr_name_
        case ("STOP")
            this % arr_name_ = ""
            this % src_ = owner
        case ("ERROR")
            this % arr_name_ = ""
            this % src_ = owner
            if (PRESENT(dst)) then 
                this % dst_ = dst
            else
                write(*,*) "Event: you need to set <dst> for ERROR action."
                call exit(1)
            end if
        case default
            write(*,*) "Event: uknown action: <"//TRIM(act)//">"
            call exit(1)
    end select 
    
    this % ga_name_ = this % make_ga_name(this % owner_, this % arr_name_, this % action_)
    
end function


!========================================================================================

integer function event_pack_size()
    integer :: ierr
    call MPI_PACK_SIZE(11*NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, event_pack_size, ierr)
end function

subroutine event_to_mpi_pack(ev, packed)
    type(Event), intent(in) :: ev
    character, intent(inout) :: packed(:)
    integer :: pos, ierr
    
    pos = 0
    call MPI_PACK(ev % time_,       NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)  
    call MPI_PACK(ev % src_,        NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)  
    call MPI_PACK(ev % dst_,        NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)
    call MPI_PACK(ev % owner_,      NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)
    call MPI_PACK(ev % action_,     NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)
    call MPI_PACK(ev % arr_name_,   NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)  
    call MPI_PACK(ev % arr_type_,   NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)  
    call MPI_PACK(ev % arr_dims_,   NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)
    call MPI_PACK(ev % ga_name_,    NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)
    call MPI_PACK(ev % time_rec_,   NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)
    call MPI_PACK(ev % info_,       NAME_LEN, MPI_CHARACTER, packed, SIZE(packed), pos, MPI_COMM_WORLD, ierr)
    
end subroutine

subroutine mpi_pack_to_event(packed, ev)
    character, intent(in) :: packed(:)
    type(Event), intent(inout) :: ev
    integer :: pos, ierr
    
    pos = 0
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % time_,    NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % src_,     NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % dst_,     NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % owner_,   NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % action_,  NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % arr_name_,NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % arr_type_,NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % arr_dims_,NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % ga_name_, NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)    
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % time_rec_,NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    call MPI_UNPACK(packed, SIZE(packed), pos, ev % info_,    NAME_LEN, MPI_CHARACTER, MPI_COMM_WORLD, ierr)
    
end subroutine

!========================================================================================

logical function less(this, that)
    class(Event) :: this
    type(Event) :: that
    if (LLT(this % time_, that % time_)) then
        less = .TRUE.
    else if (LGT(this % time_, that % time_)) then
        less = .FALSE.
    else 
        if (this % action_ /= that % action_) then
            less = ActionHandler % less(this % action_, that % action_)
        else
            less = LLT(this % arr_name_, that % arr_name_)
        end if
    end if
end function

!========================================================================================

! Should be static in normal language
character(NAME_LEN) function make_ga_name(this, owner, arr_name, action)
    class(Event) :: this
    character(*), intent(in) :: owner, arr_name, action
    make_ga_name = TRIM(owner)//"_"//TRIM(arr_name)//"_"//TRIM(action)
end function

!========================================================================================

character(100) function to_string_event(this)
    class(Event) :: this
    
    write(to_string_event, '(a,a7,a12,a)') "["//TRIM(this % time_)//"] ", TRIM(this % action_), &
                                  "<"//TRIM(this % arr_name_)//">", &
                                  " ("//TRIM(this % src_)//" -> "//TRIM(this % dst_)//")"
end function

!========================================================================================

subroutine set_time(this, time)
    class(Event) :: this
    character(*), intent(in) :: time
    this % time_ = time
end subroutine

subroutine set_time_rec(this, time_rec)
    class(Event) :: this
    character(*), intent(in) :: time_rec
    this % time_rec_ = time_rec
end subroutine

character(NAME_LEN) function get_time(this)
    class(Event) :: this
    get_time = this % time_
end function

character(NAME_LEN) function get_arr_name(this)
    class(Event) :: this
    get_arr_name = this % arr_name_
end function

character(NAME_LEN) function get_action(this)
    class(Event) :: this
    get_action = this % action_
end function

character(NAME_LEN) function get_array_type(this)
    class(Event) :: this
    get_array_type = this % arr_type_
end function

character(NAME_LEN) function get_array_dims(this)
    class(Event) :: this
    get_array_dims = this % arr_dims_
end function

character(NAME_LEN) function get_ga_name(this)
    class(Event) :: this
    get_ga_name = this % ga_name_
end function

character(NAME_LEN) function get_src(this)
    class(Event) :: this
    get_src = this % src_
end function

character(NAME_LEN) function get_dst(this)
    class(Event) :: this
    get_dst = this % dst_
end function

subroutine set_dst(this, dst)
    class(Event) :: this
    character(*), intent(in) :: dst
    this % dst_ = dst
end subroutine

character(NAME_LEN) function get_owner(this)
    class(Event) :: this
    get_owner = this % owner_
end function

character(NAME_LEN) function get_time_rec(this)
    class(Event) :: this
    get_time_rec = this % time_rec_
end function

character(NAME_LEN) function get_info(this)
    class(Event) :: this
    get_info = this % info_
end function



! character(NAME_LEN) function origin(this)
! ! 
! !         ! Define author of message
! !         select case(ev % action())
! !             case("SEND_MP", "SAVE_DG", "SAVE_CP")
! !                 origin = ev % src()
! !             case("RECV_MP")
! !                 origin = ev % dst()
! !         end select

!======================================================================================== 
    
end module

#define CONCRETE_TYPE Event
#define CONCRETE_MODULE event_module
#define CONCRETE_MODULE_FLIBS event_vector_module_
#define OUTPUT_MODULE event_vector_module
#include "VectorStorage.f90"


subroutine sort_event_vector(events)
    use event_module
    use event_vector_module, VectorEvent => VectorStorage 
    
    implicit none
    
    type(VectorEvent), intent(in) :: events
    type(Event) :: hold
    type(Event), allocatable :: array(:)
    integer, parameter :: QSORT_THRESHOLD = 96
    integer :: i
    
    allocate(array(events % length()))
    !write(*,*), "unsort_event_vector "
    do i = 1, events % length()
        array(i) = events % get(i)
    !    write(*,*), i, array (i) %  ga_name()
    end do
    call qsort_data()
    !write(*,*), "sort_event_vector " 
    do i = 1, events % length()
        call events % put_at(i, array(i))
    !    write(*,*),  i, array (i) %  ga_name()
    end do   
    deallocate(array)

 CONTAINS

    subroutine QSORT_DATA
        include "qsort_inline.f90"
    end subroutine QSORT_DATA

    logical function QSORT_COMPARE(a,b)
        integer, intent(in) :: a,b
        QSORT_COMPARE =  array(a) % less(array(b))
    end function QSORT_COMPARE

end subroutine
