!========================================================================================
!   Class-facade for work with netCDF. 
!
!   Simplify interaction with netCDF file for case "open-put/get-close",
!   because store necessary ids in private members.
!
!   Also control last record of time vars using attributes for it (because
!   different length for time vars is not supporrted in netCDF4.0 yet)
!========================================================================================

module file_handler_nc_module

use netcdf
use const_module
use mpi
use datetime_module
use utils_array_module

implicit none

integer, parameter :: NC_NULL = -1
character(*), parameter :: TIME_NAME = "time"

type :: FileHandler_NC
private
    integer :: file_id_ = NC_NULL 
    integer :: var_id_ = NC_NULL
    integer :: time_id_ = NC_NULL
    character(100) :: file_name_ = ""
    character(100) :: var_name_ = ""
    
    integer(8), allocatable :: time_axis_(:)

 contains

    !=====================
    ! ==== PUBLIC API ====
    procedure, public :: open_file
    procedure, public :: create_file
    procedure, public :: open_or_create_file
    procedure, public :: close_file
    
    procedure, public :: create_dim
    procedure, public :: create_time_dim
    
    procedure, public :: open_var
    procedure, public :: create_var
    procedure, public :: open_or_create_var
    generic,   public :: put => put_nc_1D_int, put_nc_1D_int8, put_nc_1D_real4, put_nc_1D_real8, put_nc_2D_int, put_nc_2D_real4, put_nc_2D_real8, put_nc_3D_real4, put_nc_3D_real8
    generic,   public :: get => get_nc_1D_int2, get_nc_1D_int4, get_nc_1D_int8, get_nc_1D_real4, get_nc_1D_real8, get_nc_2D_int, get_nc_2D_real4, get_nc_2D_real8, get_nc_3D_real4, get_nc_3D_real8

    generic,   public :: put_att => put_att_nc_int, put_att_nc_real4, put_att_nc_str
    generic,   public :: get_att => get_att_nc_int, get_att_nc_real4, get_att_nc_str
    
    procedure, public :: get_dim_size
    procedure, public :: get_time_axis
    procedure, public :: is_time_var
    procedure, public :: flush
        
    ! ==== PUBLIC API ====
    !=====================
    
    procedure, private :: read_time_axis
    procedure, private :: put_time
    procedure, private :: find_time
    procedure, private :: clean
    
    procedure, private :: get_nc_1D_int2
    procedure, private :: get_nc_1D_int4
    procedure, private :: get_nc_1D_int8
    procedure, private :: get_nc_1D_real4
    procedure, private :: get_nc_1D_real8
    procedure, private :: get_nc_2D_int
    procedure, private :: get_nc_2D_int8
    procedure, private :: get_nc_2D_real4
    procedure, private :: get_nc_2D_real8
    procedure, private :: get_nc_3D_real4
    procedure, private :: get_nc_3D_real8

    procedure, private :: put_nc_1D_int
    procedure, private :: put_nc_1D_int8
    procedure, private :: put_nc_1D_real4
    procedure, private :: put_nc_1D_real8
    procedure, private :: put_nc_2D_int
    procedure, private :: put_nc_2D_int8
    procedure, private :: put_nc_2D_real4
    procedure, private :: put_nc_2D_real8
    procedure, private :: put_nc_3D_real4
    procedure, private :: put_nc_3D_real8
    
    procedure, private :: put_att_nc_int
    procedure, private :: put_att_nc_real4
    procedure, private :: put_att_nc_str
    procedure, private :: get_att_nc_int
    procedure, private :: get_att_nc_real4
    procedure, private :: get_att_nc_str
        
    
end type

 CONTAINS
 
!========================================================================================
!----------------------------------------------------------------------------------------
! Description:      Try to open file
! Parameters:   
!   filename        - string name of file to open
!   mpi_comm    - <optional> MPI mpi_comm if it is parallel run
!   status          - <optional> status of operation: 0 if ok, 1 is error
!----------------------------------------------------------------------------------------

subroutine open_file(this, filename, mpi_comm, status)
    class(FileHandler_NC) :: this
    character(*), intent(in) :: filename
    integer, intent(in), optional :: mpi_comm
    integer, intent(inout), optional :: status
    integer :: status_
   
    ! Clean previous data
    call this % clean()    
   
    this % file_name_ = filename
    
    if (PRESENT(mpi_comm)) then
        status_ = NF90_OPEN(path = TRIM(filename), mode = IOR(NF90_WRITE, IOR(NF90_NETCDF4, NF90_MPIIO)),      &
                                    ncid = this % file_id_, comm = mpi_comm, info = MPI_INFO_NULL)                 
    else
        status_ = NF90_OPEN(path = TRIM(filename), mode = IOR(NF90_WRITE, NF90_NETCDF4),      &
                                    ncid = this % file_id_)                 
    end if
    
    ! Save time axis in object
    call this % read_time_axis()
    
    if(PRESENT(status)) then
        status = status_
        if (status_ == NF90_NOERR) status = NO_ERROR
    else    
        if (status_ /= NF90_NOERR) then
            call check_nc(NC_NULL, "Cant open file <"//TRIM(filename)//">.") 
        end if
    end if    
end subroutine

!========================================================================================
!----------------------------------------------------------------------------------------
! Description:      Create file or rewrite previous
! Parameters:   
!   filename        - string name of file to create
!   mpi_comm        - <optional> MPI mpi_comm if it is parallel run
!---------------------------------------------------------------------------------------- 

subroutine create_file(this, filename, mpi_comm)
    class(FileHandler_NC) :: this
    character(*), intent(in) :: filename
    integer, intent(in), optional :: mpi_comm
    
    ! Clean previous data
    call this % clean()    
    
    this % file_name_ = filename
    
    if (PRESENT(mpi_comm)) then
        call check_nc(NF90_CREATE(path = TRIM(filename), cmode = IOR(NF90_CLOBBER, IOR(NF90_NETCDF4, NF90_MPIIO)), &
                                    ncid = this % file_id_, comm = mpi_comm, info = MPI_INFO_NULL), &
                                    "CREATE")                                            
    else 
        call check_nc(NF90_CREATE(path = TRIM(filename), cmode = IOR(NF90_CLOBBER, NF90_NETCDF4), &
                                    ncid = this % file_id_), &
                                    "CREATE")
    end if

    ! Global attribute for conventions
    call check_nc( NF90_PUT_ATT(this % file_id_, NF90_GLOBAL, "conventions", "CF-1.7"), "PUT_GLOB_ATT" )      
    call check_nc( NF90_PUT_ATT(this % file_id_, NF90_GLOBAL, "_FillValue", missval), "PUT_GLOB_ATT" )      
    call check_nc( NF90_PUT_ATT(this % file_id_, NF90_GLOBAL, "missing_value", missval), "PUT_GLOB_ATT" )
    call check_nc( NF90_ENDDEF(this % file_id_), "ENDDEF" )                        
end subroutine

!========================================================================================
!----------------------------------------------------------------------------------------
! Description:      Try open and then create file
! Parameters:       Combination of parameters for open_file and create_file procedures
!---------------------------------------------------------------------------------------- 

subroutine open_or_create_file(this, filename, mpi_comm)
    class(FileHandler_NC) :: this
    character(*), intent(in) :: filename
    integer, intent(in), optional :: mpi_comm
    integer :: status
    
    call this % open_file(filename, mpi_comm, status)
    if (status /= NO_ERROR) call this % create_file(filename, mpi_comm)    
end subroutine

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      Save records and close file with current file_id
! Parameters:   
!----------------------------------------------------------------------------------------

subroutine close_file(this)
    class(FileHandler_NC) :: this
    integer :: rc
    call check_nc (NF90_CLOSE(this % file_id_), "CLOSE")    
    call this % clean()
end subroutine

!========================================================================================
!----------------------------------------------------------------------------------------
! Description:      add NC-dimension to file
! Parameters:
!   name            - name of dimension
!   length          - corresponding length of dimension
!----------------------------------------------------------------------------------------

subroutine create_dim(this, name, length)
    class(FileHandler_NC) :: this
    character(*), intent(in) :: name
    integer, intent(in) :: length
    integer :: dimid

    if (NF90_INQ_DIMID(this % file_id_, name, dimid) == NF90_NOERR) return
    
    ! No such dimension, create it
    call check_nc( NF90_REDEF(this % file_id_), "REDEF" )      
    call check_nc(NF90_DEF_DIM(this % file_id_, name, length, dimid), "DEF_DIM")
    call check_nc( NF90_ENDDEF(this % file_id_), "ENDDEF" )    
        
end subroutine

!========================================================================================
!----------------------------------------------------------------------------------------
! Description:      add time NC-dimension to file
! Parameters:
!----------------------------------------------------------------------------------------

subroutine create_time_dim(this)
    class(FileHandler_NC) :: this
    integer :: dimid
    type(DateTime) :: dt

    if (NF90_INQ_DIMID(this % file_id_, TIME_NAME, dimid) == NF90_NOERR) return
    
    call check_nc( NF90_REDEF(this % file_id_), "REDEF" )    
    call check_nc( NF90_DEF_DIM(this % file_id_, TIME_NAME, NF90_UNLIMITED, dimid), "DEF TIME DIM" )
    call check_nc( NF90_DEF_VAR(this % file_id_, TIME_NAME, NF90_DOUBLE, dimid, this % time_id_), "DEF_VAR") 
    call check_nc( NF90_PUT_ATT(this % file_id_, this % time_id_, "long_name", TIME_NAME), "PUT_ATT" )       
    call check_nc( NF90_PUT_ATT(this % file_id_, this % time_id_, "standart_name", TIME_NAME), "PUT_ATT" )
    dt = epoch_start()
    call check_nc( NF90_PUT_ATT(this % file_id_, this % time_id_, "units", "seconds since "//TRIM(dt % to_string())), "PUT_ATT" ) 
    call check_nc( NF90_ENDDEF(this % file_id_), "ENDDEF" )        
    
end subroutine

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      Try to open variable
! Parameters:   
!   var_name        - string name of variable to open
!   status          - <optional> status of operation: 0 if ok, 1 is error
!----------------------------------------------------------------------------------------    

subroutine open_var(this, var_name, status)
    class(FileHandler_NC) :: this
    character(*), intent(in) :: var_name    
    integer, intent(inout), optional :: status    
    integer :: status_
    
    this % var_name_ = var_name
    
    status_ = NF90_INQ_VARID(this % file_id_, var_name, this % var_id_)
    if(PRESENT(status)) then
        status = status_
        if (status_ == NF90_NOERR) status = NO_ERROR
        return
    else    
        if (status_ /= NF90_NOERR) then
            call check_nc(NC_NULL, "Cant find variable <"//TRIM(var_name)//"> in file <"//TRIM(this % file_name_)//">")
        end if
    end if
end subroutine

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      Try to create var. Error if this var is already exist.
! Parameters:   
!   var_name        - string name of file to create
!   var_type         - type of variable
!   dim*_name       - create variable of these dimensions
!---------------------------------------------------------------------------------------- 

subroutine create_var(this, var_name, var_type, dim1_name, dim2_name, dim3_name, dimt_name)
    class(FileHandler_NC) :: this
    character(*), intent(in) :: var_name, var_type
    character(*), intent(in), optional :: dim1_name, dim2_name, dim3_name, dimt_name
    integer :: dimids(4), dc
    integer :: var_type_
    
    this % var_name_ = var_name

    select case (var_type)
        case("real4")
            var_type_ = NF90_REAL
        case("real8")
            var_type_ = NF90_DOUBLE
        case("int")
            var_type_ = NF90_INT
        case default
            call check_nc(-1, "unknown datatype <"//TRIM(var_type)//">")
    end select

    dc = 0
    if (PRESENT(dim1_name)) then
        dc = dc + 1        
        call check_nc( NF90_INQ_DIMID(this % file_id_, dim1_name, dimids(dc)), "INQ_DIMID")
    end if
    if (PRESENT(dim2_name)) then
        dc = dc + 1
        call check_nc( NF90_INQ_DIMID(this % file_id_, dim2_name, dimids(dc)), "INQ_DIMID")
    end if
    if (PRESENT(dim3_name)) then
        dc = dc + 1
        call check_nc( NF90_INQ_DIMID(this % file_id_, dim3_name, dimids(dc)), "INQ_DIMID")
    end if
    if (PRESENT(dimt_name)) then
        dc = dc + 1
        call check_nc( NF90_INQ_DIMID(this % file_id_, TIME_NAME, dimids(dc)), "INQ_DIMID")
    end if
        
    call check_nc( NF90_REDEF(this % file_id_), "REDEF" )
    call check_nc( NF90_DEF_VAR(this % file_id_, var_name, var_type_, dimids(1:dc), this % var_id_ ), "DEF_VAR <"//var_name//">")
    call check_nc( NF90_ENDDEF(this % file_id_), "ENDDEF" )

    ! Initialize length of time variable
    if (PRESENT(dimt_name)) call this % put_att("len", 0)

end subroutine

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      try open and then create variable
! Parameters:       combination of parameters for open_var and create_var procedures
!----------------------------------------------------------------------------------------   
subroutine open_or_create_var(this, varname, var_type, dim1_name, dim2_name, dim3_name, dimt_name)
    class(FileHandler_NC) :: this
    character(*), intent(in) :: varname, var_type
    character(*), intent(in), optional :: dim1_name, dim2_name, dim3_name, dimt_name
    integer :: status
    
    call this % open_var(varname, status)
    if (status /= NO_ERROR) call this % create_var(varname, var_type, dim1_name, dim2_name, dim3_name, dimt_name)
end subroutine

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      put and get data.
! Parameters:
!   arr             - data array
!   lo, hi          - lower and upper bounds of dimensions (e.g. (/1, 1/), (/ il, jl /) )
!   dt              - (optional) DateTime corresponding to field. Necessary for time vars.
!----------------------------------------------------------------------------------------
! #TODO Replace it when Fortran will have template system

#define _PUT_NC_(_FUNC_, _TYPE_, _DIMS_)                                     \
subroutine _FUNC_ (this, arr, lo, hi, dt);                                   \
    class(FileHandler_NC) :: this;                                           \
    _TYPE_ , _DIMS_, intent(in) :: arr;                                      \
    integer, intent(in) :: lo(:), hi(:);                                     \
    type(DateTime), intent(in), optional :: dt;                              \
    integer :: cur_len, ind; \
    if (this % is_time_var()) then; \
        if (PRESENT(dt)) then; \
            call this % get_att("len", cur_len); \
            ind = this % put_time(dt, cur_len); \
            call check_nc( NF90_PUT_VAR(this % file_id_, this % var_id_, arr, (/ lo, ind /), (/ hi - lo + 1, 1 /)), "PUT_VAR <"//TRIM(this % var_name_)//">, file <"//TRIM(this % file_name_)//">"); \
            call this % put_att("len", MAX(cur_len, ind)); \
        else; \
            call check_nc(-1, "should specify <dt> parameter when put to time variable."); \
        end if; \
    else; \
        if (PRESENT(dt)) then; \
            call check_nc(-1, "try to put time to non-time variable"); \
        else; \
            call check_nc( NF90_PUT_VAR(this % file_id_, this % var_id_, arr, lo, hi - lo + 1), "PUT_VAR"); \
        end if; \
    end if;                                                                  \
end subroutine

_PUT_NC_(put_nc_1D_int,    integer, dimension(:))
_PUT_NC_(put_nc_1D_int8, integer(8),dimension(:))
_PUT_NC_(put_nc_1D_real4,  real(4), dimension(:))
_PUT_NC_(put_nc_1D_real8,  real(8), dimension(:))
_PUT_NC_(put_nc_2D_int,    integer, dimension(:,:))
_PUT_NC_(put_nc_2D_int8,integer(8), dimension(:,:))
_PUT_NC_(put_nc_2D_real4,  real(4), dimension(:,:))
_PUT_NC_(put_nc_2D_real8,  real(8), dimension(:,:))
_PUT_NC_(put_nc_3D_real4,  real(4), dimension(:,:,:))
_PUT_NC_(put_nc_3D_real8,  real(8), dimension(:,:,:))

!======================================================================================== 

#define _GET_NC_(_FUNC_, _TYPE_, _DIMS_)                                     \
subroutine _FUNC_ (this, arr, lo, hi, dt);                                   \
    class(FileHandler_NC) :: this;                                           \
    _TYPE_ , _DIMS_, intent(inout) :: arr;                                   \
    integer, intent(in) :: lo(:), hi(:);                                     \
    type(DateTime), intent(in), optional :: dt;                              \
    integer :: get_rec;                                                      \
    if (this % is_time_var()) then; \
        if (PRESENT(dt)) then;   \
            get_rec = this % find_time(dt); \
            if (get_rec == 0) call check_nc(-1, "cant find time "//dt % to_string()//" in file "// this % file_name_); \
            call check_nc( NF90_GET_VAR(this % file_id_, this % var_id_, arr, (/ lo, get_rec /), (/ hi - lo + 1, 1 /)), "GET_VAR"); \
        else; \
            write(*,*) "should specify <dt> parameter when get from time variable.",this % file_id_; \
            call check_nc( NF90_GET_VAR(this % file_id_, this % var_id_, arr, (/ lo, get_rec /), (/ hi - lo + 1, 1 /)), "GET_VAR"); \
        end if; \
    else; \
        call check_nc( NF90_GET_VAR(this % file_id_, this % var_id_, arr, (/ lo /), (/ hi - lo + 1 /)), "GET_VAR"); \
    end if; \
end subroutine
            
!write(*,*),"IAFv2 time", dt % to_string(), TRIM(this % file_name_) ; \

_GET_NC_(get_nc_1D_int2,integer(2), dimension(:))
_GET_NC_(get_nc_1D_int4,   integer, dimension(:))
_GET_NC_(get_nc_1D_int8,integer(8), dimension(:))
_GET_NC_(get_nc_1D_real4,  real(4), dimension(:))
_GET_NC_(get_nc_1D_real8,  real(8), dimension(:))
_GET_NC_(get_nc_2D_int,    integer, dimension(:,:))
_GET_NC_(get_nc_2D_int8,integer(8), dimension(:,:))
_GET_NC_(get_nc_2D_real4,  real(4), dimension(:,:))
_GET_NC_(get_nc_2D_real8,  real(8), dimension(:,:))
_GET_NC_(get_nc_3D_real4,  real(4), dimension(:,:,:))
_GET_NC_(get_nc_3D_real8,  real(8), dimension(:,:,:))


!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      Check if variable has a time axis
! Parameters:   
!----------------------------------------------------------------------------------------

logical function is_time_var(this) 
    class(FileHandler_NC) :: this
    integer :: dimnum, dimid(4), tdimid, rc
    call check_nc(NF90_INQUIRE_VARIABLE(this % file_id_, this % var_id_, ndims = dimnum), "INQ_VAR")
    call check_nc(NF90_INQUIRE_VARIABLE(this % file_id_, this % var_id_, dimids = dimid), "INQ_VAR")
    rc = NF90_INQ_DIMID(this % file_id_, TIME_NAME, tdimid)
    is_time_var = rc == NF90_NOERR .AND. dimid(dimnum) == tdimid
end function

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      Get time axis
! Parameters:   
!   time_axis       - integer(8), allocatable :: time_axis(:) - where to put time axis in 
!                     seconds from DateTime % epoch_start()
!----------------------------------------------------------------------------------------

subroutine get_time_axis(this, time_axis)
    class(FileHandler_NC) :: this
    integer(8), allocatable :: time_axis(:)
    
    allocate(time_axis(this % get_dim_size(TIME_NAME)))
    time_axis = this % time_axis_
    
end subroutine

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      Return size of interested dimension
! Parameters:   
!   dim1_name       - dimension name
!----------------------------------------------------------------------------------------

integer function get_dim_size(this, dim_name)
    class(FileHandler_NC) :: this
    character(*), intent(in) :: dim_name
    integer :: dimid
    call check_nc(NF90_INQ_DIMID(this % file_id_, dim_name, dimid), "INQ_DIMID <"//dim_name//">")
    call check_nc(NF90_INQUIRE_DIMENSION(this % file_id_, dimid, len = get_dim_size), "INQ_DIMENSION")
end function

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      put/get attribute to variable or whole file
! Parameters:   
!   att_name        - name of attribute
!   att_val         - value of attribute
!   is_global       - (optional) if .TRUE., this attribute is made NF90_GLOBAL
!---------------------------------------------------------------------------------------- 

#define _PUT_ATT_(_FUNC_, _TYPE_)                                     \
subroutine _FUNC_ (this, att_name, att_val, is_global);                      \
    class(FileHandler_NC) :: this;                                           \
    character(*), intent(in) :: att_name; \
    _TYPE_, intent(in) :: att_val; \
    logical, intent(in), optional :: is_global; \
    integer :: varid; \
    if (PRESENT(is_global) .AND. is_global) then; \
        varid = NF90_GLOBAL; \
    else; \
        varid = this % var_id_; \
    end if; \
    call check_nc( NF90_REDEF(this % file_id_), "REDEF" ); \
    call check_nc( NF90_PUT_ATT(this % file_id_, varid, att_name, att_val), "put_att" ); \
    call check_nc( NF90_ENDDEF(this % file_id_), "ENDDEF" ); \
end subroutine

_PUT_ATT_(put_att_nc_int, integer)
_PUT_ATT_(put_att_nc_real4, real(4))
_PUT_ATT_(put_att_nc_str, character(*))


#define _GET_ATT_(_FUNC_, _TYPE_)                                     \
subroutine _FUNC_ (this, att_name, att_val, is_global);                      \
    class(FileHandler_NC) :: this;                                           \
    character(*), intent(in) :: att_name; \
    _TYPE_, intent(inout) :: att_val; \
    logical, intent(in), optional :: is_global; \
    integer :: varid; \
    if (PRESENT(is_global) .AND. is_global) then; \
        varid = NF90_GLOBAL; \
    else; \
        varid = this % var_id_; \
    end if; \
    call check_nc( NF90_GET_ATT(this % file_id_, varid, att_name, att_val), "get_att "// att_name // "from "// this % file_name_); \
end subroutine

_GET_ATT_(get_att_nc_int, integer)
_GET_ATT_(get_att_nc_real4, real(4))
_GET_ATT_(get_att_nc_str, character(*))

!======================================================================================== 

subroutine flush(this)
    class(FileHandler_NC) :: this
    call check_nc( NF90_SYNC(this % file_id_), "flush")
end subroutine

!======================================================================================== 

!======================================================================================== 

!======================================================================================== 
! PRIVATE
!======================================================================================== 

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      Get time axis to parameter array <time_axis> (also allocate memory for it).
!                   Recognize data in "seconds since " and "days since ".
!                   Output data in "seconds since " format.
! Parameters:       
!   time_axis       - allocatable integer(8) array which will be filled with time info 
!----------------------------------------------------------------------------------------

subroutine read_time_axis(this)
    class(FileHandler_NC) :: this
    real(8), allocatable :: time_axis_days(:)
    integer(8), allocatable :: time_axis_hours(:)
    character(100) :: units
    integer :: t_size, date_ind, t
    type(DateTime) :: dt_since
    integer :: status
    integer, parameter :: SECS_IN_DAY = 60*60*24
    integer, parameter :: SECS_IN_HOUR = 60*60
    integer, parameter :: T_BIG_TIME_SIZE   = 90520  ! IAFv2 t_10,q_10,u_10,v_10 size
    integer, parameter :: RAD_BIG_TIME_SIZE = 22630 ! IAFv2 t_10,q_10,u_10,v_10 size
    integer i
    
    ! Open var and get units string
    call this % open_var(TIME_NAME, status = status)
    this % time_id_ = this % var_id_
    if (status /= NF90_NOERR) return
    
    t_size = this % get_dim_size(TIME_NAME)
    call this % get_att("units", units)

    allocate(this % time_axis_(t_size))
    
    ! Decide what format is used
    if (INDEX(units, "days since") /= 0) then
        date_ind = LEN("days since ") + 1
        allocate(time_axis_days(t_size))
!TODO :This is bad WorkAround for Data with big t_size (IAFv2 e.g.)
!TODO: Make Hashtable for file_handlers
        if( t_size == T_BIG_TIME_SIZE ) then
            do i =1, t_size
                time_axis_days (i) = 0.125d+0 + (i -1 ) * 0.25d+0
            enddo
        else if ( t_size == RAD_BIG_TIME_SIZE) then
            do i =1, t_size
                time_axis_days (i) = 0.5d+0 + (i -1 ) * 1.0d+0
            enddo
        else
            call check_nc( NF90_GET_VAR(this % file_id_, this % time_id_, time_axis_days, (/ 1 /), (/ t_size /)), "GET_TIME")
        endif
!        write(*,*),"Time axis days",TRIM(this % file_name_), time_axis_days(1:min(3,t_size))
        this % time_axis_ = NINT(time_axis_days*SECS_IN_DAY)
        deallocate(time_axis_days)
    else if (INDEX(units, "seconds since") /= 0) then
        date_ind = LEN("seconds since ") + 1    
        call check_nc( NF90_GET_VAR(this % file_id_, this % time_id_, this % time_axis_, (/ 1 /), (/ t_size /)), "GET_TIME");
    else if (INDEX(units, "hours since") /= 0) then
        date_ind = LEN("hours since ") + 1    
        allocate(time_axis_hours(t_size))
        call check_nc( NF90_GET_VAR(this % file_id_, this % time_id_, time_axis_hours, (/ 1 /), (/ t_size /)), "GET_TIME");
        this % time_axis_ = time_axis_hours*SECS_IN_HOUR
        deallocate(time_axis_hours)
    else
        call check_nc(-1, "unknown units of time axis. Supported: days since date, seconds since date, hours since date.") 
    end if
    
    ! Convert it to absoulute axis of seconds since NULL_YEAR (DateTime)
    dt_since = DateTime(units(date_ind:LEN(units)))
    this % time_axis_ = this % time_axis_ + date2sec(dt_since)
    
end subroutine


!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      store date in seconds from <EPOCH_START> in time axis.
!                   If found timestep in axis, just return this index.
!                   If not, check 1) increasing order of time and return next index
!                   and 2) in case of time_len > var_len <dt> should be already in axis.
! Paramaters:
!   dt              - DateTime object of time
!   var_len         - current length of variable
!----------------------------------------------------------------------------------------
integer function put_time(this, dt, var_len) 
    class(FileHandler_NC) :: this
    type(DateTime), intent(in) :: dt
    integer, intent(in) :: var_len
    integer(8) :: new_time
    integer :: ind, time_len
    
    ! Try to binary search timestamp in time_axis   
    time_len = this % get_dim_size(TIME_NAME)
    new_time = date2sec(dt)

    if (time_len == 0) then
        if (.NOT. allocated(this % time_axis_)) allocate(this % time_axis_(1)); this % time_axis_ = -1
        ind = 0
    else
        ind = search_binary(new_time, this % time_axis_, 1, time_len)
    end if
    
    ! If not found, check increasing order and put to the end
    if (ind == 0) then
        ind = time_len + 1

        ! If time_len > var_len and we dont found this time in axis, therefore this is incosistent time
        if (time_len > var_len) call check_nc(-1, "you try to put time incosistent vars into one file.") 
        
        ! Check increasing order        
        if (time_len > 0) then
            if (new_time < this % time_axis_(time_len)) call check_nc(-1, "you can add time data only in ascending order.") 
        end if
        
        ! This is ugly vector.append()
        if (ind > SIZE(this % time_axis_)) call resize_array(this % time_axis_, SIZE(this % time_axis_)*2)

        ! Put data to index and store it in file
        this % time_axis_(ind) = new_time
        call check_nc( NF90_PUT_VAR(this % file_id_, this % time_id_, (/ new_time /), (/ ind /), (/ 1 /)), "PUT_TIME");
    end if
    
    put_time = ind
    
end function

!----------------------------------------------------------------------------------------
! Description:      read time axis and try to find index of <dt> time (0 if fail)
!----------------------------------------------------------------------------------------
integer function find_time(this, dt) 
    class(FileHandler_NC) :: this
    type(DateTime), intent(in) :: dt

    ! Binary search value
    find_time = search_binary(date2sec(dt), this % time_axis_, 1, this % get_dim_size(TIME_NAME))
    !write(*,*),"FileHandler: IAFv2 Find time dt",find_time,TRIM( this % file_name_) , TRIM( this % var_name_ )
    if( this % get_dim_size(TIME_NAME) == 1 .and. find_time == 0) then
        write(*,*),"File_handler warning:wrong date, but one", TRIM(dt % to_string())
        find_time = 1
    endif
end function

subroutine clean(this)
    class(FileHandler_NC) :: this
    
    if(ALLOCATED(this % time_axis_)) deallocate(this % time_axis_)
    this % file_id_ = NC_NULL
    this % var_id_ = NC_NULL
    this % time_id_ = NC_NULL
    this % file_name_ = ""
    this % var_name_ = ""
    
end subroutine
    

!======================================================================================== 
!----------------------------------------------------------------------------------------
! Description:      Private. Check for fatal netCDF error
! Parameters:       
!   status          - status from netCDF procedure
!   msg             - string to print before exit
!----------------------------------------------------------------------------------------

subroutine check_nc(status, msg)

    integer, intent(in) :: status
    character(*), intent(in), optional :: msg
    character(300) :: msg_
    
    msg_ = MERGE(msg, "", PRESENT(msg))
    
    if (status == NC_NULL) then
        write(*,*) "FileHandler_NC: "//TRIM(msg_)
        call exit(1)
    else if (status /= NF90_NOERR) then
        write(*,*) "FileHandler_NC: "//TRIM(msg_)//": "//TRIM(NF90_STRERROR(status))
        call exit(1)
    end if

end subroutine  
       
!======================================================================================== 


end module
