
module __MODULE_NAME__

use mpi
use utils_module
use component_splitter_module
use decomposition_module
implicit none

type :: __CLASS_NAME__
private
	type(Decomposition) :: decomp

	integer :: max_halo_width
	
	integer :: srank, nrank, wrank, erank
	logical :: is_icycle
	logical :: is_bsc
	
	__DATA_TYPE__, allocatable :: buffer(:,:,:)
	__DATA_TYPE__, allocatable, dimension(:) :: vs_send, vn_send, vw_send, ve_send, vs_recv, vn_recv, vw_recv, ve_recv
	
	integer, allocatable, dimension(:,:,:) :: request_lon, request_lat
	integer, allocatable ::  status(:,:)
	
 contains
    procedure, public :: init => init_halo_updater_base
    procedure, public :: destroy => destroy_halo_updater_base
    procedure, public :: update => perform_halo_update
    
    procedure, private :: compute_neighbours
    procedure, private :: perform_mpi_inits
    procedure, private :: perform_west_east_update
    procedure, private :: perform_south_north_update
    
end type	
	
	! Work common vars
    integer, private :: i, j, k, counter, wb, eb, sb, nb, kb, ierr
    integer, private :: dim_num, change_sign_on_bipolar_, upd_w
    character(1), private :: grd_type
	
	CONTAINS

!========================================================================================
!----------------------------------------------------------------------------------------
! Initialize halo_updater using Decomposition object and max width of updates in future
!----------------------------------------------------------------------------------------

subroutine init_halo_updater_base(this, decomp, max_halo_width)
    class(__CLASS_NAME__) :: this
    type(Decomposition), intent(in) :: decomp
    integer, intent(in) :: max_halo_width
    
    call this % destroy()
    
    this % decomp = decomp
    this % max_halo_width = max_halo_width
    this % is_icycle = decomp % is_icycle()
    this % is_bsc = decomp % is_bsc()
    
    if (max_halo_width >= decomp % i_size_loc() .OR. max_halo_width >= decomp % j_size_loc()) then
        write(*,*) "HaloUpdater: error: local domain is too small for halo-update."
        call exit(1)
    end if
    
    ! Main buffer
    allocate (this % buffer (decomp % west() - max_halo_width : decomp % east() + max_halo_width + 1, &
                             decomp % south() -max_halo_width : decomp % north() +max_halo_width, &
                             decomp % k_size())); this %  buffer = 0.
                                 
    ! Send-recv buffers (T-exchanges requires special +1 cell)
    allocate (this % vw_send(decomp % k_size()*decomp % j_size_loc()*(max_halo_width+1))); this % vw_send = 0
    allocate (this % ve_send(decomp % k_size()*decomp % j_size_loc()*max_halo_width)); this % ve_send = 0
    allocate (this % vw_recv(decomp % k_size()*decomp % j_size_loc()*max_halo_width)); this % vw_recv = 0
    allocate (this % ve_recv(decomp % k_size()*decomp % j_size_loc()*(max_halo_width+1))); this % ve_recv = 0
    allocate (this % vs_send(decomp % k_size()*(decomp % i_size_loc() + 2*max_halo_width)*max_halo_width)); this % vs_send = 0
    allocate (this % vn_send(decomp % k_size()*(decomp % i_size_loc() + 2*max_halo_width)*max_halo_width)); this % vn_send = 0
    allocate (this % vs_recv(decomp % k_size()*(decomp % i_size_loc() + 2*max_halo_width)*max_halo_width)); this % vs_recv = 0
    allocate (this % vn_recv(decomp % k_size()*(decomp % i_size_loc() + 2*max_halo_width)*max_halo_width)); this % vn_recv = 0
    
    ! MPI buffers
    allocate (this % request_lon(4, 2:3, 1:max_halo_width))
    allocate (this % request_lat(4, 2:3, 1:max_halo_width))
    allocate (this % status(MPI_STATUS_SIZE,4))
    
    ! Other initializations
    call this % compute_neighbours()
    call this % perform_mpi_inits()
        
end subroutine

!========================================================================================

subroutine destroy_halo_updater_base(this)
    class(__CLASS_NAME__) :: this
    
    if (allocated(this % buffer)) then
        deallocate(this % buffer)
        deallocate(this % vs_send, this % vn_send, this % vw_send, this % ve_send, this % vs_recv, this % vn_recv, this % vw_recv, this % ve_recv)
        deallocate(this % request_lat, this % request_lon, this % status)
    end if
end subroutine

!========================================================================================

subroutine perform_halo_update (this, DIM3_arr, DIM2_arr, update_width, grid_type, change_sign_on_bipolar)
    class(__CLASS_NAME__) :: this
    integer, intent(in) :: update_width
    logical, optional :: change_sign_on_bipolar
    character(1), intent(in) :: grid_type

    ! Get array with required halo
    __DATA_TYPE__, optional, intent(inout) :: DIM2_arr(this % decomp % west() - update_width:this % decomp % east() + update_width, &
                                                        this % decomp % south()- update_width:this % decomp % north()+ update_width)
    
    __DATA_TYPE__, optional, intent(inout) :: DIM3_arr(1:this % decomp % k_size(), &
                                                        this % decomp % west() - update_width:this % decomp % east() + update_width, &
                                                        this % decomp % south()- update_width:this % decomp % north()+ update_width)
    
    ! Make shortcuts
    wb = this % decomp % west()
    eb = this % decomp % east()
    sb = this % decomp % south()
    nb = this % decomp % north()
    upd_w = update_width
    grd_type = grid_type
    
    if (PRESENT(change_sign_on_bipolar)) then
        change_sign_on_bipolar_ = MERGE(-1, 1, change_sign_on_bipolar)
    else
        change_sign_on_bipolar_ = 1
    end if
    
    
    ! Make 3D arrays from input data
    if(PRESENT(DIM3_arr)) then
        dim_num = 3
        kb = this % decomp % k_size()
        this % buffer(wb-upd_w : eb+upd_w, sb-upd_w : nb+upd_w, :) = indexing_kij_to_ijk(DIM3_arr)
    end if

    if(PRESENT(DIM2_arr)) then
        dim_num = 2
        kb = 1
        this % buffer(wb-upd_w : eb+upd_w, sb-upd_w : nb+upd_w, 1) = DIM2_arr(:,:)
    end if
    
!     call print_array(this % buffer, "buf at start")
    call this % perform_west_east_update()
    call this % perform_south_north_update()
        
    ! Now we must return changed arrays to parameters of called function
    if(PRESENT(DIM3_arr)) then
        DIM3_arr = indexing_ijk_to_kij(this % buffer(wb-upd_w : eb+upd_w, sb-upd_w : nb+upd_w, :))
    end if

    if(PRESENT(DIM2_arr)) then
        DIM2_arr(:,:) = this % buffer(wb-upd_w : eb+upd_w, sb-upd_w : nb+upd_w, 1)
    end if
    
end subroutine


!========================================================================================
! First in series of two updates: update of width <update_width> with 
! west and east neighbours (special bipolar case for T)
!========================================================================================

subroutine perform_west_east_update(this)
    class(__CLASS_NAME__) :: this

    ! Create sending-arrays for west-east(+bsc) case 
    counter = 1
    do k = 1, kb
    do j = sb, nb
    do i = 1, upd_w
        this % vw_send(counter) = this % buffer(wb+i-1, j, k)
        counter = counter + 1
    end do
    end do
    end do
    
    counter = 1
    do k = 1, kb
    do j = sb, nb
    do i = 1, upd_w
        this % ve_send(counter) = this % buffer(eb-upd_w+i, j, k)
        counter = counter + 1
    end do
    end do
    end do
    
    ! Special case
    if(nb == this % decomp % j_size() .AND. this % is_bsc) then
        if (grd_type == "T") call bsc_special_send_westeastT()
    end if
        
    ! Start INIT-operations
    call MPI_STARTALL (4, this % request_lon(:, dim_num, upd_w) , ierr)
    call MPI_WAITALL  (4, this % request_lon(:, dim_num, upd_w), this % status , ierr)

    ! Copy received belts to work arrays (or just zeroes)
    counter = 1     
    do k = 1, kb
    do j = sb, nb
    do i = 1, upd_w
        this % buffer(wb-upd_w-1+i, j, k) = this % vw_recv(counter)
        counter = counter + 1
    end do
    end do
    end do
        
    counter = 1
    do k = 1, kb
    do j = sb, nb
    do i = 1, upd_w
        this % buffer(eb+i,j,k) = this % ve_recv(counter)
        counter = counter + 1
    end do
    end do
    end do

    ! Special case
    if(nb == this % decomp % j_size() .AND. this % is_bsc) then
        if (grd_type == "T") call bsc_special_recv_westeastT()
    end if
    
 contains

!====================================================== 
 
    subroutine bsc_special_send_westeastT()

        counter = 1
        do k = 1, kb
        do j = sb, nb
        do i = 1, upd_w+1
            this % vw_send(counter) = this % buffer(wb+i-1, j, k)
            counter = counter + 1
        end do
        end do
        end do
        
    end subroutine    
!====================================================== 

    subroutine bsc_special_recv_westeastT()

        counter = 1
        do k = 1, kb
        do j = sb, nb
        do i = 1, upd_w+1
            this % buffer(eb+i,j,k) = this % ve_recv(counter)
            counter = counter + 1
        end do
        end do
        end do
        
    end subroutine  

!====================================================== 
    
end subroutine    

!========================================================================================

!=========================================================
! Now we have belts for west and east and we can get wide
! halo belts for south and north case (without diagonal !)
!=========================================================

subroutine perform_south_north_update(this)
    class(__CLASS_NAME__) :: this

    ! Create sending-arrays for wide south-north(+bsc) case 
    counter = 1
    do k = 1, kb
    do j = 1, upd_w
    do i = wb - upd_w, eb + upd_w
        this % vs_send(counter) = this % buffer(i,sb+j-1, k)
        counter = counter + 1
    end do
    end do
    end do

    counter = 1
    do k = 1, kb
    do j = 1, upd_w
    do i = wb - upd_w, eb + upd_w
        this % vn_send(counter) = this % buffer(i, nb-upd_w+j, k)
        counter = counter + 1
    end do
    end do
    end do
    
    ! Special case
    if(nb == this % decomp % j_size() .AND. this % is_bsc) then
        if (grd_type == "T") call bsc_special_send_southnorthT()
        if (grd_type == "V") call bsc_special_send_southnorthV()
    end if    
    
    ! Start INIT-operations
    call MPI_STARTALL (4, this % request_lat(:, dim_num, upd_w) , ierr)
    call MPI_WAITALL  (4, this % request_lat(:, dim_num, upd_w), this % status , ierr)
    
    ! Receive wide south-north belts (or zeroes)
    counter = 1
    do k = 1, kb
    do j = 1, upd_w
    do i = wb - upd_w, eb + upd_w
        this % buffer(i,sb-upd_w-1+j, k) = this % vs_recv(counter)
        counter = counter + 1
    end do
    end do
    end do
        
    counter = 1
    do k = 1, kb
    do j = 1, upd_w
    do i = wb - upd_w, eb + upd_w
        this % buffer(i,nb+j,k) = this % vn_recv(counter)
        counter = counter + 1
    end do
    end do
    end do
    
    ! Special case
    if(nb == this % decomp % j_size() .AND. this % is_bsc) then
        if (grd_type == "T") call bsc_special_recv_southnorthT()
        if (grd_type == "V") call bsc_special_recv_southnorthV()
    end if    
    

 contains
 
!====================================================== 

    subroutine bsc_special_send_southnorthT()

        counter = 1
        do k = 1, kb
        do j = 1, upd_w
        do i = wb - upd_w+1, eb + upd_w+1
            this % vn_send(counter) = this % buffer(i, nb-upd_w+j, k)
            counter = counter + 1
        end do
        end do
        end do
        
    end subroutine    

!====================================================== 

    subroutine bsc_special_recv_southnorthT()

        counter = 1
        do k = 1, kb
        do j = 1, upd_w
        do i = wb - upd_w, eb + upd_w
            this % buffer(wb+(eb-i), nb+upd_w+1-j, k) = change_sign_on_bipolar_*this % vn_recv(counter)
            counter = counter + 1
        end do
        end do
        end do
        
    end subroutine  

!====================================================== 

    subroutine bsc_special_send_southnorthV()

        counter = 1
        do k = 1, kb
        do j = 1, upd_w
        do i = wb - upd_w, eb + upd_w
            this % vn_send(counter) = this % buffer(i, nb-upd_w-1+j, k)
            counter = counter + 1
        end do
        end do
        end do
        
    end subroutine  

!====================================================== 

    subroutine bsc_special_recv_southnorthV()

        counter = 1
        do k = 1, kb
        do j = 1, upd_w
        do i = wb - upd_w, eb + upd_w
            this % buffer(wb+(eb-i), nb+upd_w+1-j, k) = change_sign_on_bipolar_*this % vn_recv(counter)
            counter = counter + 1
        end do
        end do
        end do
        
    end subroutine   
!====================================================== 

end subroutine

!========================================================================================
!----------------------------------------------------------------------------------------
! Compute south, west, east, north neighbours of this process
! depends on grid configuration
!----------------------------------------------------------------------------------------

subroutine compute_neighbours(this)
    class(__CLASS_NAME__) :: this
    
    integer :: neighbours(4)
    
    neighbours = this % decomp % neighbour_ranks()
    this % wrank = MERGE(neighbours(1), MPI_PROC_NULL, neighbours(1) /= -1)
    this % erank = MERGE(neighbours(2), MPI_PROC_NULL, neighbours(2) /= -1)
    this % srank = MERGE(neighbours(3), MPI_PROC_NULL, neighbours(3) /= -1)
    this % nrank = MERGE(neighbours(4), MPI_PROC_NULL, neighbours(4) /= -1)

end subroutine
    
!========================================================================================
!----------------------------------------------------------------------------------------
! Initialize persistent MPI operations
!----------------------------------------------------------------------------------------

subroutine perform_mpi_inits(this)
    class(__CLASS_NAME__) :: this
    integer :: w, dim_num, size_ext, ierr
!===================================================
! Start big cycle for all types of exchanges: 
! from 1 to comp_halo_stencil and for comp_halo_wide
!===================================================
    do w = 1, this % max_halo_width

        do dim_num = 2, 3

            size_ext = MERGE(this % decomp % k_size(), 1, dim_num == 3)
        
            call MPI_SEND_INIT(this % vw_send,                          &
                    size_ext*(w+1)*(this % decomp % j_size_loc()),          &
                    __MPI_DATATYPE__,                               &
                    this % wrank,                               &
                    201,                                    &
                    CompSplitter % comm_local(),                               &
                    this % request_lon(1,dim_num,w), ierr)
        
            call MPI_SEND_INIT(this % ve_send,                          &
                    size_ext*w*(this % decomp % j_size_loc()),          &
                    __MPI_DATATYPE__,                               &
                    this % erank,                               &
                    202,                                    &
                    CompSplitter % comm_local(),                               &
                    this % request_lon(2,dim_num,w), ierr)
                    
            call MPI_RECV_INIT(this % ve_recv,                          &
                    size_ext*(w+1)*(this % decomp % j_size_loc()),          &
                    __MPI_DATATYPE__,                               &
                    this % erank,                               &
                    MPI_ANY_TAG,                                &
                    CompSplitter % comm_local(),                               &
                    this % request_lon(3,dim_num,w), ierr)
            
            call MPI_RECV_INIT(this % vw_recv,                          &
                    size_ext*w*(this % decomp % j_size_loc()),          &
                    __MPI_DATATYPE__,                               &
                    this % wrank,                               &
                    MPI_ANY_TAG,                                &
                    CompSplitter % comm_local(),                               &
                    this % request_lon(4,dim_num,w), ierr)                 
                                        
            call MPI_SEND_INIT(this % vs_send,                          &
                    size_ext*w*(this % decomp % i_size_loc() + 2*w),    &
                    __MPI_DATATYPE__,                               &
                    this % srank,                           &
                    203,                                    &
                    CompSplitter % comm_local(),                               &
                    this % request_lat(1,dim_num,w), ierr)

            call MPI_SEND_INIT(this % vn_send,                          &
                    size_ext*w*(this % decomp % i_size_loc() + 2*w),    &
                    __MPI_DATATYPE__,                               &
                    this % nrank,                           &
                    204,                                    &
                    CompSplitter % comm_local(),                               &
                    this % request_lat(2,dim_num,w), ierr)

            call MPI_RECV_INIT(this % vn_recv,                          &
                    size_ext*w*(this % decomp % i_size_loc() + 2*w),    &
                    __MPI_DATATYPE__,                               &
                    this % nrank,                           &
                    MPI_ANY_TAG,                                &
                    CompSplitter % comm_local(),                               &
                    this % request_lat(3,dim_num,w), ierr)

            call MPI_RECV_INIT(this % vs_recv,                          &
                    size_ext*w*(this % decomp % i_size_loc() + 2*w),    &
                    __MPI_DATATYPE__,                               &
                    this % srank,                           &
                    MPI_ANY_TAG,                                &
                    CompSplitter % comm_local(),                               &
                    this % request_lat(4,dim_num,w), ierr)
        end do

    end do
end subroutine

!========================================================================================

end module

#undef __MODULE_NAME__ 
#undef __CLASS_NAME__   
#undef __DATA_TYPE__   
#undef __MPI_DATATYPE__