module mapper_module

use decomposition_module
use file_handler_nc_module
use utils_module
use communicator_ga_module
implicit none

!=======================================================================================
! Define set GA interface i4 or i8 ( recomend i8 for x64 OS)
!========================================================================================   
#define GA_IS 8     ! GA Integer Size 8 or 4


integer, parameter :: MAP_NAME_LEN = 20

type :: Mapper
private
    type(Decomposition) :: src_decomp_, dst_decomp_
    
    integer :: link_num_ = -1                                   ! number of links connecting src, dst and wts together
    integer, allocatable :: src_(:), dst_(:)                    ! addresses for source and destination arrays
    real(kind=8), allocatable :: wts_(:,:)                      ! corresponding mapping weights    
    real(kind=4), allocatable :: dst_data_(:)                   ! real physical dst data (result)

    integer, allocatable :: src_unique_coords_(:,:)             ! 2D-coordinates of points in GA of source data (optimized by cutting)
    real(4), allocatable :: src_unique_data_(:)                 ! real physical src data on every step (only unique points from <src_unique_coords_>)
    integer, allocatable :: decoder_src2unique_(:)              ! map src_addresses to src_unique_data_
    integer ::  num_of_dif_
 contains
    procedure, public :: init => init_map                       ! Pass decompositions
    procedure, public :: destroy => destroy_map                 
    procedure, public :: is_initialized                         ! Is all data mapping is read and analize by object?
 
    generic, public :: interpolate => interpolate_parallel, interpoalate_serial

    procedure, public :: analize                                ! Analize input data to choose local domain of every process
    procedure, public :: read_map_file                          ! Read all necessary information from SCRIP-file
    procedure, public :: read_and_analize                       ! Read then analize data
    procedure, public :: get_result_2D                          ! Return interpolated result
    
    procedure, private :: calc_data_coords                      ! Build request set for NGA_GATHER
    procedure, private :: interpolate_parallel                  ! Make parallel interpolation taking unique src_data from GA
    procedure, private :: interpoalate_serial                   ! Make simple interpolation
 
end type
 
 CONTAINS

!========================================================================================
 
subroutine init_map(this, src_decomp, dst_decomp)
    class(Mapper) :: this
    type(Decomposition), intent(in) :: src_decomp, dst_decomp

    this % src_decomp_ = src_decomp
    this % dst_decomp_ = dst_decomp
        
end subroutine

subroutine destroy_map(this)
    class(Mapper) :: this
    
    if (ALLOCATED(this % src_)) then
        deallocate(this % src_, this % dst_, this % wts_)
        deallocate(this % src_unique_coords_, this % src_unique_data_, this % dst_data_, this % decoder_src2unique_)
    end if
end subroutine

logical function is_initialized(this)
    class(Mapper) :: this
    is_initialized = this % link_num_ >= 0
end function

!========================================================================================

subroutine interpoalate_serial(this, src_data_global)
    class(Mapper) :: this
    real(kind=4), intent(in) :: src_data_global(:,:)
    real(kind=4) :: src_1D(SIZE(src_data_global))
    integer :: link, dst_p

    src_1D = convert_2D_to_1D(src_data_global) 
        
    this % dst_data_ = 0  

    do link = 1, this % link_num_
        dst_p = this % dst_(link)
        this % dst_data_(dst_p) = this % dst_data_(dst_p) + this % wts_(1,link)*src_1D(this % src_(link))
    end do
     
end subroutine

!========================================================================================

subroutine interpolate_parallel(this, src_ga_id)
    class(Mapper) :: this
    integer, intent(in) :: src_ga_id  
    integer :: link, dst_lo, dst_p, i, il, jl
    integer(kind=8) num_of_dif
    real(kind=4), allocatable :: src_data_global(:,:)
    real(kind=8) :: tarray(2)
    integer :: lo(2), hi(2), tmp(1,1)
!##### Patch patch patch #############
    il = this % src_decomp_ % i_size()
    jl = this % src_decomp_ % j_size()
    num_of_dif = this % num_of_dif_
    this % dst_data_ = 0
    if ( this % src_decomp_ % proc_count() == 1) then
        allocate ( src_data_global(il,jl) )
        call CPU_TIME (tarray(1))
        call NGA_GET (int(src_ga_id, GA_IS), int((/1,1/),GA_IS), int((/il,jl/),GA_IS), src_data_global, int(il,GA_IS)  ) 
        call CPU_TIME (tarray(2))
        !write(*,'(a,2f8.4,3i6)'), "interpolate_parallel: NGA_GET        ", tarray(2), tarray(2) -tarray(1) ,num_of_dif, il, jl
        do i = 1, num_of_dif
            this % src_unique_data_(i) = src_data_global( this % src_unique_coords_(1,i) , this % src_unique_coords_(2,i) )
        enddo
        !call interpoalate_serial(this, src_data_global)
        dst_lo = this % dst_decomp_ % first_global_index()    
        do link = 1, this % link_num_
            dst_p = this % dst_(link) - dst_lo + 1
            this % dst_data_(dst_p) = this % dst_data_(dst_p) + this % wts_(1,link)* this % src_unique_data_(this % decoder_src2unique_(link))
        end do

        deallocate (src_data_global)
        return
    endif

    if ( this % src_decomp_ % proc_count() > 1) then
        lo(1) = 1
        hi(1) = il
        lo(2) = this % src_unique_coords_(2,1)
        hi(2) = this % src_unique_coords_(2,num_of_dif)
        allocate ( src_data_global(lo(1):hi(1),lo(2):hi(2)) )
        call CPU_TIME (tarray(1))
        call NGA_GET (int(src_ga_id, GA_IS), int(lo,GA_IS), int(hi,GA_IS), src_data_global, int(SIZE(src_data_global, 1), GA_IS)  ) 
        call CPU_TIME (tarray(2))
        !write(*,'(a,2f8.4, 2i6)'), "interpolate_parallel2: NGA_GET        ", tarray(2), tarray(2) -tarray(1) , lo(2), hi(2) 
        
        do i = 1, num_of_dif
            this % src_unique_data_(i) = src_data_global( this % src_unique_coords_(1,i) , this % src_unique_coords_(2,i) )
        enddo
        
        dst_lo = this % dst_decomp_ % first_global_index()    
        do link = 1, this % link_num_
            dst_p = this % dst_(link) - dst_lo + 1
            this % dst_data_(dst_p) = this % dst_data_(dst_p) + this % wts_(1,link)* this % src_unique_data_(this % decoder_src2unique_(link))
        end do
        deallocate (src_data_global)
        return
    endif

end subroutine

!========================================================================================

function get_result_2D(this) result(arr_2D)
    class(Mapper) :: this
    real(kind=4) :: arr_2D(this % dst_decomp_ % i_size_loc(),this % dst_decomp_ % j_size_loc())
    arr_2D = convert_1D_to_2D(this % dst_data_, this % dst_decomp_ % i_size_loc(), this % dst_decomp_ % j_size_loc())
end function

!========================================================================================

subroutine read_and_analize(this, filename)
    class(Mapper) :: this
    character(*), intent(in) :: filename
    integer, allocatable :: src_addr(:), dst_addr(:)
    real (kind = 8), allocatable :: wts_map(:,:)
    
    call this % read_map_file(filename, src_addr, dst_addr, wts_map)
    call this % analize(src_addr, dst_addr, wts_map)    
    
    deallocate(dst_addr, src_addr, wts_map)   
    
end subroutine
!========================================================================================

! #TODO make abstract using decomp % points_set and binary search 
subroutine analize(this, src_addr, dst_addr, wts_map)
    class(Mapper) :: this
    integer, intent(in) :: src_addr(:), dst_addr(:)
    real(kind=8), intent(in) :: wts_map(:,:)
    integer, allocatable :: links(:)
    integer :: p, link, l, dst_start, dst_end
   
    ! Set boundaries of decomp
    dst_start = this % dst_decomp_ % first_global_index()    
    dst_end = dst_start + this % dst_decomp_ % points_num() - 1    

    ! Cycle through all links
    allocate(links(2))    
    this % link_num_ = 0    
    do link = 1, SIZE(dst_addr)
        p = dst_addr(link)
        ! We need only our links
        if (p >= dst_start .AND. p <= dst_end) then
            this % link_num_ = this % link_num_ + 1        
            if (this % link_num_ > SIZE(links)) call resize_array(links, SIZE(links)*2)            
            links(this % link_num_) = link        
        end if
    end do
    call resize_array(links, this % link_num_)

    ! Put only my links to work arrays
    allocate(this % dst_(this % link_num_))
    allocate(this % src_(this % link_num_))
    allocate(this % wts_(1, this % link_num_))
    
    do l = 1, this % link_num_
        this % dst_(l) = dst_addr(links(l))
        this % src_(l) = src_addr(links(l))
        this % wts_(1,l) = wts_map(1,links(l))
    end do
    
    deallocate(links)
    
    allocate(this % dst_data_(this % dst_decomp_ % points_num()))
        
    call this % calc_data_coords()
    
end subroutine

!========================================================================================

subroutine read_map_file(this, filename, src_addr, dst_addr, wts_map) 
    class(Mapper) :: this
    character(*), intent(in) :: filename    
    integer, allocatable, intent(out) :: src_addr(:), dst_addr(:)
    real (kind = 8), allocatable, intent(out) :: wts_map(:,:)
    integer :: num_links
    type(FileHandler_NC) :: handler
    
    call handler % open_file(filename)
    
    num_links = handler % get_dim_size("num_links")

    allocate(src_addr(num_links)); src_addr = 0
    allocate(dst_addr(num_links)); dst_addr = 0
    allocate(wts_map(1, num_links)); wts_map = 0
    
    call handler % open_var("src_address")
    call handler % get(src_addr, (/ 1 /), (/ num_links /))
    call handler % open_var("dst_address")
    call handler % get(dst_addr, (/ 1 /), (/ num_links /))
    call handler % open_var("remap_matrix")
    call handler % get(wts_map(1,:), (/1, 1/) , (/ 1, num_links/) )    
    call handler % close_file()

end subroutine

!========================================================================================

subroutine calc_data_coords(this)
    class(Mapper) :: this
    integer :: link
    integer, allocatable :: src_points_counter(:), unique_addresses(:)
    integer :: num_of_dif, lo, hi, counter, i
        
    ! Count all different addresses of src
    lo = MINVAL(this % src_)
    hi = MAXVAL(this % src_)
    allocate(src_points_counter(lo:hi)); src_points_counter = 0
    
    do i = 1, this % link_num_
        src_points_counter(this % src_(i)) = src_points_counter(this % src_(i)) + 1
    end do

    ! Ask to send us only different points
    num_of_dif = COUNT(src_points_counter > 0)
    write(*,*), 'num_of_dif =', num_of_dif
    allocate(unique_addresses(num_of_dif))
    allocate(this % src_unique_coords_(2, num_of_dif)) ! NGA_GATHER required not adresses but coords
    
    counter = 1
    do i = lo, hi
        if (src_points_counter(i) > 0) then
            unique_addresses(counter) = i
            this % src_unique_coords_(:, counter) = (/ this % src_decomp_ % global_to_i(i), this % src_decomp_ % global_to_j(i) /)
            counter = counter + 1
        end if
    end do    
    deallocate(src_points_counter)
    
    ! Build decoder which maps src addresses to uniqued array
    allocate(this % decoder_src2unique_(this % link_num_))
    do i = 1, this % link_num_
        this % decoder_src2unique_(i) = search_binary(this % src_(i), unique_addresses, 1, SIZE(unique_addresses))
    end do
    deallocate(unique_addresses)
    
    allocate(this % src_unique_data_(num_of_dif))
    this % num_of_dif_ = num_of_dif
end subroutine

!========================================================================================

end module



