!========================================================================================
! Class for storage of model information.
!
! Has singleton array of ModelInfo synchronized through all components.
!========================================================================================

module model_info_module

use const_module
use component_splitter_module, only: NUMBER_OF_COMPONENTS
use decomposition_module
use utils_mpi_module, only: broadcast
implicit none

type :: ModelInfo
private
    character(NAME_LEN) :: comp_name_ = ""
    character(NAME_LEN) :: model_name_ = ""
    type(Decomposition) :: decomp_
    integer :: timestep_sec_ = 0
    
 contains 
    !========================
    ! ===== PUBLIC API ======
    procedure, public :: comp_name => component_name        ! component_name (e.g. OCN or ATM)
    procedure, public :: model_name                         ! model_name (e.g. OCN_INMIO or ATM_NCAR)
    procedure, public :: decomp                             ! type(Decomposition) object of model
    procedure, public :: timestep_sec => get_timestep_sec   ! seconds in one timestep

    procedure, public :: set_model_name
    
    procedure, public :: show                               ! print information about model
    
    procedure, public :: broadcast => broadcast_info        ! broadcast object (used by synchronize client)     
    ! ===== PUBLIC API ======
    !========================
    
end type

    interface ModelInfo
        procedure :: init_comp
    end interface

!========================================================================================      

    type(ModelInfo) :: ModelsInfo(NUMBER_OF_COMPONENTS)

 CONTAINS

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description: constructor
!----------------------------------------------------------------------------------------

function init_comp(comp_name__, model_name__, decomp__, timestep_sec__) result(this)
    type(ModelInfo) :: this
    character(*), intent(in) :: comp_name__, model_name__
    type(Decomposition), intent(in) :: decomp__
    integer, intent(in) :: timestep_sec__
    
    this % comp_name_ = comp_name__
    this % model_name_ = model_name__
    this % decomp_ = decomp__    
    this % timestep_sec_ = timestep_sec__
end function

!========================================================================================      

character(NAME_LEN) function component_name(this)
    class(ModelInfo) :: this
    component_name = this % comp_name_
end function

character(NAME_LEN) function model_name(this)
    class(ModelInfo) :: this
    model_name = this % model_name_
end function 

subroutine set_model_name(this, name)
    class(ModelInfo) :: this
    character(NAME_LEN), intent(in) :: name
    this % model_name_ = name
end subroutine

function decomp(this)
    class(ModelInfo) :: this
    type(Decomposition) :: decomp
    decomp = this % decomp_
end function 

integer function get_timestep_sec(this)
    class(ModelInfo) :: this
    get_timestep_sec = this % timestep_sec_
end function 

!========================================================================================      

subroutine show(this)
    class(ModelInfo) :: this
    write(*,'(a)') '-----------------------------------------------------------------------'
    write(*,'(a, a)')  'Comp name:                    ', TRIM(this % comp_name())
    write(*,'(a, a)')  'Model name:                   ', TRIM(this % model_name())
    write(*,'(a, a)')  'Time step (sec):              ', TRIM(n2s(this % timestep_sec())) 
    call this % decomp_ % show() 
    write(*,'(a)') '-----------------------------------------------------------------------'
    
end subroutine

!========================================================================================      
!----------------------------------------------------------------------------------------
! Description:  broadcast object from root to all. Every process should call it.
! Parameters:
!   root        - rank of root broadcasting process
!   proc_count  - number of current broadcasting process (need to Decomposition)
!----------------------------------------------------------------------------------------

subroutine broadcast_info(this, root)

    class(ModelInfo) :: this
    integer, intent(in) :: root
    
    call broadcast(this % comp_name_, root, MPI_COMM_WORLD)
    call broadcast(this % model_name_, root, MPI_COMM_WORLD)
    call broadcast(this % timestep_sec_, root, MPI_COMM_WORLD)
    
    call this % decomp_ % broadcast(root, MPI_COMM_WORLD)
end subroutine

!========================================================================================      

end module

! Synchronize ModelsInfo array
subroutine model_info_synchronize()
    use model_info_module
    use component_splitter_module
    implicit none
    integer :: comp
    
    do comp = 1, CompSplitter % comp_defined()
        call ModelsInfo(comp) % broadcast(root = CompSplitter % proc_first(CompSplitter % comp_name(comp)))
    end do        
end subroutine
