module offline_module

use const_module
use netcdf
use decomposition_module
use utils_module

implicit none

character(*), parameter :: DATA_FOLDER = "./data/"

!========================================================================================

type, public :: ModelGrid
    character(NAME_LEN) :: name = ""
    character(1) :: grid_type
    integer :: dims(3)
    real (kind = 8), allocatable :: center_lat(:,:), center_lon(:,:)
    real (kind = 8), allocatable :: corner_lat(:,:,:), corner_lon(:,:,:)
    integer, allocatable :: mask2D(:,:)    
    real (kind = 4), allocatable :: level(:)
    
    integer :: dim_num
    integer :: corner_num
 contains
    procedure, public :: init => init_grid
    procedure, public :: dump_to_cdo_format
    procedure, public :: check => check_grid
    
end type

!========================================================================================

type, abstract, public :: Offline
    type(ModelGrid) :: model_grid
    character(NAME_LEN) :: experiment_name

 contains

    procedure, public :: register_model_grid
    procedure, public :: get_ic_filename
    procedure, public :: add_info_to_ic_file
    procedure, public :: analyze_land
    
    procedure(i_make_grid), deferred :: make_grid
    procedure(i_make_ic), deferred :: make_ic   

end type

abstract interface
    subroutine i_make_grid(this, name, dims, grid_type)
        import Offline
        class(Offline) :: this
        character(*), intent(in) :: name, grid_type
        integer, intent(in) :: dims(3)
    end subroutine

    subroutine i_make_ic(this, name, dims, grid_type)
        import Offline
        class(Offline) :: this
        character(*), intent(in) :: name, grid_type
        integer, intent(in) :: dims(3)
    end subroutine
    
end interface

!========================================================================================

 CONTAINS

!========================================================================================
 
subroutine init_grid(this, name, dims, grid_type)
    class(ModelGrid) :: this
    
    character(*), intent(in) :: name
    character(1), intent(in) :: grid_type
    integer, intent(in) :: dims(3)
 
    this % name = name
    this % grid_type = grid_type
    this % dims = dims
    this % dim_num = 2
    this % corner_num = 4
    
    write(*,'(A$)') "Initializing grid: "//TRIM(this % name)//'_'//TRIM(n2s(this % dims(1)))//'x'//TRIM(n2s(this % dims(2)))//"..."
    
    allocate(this % center_lat(dims(1), dims(2)))
    allocate(this % center_lon(dims(1), dims(2)))
    allocate(this % corner_lat(4, dims(1), dims(2)))
    allocate(this % corner_lon(4, dims(1), dims(2)))
    allocate(this % mask2D(dims(1), dims(2)))
    allocate(this % level(dims(3)))
    
    write(*,'(A)') ' success.'      
end subroutine 
 
!========================================================================================
!----------------------------------------------------------------------------------------
! Allocate memory for object
!----------------------------------------------------------------------------------------

subroutine check_grid(this)
	class(ModelGrid) :: this
	
	if (this % name == "") then
        write(*,*) "Error: Offline: No grid registered."
        call exit(1)
    end if
	
end subroutine

!========================================================================================
!----------------------------------------------------------------------------------------
! Make netCDF file in modified SCRIP format for CDO operators
!----------------------------------------------------------------------------------------

subroutine dump_to_cdo_format(this)
	class(ModelGrid) :: this	

    type(FileHandler_NC) :: fh
	character(NAME_LEN) :: filename
	    
    filename = DATA_FOLDER//TRIM(this % name)//"_"// &
               TRIM(n2s(this % dims(1)))//"x"//TRIM(n2s(this % dims(2)))//"_"// &
               TRIM(this % grid_type)//"_cdo.nc"
	
	! Create file and set title
	write(*,'(A$)') "Saving "//TRIM(this % name)//" grid in CDO format to file <"//TRIM(filename)//">..."
	
	call fh % create_file(TRIM(filename))
	
	! Define dimensions
	call fh % create_dim('grid_size', this % dims(1) * this % dims(2))
    call fh % create_dim('grid_xsize', this % dims(1))
    call fh % create_dim('grid_ysize', this % dims(2))
    call fh % create_dim('grid_zsize', this % dims(3))
    call fh % create_dim('grid_corners',  this % corner_num)
    call fh % create_dim('grid_rank', this % dim_num)
	
	call fh % create_var('grid_dims',       var_type = "int",   dim1_name = 'grid_rank')
    call fh % put(this % dims, (/ 1 /), (/ this % dim_num /))
    
    call fh % create_var('grid_center_lat', var_type = "real4", dim1_name = 'grid_xsize', dim2_name = 'grid_ysize' )
    call fh % put_att("units", "degrees")
    call fh % put_att("bounds", "grid_corner_lat")
    call fh % put(REAL(this % center_lat, 4), (/ 1, 1 /), (/ this % dims(1), this % dims(2) /))
    
    call fh % create_var('grid_center_lon', var_type = "real4", dim1_name = 'grid_xsize', dim2_name = 'grid_ysize' )
    call fh % put_att("units", "degrees")
    call fh % put_att("bounds", "grid_corner_lon")
    call fh % put(REAL(this % center_lon, 4), (/ 1, 1 /), (/ this % dims(1), this % dims(2) /))

    call fh % create_var("grid_imask",      var_type = "int",   dim1_name = 'grid_xsize', dim2_name = 'grid_ysize' )
    call fh % put_att("units", "unitless")
    call fh % put_att("coordinates", "grid_center_lon grid_center_lat")
        ! Cause strange output 0 0 0 0 when int
    call fh % put(this % mask2D, (/ 1, 1 /), (/ this % dims(1), this % dims(2) /) )
    
    call fh % create_var("grid_corner_lat", var_type = "real4", dim1_name = 'grid_corners', dim2_name = 'grid_xsize', dim3_name = 'grid_ysize')
    call fh % put_att("units", "degrees")
    call fh % put(REAL(this % corner_lat, 4), (/ 1, 1, 1 /), (/ this % corner_num, this % dims(1), this % dims(2) /))
    
    call fh % create_var("grid_corner_lon", var_type = "real4", dim1_name = 'grid_corners', dim2_name = 'grid_xsize', dim3_name = 'grid_ysize')
    call fh % put_att("units", "degrees")
    call fh % put(REAL(this % corner_lon, 4), (/ 1, 1, 1 /), (/ this % corner_num, this % dims(1), this % dims(2) /))

    !call fh % create_var("grid_imask",      var_type = "int",   dim1_name = 'grid_xsize', dim2_name = 'grid_ysize' )
    !call fh % put_att("coordinates", "grid_center_lon grid_center_lat")
    !call fh % put_att("units", "unitless")
    !! Cause strange output 0 0 0 0 when int
    !call fh % put(this % mask2D, (/ 1, 1 /), (/ this % dims(1), this % dims(2) /) )
    
    call fh % create_var('z',           var_type = "real4", dim1_name = 'grid_zsize')
    call fh % put_att("units", "meters")
    call fh % put(this % level, (/ 1 /), (/ this % dims(3) /))

    call fh % put_att("_FillValue", missval, is_global = .TRUE.)
    call fh % put_att("missing_value", missval, is_global = .TRUE.)
    call fh % put_att("title", "CDO grid file", is_global = .TRUE.)
    
    call fh % close_file()
    
	write(*,'(A)') ' success.'	
	
end subroutine

!========================================================================================

subroutine register_model_grid(this, grid)
    class(Offline) :: this
    type(ModelGrid), intent(in) :: grid
    
    call grid % check()
    this % model_grid = grid
end subroutine    

!========================================================================================
    
character(NAME_LEN) function get_ic_filename(this)
    class(Offline) :: this
    get_ic_filename = "../data/"//TRIM(this % model_grid % name)//"_"//TRIM(this % experiment_name)//"_"//"IC.nc"
end function

!========================================================================================
!----------------------------------------------------------------------------------------
! Add necessary information to IC-file
!----------------------------------------------------------------------------------------

subroutine add_info_to_ic_file(this)
    class(Offline) :: this
    type(FileHandler_NC) :: fh
    type(ModelGrid) :: mg
    
    write(*,'(A$)') "Saving additional grid information to file <"//TRIM(this % get_ic_filename())//">..."
    
    mg = this % model_grid   
    
    call fh % create_file(this % get_ic_filename())
    call fh % create_dim("i", mg % dims(1)) 
    call fh % create_dim("j", mg % dims(2)) 
    call fh % create_dim("k", mg % dims(3))       

    call fh % create_var("center_lat", "real4", "i", "j")
    call fh % put(REAL(mg % center_lat, kind = 4),  (/ 1, 1 /), (/ mg % dims(1), mg % dims(2) /))
    call fh % put_att("units", "degrees_north")
    call fh % put_att("long_name", "lat")
    
    call fh % create_var("center_lon", "real4", "i", "j")
    call fh % put(REAL(mg % center_lon, kind = 4),  (/ 1, 1 /), (/ mg % dims(1), mg % dims(2) /))
    call fh % put_att("units", "degrees_east")
    call fh % put_att("long_name", "lon")

    call fh % create_var("mask2D", "real4", "i", "j")
    call fh % put(REAL(mg % mask2D, 4),  (/ 1, 1 /), (/ mg % dims(1), mg % dims(2) /))
    call fh % put_att("units", "unitless")
    call fh % put_att("long_name", "mask")
    
    call fh % open_or_create_var("z", "real4", "k")
    call fh % put(mg % level,  (/ 1 /), (/ mg % dims(3) /))
    call fh % put_att("units", "meters")
    call fh % put_att("long_name", "vertical level")
    
    call fh % close_file()
    
    write(*,'(A)') ' success.'      

end subroutine

!========================================================================================
!----------------------------------------------------------------------------------------
! Build earth analize matrix
!----------------------------------------------------------------------------------------

subroutine analyze_land(this, proc_num, decomp_type)
    class(Offline) :: this
    integer, intent(in) :: proc_num
    character(*), intent(in) :: decomp_type
    
    type(FileHandler_NC) :: fh
    type(Decomposition) :: dec
    integer :: rank, i, j
    integer, allocatable :: land_mask(:,:)
    type(ModelGrid) :: mg
    character(NAME_LEN) :: filename
    integer :: non_masked
    
    mg = this % model_grid
    
    ! Build decomposition 
    dec = Decomposition(mg % dims(1), mg % dims(2), mg % dims(3), decomp_type, proc_num, 0)
                
    ! Using decomposition, analize land and build mask of earth
    allocate(land_mask(dec % inp(), dec % jnp()))
    do j = 1, dec % jnp()
    do i = 1, dec % inp()
        rank = dec % rank_by_ij(i, j)
        if (ANY(mg % mask2D(dec % west(rank) : dec % east(rank), dec % south(rank) : dec % north(rank)) == 1)) then
            land_mask(i,j) = 1
        else
            land_mask(i,j) = 0
        end if
    end do
    end do    
        
    ! Rebuild processor decomposition using land mask
    call dec % build_proc_decomp(land_mask)
    
    ! Save processor decomposition to file to read it on run stage
    call dec % dump_proc_decomp_to_file(this % get_ic_filename())
    
    ! Output info
     write(*,'(a,i4,a)') "Grid was decomposed on ",proc_num," processor domains. "
     non_masked = dec % proc_count() 
     write(*,'(i4,a,f7.1,a)') non_masked ," of them are non-masked (",real(non_masked)/real(proc_num)," %)."
     write(*,'(a,i4,a,a)') "Use exactly this number of processors (",non_masked,") to run ",mg % name," component."
!     
    deallocate(land_mask)
    
end subroutine

!========================================================================================

end module










	
