
module service_module

use event_module
use datetime_module
use component_splitter_module
use mpi
use communicator_ga_module
use model_info_module
use experiment_info_module

implicit none


type, abstract, public :: Service
    integer :: running_count_
    character, allocatable :: ev_packed_(:)    
    type(Communicator) :: comm_    
    
 contains
    procedure, public :: init_base                                          ! Constructor of base class
    procedure, public :: destroy_base                                       ! Destructor of base class

    procedure, public :: request_cycle                                      ! Main cycle of receiving requests

    procedure(I_handle_request), DEFERRED :: handle_request                 ! Virtual method for raising event
    procedure(I_init_service),  DEFERRED :: init                            ! Virtual method for initialization
    procedure(I_destroy_service),  DEFERRED :: destroy                      ! Virtual method for destruction
    procedure(I_after_init_service),  DEFERRED :: after_init                ! Virtual method once after init and models start

    
    ! Helpers
    procedure, public :: try_register_service_ga                            ! Shortcut for ga array registration 
    procedure, public :: report_handle                                      ! Print message
    procedure, public :: morph_to_decomp_1D                                 ! Decompose Decomposition object to service 1D decomposition

    procedure, private :: receive_request                                   ! Receive and unpack next event
    
end type

abstract interface
    
    subroutine I_handle_request(this, ev)
        import Service
        import Event
        class(Service) :: this
        type(Event), intent(in) :: ev        
    end subroutine

    subroutine I_init_service(this)
        import Service
        class(Service) :: this
    end subroutine

    subroutine I_destroy_service(this)
        import Service
        class(Service) :: this
    end subroutine

    subroutine I_after_init_service(this)
        import Service
        class(Service) :: this
    end subroutine
    
    
end interface

 contains

!======================================================================================== 
subroutine init_base(this, comm)
    class(Service) :: this
    type(Communicator), intent(in) :: comm
    
    if (.NOT. ALLOCATED(this % ev_packed_)) allocate(this % ev_packed_(event_pack_size()))    

    this % running_count_ = CompSplitter % model_launched()

    this % comm_ = comm
     
end subroutine

subroutine destroy_base(this)
    class(Service) :: this
    if (ALLOCATED(this % ev_packed_)) deallocate(this % ev_packed_)   
    
end subroutine
 
!======================================================================================== 

subroutine receive_request(this, ev, tag)
    class(Service) :: this
    type(Event), intent(out) :: ev
    integer, intent(in), optional :: tag
    integer :: ierr, status(MPI_STATUS_SIZE)
    
    call MPI_RECV(this % ev_packed_, SIZE(this % ev_packed_), MPI_PACKED, &
                  MPI_ANY_SOURCE, MERGE(tag, MPI_ANY_TAG, PRESENT(tag)), MPI_COMM_WORLD, status, ierr)
    call mpi_pack_to_event(this % ev_packed_, ev)
        
end subroutine
 
!======================================================================================== 

subroutine request_cycle(this)
    class(Service) :: this
    type(Event) :: ev

    ! Registration phase
    do while (this % running_count_ > 0)
    
        ! Receive any request
        call this % receive_request(ev)
        call this % handle_request(ev)
        
        select case(ev % action())                
            ! One component finish work
            case("STOP")
                this % running_count_ = this % running_count_ - 1
                CYCLE
        end select        
    end do
    ! Signal that registration is complete
    call CompSplitter % barrier("GA registration phase...", complete = .TRUE.)
    
    this % running_count_ = CompSplitter % model_launched()
    call this % after_init()
    ! Run phase
    do while (this % running_count_ > 0)
     
        ! Receive any request
        call this % receive_request(ev)
        call this % handle_request(ev)
        
        select case(ev % action())                
            ! One component finish work
            case("STOP")
                call this % report_handle(ev)                 
                this % running_count_ = this % running_count_ - 1
                CYCLE
        end select        
    end do

    
end subroutine

subroutine try_register_service_ga(this, ev)
    class(Service) :: this
    type(Event) :: ev
    ! Lazy: if not registered: register, return
    if (.NOT. this % comm_ % is_registered(ev % ga_name())) then
        call this % comm_ % init_array(ev % ga_name(), ev % arr_type(), ev % arr_dims(), &
                                        CompSplitter % comp_id(ev % owner()), ModelsInfo(CompSplitter % comp_id(ev % owner())) % decomp(), &
                                        CompSplitter % i_am_id())
    end if
end subroutine



!========================================================================================

subroutine report_handle(this, ev)
    class(Service) :: this
    type(Event), intent(in) :: ev
    
    if (CompSplitter % is_first_rank() .AND. .not. ExpInfo % silent()) &
        write(*,'(A)') CompSplitter % i_am()//" :: "//TRIM(ev % to_string())
end subroutine

!========================================================================================

function morph_to_decomp_1D(this, comp_name) result(cpl_decomp)
    class(Service) :: this
    character(*), intent(in) :: comp_name
    type(Decomposition) :: cpl_decomp, comp_decomp

    comp_decomp = ModelsInfo(CompSplitter % comp_id(comp_name)) % decomp()
    cpl_decomp = Decomposition(comp_decomp % i_size(), comp_decomp % j_size(), comp_decomp % k_size(),  & 
                        "1D", CompSplitter % comm_local_size(CompSplitter % i_am()), CompSplitter % rank_local()) 

end function
!========================================================================================

end module
