! 
!========================================================================================
! Derived class for IOD service.
!========================================================================================

module service_iod_module

use utils_module
use actions_module
use service_module
use event_module
use communicator_ga_module
use component_splitter_module
use model_info_module
use file_handler_nc_module
use cpl_comp_clients_module
use decomposition_module
#ifdef DEBUG
use debug_module
#endif

implicit none

type, extends(Service) :: ServiceIOD
private

 contains  
    !=======================
    ! ===== PUBLIC API =====
    procedure, public :: init => init_iod
    procedure, public :: destroy => destroy_iod
    procedure, public :: handle_request => handle_iod
    procedure, public :: after_init => after_init_iod

    procedure, private :: handle_put
    procedure, private :: handle_get
    
    ! ===== PUBLIC API =====
    !=======================    
end type

 real(4), allocatable, private :: buffer4(:,:,:)
 real(8), allocatable, private :: buffer8(:,:,:)

 CONTAINS

subroutine init_iod(this)
    class(ServiceIOD) :: this
    integer :: s, r
    integer, allocatable :: serv_ranks(:)
     
end subroutine

subroutine destroy_iod(this)
    class(ServiceIOD) :: this

   call this % destroy_base()
    
end subroutine

subroutine after_init_iod(this)
    class(ServiceIOD) :: this
         
end subroutine
 
!========================================================================================

subroutine handle_iod(this, ev)
    class(ServiceIOD) :: this
    type(Event), intent(in) :: ev
                
    select case (ev % action())
        case("SAVE_CP", "SAVE_DG")
            if (.NOT. this % comm_ % is_registered(ev % ga_name())) then
                call this % try_register_service_ga(ev)
                call Debugger % log_msg ( "Registered: "//TRIM(ev % to_string()))
            else
                call this % handle_put(ev)
                call this % report_handle(ev)
                
                call Debugger % log_msg ( "Processed: "//TRIM(ev % to_string()))
            end if
            
            
        case("READ_FD")        
            if (.NOT. this % comm_ % is_registered(ev % ga_name())) then
                call this % try_register_service_ga(ev)
                call Debugger % log_msg ( "Registered: "//TRIM(ev % to_string()))
            else
                call this % handle_get(ev)    
                call this % report_handle(ev)
                
                call Debugger % log_msg ( "Processed: "//TRIM(ev % to_string()))
            end if
    
        case("ERROR")
            call this % report_handle(ev)
!             call sleep(1)
            call exit(1)
        
    end select
    
end subroutine
 
!========================================================================================

! Macros: get data from GA, ensure that all made it, mark array as free, put to file

#define _GET_GA_PUT_NC_(_BUF_, _LO_, _HI_, _ALLOC_)                 \
    allocate(_BUF_ _ALLOC_);                                        \
    call this % comm_ % get(ev % ga_name(), _LO_, _HI_, _BUF_);            \
    call this % comm_ % sync(service_id);                           \
    if(CompSplitter % is_first_rank()) call this % comm_ % put_info(ev % ga_name(), COMM_GA_STATUS, "free"); \
    call hdr_file % put(_BUF_, _LO_, _HI_, time_put);               \
    deallocate(_BUF_);

!========================================================================================
    
subroutine handle_put(this, ev)
    use decomposition_module
    use datetime_module
    class(ServiceIOD) :: this
    type(Event), intent(in) :: ev
    type(Decomposition) :: cpl_src_dec
    integer :: lo(3), hi(3), service_id
    type(FileHandler_NC) :: hdr_file
    type(DateTime) :: time_put
    
    ! Shortcuts
    cpl_src_dec = this % morph_to_decomp_1D(ev % src())
    lo = cpl_src_dec % lower_bound_3D()
    hi = cpl_src_dec % upper_bound_3D()
    service_id = CompSplitter % i_am_id()
    time_put = DateTime(MERGE(ev % time_rec(), ev % time(), ev % time_rec() /= ""))
    !write(*,*),"IOD PUT Time_put",  TRIM( time_put % to_string()) , TRIM(ev % time_rec() )  , TRIM( ev % time() ) , TRIM(ev % ga_name())
    
    ! Open file
    call hdr_file % open_or_create_file(ev % dst(), CompSplitter % comm_local())                                                
    call hdr_file % create_dim("i", cpl_src_dec % i_size())
    call hdr_file % create_dim("j", cpl_src_dec % j_size())
    call hdr_file % create_dim("k", cpl_src_dec % k_size())
    call hdr_file % create_time_dim()
    
    call hdr_file % put_att("_FillValue", missval, is_global = .TRUE.)
    call hdr_file % put_att("missing_value", missval, is_global = .TRUE.)
    
    ! Use macros to get from ga, mark it as free and put to file
    select case (ev % arr_dims()) 
    case("2D")
        call hdr_file % open_or_create_var(ev % arr_name(), ev % arr_type(), dim1_name = "i", dim2_name = "j", dimt_name = "default")
                        
        if (ev % arr_type() == "real4") then
            _GET_GA_PUT_NC_(buffer4, lo(1:2), hi(1:2), (lo(1):hi(1), lo(2):hi(2), 1:1))            
        else if (ev % arr_type() == "real8") then
            _GET_GA_PUT_NC_(buffer8, lo(1:2), hi(1:2), (lo(1):hi(1), lo(2):hi(2), 1:1))
        end if
    case("3D")
        call hdr_file % open_or_create_var(ev % arr_name(), ev % arr_type(), "i", "j", "k", "default")

        if (ev % arr_type() == "real4") then
            _GET_GA_PUT_NC_(buffer4, lo, hi, (lo(1):hi(1), lo(2):hi(2), lo(3):hi(3)))            
        else if (ev % arr_type() == "real8") then
            _GET_GA_PUT_NC_(buffer8, lo, hi, (lo(1):hi(1), lo(2):hi(2), lo(3):hi(3))) 
        end if
    end select
          
    call hdr_file % close_file()  
end subroutine

!========================================================================================

! Macros: get data from NC, wait untill free, put to GA and sync, mark as full

#define _GET_NC_PUT_GA_(_BUF_, _LO_, _HI_, _ALLOC_)                 \
    allocate(_BUF_ _ALLOC_);                                        \
    call hdr_file % get(_BUF_, _LO_, _HI_, time_get);               \
    call this % comm_ % put(ev % ga_name(), _LO_, _HI_, _BUF_);            \
    call this % comm_ % sync(service_id);                           \
    if(CompSplitter % is_first_rank()) call this % comm_ % put_info(ev % ga_name(), COMM_GA_STATUS, "full");           \
    deallocate(_BUF_);

!========================================================================================

subroutine handle_get(this, ev)
    use decomposition_module
    
    class(ServiceIOD) :: this
    type(Event), intent(in) :: ev
    type(Decomposition) :: cpl_dst_dec
    integer :: lo(3), hi(3), service_id, comp_id
    type(FileHandler_NC) :: hdr_file
    type(DateTime) :: time_get
    
    ! Shortcuts
    cpl_dst_dec = this % morph_to_decomp_1D(ev % dst())
    lo = cpl_dst_dec % lower_bound_3D()
    hi = cpl_dst_dec % upper_bound_3D()
    service_id = CompSplitter % i_am_id()
    comp_id = CompSplitter % comp_id(ev % dst())
    !time_get = DateTimeFromString(MERGE(ev % time_rec(), ev % time(), ev % time_rec() /= ""))
    time_get = DateTimeFromString( ev % time() )
    !write(*,*)," GET Time_get",  TRIM( time_get % to_string()) , TRIM(ev % time_rec() )  , TRIM( ev % time() ) , TRIM(ev % ga_name())
    ! Try to open file
    call hdr_file % open_file(ev % src(), CompSplitter % comm_local())

    ! Try to open var
    call hdr_file % open_var(ev % arr_name())

    ! Use macros to get from ga, mark it as free and put to file
    select case (ev % arr_dims()) 
    case("2D")                        
        if (ev % arr_type() == "real4") then
            _GET_NC_PUT_GA_(buffer4, lo(1:2), hi(1:2), (lo(1):hi(1), lo(2):hi(2), 1:1))            
        else if (ev % arr_type() == "real8") then
            _GET_NC_PUT_GA_(buffer8, lo(1:2), hi(1:2), (lo(1):hi(1), lo(2):hi(2), 1:1))
        end if
    case("3D")
        if (ev % arr_type() == "real4") then
            _GET_NC_PUT_GA_(buffer4, lo, hi, (lo(1):hi(1), lo(2):hi(2), lo(3):hi(3)))            
        else if (ev % arr_type() == "real8") then
            _GET_NC_PUT_GA_(buffer8, lo, hi, (lo(1):hi(1), lo(2):hi(2), lo(3):hi(3))) 
        end if
    end select
          
    call hdr_file % close_file()  
  
end subroutine

!========================================================================================


end module
