!========================================================================================
! All utils for the system.
! Now contains utils for: arrays, MPI operations, strings.
!========================================================================================

!========================================================================================
! Useful utils for work with arrays
!========================================================================================
module utils_array_module

implicit none

    !======================
    ! ==== PUBLIC API =====
    public :: resize_array          ! (a, newsize): Resize allocatable array <a> to new size <newsize>
    
    public :: search_linear         ! (elem, a, lo, hi) : Find index of <elem> in array <a(lo:hi)> or return <lo>-1 as fail, T = O(N)
    public :: search_binary         ! (elem, a, lo, hi) : the same for sorted(!) array. T = O(logN)
    
    public :: indexing_ijk_to_kij   ! (arr_ijk): Return array of shape (k,i,j) with acoording elements
    public :: indexing_kij_to_ijk   ! opposite to indexing_ijk_to_kij
    
    public :: convert_1D_to_2D      ! (arr_1D, dim1, dim2) : Return array of shape (dim1,dim2) built from vector <arr_1D>
    public :: convert_2D_to_1D      ! (arr_2D) : linearize 2D array <arr_2D> to 1D array of shape (SIZE(arr_2D))
    ! ==== PUBLIC API =====
    !======================    

    interface print_array
        module procedure print_array_2D, print_array_3D
    end interface   
    
    interface resize_array   
        module procedure resize_int4, resize_int8, resize_real4, resize_real8
    end interface

    interface search_linear
        module procedure search_linear_int, search_linear_str, search_linear_real4, search_linear_real8
    end interface


    interface search_binary
        module procedure search_binary_int, search_binary_int8, search_binary_real, search_binary_real8
    end interface

    interface indexing_kij_to_ijk
        module procedure indexing_kij_to_ijk_int, indexing_kij_to_ijk_real4,  indexing_kij_to_ijk_real8
    end interface

    interface indexing_ijk_to_kij
        module procedure indexing_ijk_to_kij_int, indexing_ijk_to_kij_real4,  indexing_ijk_to_kij_real8
    end interface

    interface convert_1D_to_2D
        module procedure convert_1D_to_2D_int, convert_1D_to_2D_real4, convert_1D_to_2D_real8
    end interface

    interface convert_2D_to_1D
        module procedure convert_2D_to_1D_int, convert_2D_to_1D_real4, convert_2D_to_1D_real8
    end interface
    
 CONTAINS

!========================================================================================
     
#define _RESIZE_ARRAY_(_FUNC_,_TYPE_, _DIMS_)                       \
subroutine _FUNC_ (a, newsize);                                     \
    _TYPE_, _DIMS_, allocatable, intent(inout) :: a;                \
    integer, intent(in) :: newsize;                                 \
    _TYPE_, _DIMS_, allocatable :: b;                               \
    allocate(b(SIZE(a)));                                           \
    b = a;                                                          \
    deallocate(a);                                                  \
    allocate(a(newsize));                                           \
    a(1: MIN(newsize, SIZE(b))) = b;                                \
    deallocate(b);                                                  \
end subroutine  
    
_RESIZE_ARRAY_(resize_int4, integer, dimension(:))
_RESIZE_ARRAY_(resize_int8, integer(8), dimension(:))
_RESIZE_ARRAY_(resize_real4, real(4), dimension(:))
_RESIZE_ARRAY_(resize_real8, real(8), dimension(:))

!========================================================================================

#define _LINEAR_SEARCH_(_FUNC_,_TYPE_)                              \
integer function _FUNC_ (elem, a, lo, hi);                          \
    integer, intent(in) :: lo, hi;                                  \
    _TYPE_, intent(in) :: elem, a(lo:hi);                           \
    do _FUNC_ = lo, hi;                                             \
        if (a(_FUNC_) == elem) return;                              \
    end do;                                                         \
    _FUNC_ = lo - 1;                                                \
end function

_LINEAR_SEARCH_(search_linear_int, integer)
_LINEAR_SEARCH_(search_linear_str, character(*))

#define _LINEAR_SEARCH_FLOAT_(_FUNC_,_TYPE_)                        \
integer function _FUNC_ (elem, a, lo, hi);                          \
    integer, intent(in) :: lo, hi;                                  \
    _TYPE_, intent(in) :: elem, a(lo:hi);                           \
    _TYPE_  :: tmp, dist;                                           \
    integer i;                                                      \
    dist = 9999999.;                                                \
    do i = lo, hi;                                                  \
        tmp = abs(a(i) - elem );                                    \
        if ( tmp < dist )  then;                                    \
            dist = tmp;                                             \
            _FUNC_ = i;                                             \
        endif;                                                      \
    end do;                                                         \
    return;                                                         \
end function

_LINEAR_SEARCH_FLOAT_(search_linear_real4, real(4))
_LINEAR_SEARCH_FLOAT_(search_linear_real8, real(8))




!========================================================================================

#define _BINARY_SEARCH_(_FUNC_, _TYPE_) \
integer function _FUNC_ (elem, a, lo, hi); \
    integer, intent(in) :: lo, hi; \
    _TYPE_, intent(in) :: elem, a(lo:hi); \
    integer mid, lo_, hi_; \
    lo_ = lo; \
    hi_ = hi; \
    _FUNC_ = lo - 1; \
    if (elem > a(hi) .OR. elem < a(lo)) return; \
    do while (lo_ <= hi_); \
        mid = (lo_ + hi_)/2; \
        if (a(mid) > elem) then; \
            hi_ = mid - 1; \
        else if (a(mid) < elem) then; \
            lo_ = mid + 1; \
        else; \
            _FUNC_ = mid; \
            return; \
        end if; \
    end do; \
    _FUNC_ = mid; \
    return; \
end function

_BINARY_SEARCH_(search_binary_int, integer)
_BINARY_SEARCH_(search_binary_int8, integer(8))

#define _BINARY_SEARCH_FLOAT_(_FUNC_, _TYPE_) \
integer function _FUNC_ (elem, a, lo, hi); \
    integer, intent(in) :: lo, hi; \
    _TYPE_, intent(in) :: elem, a(lo:hi); \
    integer mid, lo_, hi_; \
    lo_ = lo; \
    hi_ = hi; \
    _FUNC_ = lo - 1; \
    if (elem > a(hi) .OR. elem < a(lo)) return; \
    do while (lo_ <= hi_); \
        mid = (lo_ + hi_)/2; \
        if (a(mid) > elem) then; \
            hi_ = mid - 1; \
        else if (a(mid) < elem) then; \
            lo_ = mid + 1; \
        else; \
            _FUNC_ = mid; \
            return; \
        end if; \
    end do; \
    if ((a(lo_) - elem) < (elem - a(hi_))) then; \
        _FUNC_ = lo_; \
    else; \
        _FUNC_ = hi_; \
    end if; \
    return; \
end function

_BINARY_SEARCH_FLOAT_(search_binary_real, real(4))
_BINARY_SEARCH_FLOAT_(search_binary_real8, real(8))

!========================================================================================

#define _INDEXING_KIJ_TO_IJK_(_FUNC_,_TYPE_)                                \
function _FUNC_ (arr_kij) result(arr_ijk);                                  \
    _TYPE_ , intent(in) :: arr_kij(:,:,:);                                  \
    _TYPE_ :: arr_ijk(SIZE(arr_kij,2), SIZE(arr_kij,3), SIZE(arr_kij,1));   \
    integer :: k;                                                           \
    do k = LBOUND(arr_kij,1), UBOUND(arr_kij,1);                            \
        arr_ijk(:,:,k) = arr_kij(k,:,:);                                    \
    end do;                                                                 \
end function

_INDEXING_KIJ_TO_IJK_(indexing_kij_to_ijk_int, integer)
_INDEXING_KIJ_TO_IJK_(indexing_kij_to_ijk_real4, real(4))
_INDEXING_KIJ_TO_IJK_(indexing_kij_to_ijk_real8, real(8))

#define _INDEXING_IJK_TO_KIJ_(_FUNC_,_TYPE_)                                \
function _FUNC_(arr_ijk) result(arr_kij);                                   \
    _TYPE_, intent(in) :: arr_ijk(:,:,:);                                   \
    _TYPE_ :: arr_kij(SIZE(arr_ijk,3), SIZE(arr_ijk,1), SIZE(arr_ijk,2));   \
    integer :: k;                                                           \
    do k = LBOUND(arr_ijk,3), UBOUND(arr_ijk,3);                            \
        arr_kij(k,:,:) = arr_ijk(:,:,k);                                    \
    end do;                                                                 \
end function

_INDEXING_IJK_TO_KIJ_(indexing_ijk_to_kij_int, integer)
_INDEXING_IJK_TO_KIJ_(indexing_ijk_to_kij_real4, real(4))
_INDEXING_IJK_TO_KIJ_(indexing_ijk_to_kij_real8, real(8))

!========================================================================================

#define _CONVERT_1D_to_2D_(_FUNC_,_TYPE_)                                   \
function _FUNC_(arr_1D, dim1, dim2) result(arr_2D);                         \
    _TYPE_, intent(in) :: arr_1D(:);                                        \
    integer, intent(in) :: dim1, dim2;                                      \
    _TYPE_ :: arr_2D(dim1, dim2);                                           \
    integer :: c, i, j;                                                     \
    c = 1;                                                                  \
    do j = 1, dim2;                                                         \
    do i = 1, dim1;                                                         \
        arr_2D(i,j) = arr_1D(c);                                            \
        c = c + 1;                                                          \
    end do;                                                                 \
    end do;                                                                 \
end function

_CONVERT_1D_to_2D_(convert_1D_to_2D_int, integer)
_CONVERT_1D_to_2D_(convert_1D_to_2D_real4, real(4))
_CONVERT_1D_to_2D_(convert_1D_to_2D_real8, real(8))

#define _CONVERT_2D_to_1D_(_FUNC_,_TYPE_)                                   \
function _FUNC_(arr_2D) result(arr_1D);                                     \
    _TYPE_, intent(in) :: arr_2D(:,:);                                      \
    _TYPE_ :: arr_1D(SIZE(arr_2D));                                         \
    integer :: c, i, j;                                                     \
    c = 1;                                                                  \
    do j = 1, SIZE(arr_2D,2);                                               \
    do i = 1, SIZE(arr_2D,1);                                               \
        arr_1D(c) = arr_2D(i,j);                                            \
        c = c + 1;                                                          \
    end do;                                                                 \
    end do;                                                                 \
end function

_CONVERT_2D_to_1D_(convert_2D_to_1D_int, integer)
_CONVERT_2D_to_1D_(convert_2D_to_1D_real4, real(4))
_CONVERT_2D_to_1D_(convert_2D_to_1D_real8, real(8))

! REWRITE TODO
!========================================================================================

! subroutine sort_bubble(vect_in, vect_out, vect_size)
! 
!     integer vect_size
!     integer, dimension(vect_size) :: vect_in, vect_out
!     integer i, cop
!     integer :: was_exch
! 
!     was_exch = 1
!     vect_out = vect_in
! 
!     do while(was_exch == 1)
! 
!         was_exch = 0
!         do i = 1, vect_size - 1
!             if (vect_out(i) > vect_out(i+1)) then
!                 cop = vect_out(i+1)
!                 vect_out(i+1) = vect_out(i)
!                 vect_out(i) = cop
!                 was_exch = 1
!             end if
!         end do
!     end do
! 
! end subroutine
! 
! !========================================================================================
! 
! recursive subroutine sort_merge(A,N,T)
! 
!     integer, intent(in) :: N
!     integer, dimension(N), intent(in out) :: A
!     integer, dimension((N+1)/2), intent (out) :: T
! 
!     integer :: NA,NB,V
! 
!     if (N < 2) return
!     if (N == 2) then
!     if (A(1) > A(2)) then
!         V = A(1)
!         A(1) = A(2)
!         A(2) = V
!     endif
!     return
!     endif      
!     NA=(N+1)/2
!     NB=N-NA
! 
!     call sort_merge(A,NA,T)
!     call sort_merge(A(NA+1),NB,T)
! 
!     if (A(NA) > A(NA+1)) then
!     T(1:NA)=A(1:NA)
!     call Merge_(T,NA,A(NA+1),NB,A,N)
!     endif
!     return
! 
!  contains
!  
!     subroutine Merge_(A,NA,B,NB,C,NC)
! 
!         integer, intent(in) :: NA,NB,NC         ! Normal usage: NA+NB = NC
!         integer, intent(in out) :: A(NA)        ! B overlays C(NA+1:NC)
!         integer, intent(in)     :: B(NB)
!         integer, intent(in out) :: C(NC)
! 
!         integer :: I,J,K
! 
!         I = 1; J = 1; K = 1;
!         do while(I <= NA .and. J <= NB)
!         if (A(I) <= B(J)) then
!             C(K) = A(I)
!             I = I+1
!         else
!             C(K) = B(J)
!             J = J+1
!         endif
!         K = K + 1
!         enddo
!         do while (I <= NA)
!         C(K) = A(I)
!         I = I + 1
!         K = K + 1
!         enddo
!         return
! 
!     end subroutine
!  
! end subroutine


!========================================================================================

subroutine print_array_2D(arr, msg)
    real(4), intent(in) :: arr(:,:)
    character(*), intent(in), optional :: msg
    integer :: i, j
    
    if (PRESENT(msg)) write(*,*) "Print 2D array message: <"//msg//">"
    do j = ubound(arr,2), lbound(arr,2), -1
        write(*,'(1000f9.3)')(arr(i,j), i = lbound(arr,1), ubound(arr,1))
    end do
    write(*,*)
end subroutine

subroutine print_array_3D(arr, msg)
    real(4), intent(in) :: arr(:,:,:)
    character(*), intent(in), optional :: msg
    integer :: k
    
    if (PRESENT(msg)) write(*,*) "Print 3D array message: <"//msg//">"
    do k = lbound(arr,3), ubound(arr,3)
        write(*,*) "Level k = ", k
        call print_array_2D(arr(:,:,k))
    end do
    write(*,*)
end subroutine
    
!========================================================================================
        
        
    end module 

    
!========================================================================================
!========================================================================================
!========================================================================================
    
 !
! Class-facade for popular MPI-operations
!

module utils_mpi_module

    use mpi

    implicit none

    integer, private, parameter :: MPI_SOME_TAG = 33    
    integer, private :: status(MPI_STATUS_SIZE)
    integer, private :: ierr

    !========================
    ! ====== PUBLIC API =====
    interface broadcast
        module procedure broadcast_bool, broadcast_int4, broadcast_string, broadcast_arr_int4, broadcast_arr_2D_int4
    end interface
 
    interface sendrecv
        module procedure sendrecv_int4, sendrecv_int8, sendrecv_real4, sendrecv_real8, sendrecv_string
    end interface
    
    interface cmf_global_itg
        module procedure UtilsAllReduce
    end interface
    
    ! ====== PUBLIC API =====
    !========================
 
 CONTAINS
    
!========================================================================================
!----------------------------------------------------------------------------------------
! Description:       perform combined send and receive of data
! Parameters:
!   var             - buffer for data to get (for src_rank) and to put (for dst_rank)
!   my_rank         - client pass his rank to filter operation only for src and dst
!   src_rank        - get data from var on rank src_rank
!   dst_rank        - put data to var on rank dst_rank
!----------------------------------------------------------------------------------------

#define _SEND_RECV_(_FUNC_, _TYPE_, _DIMS_, _MPI_SIZE_, _MPI_TYPE_)                                                             \
subroutine _FUNC_(var, my_rank, src_rank, dst_rank);                                                                            \
    integer, intent(in) :: my_rank, src_rank, dst_rank;                                                                         \
    _TYPE_ :: var _DIMS_;                                                                                                       \
    if(my_rank == src_rank) call MPI_SEND(var, _MPI_SIZE_, _MPI_TYPE_, dst_rank, MPI_SOME_TAG, MPI_COMM_WORLD, ierr);           \
    if(my_rank == dst_rank) call MPI_RECV(var, _MPI_SIZE_, _MPI_TYPE_, src_rank, MPI_SOME_TAG, MPI_COMM_WORLD, status, ierr);   \
end subroutine

_SEND_RECV_(sendrecv_int4,      integer(4),     ,               1,         MPI_INTEGER)
_SEND_RECV_(sendrecv_int8,      integer(8),     ,               1,         MPI_DOUBLE_PRECISION)
_SEND_RECV_(sendrecv_real4,     real(4),        ,               1,         MPI_REAL)
_SEND_RECV_(sendrecv_real8,     real(8),        ,               1,         MPI_DOUBLE_PRECISION)
_SEND_RECV_(sendrecv_string,    character(*),   ,               LEN(var),  MPI_CHARACTER)
        
!========================================================================================
!----------------------------------------------------------------------------------------
! Description:       perform data broadcasting
! Parameters:
!   var             - buffer for data to get (for src_rank) and to put (for dst_rank)
!   root            - who broadcast data
!   comm            - communicator for broadcasting
!----------------------------------------------------------------------------------------

#define _BROADCAST_(_FUNC_, _TYPE_, _DIMS_, _MPI_SIZE_, _MPI_TYPE_)                                                             \
subroutine _FUNC_(var, root, comm);                                                                                             \
    _TYPE_, intent(inout) :: var _DIMS_;                                                                                           \
    integer, intent(in) :: root, comm;                                                                                          \
    call MPI_BCAST(var, _MPI_SIZE_, _MPI_TYPE_, root, comm, ierr);                                                              \
end subroutine

_BROADCAST_(broadcast_bool,         logical,        ,               1,          MPI_CHAR)
_BROADCAST_(broadcast_int4,         integer,        ,               1,          MPI_INTEGER)
_BROADCAST_(broadcast_string,       character(*),   ,               LEN(var),   MPI_CHARACTER)
_BROADCAST_(broadcast_arr_int4,     integer,        (:),            SIZE(var),  MPI_INTEGER)
_BROADCAST_(broadcast_arr_2D_int4,  integer,        (:,:),          SIZE(var),  MPI_INTEGER)
  
!========================================================================================
!----------------------------------------------------------------------------------------
! Description:       calculate global integral or jsut global sum
! Parameters:
!   local_area      - (optional) area of local domain          
!   local_val       - value of local integral
!   comm            - communicator for all processes in reduce
!----------------------------------------------------------------------------------------
real(kind=8) function UtilsAllReduce(local_area, local_val, comm)
    real(kind=8), intent(in) :: local_val
    real(kind=8), intent(in), optional :: local_area
    integer, intent(in) :: comm
    real(kind=8) :: rbuf(2)
    integer :: ierr

    rbuf = 0
    if (PRESENT(local_area)) then
        call MPI_ALLREDUCE([local_area, local_val], rbuf, SIZE(rbuf), MPI_DOUBLE_PRECISION, MPI_SUM, comm, ierr)
        UtilsAllReduce = rbuf(2)/rbuf(1)
    else
        call MPI_ALLREDUCE([local_val], rbuf(1), 1, MPI_DOUBLE_PRECISION, MPI_SUM, comm, ierr)
        UtilsAllReduce = rbuf(1)
    end if
end function



end module


!========================================================================================
!========================================================================================
!========================================================================================


module utils_string_module

implicit none

interface n2s
    module procedure i2s_no_len, i2s_len, r2s
end interface

 CONTAINS

!========================================================================================

integer function length(int_in)
    integer, intent(in) :: int_in
    integer :: int_remain
    
    length = 0
    int_remain = int_in
    do 
        if (int_remain == 0) then
            EXIT
        else
            length = length + 1
            int_remain = int_remain/10
        end if
    end do
    if (length == 0) length = 1
end function

!========================================================================================


function i2s_len(int_in, out_len)
    integer, intent(in) :: int_in, out_len
    character(out_len) :: i2s_len
    character(len=2) :: string
    
    write(unit = string, fmt = '(i2)') out_len
    write(unit = i2s_len, fmt = '(i'//string//'.'//string//')') int_in
end function


function i2s_no_len(int_in)
    integer, parameter :: STR_LEN = 20
    integer, intent(in) :: int_in
    character(STR_LEN) :: i2s_no_len
    character(len=2) :: string
    
    write(unit = string, fmt = '(i2)') length(int_in)
    write(unit = i2s_no_len, fmt = '(i'//string//'.'//string//')') int_in
end function
    
!========================================================================================

function r2s(number)
    integer, parameter :: STR_LEN = 20
    character(STR_LEN) :: r2s
    
    real(8), intent(in) :: number
   
    write (r2s, "(f6.2)") number
end function
    
!========================================================================================

integer function s2i(str)
    character(*) :: str

    if(str == "") then 
        s2i = 0
    else
        if (is_numeric(str)) then
            read(str,*) s2i
        else
            call assert_error("String <"//str//"> is not a number.")
        endif
    end if
end function

!========================================================================================

FUNCTION is_numeric(string)
  IMPLICIT NONE
  CHARACTER(len=*), INTENT(IN) :: string
  LOGICAL :: is_numeric
  REAL :: x
  INTEGER :: e
  READ(string,*,IOSTAT=e) x
  is_numeric = e == 0
END FUNCTION is_numeric

end module


!========================================================================================
! Datatypes
!========================================================================================

module int4_data_module
type :: Int4_data
    integer :: value = -32000
end type

type(Int4_data) :: EMPTY_DATA
end module

#define CONCRETE_TYPE Int4_data
#define CONCRETE_MODULE int4_data_module
#define CONCRETE_MODULE_FLIBS int4_data_module_
#define OUTPUT_MODULE Int4_vector_module
#include "VectorStorage.f90"

module string_data_module
type :: string_data
    character(100) :: value = ""
end type

end module

#define CONCRETE_TYPE string_data
#define CONCRETE_MODULE string_data_module
#define CONCRETE_MODULE_FLIBS string_data_module_
#define OUTPUT_MODULE string_hash_module
#include "HashStorage.f90"

!========================================================================================


module utils_module

use utils_array_module
use utils_mpi_module
use utils_string_module

use int4_vector_module


end module

!========================================================================================

subroutine memory_usage_short(msg)
    use utils_string_module
    implicit none

    character(*), intent(in) :: msg
    character(100) :: vmSize
    integer :: i

    open(111, file = '/proc/self/status')

    vmSize = ''
    do while(INDEX(vmSize,'VmSize') == 0)
        read(111,'(a100)') vmSize
    end do
    close(111)
    write(*, '(a)') 'Virtual memory usage: '//TRIM(msg)//" "//TRIM(vmSize)
end subroutine

!========================================================================================

subroutine memory_usage(msg)
    use utils_string_module
    implicit none

    character(*), intent(in) :: msg
    character(100) :: vmPeak, vmSize
    integer :: i

    open(111, file = '/proc/self/status')

    write(*, '(a/a/a/)') 'Memory characteristics: '//TRIM(msg) 
    do i = 1,21
        read(111, '(a100)') vmSize
        write(*, '(a100)') vmSize
    end do
    write(*,*)

    close(111)


    !   vmPeak = ''
    !   do while(INDEX(vmPeak,'VmPeak') == 0)
    !       read(111,'(a100)') vmPeak
    !   end do
    !   
    !   vmSize = ''
    !   do while(INDEX(vmSize,'VmSize') == 0)
    !       read(111,'(a100)') vmSize
    !   end do
    !   
    !   close(111)
    ! 
    !   write(run_log_file_num, '(a/a/a/)') 'Virtual memory characteristics of '//TRIM(compName)//' kernel '//TRIM(I2S(rank_world))//' '//strMessage//': ', TRIM(vmPeak), TRIM(vmSize)

end subroutine


subroutine assert_error(msg)
    character(*), intent(in) :: msg
    
    integer :: test(1), i
    write(*,*) "Assert error: ", msg
    write(*,*) "Error message after it is not real error - it is introduced to make traceback."
    
    ! Crazy way to create exception in Fortran
    i = 2
    test(i) = 1
    
    call exit(1)
    
end subroutine








