!========================================================================================




!========================================================================================
!========================================================================================
! Abstract class for VarEvent
!========================================================================================
module var_event_module

use event_module, ONLY : Event
use datetime_module

implicit none

type, abstract, public :: VarEvent
!     type(VectorStorageEvent) :: events_storage_    
    type(Event) :: ev_

 contains
    procedure, public :: get_event => get_event_ve
 
    procedure(I_update_ve), DEFERRED :: update         ! Virtual method for raising event
    procedure(I_destroy_ve), DEFERRED :: destroy       ! Virtual destuctor
end type

abstract interface
    
    logical function I_update_ve(this, cur_time, ev)
        import VarEvent
        import DateTime
        import Event
        class(VarEvent) :: this
        type(DateTime), intent(in) :: cur_time
        type(Event), intent(inout) :: ev        
    end function

    subroutine I_destroy_ve(this)
        import VarEvent
        class(VarEvent) :: this
    end subroutine
end interface

 contains

type(Event) function get_event_ve(this)
    class(VarEvent) :: this
    get_event_ve = this % ev_
end function

end module

!========================================================================================
!========================================================================================
! Concrete class for VarEvent: normal event (periodic)
!========================================================================================

module var_event_normal_module

use var_event_module

type, extends(VarEvent) :: VarEventNormal
private
    integer :: period_sec_ = 0
    type(DateTime) :: start_date_
    
 contains
 
    procedure :: update
    procedure :: destroy    
end type


 CONTAINS

!========================================================================================
 
function new_VarEventNormal(ev, start_date, dd, dh, dm, ds) result (obj)
    type(Event), intent(in) :: ev
    type(DateTime), intent(in) :: start_date
    integer, intent(in) :: dd, dh, dm, ds
    ! Intel compiler bug (reported)
    ! https://software.intel.com/en-us/forums/topic/390944      
    ! This workaround is work
    class(VarEventNormal), pointer :: obj

    allocate(VarEventNormal::obj)
    
    obj % ev_ = ev
    obj % start_date_ = start_date
    obj % period_sec_ = dd*60*60*24 + dh*60*60 + dm*60 + ds
end function

!========================================================================================

logical function update(this, cur_time, ev)
    class(VarEventNormal) :: this
    type(DateTime), intent(in) :: cur_time
    type(Event), intent(inout) :: ev
    
    integer(8) :: sec_from_start
    
    sec_from_start = date2sec(cur_time) - date2sec(this % start_date_)
    
    ! Update only on start time
    if (this % period_sec_ /= 0) then
        update = sec_from_start >= 0 .AND. MOD(sec_from_start, this % period_sec_) == 0
    else
        update = sec_from_start == 0
    end if
   
    if (update) then
        ev = this % ev_
    end if
    
end function

!========================================================================================

subroutine destroy(this)
    class(VarEventNormal) :: this
end subroutine 

!========================================================================================

end module

!========================================================================================
!========================================================================================
! Concrete class for VarEvent: file_synced
!========================================================================================

module var_event_synced_module

use var_event_module
use file_handler_nc_module

implicit none

type, extends(VarEvent) :: VarEventSynced
private
    integer(8) :: cur_rec_
    integer(8), allocatable :: times_(:)
    
 contains
 
    procedure :: update => update_synced
    procedure :: destroy => destroy_synced
end type


 CONTAINS

!========================================================================================
 
! Repeat year condition: ignore year, look on just remainder of date.
! Used in CORE experiments where data has year period.
! Make it more flexible in future #TODO.

function new_VarEventSynced(ev, start_date) result(obj)
    type(DateTime), intent(in) :: start_date
    type(Event), intent(in) :: ev
    class(VarEventSynced), pointer :: obj
    type(DateTime) :: dt
    
    type(FileHandler_NC) :: fh
    integer :: t

    allocate(VarEventSynced::obj)    
    obj % ev_ = ev
    
    ! Try to open var, read time axis and close file
    call fh % open_file(ev % src())
    call fh % open_var(ev % arr_name())
    call fh % get_time_axis(obj % times_)
    call fh % close_file()
    
    ! Repeat year condition: ignore year
    do t = 1, SIZE(obj % times_)
        dt = sec2date(obj % times_(t))
        if (dt % year_sec() >= start_date % year_sec()) EXIT 
    end do  
    obj % cur_rec_ = t

end function

!========================================================================================

logical function update_synced(this, cur_time, ev)
    class(VarEventSynced) :: this
    type(DateTime), intent(in) :: cur_time
    type(Event), intent(inout) :: ev
    type(DateTime) :: dt

    ! Repeat year condition: ignore year
    dt = sec2date(this % times_(this % cur_rec_))
    update_synced = cur_time % year_sec() == dt % year_sec()
    
    if(update_synced) then
        ev = this % ev_
        call ev % set_time_rec(dt % to_string())
        this % cur_rec_ = this % cur_rec_ + 1
        if (this % cur_rec_ > SIZE(this % times_)) this % cur_rec_ = 1
        
!         write(*,*) "Raise at: ", cur_time % to_string(), this % cur_rec_
    end if
    
end function

!========================================================================================

subroutine destroy_synced(this)
    class(VarEventSynced) :: this
    deallocate(this % times_)
end subroutine

!========================================================================================

end module


!========================================================================================
!========================================================================================
! Container, which stores base class pointers and update them all 
!========================================================================================

!*********************************************************
! Magic for enabling vector in Fortran #BAD #TODO
module var_event_wrapper_module
use var_event_module
! Magic because Fortran doesnt have array of pointers
type :: VarEventWrapper
    class(VarEvent), pointer :: obj
end type

type(VarEventWrapper) :: EMPTY_DATA
end module

#define CONCRETE_TYPE VarEventWrapper
#define CONCRETE_MODULE var_event_wrapper_module
#define CONCRETE_MODULE_FLIBS var_event_wrapper_module_
#define OUTPUT_MODULE var_event_wrapper_vector_module
#include "VectorStorage.f90"
!*********************************************************

module event_scheduler_module

use var_event_module
use var_event_wrapper_vector_module, VarEventVector => VectorStorage
use datetime_module
use var_event_wrapper_module

implicit none

type :: EventScheduler
private    
    type(VarEventVector) :: ves_
    
 contains
    procedure, public :: init
    procedure, public :: add => add_ves
    procedure, public :: gather_events
    procedure, public :: get_all_events
    procedure, public :: destroy => destroy_ves
    
end type

 contains

!========================================================================================

subroutine init(this)
    class(EventScheduler) :: this
    
    call this % ves_ % init()
end subroutine

!========================================================================================

subroutine add_ves(this, ve)
    class(EventScheduler) :: this
    class(VarEvent), target :: ve
    type(VarEventWrapper) :: vew
    
    vew % obj => ve
    call this % ves_ % append(vew)

end subroutine

!========================================================================================

subroutine gather_events(this, cur_time, new_events)
    use event_vector_module, VectorEvent => VectorStorage    
    class(EventScheduler) :: this
    type(DateTime), intent(in) :: cur_time
    type(VectorEvent), intent(inout) :: new_events
    type(Event)  :: ev
    integer :: i
    type(VarEventWrapper) :: vew

    call new_events % init()
    do i = 1, this % ves_ % length()
        vew = this % ves_ % get(i)        
        if(vew % obj % update(cur_time, ev)) then
            call ev % set_time(cur_time % to_string())
            call new_events % append(ev)
        end if
    end do 
    
    ! Sort vector
    call sort_event_vector(new_events)
  
end subroutine

!========================================================================================

subroutine get_all_events(this, all_events)
    use event_vector_module, VectorEvent => VectorStorage    
    class(EventScheduler) :: this
    type(VectorEvent), intent(inout) :: all_events
    integer :: i
    type(VarEventWrapper) :: vew
  
    call all_events % init()
  
    do i = 1, this % ves_ % length()
        vew = this % ves_ % get(i)        
        call all_events % append(vew % obj % get_event())
    end do 

end subroutine

!========================================================================================


subroutine destroy_ves(this)
    class(EventScheduler) :: this
    integer :: i
    type(VarEventWrapper) :: vew

    do i = 1, this % ves_ % length()
        vew = this % ves_ % get(i)        
        call vew % obj % destroy()
        deallocate(vew % obj)
    end do     
     
    call this % ves_ % destroy()

end subroutine
  
!========================================================================================

end module

  
  
  