!======================================================================================== 
! This module contains help procedures for imitatin global model grid.
! It used by unit tests to compare some local results with etalon fields.
!======================================================================================== 

module test_global_domain_module
use utils_module
use decomposition_module
implicit none

type :: TestGlobalDomain
    real(kind=4), allocatable :: glob(:,:,:,:)
    real(kind=4), allocatable :: loc_2D(:,:)
    real(kind=8), allocatable :: loc_2D8(:,:)    
    real(kind=4), allocatable :: loc_3D(:,:,:)
    type(Decomposition) :: dec
    integer :: il, jl, kl, tl
    integer :: w, e, s, n
    integer :: max_halo, rank_local
    
 contains
    procedure, public :: init => init_tgd
    procedure, public :: destroy => destroy_tgd
    procedure, public :: is_correct
    procedure, public :: reset_local_arrays
    procedure, public :: halo_update
    
    procedure, private :: reset_global_array
   
end type


 CONTAINS
 
!======================================================================================== 
!

subroutine init_tgd(this, dec, t_size, rank_local, max_halo_width)
    class(TestGlobalDomain) :: this
    type(Decomposition), intent(in) :: dec
    integer, intent(in) :: t_size
    integer :: rank_local
    integer, intent(in), optional :: max_halo_width

    call this % destroy()
    
    this % dec = dec
    this % max_halo = MERGE(max_halo_width, 0, PRESENT(max_halo_width))
    this % rank_local = rank_local
    
    ! Make shortcuts
    this % il = dec % i_size()
    this % jl = dec % j_size()
    this % kl = dec % k_size()
    this % tl = t_size
    this % w = dec % west(rank_local)
    this % e = dec % east(rank_local)
    this % s = dec % south(rank_local)
    this % n = dec % north(rank_local)
    
!     write(*,*) "Rank", rank_local, ":", this % w, this % e, this % s, this % n
    
    ! Allocate buffers
    allocate(this % glob(1-this % max_halo : this % il +this % max_halo,   1-this % max_halo : this % jl +this % max_halo, this % kl, t_size))
    allocate(this % loc_2D(  this % w - this % max_halo : this % e + this % max_halo, this % s -this % max_halo : this % n + this % max_halo))  
    allocate(this % loc_2D8( this % w - this % max_halo : this % e + this % max_halo, this % s -this % max_halo : this % n + this % max_halo))  
    allocate(this % loc_3D(  this % w - this % max_halo : this % e + this % max_halo, this % s -this % max_halo : this % n + this % max_halo, this % kl))  
    call this % reset_global_array()
            
end subroutine

!======================================================================================== 

subroutine reset_global_array(this)
    class(TestGlobalDomain) :: this
    integer :: i, j, k, t, c, in, jn, r
    
    this % glob = 0
        
    ! Fill only non-masked domains
    c = 1
    do in = 1, this % dec % inp()
    do jn = 1, this % dec % jnp()
        r = this % dec % rank_by_ij(in, jn)
        if (r /= -1) then
            do t = 1, this % tl
            do k = 1, this % kl  
            do i = this % dec % west(r), this % dec % east(r) 
            do j = this % dec % south(r), this % dec % north(r) 
                this % glob(i,j,k,t) = t + 0.01*c   
                c = c + 1
            end do
            end do
            end do
            end do
        end if
    end do
    end do
    
end subroutine  

!======================================================================================== 

subroutine reset_local_arrays(this, mode)
    class(TestGlobalDomain) :: this
    character(*), intent(in), optional :: mode
    
    this % loc_2D = 0
    this % loc_2D8 = 0
    this % loc_3D = 0
    
    if (PRESENT(mode) .AND. mode == "halo") then
        this % loc_2D  (this % w : this % e, this % s : this % n)    = this % glob(this % w : this % e, this % s : this % n, 1, 1)
        this % loc_2D8 (this % w : this % e, this % s : this % n)    = this % glob(this % w : this % e, this % s : this % n, 1, 1)
        this % loc_3D  (this % w : this % e, this % s : this % n, :) = this % glob(this % w : this % e, this % s : this % n, :, 1)
    end if
end subroutine

!======================================================================================== 

subroutine destroy_tgd(this)
    class(TestGlobalDomain) :: this
    if (ALLOCATED(this % glob)) deallocate(this % glob, this % loc_2D, this % loc_2D8, this % loc_3D)
end subroutine

!======================================================================================== 

subroutine halo_update(this, grid_type, change_sign_on_bipolar)
    class(TestGlobalDomain) :: this
    logical, optional, intent(in) :: change_sign_on_bipolar 
    character(1), intent(in) :: grid_type
    integer :: i, j, k, t, c
    integer :: sign_

    ! Clear array
    call this % reset_global_array()    
    
    if (PRESENT(change_sign_on_bipolar)) then
        sign_ = MERGE(-1, 1, change_sign_on_bipolar)
    else
        sign_ = 1
    end if
    
    if (this % dec % is_icycle()) then
        ! West border
        do k = 1, this % kl
        do i = 1, this % max_halo
        do j = 1-this % max_halo, this % jl+this % max_halo
            this % glob(1-i, j, k, 1) =  this % glob(this % il-i+1, j, k, 1)
        end do
        end do
        end do    

        ! East border
        do k = 1, this % kl
        do i = 1, this % max_halo
        do j = 1-this % max_halo, this % jl+this % max_halo
            this % glob(this % il+i, j, k, 1) =  this % glob(1+i-1, j, k, 1)
        end do
        end do
        end do    
    end if
        
    ! North border
    if (this % dec % is_bsc()) then
        do k = 1, this % kl
        do j = 1, this % max_halo
        do i = 1, this % il
            if (grid_type == "T") &
                this % glob(i, this % jl+j, k, 1) = sign_*this % glob(this % il+1-i+1, this % jl-j+1, k, 1)
            if (grid_type == "V") &                
                this % glob(i, this % jl+j, k, 1) = sign_*this % glob(this % il+1-i,   this % jl-j,   k, 1)
        end do
        end do
        end do
    end if
    
end subroutine

!======================================================================================== 

logical function is_correct(this, arr_type, t, halo_width)
    class(TestGlobalDomain) :: this
    character(*), intent(in) :: arr_type
    integer, intent(in), optional :: t
    integer, intent(in), optional :: halo_width
    integer :: halo_width_, t_, i, j, k
    
    halo_width_ = MERGE(halo_width, 0, PRESENT(halo_width))
    
    if (PRESENT(t)) then
        t_ = t
    else
        t_ = 1
    end if
    
    is_correct = .TRUE.
        
    do i = this % w - halo_width_, this % e + halo_width_
    do j = this % s - halo_width_, this % n + halo_width_
        if(.NOT.( (i > this % il-1 .AND. j > this % jl) .OR. (i < 1+1 .AND. j > this % jl) )) then
    
            select case (arr_type)
            case("2D")
                ! Compare axcept halo top corners
                    if (this % loc_2D(i,j) /= this % glob(i,j,1,1)) then
                        write(*,'(a,i5,i5,f9.3,f9.3)') "Error in point 2D: ", i, j, this % glob(i,j,1,t_), this % loc_2D(i,j), "rank ", this % rank_local 
                        is_correct = .false.
                    end if
                
            case("2D8")
                ! Compare axcept halo top corners
                    if (this % loc_2D8(i,j) /= this % glob(i,j,1,1)) then
                        write(*,'(a,i5,i5,f9.3,f9.3)') "Error in point 2D8: ", i, j, this % glob(i,j,1,t_), this % loc_2D8(i,j), "rank ", this % rank_local 
                        is_correct = .false.
                    end if
                                                    
            case("3D")   
                do k = 1, this % kl
                    ! Compare axcept halo top corners
                    if (this % loc_3D(i,j,k) /= this % glob(i,j,k,1)) then
                        write(*,'(a,i5,i5,i3,f9.3,f9.3)') "Error in point 3D: ", i, j, k, this % glob(i,j,k,t_), this % loc_3D(i,j,k), "rank ", this % rank_local 
                        is_correct = .false.
                    end if
                end do
            end select
        end if
    end do
    end do
    
    call this % reset_local_arrays()
   
end function

!======================================================================================== 
logical function compare_arrays(global_arr, local_arr, w, e, s, n, eps)
    integer, intent(in) :: w, e, s, n
    real(4), intent(in) :: global_arr(:,:), local_arr(w:e,s:n), eps
    integer :: i, j, c
    
    c = 1
    do i = w, e
    do j = s, n
        if (ABS(global_arr(i,j) - local_arr(i,j) > eps)) then
            write(*,*) "Error in point: ", i, j, c, global_arr(i,j), local_arr(i,j)
            compare_arrays = .FALSE.
            return
        end if
        c = c + 1
    end do
    end do
    compare_arrays = .TRUE.

end function

!======================================================================================== 

end module
