module test_manager_module

use mpi
use utils_string_module
implicit none

type :: TestManager    
private
    integer :: test_results_(1000) = -1
    integer :: test_number_ = 0

    character(3) :: mode_
    integer :: comm_
    integer :: comm_size_
    integer :: rank_
    
 contains
    procedure, public :: init => init_manager
    procedure, public :: destroy
    
    procedure, public :: assert
    procedure, public :: report_tests

    procedure, public :: barrier
    procedure, public :: rank
    procedure, public :: comm_size
    procedure, public :: comm
    
    procedure, private :: print_message    
END TYPE

    private :: n2s

 CONTAINS

!========================================================================================

subroutine init_manager(this, mode)
    class(TestManager) :: this
    character(*) :: mode
    integer :: ierr
    logical :: is_init

    if (mode == "par") then
        call MPI_INITIALIZED(is_init, ierr)
        if (.NOT. is_init) call MPI_INIT(ierr)
        call MPI_COMM_RANK(MPI_COMM_WORLD, this % rank_, ierr)
        this % comm_ = MPI_COMM_WORLD
        this % mode_ = "par"
        call MPI_COMM_SIZE(MPI_COMM_WORLD, this % comm_size_, ierr)
        if (this % rank_ == 0) then
            write(*,'(A,i3)') NEW_LINE('A')//"Initialize parallel TestManager with communicator size: ", this % comm_size()
            write(unit=6,fmt='(91("-"))')
        end if
    else
        this % mode_ = "seq"
        this % rank_ = 0
        this % comm_size_ = 0        
        write(*,*) NEW_LINE('A')//'Initialize sequential TestManager' 
        write(unit=6,fmt='(91("-"))')        
    end if
end subroutine

subroutine destroy(this)
    class(TestManager) :: this
    integer :: ierr
    if (this % mode_ == "par") call MPI_FINALIZE(ierr)
end subroutine

!========================================================================================

subroutine assert(this, condition, msg)
    class(TestManager) :: this
    logical, intent(in) :: condition
    character(*), intent(in) :: msg
    integer :: ierr, p, i
    logical :: answers(0:this % comm_size_-1)
    
    this % test_number_ = this % test_number_ + 1
    this % test_results_(this % test_number_) = MERGE(1, 0, condition) 
    
    if (this % mode_ == "seq") then
        call this % print_message(condition, msg)
        return
    end if
            
    call MPI_GATHER(condition, 1, MPI_LOGICAL, answers, 1, MPI_LOGICAL, 0, this % comm_, ierr)
    if (this % rank_ == 0) then
        if (ALL(answers == .TRUE.)) then
            call this % print_message(.TRUE., msg)
        else    
            do i = 0, this % comm_size_-1
                if (answers(i) /= .TRUE.) call this % print_message(answers(i), msg, i) 
            end do
        end if
    end if
    call this % barrier()
end subroutine

!========================================================================================
    
subroutine report_tests(this)
    class(TestManager) :: this
    integer :: n, nsuccess, nfailure
    
    nsuccess = 0
    nfailure = 0
    do n = 1, this % test_number_
        if(this % test_results_(n) == 1) then
            nsuccess = nsuccess + 1
        else
            nfailure = nfailure + 1
        end if
    end do

    if (this % rank_ == 0) then
        write(unit=6,fmt='(91("-"))')
        write(unit=6,fmt='(a,i3,a)')'Ran a total of ',this % test_number_,' tests.'
        write(unit=6,fmt='(i3,a,i3,a)')nsuccess,' tests passed, ',nfailure,' tests failed (stats for rank 0)'
        write(unit=6,fmt='(91("-"))')
    end if
        
    if (this % test_number_ /= nsuccess) then
        call this % destroy()
        call exit(1)
    end if
    
end subroutine 

!========================================================================================

subroutine barrier(this)
    class(TestManager) :: this
    integer :: ierr
    if (this % mode_ == "par") call MPI_BARRIER(this % comm_, ierr)
end subroutine
 
integer function rank(this)
    class(TestManager) :: this
    rank = this % rank_
end function

integer function comm_size(this)
    class(TestManager) :: this
    comm_size = this % comm_size_
end function

integer function comm(this)
    class(TestManager) :: this
    comm = this % comm_
end function

!========================================================================================

subroutine print_message(this, condition, msg, check_rank)
    class(TestManager) :: this
    logical, intent(in) :: condition
    character(*), intent(in) :: msg
    integer, intent(in), optional :: check_rank
    character(80) :: outlen_msg
    outlen_msg = msg
        
    if (condition) then
        write(unit=6,fmt='(a)')'test '//outlen_msg//': '//char(27)//'[32mpass'//char(27)//'[0m'
    else
        if (this % mode_ == "seq") then
            write(unit=6,fmt='(a)')'test '//outlen_msg//': '//char(27)//'[31mfail'//char(27)//'[0m'
        else 
            write(unit=6,fmt='(a)')'test '//outlen_msg//': '//char(27)//'[31mfail'//char(27)//'[0m'//" on rank:"//n2s(check_rank)
        end if
    end if  
end subroutine

!========================================================================================

end module
