
program test_decomposition
    use decomposition_module
    use test_manager_module
    implicit none 
    
    type(Decomposition) :: dec
    type(TestManager) :: tm
    
    integer, parameter :: il = 180, jl = 160, kl = 49
    integer, allocatable :: land_mask(:,:)
    
    call tm % init("par")
    
    ! Decompose on different number of parts
    dec = Decomposition(il, jl, kl, "2D", 1, 0)
    call tm % assert(.TRUE., "create Decomposition object for 1 domain")
    call tm % assert(ALL(dec % global_dims() == (/ il, jl, kl /)), "check sizes" )

    dec = Decomposition(il, jl, kl, "2D", 2, 0)
    call tm % assert(.TRUE., "create Decomposition object for 2 domains")

    call tm % assert(ALL(dec % borders(0) == (/ 1, il/2, 1, jl /)), "check bounds for 0 domain...")
    call tm % assert(ALL(dec % borders(1) == (/ il/2 + 1, il, 1, jl /)), "check bounds for 1 domain...")
    
    dec = Decomposition(il, jl, kl, "2D", 4, 0)
    call tm % assert(.TRUE., "create Decomposition object for 4 domains")
    
    call tm % assert(ALL(dec % borders(0) == (/ 1,      il/2,   1,      jl/2 /)),   "check bounds for 0 domain...")
    call tm % assert(ALL(dec % borders(1) == (/ il/2+1, il,     1,      jl/2 /)),   "check bounds for 1 domain...")
    call tm % assert(ALL(dec % borders(2) == (/ 1,      il/2,   jl/2+1, jl /)),     "check bounds for 2 domain...")
    call tm % assert(ALL(dec % borders(3) == (/ il/2+1, il,     jl/2+1, jl /)),       "check bounds for 3 domain...")
    
    

    ! Check for parallel case
    dec = Decomposition(il, jl, kl, "2D", tm % comm_size(), tm % rank())
    call tm % assert(.TRUE., "create Decomposition object for parallel case")
    
    ! Check broadcast
    call dec % broadcast(0, tm % comm())
    call tm % assert(dec % south(0) == 1, "check how broadcast works")
    call tm % assert(dec % global_to_i(10) == 10 .AND. dec % global_to_i(il*jl) == il, "check global_to_i")
    call tm % assert(dec % global_to_j(10) == 1 .AND. dec % global_to_j(il*jl) == jl, "check global_to_j...")
!     
    ! Test masking land
    dec = Decomposition(180, 90, kl, "2D", 8, 0)
    call tm % assert(ALL(dec % proc_decomp_np() ==  (/ 4, 2 /)), "dec (/4, 2/) on 8 cores")
    
    ! Check neighbours for non-mask case:   4567
    !                                       0123
    dec = Decomposition(180, 90, kl, "2D", 8, 0, is_icycle = .true., is_bsc = .true.)    
    
    call tm % assert(ALL(dec % neighbour_ranks(0) == (/ 3, 1, -1, 4 /)), "check neighbours for 0 for unmasked case with icycle, bsc")
    call tm % assert(ALL(dec % neighbour_ranks(2) == (/ 1, 3, -1, 6 /)), "check neighbours for 2 for unmasked case with icycle, bsc")
    call tm % assert(ALL(dec % neighbour_ranks(3) == (/ 2, 0, -1, 7 /)), "check neighbours for 3 for unmasked case with icycle, bsc")
    call tm % assert(ALL(dec % neighbour_ranks(4) == (/ 7, 5, 0,  7 /)), "check neighbours for 4 for unmasked case with icycle, bsc")
    call tm % assert(ALL(dec % neighbour_ranks(5) == (/ 4, 6, 1,  6 /)), "check neighbours for 5 for unmasked case with icycle, bsc")
    call tm % assert(ALL(dec % neighbour_ranks(6) == (/ 5, 7, 2,  5 /)), "check neighbours for 6 for unmasked case with icycle, bsc")
    call tm % assert(ALL(dec % neighbour_ranks(7) == (/ 6, 4, 3,  4 /)), "check neighbours for 7 for unmasked case with icycle, bsc")    
    
    ! Create mask:   0111  => *234
    !                0011     **01
    allocate(land_mask(dec % inp(), dec % jnp())); land_mask = 1
    land_mask(1,1) = 0
    land_mask(2,1) = 0
    land_mask(1,2) = 0
    write(*,*) "Proc decomp before land apply:", dec % proc_decomp_
    
    call dec % build_proc_decomp(land_mask)

    write(*,*) "Proc decomp after land apply:", dec % proc_decomp_
    
    ! Check how dumps proc_decomp and reads it
    call dec % dump_proc_decomp_to_file("test_proc_decomp.nc")
    dec = Decomposition(180, 90, kl, "2D", 8, 0, proc_decomp_file = "test_proc_decomp.nc", is_icycle = .true., is_bsc = .true.)      
          
    call tm % assert(dec % rank_by_ij(1,1) == -1, "rank at land")
    call tm % assert(dec % rank_by_ij(2,1) == 0, "rank at (2,1)")
    call tm % assert(dec % rank_by_ij(2,2) == 4, "rank at (2,2)")
    call tm % assert(dec % rank_by_ij(4,2) == 6, "rank at (4,2)")
    deallocate(land_mask)
    
    ! Check neighbours for masked case
    call tm % assert(ALL(dec % neighbour_ranks(0) == (/ -1, 1, -1, 4 /)), "check neighbours for 0 for masked case")
    call tm % assert(ALL(dec % neighbour_ranks(2) == (/ 1, -1, -1, 6 /)), "check neighbours for 2 for masked case")
    call tm % assert(ALL(dec % neighbour_ranks(3) == (/ 6, 4, -1, 6 /)), "check neighbours for 3 for masked case")
    

    
    ! Finalizing
    call tm % report_tests()    
    call tm % destroy()
    
end
