    
program test_file_handler_nc
    use file_handler_nc_module
    use test_global_domain_module
    use test_manager_module
    use decomposition_module
    implicit none

    type(FileHandler_NC) :: handler
    type(TestManager) :: tm
    type(TestGlobalDomain) :: tgd
    type(Decomposition) :: decomp
    integer, parameter :: il = 90, jl = 60, kl = 10, rl = 100
    integer :: att_val_int
    character(20) :: att_val_str
    integer :: status
    integer(8), allocatable :: time_axis(:)
    
    call tm % init("par")

    !=====================================================
    ! Initialize
    !=====================================================
    decomp = Decomposition(il, jl, kl, "1D", tm % comm_size(), tm % rank())
    call tgd % init(decomp, t_size = 10, rank_local = tm % rank())

    if (tm % rank() == 0) call SYSTEM("rm -f *.nc")

    !=====================================================
    ! Open and close files and vars
    !=====================================================
    call handler % create_file("test_1D.nc", mpi_comm = tm % comm())
    call handler % create_dim("dim1", kl)
    call tm % assert(handler % get_dim_size("dim1") == kl, "create 1D parallel file")
    call handler % close_file()
    
    call handler % open_file("test_1D_ser.nc", status = status)
    call tm % assert(status /= NO_ERROR, "try to open nonexisting file but ask not to crash")
 
    if (tm % rank() == 0) then 
        call handler % create_file("test_1D_ser.nc")
        call handler % create_dim("dim1", kl)       
        call tm % assert(handler % get_dim_size("dim1") == kl, "create 1D serial file")
    else
        call tm % assert(.TRUE., "create 1D serial file")
    end if

    if (tm % rank() == 0) then     
        call handler % create_var("test_1D", var_type = "int", dim1_name = "dim1")
        call tm % assert(.NOT. handler % is_time_var(), "create 1D var")
        call handler % close_file()
    else
         call tm % assert(.TRUE., "create 1D var")
    end if        
 
    call handler % create_file("test_4D.nc", mpi_comm = tm % comm())
    call handler % create_dim("i", il) 
    call handler % create_dim("j", jl) 
    call handler % create_dim("k", kl) 
    call tm % assert(.TRUE., 'create 4D file')

    call handler % create_dim("k", kl) 
    call handler % create_time_dim()
    call tm % assert(.TRUE., "add already existing dims")
    
    call handler % close_file()
    
        
        
    call handler % open_file("test_4D.nc", mpi_comm = tm % comm(), status = status)    
    call tm % assert(status == NO_ERROR, "open 4D file")

    call tm % assert(handler % get_dim_size("i") == il .AND. handler % get_dim_size("k") == kl, "check dimension sizes")
    
    call handler % open_var("test_2D", status = status)
    call tm % assert(status /= NO_ERROR, "try to open nonexistent var")
    
    call handler % create_var("test_2D", "real4", "i", "j")
    call tm % assert(.NOT. handler % is_time_var(), "create 2D var")

    !=====================================================
    ! Put and get data to non-time vars
    !=====================================================    
    call handler % put(tgd % glob(decomp % west():decomp % east(), decomp % south():decomp % north(), 1, 1), &
                       lo = decomp % lower_bound_2D() , hi = decomp % upper_bound_2D())
    call tm % assert(.TRUE., "put 2D data")

    call handler % create_var("test_3D", "real4", "i", "j", "k")
    call tm % assert(.TRUE., "create 3D var")
    
    call handler % open_var("test_3D", status = status)
    call tm % assert(status == NO_ERROR, "open 3D var")
        
    call handler % put(tgd % glob(decomp % west():decomp % east(), decomp % south():decomp % north(), :, 1), & 
                       lo = decomp % lower_bound_3D(), hi = decomp % upper_bound_3D())
    call tm % assert(.TRUE., "put 3D data")
    
    call handler % open_var("test_2D")
    call handler % get(tgd % loc_2D, lo = decomp % lower_bound_2D() , hi = decomp % upper_bound_2D())
    call tm % assert(tgd % is_correct(arr_type = "2D"), "get 2D var and compare") 

    call handler % open_var("test_3D")    
    call handler % get(tgd % loc_3D, lo = decomp % lower_bound_3D(), hi = decomp % upper_bound_3D())
    call tm % assert(tgd % is_correct(arr_type = "3D"), "get 3D var and compare") 

    call handler % close_file()
    
    !=====================================================
    ! Put and get data to time vars
    !=====================================================    

    call handler % create_file("test_4D_t.nc", mpi_comm = tm % comm())
    call handler % create_dim("i", il) 
    call handler % create_dim("j", jl) 
    call handler % create_dim("k", kl) 
    call handler % create_time_dim()
    
    call handler % create_var("test_2D_t", "real4", "i", "j", dimt_name = "TIME")    
    call tm % assert(handler % is_time_var(), "create 2D time var")
    
   
    call handler % put(tgd % glob(decomp % west():decomp % east(), decomp % south():decomp % north(),1, 1), &
                       lo = decomp % lower_bound_2D(), hi = decomp % upper_bound_2D(), &
                       dt = DateTime(1988, 03, 15, 12, 0, 0))
    call tm % assert(handler % get_dim_size("TIME") == 1, "put 2D data to time var")

    call handler % create_var("test_2D_t2", "real4", "i", "j", dimt_name = "TIME")    

    call handler % put(tgd % glob(decomp % west():decomp % east(), decomp % south():decomp % north(),1, 1), &
                       lo = decomp % lower_bound_2D(), hi = decomp % upper_bound_2D(), &
                       dt = DateTime(1988, 03, 15, 12, 0, 0))
    
    call tm % assert(handler % get_dim_size("TIME") == 1, "put one more with the same time stamp") 
                       
    
    call handler % open_var("test_2D_t")    
    call handler % get(tgd % loc_2D, lo = decomp % lower_bound_2D() , hi = decomp % upper_bound_2D(), dt = DateTime(1988, 03, 15, 12, 0, 0))
    call tm % assert(tgd % is_correct(arr_type = "2D", t = 1), "get 2D time var and compare") 
    
    call handler % create_var("test_3D_t", "real4", "i", "j", "k", dimt_name = "TIME")    
    call tm % assert(handler % is_time_var(), "create 3D time var")
    
    call handler % get_att("len", att_val_int)                       
    call tm % assert(att_val_int == 0, "now length of var equal 0")
    
    call handler % put(tgd % glob(decomp % west():decomp % east(), decomp % south():decomp % north(), :, 1), & 
                       lo = decomp % lower_bound_3D(), hi = decomp % upper_bound_3D(), &
                       dt = DateTime(1988, 03, 15, 12, 0, 0))
    call handler % put(tgd % glob(decomp % west():decomp % east(), decomp % south():decomp % north(), :, 1), & 
                       lo = decomp % lower_bound_3D(), hi = decomp % upper_bound_3D(), &
                       dt = DateTime(1988, 03, 15, 12, 0, 1))

    call handler % get_att("len", att_val_int)                       
    call tm % assert(att_val_int == 2, "put 3D time var 2 times and check record number")

    call handler % close_file()
    call handler % open_file("test_4D_t.nc", mpi_comm = tm % comm())
    call handler % open_var("test_3D_t")
    call handler % get_att("len", att_val_int)       
    call tm % assert(att_val_int == 2, "after close/open too")
     
    call handler % open_var("test_3D_t")    
    call handler % get(tgd % loc_3D, lo = decomp % lower_bound_3D(), hi = decomp % upper_bound_3D(), dt = DateTime(1988, 03, 15, 12, 0, 1))
    call tm % assert(tgd % is_correct(arr_type = "3D", t = 1), "get 3D time var and compare") 

    call handler % close_file()
    call tm % assert(.TRUE., "close file")
    
    !=====================================================
    ! Put and get data to time vars with DateTime
    !=====================================================    
    
    call handler % create_file("test_2D_dt.nc", mpi_comm = tm % comm())
    call handler % create_dim("i", il) 
    call handler % create_dim("j", jl) 
    call handler % create_dim("k", kl) 
    call handler % create_time_dim()

    call handler % create_var("test_2D_dt", "real4", "i", "j", dimt_name = "TIME")    

    call handler % put(tgd % glob(decomp % west():decomp % east(), decomp % south():decomp % north(),1, 1), &
                       lo = decomp % lower_bound_2D(), hi = decomp % upper_bound_2D(), dt = DateTime(1988, 03, 15, 0, 0, 0))
    call handler % put(tgd % glob(decomp % west():decomp % east(), decomp % south():decomp % north(),1, 1), &
                       lo = decomp % lower_bound_2D(), hi = decomp % upper_bound_2D(), dt = DateTime(1988, 03, 15, 12, 0, 0))
                       
    call tm % assert(handler % get_dim_size("TIME") == 2, "put 2D data to time var with date")
        
    call handler % close_file()
    
    !=====================================================
    ! Get time axis
    !=====================================================    
    
    call handler % open_file("test_2D_dt.nc", mpi_comm = tm % comm())    
    call handler % get_time_axis(time_axis)
    call handler % close_file()
    
    call tm % assert(ALL(time_axis == (/ date2sec(DateTime(1988, 03, 15, 0, 0, 0)), date2sec(DateTime(1988, 03, 15, 12, 0, 0)) /)), "get_time_axis")
    

    !=====================================================
    ! Put/Get attributes
    !=====================================================    
    
    call handler % open_file("test_2D_dt.nc", mpi_comm = tm % comm())

    ! Global attribute
    call handler % put_att("test_att_str", "hello", is_global = .TRUE.)
    call handler % get_att("test_att_str", att_val_str, is_global = .TRUE.)
    call tm % assert(att_val_str == "hello", "put/get global att str")
    
    call handler % put_att("test_att_int", 123, is_global = .TRUE.)
    call handler % get_att("test_att_int", att_val_int, is_global = .TRUE.)
    call tm % assert(att_val_int == 123, "put/get global att int")
    
    ! Var attribute
    call handler % open_var("test_2D_dt")
    call handler % put_att("test_att_str", "hello")
    call handler % get_att("test_att_str", att_val_str)
    call tm % assert(att_val_str == "hello", "put/get att str")
    
    call handler % put_att("test_att_int", 123)
    call handler % get_att("test_att_int", att_val_int)
    call tm % assert(att_val_int == 123, "put/get att int")
    
    call handler % close_file()
    
    !=====================================================
    ! Finalize
    !=====================================================    
    call tm % report_tests()
    !if (tm % rank() == 0) call SYSTEM("rm -f *.nc")    
    
    call tm % destroy()    
        
end
