program test_mapper
    use test_manager_module
    use mapper_module
    use communicator_ga_module
    use test_global_domain_module
    use file_handler_nc_module
    implicit none
    
    type(FileHandler_NC) :: handler
    type(Mapper) :: map_par, map_ser, map_hand
    type(TestManager) :: tm
    type(Decomposition) :: src_decomp_par, dst_decomp_par, src_decomp_ser, dst_decomp_ser    
    type(Communicator) :: comm_ga
   
    character(100) :: size_arg, filename
    character(100) :: src_name, dst_name
    integer :: src_i, src_j, dst_i, dst_j
    
    real, allocatable :: src_data_global(:,:), src_data_global_1D(:), dst_data_global_1D(:)
    integer :: i, link
    
    integer, allocatable :: src_addr(:), dst_addr(:)
    real (kind = 8), allocatable :: wts_map(:,:)
        
    call tm % init("par")
    
    ! Read command line args
    if (tm % rank() == 0 .AND. COMMAND_ARGUMENT_COUNT() /= 6) then
        write(*,*) "Wrong arguments. Usage: ./exe OCN 180 90 ATM_NCAR 192 94"
        call exit(1)
    end if
    
    call GET_COMMAND_ARGUMENT(1, src_name)
    call GET_COMMAND_ARGUMENT(2, size_arg)    
    read(size_arg, *) src_i
    call GET_COMMAND_ARGUMENT(3, size_arg)    
    read(size_arg, *) src_j
    call GET_COMMAND_ARGUMENT(4, dst_name)
    call GET_COMMAND_ARGUMENT(5, size_arg)    
    read(size_arg, *) dst_i
    call GET_COMMAND_ARGUMENT(6, size_arg)    
    read(size_arg, *) dst_j


    ! Allocate global data
    allocate(src_data_global(src_i, src_j))
    allocate(src_data_global_1D(src_i*src_j))
    allocate(dst_data_global_1D(dst_i*dst_j))
       
    ! Build decompositions and filename
    src_decomp_par = Decomposition(src_i, src_j, 10, "1D", tm % comm_size(), tm % rank())
    dst_decomp_par = Decomposition(dst_i, dst_j, 10, "1D", tm % comm_size(), tm % rank())
    src_decomp_ser = Decomposition(src_i, src_j, 10, "1D", 1, 0)
    dst_decomp_ser = Decomposition(dst_i, dst_j, 10, "1D", 1, 0)
    
    filename = "./data/"//TRIM(src_name)//'_'//TRIM(src_decomp_par % to_string_2D())// &
            '_to_'//TRIM(dst_name)//'_'//TRIM(dst_decomp_par % to_string_2D())//".nc"

    call tm % assert(.TRUE., "Data allocation")
            
                
    ! Prepare communicator
    call comm_ga % init(1, tm % comm_size())
    call comm_ga % init_group(1, (/ (i, i = 0, tm % comm_size()-1) /))    
    call comm_ga % init_array("src_data_global", "real4", "2D", 1, src_decomp_par, 1)

    call tm % assert(.TRUE., "Init communicator")
    
    ! Put test data to src array
    src_data_global = 101
    call comm_ga % put("src_data_global", (/ 1, 1 /), (/ src_i, src_j /), src_data_global)
    call comm_ga % sync(1, 1)    

    call tm % assert(.TRUE., "Put test data to GA")

    !=================================
    ! Serial interpolation by hand
    !=================================
    call map_hand % init(src_decomp_ser, dst_decomp_ser) 
    call tm % assert(.TRUE., "map_hand init")
    call map_hand % read_map_file(TRIM(filename), src_addr, dst_addr, wts_map)
    call tm % assert(.TRUE., "map_hand read_map_file")
    dst_data_global_1D = 0
    src_data_global_1D = convert_2D_to_1D(src_data_global)
    do link = 1, SIZE(src_addr)
        dst_data_global_1D(dst_addr(link)) = dst_data_global_1D(dst_addr(link)) + wts_map(1,link)*src_data_global_1D(src_addr(link))
    end do  
    call tm % assert(.TRUE., "Interpolate by hand")

    !=================================
    ! Serial interpolation by Mapper
    !=================================
    call map_ser % init(src_decomp_ser, dst_decomp_ser)
    call tm % assert(.TRUE., "Initialize serial")
    
    call map_ser % read_and_analize(filename)
    call tm % assert(.TRUE., "Read and analize")    
  
    call map_ser % interpolate(src_data_global)
    call tm % assert(compare_arrays(convert_1D_to_2D(dst_data_global_1D, dst_i, dst_j), & 
                                      map_ser % get_result_2D(), &
                                      1, dst_i, 1, dst_j, 0.0001), &
                     "Interpolate serial and compare")
                                      
    !=================================
    ! Parallel interpolation by Mapper
    !=================================
    call map_par % init(src_decomp_par, dst_decomp_par)
    call tm % assert(.TRUE., "Initialize parallel")
    
    call map_par % read_and_analize(filename)
    call tm % assert(.TRUE., "Read and analize")
  
    call map_par % interpolate(comm_ga % id("src_data_global"))
    call tm % assert(compare_arrays(convert_1D_to_2D(dst_data_global_1D, dst_i, dst_j), & 
                                      map_par % get_result_2D(), &
                                      dst_decomp_par % west(), dst_decomp_par % east(), dst_decomp_par % south(), dst_decomp_par % north(), 0.0001), &
                     "Interpolate parallel and compare...")
    
    ! Write serial result to netCDF file 
     if (tm % rank() == 0) then
        call handler % create_file("interp_src.nc")
        call handler % create_dim("il", src_i)
        call handler % create_dim("jl", src_j)
        call handler % create_var("src", "real4", "il", "jl")
        call handler % put(src_data_global_1D, (/ 1, 1 /), (/ src_i, src_j /) )    
        call handler % close_file()
      
        call handler % create_file("interp_dst.nc")
        call handler % create_dim("il", dst_i)
        call handler % create_dim("jl", dst_j)
        call handler % create_var("dst", "real4", "il", "jl")
        call handler % put(convert_1D_to_2D(dst_data_global_1D, dst_i, dst_j), (/ 1, 1 /), (/ dst_i, dst_j /) )    
        call handler % close_file()
     end if
     call tm % assert(.TRUE.,"Write src and dst data to netCDF-file")    

    call map_hand % destroy()
    call map_par % destroy()
    call map_ser % destroy()
    call tm % assert(.TRUE., "Destroy")     

    call tm % report_tests()
    
    call comm_ga % destroy()
    call tm % destroy()    
    
end program