
!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!		Compact Modelling Framework	
!			CMF2.0
!
! authors: 
!	2013.08.07 / Kalmykov, Vladimir, IO RAS, Russia, vvk88@mail.ru
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!	2016.08.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru !
!
! reference:
!	K. V. Ushakov, R. A. Ibraev, and V. V. Kalmykov, 2015.
!		Simulation of the world ocean climate with a massively parallel numerical model, 
!		Izv., Atmos. Ocean. Phys. 51 (4), 362–380.
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Vychisl. Metody Programm., 2013, v. 14, iss. 3, 88-95. 
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Certificate of state registration of the computer progam number 2013619320. 
!		Date of state registration in the Registry of the computer programs October 1, 2013
!	R. A. Ibraev, K. V. Ushakov, and R. N. Khabeev, 2012.
!		Eddy-resolving 1/10° model of the World Ocean,
!		Izv., Atmos. Ocean. Phys. 48 (1), 37–46.
!	Ibrayev R.A., 2001: Model of enclosed and semi-enclosed sea hydrodynamics. 
!		Russ. J. Numer. Anal. Math. Modelling, 2001, vol. 16, No 4, p. 291-304.
!--------------------------------------

	module off_ocn_module
	
	use off_module
	use o_grid_module
	use o_driver_module

	implicit none
	integer :: additional_ny=0	!0 - to make full grid, <0 - to make reduced grid (eg. for Arctic only)
	
	TYPE, EXTENDS(GridConstructor) :: off_ocn
	
	CONTAINS
	
	PROCEDURE :: make_grid => o_make_grid
	PROCEDURE :: make_mask => o_make_mask
	PROCEDURE :: make_ic => o_make_ic
	
	END TYPE off_ocn

	!====> ice start <====!
	logical :: ice_grid = .true.
	!====>  ice end  <====!

	CONTAINS

!================================================================================================================================================================

	subroutine o_make_grid(this)

	class(off_ocn) :: this
	
	! Work variables
	integer spot_size
	integer counter
	integer coupler_worker, num_couplers
	
	write(*,'(/,A$)') 'Building global OCN grid'
	
	! Set grid size
	il = this % nx + 2
	jl = this % ny + 2
	kl = this % nz
		
	write(*,'(A,i3,i3,i3,A$,/)') ' with dimensions: ',il,jl,kl,'...'
	write(*,*) 'Grid type = '//this % gType
	
! Initiate parameters (WARNING: it overrides il, jl, kl)
	call o_driver_ini_params()
	
! Initiate domains for global serial case
	call set_grid_subdomains(rank = 1, iw = 2, ie = il-1, js = 2, jn = jl-1)
	
! Allocate grid arrays
	call allocate_o_grid_coord_module
	
! Make grid arrays x1_t, x2_t, x1_v, x2_v
	call o_ini_set_grid_coordinate

! Fill output 'GridType' structure with our information 
	this % mask2D = 1
	this % z(1:this % nz) = z(1:this % nz)

	select case (this % gType)
	
	case("T")		
		do j = 2, jl-1
		do i = 2, il-1
			this % center_lat(i-1,j-1)   = x2_t(i,j)
			this % corner_lat(1,i-1,j-1) = x2_v(i,j)
			this % corner_lat(2,i-1,j-1) = x2_v(i-1,j)
			this % corner_lat(3,i-1,j-1) = x2_v(i-1,j-1)
			this % corner_lat(4,i-1,j-1) = x2_v(i,j-1)

			this % center_lon(i-1,j-1  ) = x1_t(i,j)
			this % corner_lon(1,i-1,j-1) = x1_v(i,j)
			this % corner_lon(2,i-1,j-1) = x1_v(i-1,j)
			this % corner_lon(3,i-1,j-1) = x1_v(i-1,j-1)
			this % corner_lon(4,i-1,j-1) = x1_v(i,j-1)
		end do
		end do
		
	case("V")
		do j = 2, jl-1
		do i = 2, il-1
			this % center_lat(i-1,j-1)   = x2_v(i,j)
			this % corner_lat(1,i-1,j-1) = x2_t(i,j)
			this % corner_lat(2,i-1,j-1) = x2_t(i+1,j)
			this % corner_lat(3,i-1,j-1) = x2_t(i+1,j+1)
			this % corner_lat(4,i-1,j-1) = x2_t(i,j+1)

			this % center_lon(i-1,j-1  ) = x1_v(i,j)
			this % corner_lon(1,i-1,j-1) = x1_t(i,j)
			this % corner_lon(2,i-1,j-1) = x1_t(i+1,j)
			this % corner_lon(3,i-1,j-1) = x1_t(i+1,j+1)
			this % corner_lon(4,i-1,j-1) = x1_t(i,j+1)
		end do
		end do
		
	case default
		write(*,*) "Error: no grid type defined for "//this % gType// " case."
		stop
	
	end select
	
	!====> ice start <====!	
	if (ice_grid .and. this % gType .eq. "V") then
		write(*, '(A)') 'Writing additional grid data...'
		call additional_dump_grid(this)
	endif
	!====>  ice end  <====!

	write(*,'(A)') ' success.'
	
	end subroutine o_make_grid
	
!================================================================================================================================================================
!=========================================================
! Add mask for SCRIP, because now we know ETOPO topography
!=========================================================
	
	subroutine o_make_mask(this)
	
	class(off_ocn) :: this
	character(100) :: filename
	character(100) :: filename_mask
	logical exists
	character (kind=1), allocatable:: bt    (:,:)
	
	filename = TRIM(data_folder)//TRIM(this % name)//"_"//TRIM(I2S(this % nx))//'x'//TRIM(I2S(this % ny))//"_V"//"_topo.nc"
	INQUIRE(file = filename, exist = exists)
	if(.NOT. exists) then
		write(*,'(A)') 'Adding ocean mask... failed: no interpolated topography file <'//TRIM(filename)//">. Mask will be empty."
	else
		! Read topography
		call this % get_var(filename, "topo", arr_2D = hv(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank)))
		
		! Make hv,kv,kt
		hv = MERGE(-hv, REAL(0, kind=4), hv < 0)	
		
		call o_user_redefine_topo()
				
		call o_ini_bottom(hv, .false.)

		! Add mask for T-points
		this % mask2D = MERGE(1,0        ,kt(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank)) > 0)
!		this % mask2D = MERGE(1,misValInt,kt(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank)) > 0)
		!this % mask2D = MERGE(1.,missval,kt(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank)) > 0)

		!====> ice start <====!
		! Dump additional grid information of kv and kt with halo data
		if (ice_grid) then
			write(*, '(A)') ' writing additional grid mask data...'
			call additional_dump_mask(this)
		endif
		!====>  ice end  <====!
	filename_mask = TRIM(data_folder) // "OCN.mask.nc"
	call this % put_var(var="center_lat", long_name="latitude", units="degrees_north",axis = "Y", external_filename =filename_mask,arr_2D=REAL(this % center_lat, kind = 4))
	call this % put_var(var="center_lon", long_name="longitude",units="degrees_east",axis = "X", external_filename =filename_mask, arr_2D=REAL(this % center_lon, kind = 4))
	call this % put_var(var="mask2D", long_name = "mask", units = "unitless", external_filename = filename_mask, arr_2D = this % mask2D)

		write(*,'(A)') 'Adding ocean mask... success.'
!-------
!	print out this % mask2D
!-------
!	if (is_print) then
        write(*,'(A)') 'start printing INMIO this % mask2D'
        allocate (bt    (iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank)))
        call charar(int(this % mask2D),bt,ieast(myrank)-iwest(myrank)+1, jnorth(myrank)-jsouth(myrank)+1)
!        call out_map('t02','_this % mask2D',bt,x1_v,x2_v,ieast(myrank)-iwest(myrank)+1, jnorth(myrank)-jsouth(myrank)+1)
!        call charar(kt,bt,il,jl)
!        call out_map('t02','_kt',bt,x1_t,x2_t,il,jl)
        deallocate(bt)
        write(*,'(A)') 'finish printing kv'
!	end if

!-------
!	analyse grid
!-------
		!call grid_analysis (this)


	end if
	
	end subroutine o_make_mask
	
!================================================================================================================================================================
!===============================================================
! This procedure include user code for any 2D and 3D data, which
! should be included to IC-file. Could be empty.
!===============================================================

	subroutine o_make_ic(this)

	class(off_ocn) :: this
	character(100) :: lev_filename

	! Put topo, kv, kt to IC-file
	call this % put_var(var = "topo", long_name = "Bottom topography", units = "meters", arr_2D = hv(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank)))	
	call this % put_var(var = "kv", long_name = "kv-array", units = "unitless", arr_2D = REAL(kv(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank)), KIND=4))
	call this % put_var(var = "kt", long_name = "kt-array", units = "unitless", arr_2D = REAL(kt(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank)), KIND=4))
	
	! Read interpolated initial conditions, modify it and put to IC-file
    lev_filename = TRIM(DATA_FOLDER)//TRIM(this % name)//"_"//TRIM(I2S(this % nx))//'x'//TRIM(I2S(this % ny))//'_T'
	call process_levitus(this, TRIM(lev_filename)//"_t_an_z.nc", "t_an")
	call process_levitus(this, TRIM(lev_filename)//"_s_an_z.nc", "s_an")
		

!================================================================================================================================================================
    contains

        subroutine process_levitus(this, filename, var_name)
            class(off_ocn) :: this
            character(*), intent(in) :: filename, var_name
            real(kind=4), dimension(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank), kl) :: var
            real(4) :: misVal
            
            ! Read var from interpolated file
            call this % get_var(filename, var_name, arr_3D = var, misVal = misVal)    
            
            ! Process it
            call process_levitus_var(var, var_name, misVal)
            
            ! Save processed var to IC-file
            call this % put_var(var = var_name, long_name = "Levitus initial conditions", units = "", arr_3D = var)
        end subroutine

    end subroutine




!====> ice start <====!
!===============================================================
! This procedures dump additional ocean grid information
! for subsequent read by ice offline module
!===============================================================


!~~~~~~~
	function add_grid_info_filename(this)
!~~~~~~~
	class(off_ocn) :: this
	character(len=256) add_grid_info_filename
	add_grid_info_filename = TRIM(data_folder)//"ADD_GRID_INFO_"//TRIM(this % name)//"_"//TRIM(I2S(this % nx))//'x'//TRIM(I2S(this % ny))//".nc"
	end function add_grid_info_filename


!~~~~~~~
	subroutine additional_dump_grid(this)
!~~~~~~~

!-------
!	2016.08.05
!	editted by: Ibrayev, Rashit
!-------
!##	use o_bsc_module

	class(off_ocn) :: this

	!vectors to compute angle between x direction and true east (of spherical long-lat system)
	real(kind=4), allocatable, dimension(:,:) :: tsc_u, tsc_v, ssc_u, ssc_v
	real(kind=8), allocatable, dimension(:,:) :: angle
	allocate(angle(il, jl))

	! Make grid arrays dx1_**_ - required for ice cice grid generation
	call allocate_o_grid_step_module
	call o_ini_set_grid_step

call this % put_var('x1_v'    ,"x1_v","degrees"	,arr_2D_8=x1_v    (1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
call this % put_var('x2_v'    ,"x2_v","degrees"	,arr_2D_8=x2_v    (1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
call this % put_var('x1_t'    ,"x1_t","degrees"	,arr_2D_8=x1_t    (1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
call this % put_var('x2_t'    ,"x2_t","degrees"	,arr_2D_8=x2_t    (1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))

call this % put_var('dx1_nw_v',"dx1_nw_v","m"	,arr_2D_8=dx1_nw_v(1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
call this % put_var('dx1_ne_v',"dx1_ne_v","m"	,arr_2D_8=dx1_ne_v(1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
call this % put_var('dx2_en_v',"dx2_en_v","m"	,arr_2D_8=dx2_en_v(1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
call this % put_var('dx2_es_v',"dx2_es_v","m"	,arr_2D_8=dx2_es_v(1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))

call this % put_var('dx1_nw_t',"dx1_nw_t","m"	,arr_2D_8=dx1_nw_t(1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
call this % put_var('dx1_ne_t',"dx1_ne_t","m"	,arr_2D_8=dx1_ne_t(1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
call this % put_var('dx2_en_t',"dx2_en_t","m"	,arr_2D_8=dx2_en_t(1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
call this % put_var('dx2_es_t',"dx2_es_t","m"	,arr_2D_8=dx2_es_t(1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
	
	!angle = angle to rotate FROM true east TO bipolar east (lies in [-pi,pi])
	if (bsc_flag .eq. 0) then
		angle=0.
	else
		allocate(tsc_u(il, jl), tsc_v(il, jl))
		allocate(ssc_u(il, jl), ssc_v(il, jl))
		ssc_u = 10.0
		ssc_v = 0.0
		call ssc_tsc_rotate_t(ssc_u, ssc_v, tsc_u, tsc_v, 1, il, 1, jl)
		angle=-asin(tsc_v/ssc_u)
		deallocate(tsc_u,tsc_v,ssc_u,ssc_v)
	end if

call this % put_var('angle',"angle FROM east TO bipolar east in T-points","radians"	,arr_2D_8=angle   (1:il,1:jl),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))

	end subroutine additional_dump_grid


!~~~~~~~
	subroutine additional_dump_mask(this)
!~~~~~~~

!-------
!	2016.08.05
!	editted by: Ibrayev, Rashit
!-------

	class(off_ocn) :: this

	call this % put_var('kv',"kv","levels",arr_2D_8=DBLE(kv(1:il,1:jl)),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))
	call this % put_var('kt',"kt","levels",arr_2D_8=DBLE(kt(1:il,1:jl)),add_nx=2,add_ny=2,external_filename=TRIM(add_grid_info_filename(this)))

	end subroutine additional_dump_mask

!====>  ice end  <====!



!--------------------------------------
!	The following procedures analyse model grid and
!	- output horisontal grid step array to file
!	- output horisontal grid statistics
!	2016.08.05
!	author: Ibrayev, Rashit
!-------
!--------------------------------------

!~~~~~~~
	function dgrid_filename(this)
!~~~~~~~

	class(off_ocn) :: this
	character(len=256) dgrid_filename
	dgrid_filename = TRIM(data_folder)//"grid_step_"//TRIM(this % name)//"_"//TRIM(I2S(this % nx))//'x'//TRIM(I2S(this % ny))//".nc"
	end function dgrid_filename


!~~~~~~~
	subroutine grid_analysis(this)
!~~~~~~~

!-------
!	Analyse model grid and
!	- output horisontal grid step array to file
!	- output horisontal grid statistics
!	2016.08.05
!	author: Ibrayev, Rashit
!-------
!##	use o_bsc_module

	class(off_ocn) :: this

	integer	(kind=4) j0_grid, j1_grid
	
	real (kind=8), allocatable:: dx	(:,:)
	real (kind=8), allocatable:: dy	(:,:)
	real (kind=8), allocatable:: dgrid(:,:)

	write(*,*) 'subroutine grid_analysis start '
	write(*,*) 'subroutine grid_analysis (il,jl)= ',il,jl

	allocate (dx (1:il,1:jl))
	allocate (dy (1:il,1:jl))
	allocate (dgrid (1:il,1:jl))

	dx = 0.
	dy = 0.
	dgrid = 0.

	call allocate_o_grid_step_module
	call o_ini_set_grid_step


!	write(*,*) 'subroutine grid_analysis #01'
!-------
!	compute 'normal' grid steps for V-cells
!-------
	area_tbox = 0.
	do j = 2, jl-1
	do i = 2, il-1
	if(kv(i,j) > 0) then
     	dx(i,j) = dx1_nw_v(i,j) + dx1_ne_v(i,j)
	dy(i,j) = dx2_en_v(i,j) + dx2_es_v(i,j)
	end if
	end do
	end do

!	write(*,*) 'subroutine grid_analysis #02'
!-------
!	mean grid step of V-boxes
!-------
	area_vbox = 0.
	do j = 2, jl-1
	do i = 2, il-1
		if(kv(i,j) > 0) then
			area_vbox (i,j) =				&
     				(dx1_nw_v(i,j)*dx2_en_v(i-1,j)+	&
     				 dx2_es_v(i,j)*dx1_ne_v(i,j-1))+	&
     				(dx1_ne_v(i,j)*dx2_en_v(i,j)+		&
     				 dx2_es_v(i-1,j)*dx1_nw_v(i,j-1))

	dgrid (i,j) = sqrt(area_vbox (i,j))
		end if
	end do
	end do

	open (888,file = TRIM(data_folder)//"grid_mmm_"//TRIM(this % name)//"_"//TRIM(I2S(this % nx))//'x'//TRIM(I2S(this % ny)))
	write(888,'(a,2i5)') '         (il,jl)=', il,jl
	write(888,*) '        min,      max,     aver,   i_min,   j_min,   i_max,   j_max,  n_aver'

!-------
!	compute min, max, average grid steps (dx,dy,sqrt(area_vbox)) for V-cells. all grid
!-------
	j0_grid = 2
	j1_grid = jl-1
	write(888,'(30x,a,2i5)') 'grid all, (j0_grid, j1_grid)=',j0_grid, j1_grid
	call step_mmm('dx',dx,j0_grid, j1_grid)
	call step_mmm('dy',dy,j0_grid, j1_grid)
	call step_mmm('dm',dgrid,j0_grid, j1_grid)

	if (bsc_flag == 1) then
!-------
!	compute min, max, average grid steps (dx,dy,sqrt(area_vbox)) for V-cells. ssc grid
!-------
	j0_grid = 2
	j1_grid = jl_ssc-1
	write(888,'(30x,a,2i5)') 'grid ssc, (j0_grid, j1_grid)=',j0_grid, j1_grid
	call step_mmm('dx',dx,j0_grid, j1_grid)
	call step_mmm('dy',dy,j0_grid, j1_grid)
	call step_mmm('dm',dgrid,j0_grid, j1_grid)

	j0_grid = 2
	j1_grid = jl_ssc
	write(888,'(30x,a,2i5)') 'grid ssc, (j0_grid, j1_grid)=',j0_grid, j1_grid
	call step_mmm('dx',dx,j0_grid, j1_grid)
	call step_mmm('dy',dy,j0_grid, j1_grid)
	call step_mmm('dm',dgrid,j0_grid, j1_grid)

!-------
!	compute min, max, average grid steps (dx,dy,sqrt(area_vbox)) for V-cells. bsc grid
!-------
	j0_grid = jl_ssc+2
	j1_grid = jl-1
	write(888,'(30x,a,2i5)') 'grid bsc, (j0_grid, j1_grid)=',j0_grid, j1_grid
	call step_mmm('dx',dx,j0_grid, j1_grid)
	call step_mmm('dy',dy,j0_grid, j1_grid)
	call step_mmm('dm',dgrid,j0_grid, j1_grid)

	j0_grid = jl_ssc+1
	j1_grid = jl-1
	write(888,'(30x,a,2i5)') 'grid bsc, (j0_grid, j1_grid)=',j0_grid, j1_grid
	call step_mmm('dx',dx,j0_grid, j1_grid)
	call step_mmm('dy',dy,j0_grid, j1_grid)
	call step_mmm('dm',dgrid,j0_grid, j1_grid)

	end if	!	if (bsc_flag = 1) then

	close(888)

!-------
!	save grid steps
!-------
	write(*,*) 'subroutine grid_analysis printing '

	call this % put_var("dgrid","grid step","m",arr_2D_8=dgrid(1:il, 1:jl),add_nx=2,add_ny=2,external_filename=TRIM(dgrid_filename(this)))
	call this % put_var("dx"   ,"x step"   ,"m",arr_2D_8=dx   (1:il, 1:jl),add_nx=2,add_ny=2,external_filename=TRIM(dgrid_filename(this)))
	call this % put_var("dy"   ,"y step"   ,"m",arr_2D_8=dy   (1:il, 1:jl),add_nx=2,add_ny=2,external_filename=TRIM(dgrid_filename(this)))

!-------
!	save model bottom topography for plot
!-------
	hv = 0.
	do j = 1, jl
	do i = 1, il
	if (kv(i,j) >= 1) then
	hv(i,j) = zw(kv(i,j)+1)
	end if
	end do
	end do

	call this % put_var('hv'   ,"hv"       ,"m",arr_2D_8=dble(hv(1:il,1:jl)),add_nx=2,add_ny=2,external_filename = TRIM(dgrid_filename(this)))

	end subroutine grid_analysis


!~~~~~~~
	subroutine step_mmm(dname,dstep,j0_grid, j1_grid)
!~~~~~~~

!-------
!	find min, max, mean of 'dstep' array
!	basic subroutine for analysing model grid and
!	- output horisontal grid step array to file
!	- output horisontal grid statistics
!	2016.08.05
!	author: Ibrayev, Rashit
!-------

!	class(off_ocn) :: this

	character(len=2) dname
	real	(kind=4) dstep_min, dstep_max, dstep_average
	integer	(kind=4) i_dstep_min, j_dstep_min, i_dstep_max, j_dstep_max, n_dstep_average
	integer	(kind=4) j0_grid, j1_grid
	real (kind=8) dstep (il,jl)

!-------
!	compute min, max, average of 'dstep' array
!-------
	dstep_min = 999999.
	i_dstep_min = 0
	j_dstep_min = 0

	dstep_max = 0.
	i_dstep_max = 0
	j_dstep_max = 0

	dstep_average = 0.
	n_dstep_average = 0

	do j = j0_grid, j1_grid
	do i = 2, il-1
	if(kv(i,j) > 0) then

	if ( dstep(i,j) < dstep_min ) then
		dstep_min = dstep(i,j)
		i_dstep_min = i
		j_dstep_min = j
	end if

	if ( dstep(i,j) > dstep_max ) then
		dstep_max = dstep(i,j)
		i_dstep_max = i
		j_dstep_max = j
	end if

	dstep_average = dstep_average + dstep(i,j)
	n_dstep_average = n_dstep_average + 1

	end if
	end do
	end do

	if (n_dstep_average /= 0 ) dstep_average = dstep_average / real(n_dstep_average)

!-------
!	save grid steps
!-------
	write(888,'(a2,3f10.2,5i9)') dname,dstep_min, dstep_max, dstep_average, i_dstep_min, j_dstep_min, i_dstep_max, j_dstep_max, n_dstep_average

	end subroutine step_mmm


	end module off_ocn_module
