
!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	module for sea ice coupling
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru 
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!		(based on the code provided by Schrum, Corinna, 
!		Institute for Oceanography, University of Hamburg, Germany, Corinna.Schrum@hzg.de)
!
! reference:
!	Ushakov K.V., Grankina T.B., Ibrayev R.A., Gromov I.V., 2016. 
!		Simulation of Arctic and North Atlantic ocean water and ice seasonal characteristics 
!		by the INMIO-CICE coupled model. IOP Conf. Series: Earth and Environmental Science. 48(2016) 012013.
!	Schrum C., and J. Backhaus, 1999. 
!		Sensitivity of atmosphere-ocean heat exchange and heat content in North Sea and Baltic Sea. A comparative Assessment. 
!		Tellus 51A, 1999, 526-549.
!--------------------------------------

!============================
! Arrays for SCHRUM and CICE5.1 ice models
!============================

	module ice_module
	save

	real (kind=4), allocatable:: hice	(:,:)
	real (kind=4), allocatable:: tice	(:,:)
	real (kind=4), allocatable:: aice	(:,:)

!------
!	air-ice flux.
!------
	real (kind=4), allocatable:: ssr_ice		(:,:)
	real (kind=4), allocatable:: str_ice		(:,:)
	real (kind=4), allocatable:: sns_ice		(:,:)
	real (kind=4), allocatable:: ltn_ice		(:,:)
	real (kind=4), allocatable:: qh_ai			(:,:)
	integer (kind=4), allocatable:: mask_ice	(:,:)
!-------
!	ice-schrum. list of constants for thermodynamic ice model
!-------
	real (kind=4) q_ice,d_ice,k_ice,epsice,h0_ice,hice0

!------
!	fluxes received from CICE
!------
	real (kind=4), allocatable:: ice_ocn_ssr	(:,:)
	real (kind=4), allocatable:: ice_ocn_heat	(:,:)
	real (kind=4), allocatable:: ice_ocn_salt	(:,:)
	real (kind=4), allocatable:: ice_ocn_stress_x_v (:,:)
	real (kind=4), allocatable:: ice_ocn_stress_y_v (:,:)
	real (kind=4), allocatable:: ice_ocn_water	(:,:)
	real (kind=4), allocatable:: ice_ocn_aice	(:,:)
	real (kind=4), allocatable:: ice_ocn_hice	(:,:)
	real (kind=4), allocatable:: ice_ocn_tice	(:,:)
	
	real (kind=4), allocatable:: ice_ocn_ssr_ai	(:,:)
	real (kind=4), allocatable:: ice_ocn_heat_ai	(:,:)
	real (kind=4), allocatable:: ice_ocn_salt_ai	(:,:)
	real (kind=4), allocatable:: ice_ocn_water_ai	(:,:)
!------
!	same arrays + mask. Use for saving at DG, CP file
!------
	real (kind=4), allocatable:: ice_ocn_ssr_dg	(:,:)
	real (kind=4), allocatable:: ice_ocn_hea_dg	(:,:)
	real (kind=4), allocatable:: ice_ocn_sal_dg	(:,:)
	real (kind=4), allocatable:: ice_ocn_stx_dg	 (:,:)
	real (kind=4), allocatable:: ice_ocn_sty_dg	(:,:)
	real (kind=4), allocatable:: ice_ocn_wat_dg	(:,:)
	real (kind=4), allocatable:: ice_ocn_aic_dg	(:,:)


!------
!	fluxes sent to CICE
!------
	real (kind=4), allocatable:: ocn_ice_frzmlt	(:,:)
	real (kind=4), allocatable:: ocn_ice_tltx	(:,:)
	real (kind=4), allocatable:: ocn_ice_tlty	(:,:)
	real (kind=4), allocatable:: ocn_ice_uocn	(:,:) ! need, because of coupler bug with send-receive
	real (kind=4), allocatable:: ocn_ice_vocn	(:,:) ! of real*8 arrays
	real (kind=4), allocatable:: ocn_ice_sst	(:,:) 
	real (kind=4), allocatable:: ocn_ice_sss	(:,:) 
!------
!	same arrays + mask. Use for saving at DG, CP file
!------
	real (kind=4), allocatable:: ocn_ice_frz_dg	(:,:)
	real (kind=4), allocatable:: ocn_ice_tlx_dg	(:,:)
	real (kind=4), allocatable:: ocn_ice_tly_dg	(:,:)
	real (kind=4), allocatable:: ocn_ice_uoc_dg	(:,:) ! need, because of coupler bug with send-receive
	real (kind=4), allocatable:: ocn_ice_voc_dg	(:,:) ! of real*8 arrays
	real (kind=4), allocatable:: ocn_ice_sst_dg	(:,:) 
	real (kind=4), allocatable:: ocn_ice_sss_dg	(:,:) 

end module ice_module


!=======
	subroutine allocate_ice_module
!=======
	use o_grid_module
	use ice_module
	use o_par_module
	
	implicit none
	namelist /OCN_ICE/ q_ice, d_ice, k_ice, epsice, h0_ice, hice0
	
	allocate (hice 		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (tice 		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (aice 		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ssr_ice	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (str_ice	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (sns_ice	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ltn_ice	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (qh_ai		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (mask_ice	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )

	hice	= 0.
	tice	= 0.
	aice	= 0.
	ssr_ice	= 0.
	str_ice	= 0.
	sns_ice	= 0.
	ltn_ice	= 0.
	qh_ai	= 0.
	mask_ice=0.

	allocate (ice_ocn_ssr		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_heat		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_salt		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_stress_x_v 	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_stress_y_v 	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_water		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_aice		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_hice		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_tice		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )

	allocate (ice_ocn_ssr_ai	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_heat_ai	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_salt_ai	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_water_ai	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	
	allocate (ice_ocn_ssr_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_hea_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_sal_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_stx_dg 	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_sty_dg 	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_wat_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ice_ocn_aic_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )

	allocate (ocn_ice_frzmlt	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_tltx		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_tlty		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_uocn		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_vocn		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_sst		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_sss		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )

	allocate (ocn_ice_frz_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_tlx_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_tly_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_uoc_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_voc_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_sst_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ocn_ice_sss_dg	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	
	ice_ocn_ssr	= 0.0
	ice_ocn_heat	= 0.0
	ice_ocn_salt	= 0.0
	ice_ocn_stress_x_v	= 0.0
	ice_ocn_stress_y_v	= 0.0
	ice_ocn_water	= 0.0
	ice_ocn_aice	= 0.0
	ice_ocn_hice	= 0.0
	ice_ocn_tice	= 0.0
	
	ice_ocn_ssr_ai		= 0.0
	ice_ocn_heat_ai		= 0.0
	ice_ocn_salt_ai		= 0.0
	ice_ocn_water_ai	= 0.0	

	ice_ocn_ssr_dg	= 0.0
	ice_ocn_hea_dg	= 0.0
	ice_ocn_sal_dg	= 0.0
	ice_ocn_stx_dg	= 0.0
	ice_ocn_sty_dg	= 0.0
	ice_ocn_wat_dg	= 0.0
	ice_ocn_aic_dg	= 0.0

	ocn_ice_frzmlt		= 0.0
	ocn_ice_tltx		= 0.0
	ocn_ice_tlty		= 0.0
	ocn_ice_uocn		= 0.0
	ocn_ice_vocn		= 0.0
	ocn_ice_sst		= 0.0
	ocn_ice_sss		= 0.0
	
	ocn_ice_frz_dg		= 0.0
	ocn_ice_tlx_dg		= 0.0
	ocn_ice_tly_dg		= 0.0
	ocn_ice_uoc_dg		= 0.0
	ocn_ice_voc_dg		= 0.0
	ocn_ice_sst_dg		= 0.0
	ocn_ice_sss_dg		= 0.0
	
	
	open(8,file = './configure/ocn_list.in',err=1)
	read(8, nml = OCN_ICE,err=1)
	close(8)
	!print *, 'q_ice, d_ice, k_ice, epsice, h0_ice, hice0 = ',q_ice, d_ice, k_ice, epsice, h0_ice, hice0
	
	return
1	call ocn_error('Cannot read ice nameleist! Stop.','LCL')
	end


!=======
	subroutine make_ocn_ice_fluxes	!Data to be sent to ice
!=======
	use o_grid_module
	use o_basic_module
	use ice_module
	use o_tf_module
	use o_td_module
!	use shared_module
	use o_par_module
	implicit none
	
	real(4),parameter :: c_water = 4217.			!water heat capacity, J/kg/K
	real(4) tfreez,freez_point
	
!-------
!	freez-melt potential (W/m2) See CICE manual p.8 and POP manual p. 106
!		If the freez-melt potential is positive then frazil ice in ocn component forms immediately. 
!		If negative then ice is immediately melt by heat fluxes collected in CICE during coupling interval 
!-------
	do j=jsouth(myrank),jnorth(myrank)
	do i=iwest (myrank),ieast (myrank)
	tfreez=freez_point(sngl(t_c(1,i,j,2)),0.)

	ocn_ice_frzmlt(i,j)=(tfreez-t_c(1,i,j,1))*c_water*densit*dx3_t_c(2,i,j)/(ocn_ice_mapping*60.)

	if(ocn_ice_frzmlt(i,j).gt.0) then
		!hice(i,j)=hice(i,j) + (tfreez-t_c(1,i,j,1))*c_water*densit*dx3_t_c(2,i,j)/q_ice/d_ice
		t_l(1,i,j,1)=tfreez
		t_c(1,i,j,1)=tfreez
		t_f(1,i,j,1)=tfreez
	else
		!hice(i,j)=hice(i,j) + ice_ocn_heat(i,j)*(ocn_ice_mapping*60.)/q_ice/d_ice
	end if


	end do
	end do
!-------
!	sea surface topography gradient. Normal component = 0 next to land.
!-------
	ocn_ice_tltx = 0.
	ocn_ice_tlty = 0.
	do j=jsouth(myrank),jnorth(myrank)
	do i=iwest (myrank),ieast (myrank)
		if(minval(kt(i-1:i+1,j))>0) & 
			ocn_ice_tltx(i,j)= (ssl_c(i+1,j)-ssl_c(i-1,j))/(dx1_nw_v(i,j-1)+dx1_ne_v(i,j-1)+dx1_nw_v(i-1,j-1)+dx1_ne_v(i-1,j-1))
		if(minval(kt(i,j-1:j+1))>0) &
			ocn_ice_tlty(i,j)= (ssl_c(i,j+1)-ssl_c(i,j-1))/(dx2_en_v(i-1,j)+dx2_es_v(i-1,j)+dx2_en_v(i-1,j-1)+dx2_es_v(i-1,j-1))
	end do
	end do

!-------
!	SST & SSS
!-------
	ocn_ice_sst=sngl(t_c(1,:,:,1))
	ocn_ice_sss=sngl(t_c(1,:,:,2))
!-------
!	SSU & SSV
!-------
	ocn_ice_uocn=sngl(u_c(1,:,:))
	ocn_ice_vocn=sngl(v_c(1,:,:))

	if(mod(time_l,1).eq.0	.and.	j_solution_control.ge.jsouth(myrank)		&
				.and.	j_solution_control.le.jnorth(myrank)		&
				.and.	i_solution_control.ge. iwest(myrank)		&
				.and.	i_solution_control.le. ieast(myrank) ) then
		i=i_solution_control
		j=j_solution_control

		write(*,'(a3,a21,i7, i5,i2.2,i2.2, a1,i2.2,a1,i2.2,  6f18.8,1x,e16.8)')						& 
			run_experiment,												&
			' make_ocn_ice_fluxes ',										&
			time_l_in_run,												&
			time_year,time_month,time_day,'/',time_hour,':',time_min,			&
			tfreez,tfreez-t_c(1,i,j,1),ocn_ice_frzmlt(i,j)
	end if
	end	

!=======
	subroutine make_ice_ocn_fluxes	!Data received from ice
!=======	
	use o_grid_module
	use ice_module
	use o_par_module
!	write (*,'(a20,i5,a1,i4,a1,i2.2,a1,i2.2,a1,i2.2,a1)') 'make_ice_ocn_fluxes ',time_l,'/',time_year,'.',time_month,'.',time_day,'.',time_hour,'h'

	aice = ice_ocn_aice
	hice = ice_ocn_hice
	tice = ice_ocn_tice
	
	if(ice_remapping_mode==0) then
		ice_ocn_ssr=ice_ocn_ssr_ai*aice
		ice_ocn_heat=ice_ocn_heat_ai*aice
		ice_ocn_salt=ice_ocn_salt_ai*aice
		ice_ocn_water=ice_ocn_water_ai*aice
	else
		ice_ocn_ssr=ice_ocn_ssr_ai
		ice_ocn_heat=ice_ocn_heat_ai
		ice_ocn_salt=ice_ocn_salt_ai
		ice_ocn_water=ice_ocn_water_ai
	end if
		
!	The sign is changed not here, but just after CICE_to_CPL rearrange
!	ice_ocn_stress_x_v = -ice_ocn_stress_x_v
!	ice_ocn_stress_y_v = -ice_ocn_stress_y_v	
!
!	do j=jsouth(myrank),jnorth(myrank)
!	do i=iwest (myrank),ieast (myrank)
!	if(kt(i,j) > 0 ) aice(i,j) = 1.	
!	if(kt(i,j) == 0 ) aice(i,j) = 0.	
!!	if(kt(i,j) == 0 ) aice(i,j) = -99999.	
!	end do
!	end do

!-------
!	transfer from CICE units to INMIO units
!-------
!			==> [kg / m2 / s]


!-------
!	eliminate spurious polynyas in case of inaccurate interpolation from CICE
!-------
!	do j=jsouth(myrank),jnorth(myrank)
!	do i=iwest (myrank),ieast (myrank)
!		if(	minval(kt(i-1:i+1,j-1:j+1))==0 .and. maxval(aice(i-1:i+1,j-1:j+1)).gt.0.9 .and. kt(i,j).gt.0) aice(i,j)=1.	
!	end do
!	end do

!-------
!	Fixing ice that can appear in Indian ocean due to overland extrapolation on reduced (only!) CICE grid
!-------
	if(cice_and_schrum) then
		!where(x2_t.lt.35.) aice=0.
		do j=jsouth(myrank)-1,jnorth(myrank)+1
		do i=iwest (myrank)-1,ieast (myrank)+1
			if(x2_t(i,j).lt.35.) then 
				aice(i,j)=0.
				hice(i,j)=0.
			end if
		end do
		end do
	end if

	if(mod(time_l,1).eq.0	.and.	j_solution_control.ge.jsouth(myrank)		&
				.and.	j_solution_control.le.jnorth(myrank)		&
				.and.	i_solution_control.ge. iwest(myrank)		&
				.and.	i_solution_control.le. ieast(myrank)) then
		write(*,'(a3,a21,i7, i5,i2.2,i2.2, a1,i2.2,a1,i2.2,  6f13.8,1x,e16.8)')						& 
			run_experiment,												&
			' make_ice_ocn_fluxes ',										&
			time_l_in_run,												&
			time_year,time_month,time_day,'/',time_hour,':',time_min
	end if

	end
	
!=======
	real(4) function freez_point(s1,p) 
!=======
	use o_basic_module
	IMPLICIT REAL(A-H, O-Z), INTEGER(I-N)

!	freezing point of sea water (millero 1978) !p-term commnted out!!
	data ca/-0.0575/,cb/1.710523e-3/,cc/-2.154996e-4/,cd/-7.53e-3/
	s=s1+sngl(S_ref)
	!--Schrum formula--
	!freez_point = s*(ca+cb*sqrt(s)+cc*s)  !+cd*p 
	
	!--CICE mushy formula--
	freez_point =  s / (-18.48  + 18.48e-3 * s)
	return 
	end 
