#include "o_defines.f90"


!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	ocean dynamic equations. 
!	advection terms for momentum equations
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru
! 
! reference:
!	Ibrayev R.A., 2001: Model of enclosed and semi-enclosed sea hydrodynamics. 
!		Russ. J. Numer. Anal. Math. Modelling, 2001, vol. 16, No 4, p. 291-304.
!
!--------------------------------------

!--------------------------------------
!	compute advection terms for momentum equations
!--------------------------------------

	subroutine o_d_am_cd
	use o_basic_module
	use o_grid_module
	use o_d_brcl_module
	implicit none
	real (kind=main_kind) flow_local	!registry variable for optimization
!-------
!	velocities at the eastern walls of v-box
!-------
	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)-1	,ieast (myrank)
	do k=1,kl
		func_u_e (k,i,j)=(cv(k,i+1,j)*cv(k,i,j))*			&
     		(u_c(k,i,j)*dx1_ne_t(i+1,j)+u_c(k,i+1,j)*dx1_nw_t(i+1,j))/	&
     		(           dx1_ne_t(i+1,j)+             dx1_nw_t(i+1,j))

		func_v_e (k,i,j)=(cv(k,i+1,j)*cv(k,i,j))*			&
     		(v_c(k,i,j)*dx1_ne_t(i+1,j)+v_c(k,i+1,j)*dx1_nw_t(i+1,j))/	&
     		(           dx1_ne_t(i+1,j)+             dx1_nw_t(i+1,j))
	end do
	end do
	end do

!-------
!		velocities at the northern walls of v-box
!-------
	do j=jsouth(myrank)-1	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
	do k=1,kl
		func_u_n (k,i,j)=(cv(k,i,j+1)*cv(k,i,j))*			&
     		(u_c(k,i,j)*dx2_en_t(i,j+1)+u_c(k,i,j+1)*dx2_es_t(i,j+1))/	&
     		(           dx2_en_t(i,j+1)+             dx2_es_t(i,j+1))

		func_v_n (k,i,j)=(cv(k,i,j+1)*cv(k,i,j))*			&
     		(v_c(k,i,j)*dx2_en_t(i,j+1)+v_c(k,i,j+1)*dx2_es_t(i,j+1))/	&
     		(           dx2_en_t(i,j+1)+             dx2_es_t(i,j+1))
	end do
	end do
	end do

!-------
!	velocities at the bottom walls of v-box
!-------
	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
	do k=1,kv(i,j)-1
		func_u_b(k+1,i,j)=0.5*( u_c(k,i,j)+u_c(k+1,i,j) )	!func_u_b(k+1) is velocity on the bottom of the k-th level
		func_v_b(k+1,i,j)=0.5*( v_c(k,i,j)+v_c(k+1,i,j) )
	end do
	end do
	end do

	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
		do k=1,kv(i,j)
			func_w_b(k,i,j)=0.25 * ( (w(k,i,j)+w(k,i+1,j+1)) + (w(k,i,j+1)+w(k,i+1,j)) )	!func_w_b(k+1) is also velocity on the bottom of the k-th level
		end do
		func_w_b(kv(i,j)+1,i,j)=0.		!to make advection zero on surface and bottom
		func_w_b(	 1,i,j)=0.
	end do
	end do

!-------
!	flows at the eastern walls of v-box
!-------
	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)-1	,ieast (myrank)
	do k=1,kl
		!volume flow [m3/s]
		flow_local=func_u_e (k,i,j)*						&
     			(dx2_en_v(i,j)+dx2_es_v(i,j))*					&
     			(dx3_t_c(k+1,i+1,j+1)+dx3_t_c(k+1,i+1,j  )) * 0.5
		
		!velocity-volume flow [(m/s)*m3/s]
		flow_u_e(k,i,j)=func_u_e (k,i,j)*flow_local
		flow_v_e(k,i,j)=func_v_e (k,i,j)*flow_local
	end do
	end do
	end do

!-------
!		flows at the northern walls of v-box
!-------
	do j=jsouth(myrank)-1	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
	do k=1,kl
		!volume flow [m3/s]
		flow_local=func_v_n (k,i,j)*						&
     			(dx1_nw_v(i,j)+dx1_ne_v(i,j))*					&
     			(dx3_t_c(k+1,i  ,j+1)+dx3_t_c(k+1,i+1,j+1)) * 0.5
	
		!velocity-volume flow [(m/s)*m3/s]
		flow_u_n(k,i,j)=func_u_n (k,i,j)*flow_local
		flow_v_n(k,i,j)=func_v_n (k,i,j)*flow_local
	end do
	end do
	end do

!-------
!	advection (WARNING: it is stored in unfamilar array names, accessed later in o_d_motor)
!-------
	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
	do k=1,kv(i,j)
		flux_u_n(k,i,j) = area_vbox_v(i,j) * (				&	!horizontal advection of U
			(flow_u_e(k,i  ,j  ) - flow_u_e(k,i-1,j  )) +		&
			(flow_u_n(k,i  ,j  ) - flow_u_n(k,i  ,j-1)))

		flux_u_e(k,i,j) = area_vbox_v(i,j) * (				&	!horizontal advection of V
			(flow_v_e(k,i  ,j  ) - flow_v_e(k,i-1,j  )) +		&
			(flow_v_n(k,i  ,j  ) - flow_v_n(k,i  ,j-1)))

		flux_v_n(k,i,j)=(func_w_b(k+1,i,j)*func_u_b(k+1,i,j) 		&	!vertical advection of U
     				-func_w_b(k  ,i,j)*func_u_b(k  ,i,j) )

		flux_v_e(k,i,j)=(func_w_b(k+1,i,j)*func_v_b(k+1,i,j) 		&	!vertical advection of V
     				-func_w_b(k  ,i,j)*func_v_b(k  ,i,j) )
	end do
	end do
	end do

	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
	do k=1,kv(i,j)
		flux_u_n(k,i,j)=flux_u_n(k,i,j) + flux_v_n(k,i,j)		!full advection of U
		flux_u_e(k,i,j)=flux_u_e(k,i,j) + flux_v_e(k,i,j)		!full advection of V
	end do
	end do
	end do		
	end
