#include "o_defines.f90"


!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	ocean dynamic equations. 
!	module for baroclinic components
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru 
!--------------------------------------


!======================================
! Contains baroclinic dymamic variables
!======================================

	module o_d_brcl_module

	save

!-------
!	basic functions at non-current time-step
!-------
	real (kind=main_kind), allocatable:: u_l		(:,:,:)
	real (kind=main_kind), allocatable:: u_f		(:,:,:)
	real (kind=main_kind), allocatable:: v_l		(:,:,:)
	real (kind=main_kind), allocatable:: v_f		(:,:,:)

!-------
!	sea surface topography arrays
!-------
	!real (kind=8), allocatable:: ssl_v_c		(:,:)
	real (kind=8), allocatable:: ssl_v_f		(:,:)
	real (kind=8), allocatable:: ssl_f		(:,:)
	real (kind=8), allocatable:: water_ssf		(:,:)
	real (kind=8), allocatable:: rx		(:,:)
	real (kind=8), allocatable:: ry		(:,:)

!-------
!	arrays for discontinuity equation
!-------
!z	real (kind=4), allocatable:: wbott		(:,:)
	real (kind=main_kind), allocatable:: w_surf_c		(:,:)
!z	real (kind=main_kind), allocatable:: w_surf_f		(:,:)

	real (kind=main_kind), allocatable:: flow_u_e		(:,:,:)
	real (kind=main_kind), allocatable:: flow_v_n		(:,:,:)
	real (kind=main_kind), allocatable:: flow_u_w		(:,:,:)
	real (kind=main_kind), allocatable:: flow_v_s		(:,:,:)
	real (kind=main_kind), allocatable:: flow_u_n		(:,:,:)
	real (kind=main_kind), allocatable:: flow_v_e		(:,:,:)
	real (kind=main_kind), allocatable:: flow_w_b		(:,:,:)

!-------
!	arrays for momentum advection operator
!-------
	real (kind=main_kind), allocatable:: func_u_e		(:,:,:)
	real (kind=main_kind), allocatable:: func_v_e		(:,:,:)
	real (kind=main_kind), allocatable:: func_u_n		(:,:,:)
	real (kind=main_kind), allocatable:: func_v_n		(:,:,:)
	real (kind=main_kind), allocatable:: func_u_b		(:,:,:)
	real (kind=main_kind), allocatable:: func_v_b		(:,:,:)
	real (kind=main_kind), allocatable:: func_w_b		(:,:,:)

!-------
!	momentum lateral mixing operator arrays
!-------
	real (kind=main_kind), allocatable:: flux_u_e		(:,:,:)
	real (kind=main_kind), allocatable:: flux_v_e		(:,:,:)
	real (kind=main_kind), allocatable:: flux_u_n		(:,:,:)
	real (kind=main_kind), allocatable:: flux_v_n		(:,:,:)

	real (kind=main_kind), allocatable:: u_laplace		(:,:,:)
	real (kind=main_kind), allocatable:: v_laplace		(:,:,:)
	real (kind=main_kind), allocatable:: u_laplace2	(:,:,:)
	real (kind=main_kind), allocatable:: v_laplace2	(:,:,:)
	real (kind=main_kind), allocatable:: u_laplace_pure	(:,:,:)
	real (kind=main_kind), allocatable:: v_laplace_pure	(:,:,:)
		
	real (kind=main_kind), allocatable:: derivative_d_x1	(:,:,:)
	real (kind=main_kind), allocatable:: derivative_d_x2	(:,:,:)
	real (kind=main_kind), allocatable:: force_x1		(:,:,:)
	real (kind=main_kind), allocatable:: force_x2		(:,:,:)

!-------
!	bottom friction coeff.
!-------
	real (kind=main_kind), allocatable:: cbot		(:,:)

!-------
!	mixing, vertical, coefficient
!-------
	real (kind=4), allocatable:: km		(:,:,:)

!------------------
!	bt_sweq function
!------------------
	real (kind=16), allocatable:: flow_v_n_2d	(:,:)
	real (kind=16), allocatable:: flow_v_s_2d	(:,:)
	real (kind=16), allocatable:: flow_u_e_2d	(:,:)
	real (kind=16), allocatable:: flow_u_w_2d	(:,:)

	real (kind=8), allocatable:: u_ksum_c	 	 (:,:)
	real (kind=8), allocatable:: u_ksum_f	 	 (:,:)
	real (kind=8), allocatable:: v_ksum_c	 	 (:,:)
	real (kind=8), allocatable:: v_ksum_f	 	 (:,:)

   	real (kind=8), allocatable:: u_bt_a_f		 (:,:)
	real (kind=8), allocatable:: v_bt_a_f		 (:,:)
	real (kind=8), allocatable:: h_bt_a_f		 (:,:)
	real (kind=8), allocatable:: div_2d		(:,:)

	real (kind=main_kind), allocatable:: func_h_e_vbox	(:,:)
	real (kind=main_kind), allocatable:: func_h_n_vbox	(:,:)
	real (kind=main_kind), allocatable:: grad_h_x		(:,:)
	real (kind=main_kind), allocatable:: grad_h_y		(:,:)

	real (kind=main_kind), allocatable:: gh_mask		(:,:)
	real (kind=4), allocatable:: ct3v		(:,:,:)

!-------
!	physical parameters
!-------
	real (kind=8), allocatable:: coriolis		(:,:)
	real (kind=4), allocatable:: am_lapl_e(:,:,:),am_lapl_n(:,:,:)
	real (kind=4), allocatable:: am_bh_e  (:,:,:),am_bh_n  (:,:,:)
	real (kind=4) am_lapl,am_lapl_2d,am_bh
	real (kind=4) c_smag_lapl, c_smag_bh
	integer am_lapl_mode,am_bh_mode
!-------
!	numerical scheme parameters
!-------
	real (kind=4)	coriolis_alfa,coriolis_gamma,coriolis_betta
	end module o_d_brcl_module
!***********************************************************************

	subroutine allocate_o_d_brcl_module
	use o_d_brcl_module
	use o_grid_module
	implicit none

!-------
!	basic functions at non-current time-step
!-------
	allocate (u_l		(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (u_f		(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (v_l		(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (v_f		(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	u_f	= 0.
	v_f	= 0.
	u_l	= 0.
	v_l	= 0.

!-------
!	sea surface topography arrays
!-------
	allocate (ssl_f		(	iwest (myrank)-2:ieast (myrank)+2, jsouth(myrank)-2:jnorth(myrank)+2) )
	!allocate (ssl_v_c	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ssl_v_f	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (water_ssf	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (rx		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ry		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	ssl_f	= 0.
	!ssl_v_c	= 0.
	ssl_v_f = 0.
	water_ssf = 0.
	rx	= 0.
	ry	= 0.

!-------
!	arrays for discontinuity equation
!-------
!z	allocate (wbott		(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (w_surf_c	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
!z	allocate (w_surf_f	(	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
!z	wbott	 = 0.
	w_surf_c = 0.
!z	w_surf_f = 0.

!-------
!	arrays for momentum advection operator
!-------
	allocate (func_u_e	(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (func_v_e	(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (func_u_n	(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (func_v_n	(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (func_u_b	(kl+1,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (func_v_b	(kl+1,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (func_w_b	(kl+1,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	func_u_e = 0.
	func_v_e = 0.
	func_u_n = 0.
	func_v_n = 0.
	func_u_b = 0.
	func_v_b = 0.
	func_w_b = 0.

!-------
!	momentum lateral mixing operator arrays
!-------
	allocate (flux_u_e	(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (flux_v_e	(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (flux_u_n	(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (flux_v_n	(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	flux_u_e = 0.
	flux_v_e = 0.
	flux_u_n = 0.
	flux_v_n = 0.


	allocate (u_laplace	(kl  ,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (v_laplace	(kl  ,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (u_laplace2	(kl  ,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (v_laplace2	(kl  ,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (u_laplace_pure	(kl  ,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (v_laplace_pure	(kl  ,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	u_laplace	=0.
	v_laplace	=0.
	u_laplace2	=0.
	v_laplace2	=0.
	u_laplace_pure	=0.
	v_laplace_pure	=0.
	
	
	allocate (derivative_d_x1(kl+1,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (derivative_d_x2(kl+1,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (force_x1	(kl  ,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (force_x2	(kl  ,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	derivative_d_x1	= 0.
	derivative_d_x2	= 0.
	force_x1	= 0.
	force_x2	= 0.

!-------
!	bottom friction coeff.
!-------
	allocate (cbot		(	iwest (myrank)  :ieast (myrank)  , jsouth(myrank)  :jnorth(myrank)  ) )
	cbot	=0.

!-------
!	mixing, vertical, coefficient
!-------
	allocate (km		(kl  ,	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	km	= 0.

!	bt_sweq function

	allocate (flow_u_e	(kl+1, 	iwest (myrank)-2:ieast (myrank)+2, jsouth(myrank)-2:jnorth(myrank)+2) )
	allocate (flow_v_n	(kl,   	iwest (myrank)-2:ieast (myrank)+2, jsouth(myrank)-2:jnorth(myrank)+2) )
	allocate (flow_u_w	(kl,   	iwest (myrank)-2:ieast (myrank)+2, jsouth(myrank)-2:jnorth(myrank)+2) )
	allocate (flow_v_s	(kl,   	iwest (myrank)-2:ieast (myrank)+2, jsouth(myrank)-2:jnorth(myrank)+2) )
	allocate (flow_u_n	(kl+1,  iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (flow_v_e	(kl,    iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (flow_w_b	(kl+1,  iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )	
	flow_u_n = 0.
	flow_u_e = 0.
	flow_u_w = 0.
	flow_v_n = 0.
	flow_v_e = 0.
	flow_v_s = 0.
	flow_w_b = 0.

	
	allocate (flow_v_n_2d	(   	iwest (myrank)-2:ieast (myrank)+2,	jsouth(myrank)-2:jnorth(myrank)+2) )
	allocate (flow_u_w_2d	(   	iwest (myrank)-2:ieast (myrank)+2,	jsouth(myrank)-2:jnorth(myrank)+2) )
	allocate (flow_u_e_2d	(   	iwest (myrank)-2:ieast (myrank)+2,	jsouth(myrank)-2:jnorth(myrank)+2) )
	allocate (flow_v_s_2d	(   	iwest (myrank)-2:ieast (myrank)+2,	jsouth(myrank)-2:jnorth(myrank)+2) )
	allocate (div_2d	(	iwest (myrank)-2:ieast (myrank)+2,	jsouth(myrank)-2:jnorth(myrank)+2) )
	flow_v_n_2d = 0.
	flow_v_s_2d = 0.
	flow_u_e_2d = 0.
	flow_u_w_2d = 0.
	div_2d	 = 0.

	allocate (u_ksum_c	(   	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (v_ksum_c	(   	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (u_ksum_f	(   	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (v_ksum_f	(   	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	u_ksum_c = 0.
	v_ksum_c = 0.
	u_ksum_f = 0.
	v_ksum_f = 0.

	allocate (u_bt_a_f	(   	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (v_bt_a_f	(   	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (h_bt_a_f	(   	iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	u_bt_a_f = 0.
	v_bt_a_f = 0.
	h_bt_a_f = 0.

	allocate (func_h_e_vbox	(	iwest (myrank)-1:ieast (myrank)+1,  	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (func_h_n_vbox	(	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (grad_h_x	(	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (grad_h_y	(	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (gh_mask	(	iwest (myrank)-1:ieast (myrank)+1,  	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (ct3v		(kl,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	
	func_h_e_vbox	= 0.
	func_h_n_vbox	= 0.
	grad_h_x	= 0.
	grad_h_y	= 0.
	gh_mask		= 0.
	ct3v		= 0.

!-------
!	physical parameters
!-------
	allocate (coriolis	(   	iwest (myrank)  :ieast (myrank)  , jsouth(myrank)  :jnorth(myrank)  ) )
	allocate (am_lapl_e	(kl,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (am_lapl_n	(kl,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (am_bh_e	(kl,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (am_bh_n	(kl,	iwest (myrank)-1:ieast (myrank)+1,	jsouth(myrank)-1:jnorth(myrank)+1) )
	coriolis = 0.
		
	!Initialization of viscosities. Will be overwritten if Smagorinsky is on.
	am_lapl_e=am_lapl
	am_lapl_n=am_lapl
	am_bh_e=am_bh
	am_bh_n=am_bh
		
	return
	end
