#include "o_defines.f90"


!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	ocean dynamic equations.
!	Right side for Equations of Motion and for sea level equation
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru 
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!
! reference:
!	Ibrayev R.A., 2001: Model of enclosed and semi-enclosed sea hydrodynamics. 
!		Russ. J. Numer. Anal. Math. Modelling, 2001, vol. 16, No 4, p. 291-304.
!--------------------------------------

!--------------------------------------
!	Compute Right side for Equations of Motion and for sea level equation
!	on time step  l   ,with known currents	on time steps (l),(l-1).
!	vertical mixing:	Crank-Nickolson scheme
!	wind stress:		defined at 'l' timestep
!	bottom stress:		implicit formulation (at 'l-1' timestep)
!	lateral mixing:		calculated in Sudr.Mflux9
!--------------------------------------

	subroutine o_d_motor (ctime,ul,vl,uc,vc,u0,v0,dx30)

	use o_grid_module
	use o_basic_module
	use o_d_brcl_module
	use o_tf_module
	use ice_module
	use o_td_module
	use o_par_module
	
	implicit none

	real (kind=main_kind) ul	(kl, iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1)
	real (kind=main_kind) vl	(kl, iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1)
	real (kind=main_kind) uc	(kl, iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1)
	real (kind=main_kind) vc	(kl, iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1)
	real (kind=main_kind) u0	(kl, iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1)
	real (kind=main_kind) v0	(kl, iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1)
	real (kind=main_kind) dx30	(kl+1,iwest(myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1)	!This is dx3_v_l or dx3_v_c (according to leapfrog or euler step)

	real (kind=4)	aice_vv
	real (kind=4)	karman  
	real (kind=main_kind)	u_impl(kl),v_impl(kl),ru_impl(kl),rv_impl(kl),factor_a
	real (kind=main_kind)	arix(kl),brix(kl),crix(kl),rrix(kl),frix(kl)
	real (kind=main_kind)	bbrix(kl),rrrix(kl)

	real (kind=4) dh, argument, Cdbot, Cdbot1, Cdbot2, ubot, vbot, botcur, c
	real (kind=4) b, e, ccrank, ctime
	integer kbot, kz1, kz2

!-------
!		density at the walls of v-box
!-------
	do j=jsouth(myrank)-1	,jnorth(myrank)
	do i=iwest (myrank)-1	,ieast (myrank)
	do k=1,kl
		func_u_e (k,i,j) = ( d_c(k,i+1,j+1) + d_c(k,i+1,j) ) * 0.5
		func_u_n (k,i,j) = ( d_c(k,i,j+1) + d_c(k,i+1,j+1) ) * 0.5
	end do
	end do
	end do

!-------
!		density gradients
!-------
	do j = jsouth(myrank)	,jnorth(myrank)
	do i = iwest (myrank)	,ieast (myrank)
	do k = 1,	kv(i,j)
	derivative_d_x1 (k,i,j) =									&
     		(func_u_e (k,i  ,j  ) - func_u_e (k,i-1,j  ) ) /	&
     		(dx1_ne_t (  i  ,j  ) + dx1_nw_t (  i+1,j  ) )

	derivative_d_x2 (k,i,j) =									&
     		(func_u_n (k,i  ,j  ) - func_u_n (k,i  ,j-1) ) /	&
     		(dx2_es_t (  i  ,j+1) + dx2_en_t (  i  ,j  ) )
	end do
	end do
	end do

!-------
!#1.3		baroclinic pressure gradient terms
!-------
	k = 1
	do j = jsouth(myrank)	,jnorth(myrank)
	do i = iwest (myrank)	,ieast (myrank)
	if( k.le.kv(i,j) ) then
		force_x1(k,i,j) = (0.5*hz(k)) * derivative_d_x1(k,i,j)
		force_x2(k,i,j) = (0.5*hz(k)) * derivative_d_x2(k,i,j)
	end if
	end do
	end do

	do j = jsouth(myrank)	,jnorth(myrank)
	do i = iwest (myrank)	,ieast (myrank)
	do k = 2,	kv(i,j)
	force_x1(k,i,j) = force_x1(k-1,i,j)+					&
     			(hz(k)*0.5) * (	derivative_d_x1(k-1,i,j) +		&
     					derivative_d_x1(k  ,i,j)  )
	
	force_x2(k,i,j) = force_x2(k-1,i,j)+					&
     			(hz(k)*0.5) * (	derivative_d_x2(k-1,i,j) +		&
     					derivative_d_x2(k  ,i,j)  )
	end do
	end do
	end do

!-------
!	compute explicit part (1/2) of vertical diffusion terms without b.c.
!-------
	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
		do k=2,kv(i,j)
			flow_u_n(k,i,j)=hzv(k)*km(k,i,j)*( ul(k,i,j)-ul(k-1,i,j) )
			flow_u_e(k,i,j)=hzv(k)*km(k,i,j)*( vl(k,i,j)-vl(k-1,i,j) )
		end do
		flow_u_n(	 1,i,j)=0.
		flow_u_e(	 1,i,j)=0.
		flow_u_n(kv(i,j)+1,i,j)=0.
		flow_u_e(kv(i,j)+1,i,j)=0.
	end do
	end do

	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
	do k=1,kv(i,j)
			func_u_b (k,i,j)=0.5*(flow_u_n(k+1,i,j)-flow_u_n(k,i,j))
			func_v_b (k,i,j)=0.5*(flow_u_e(k+1,i,j)-flow_u_e(k,i,j))
	end do
	end do
	end do

!-------
!	Compute coefficient for Stress Terms at the bottom
!-------
!-------
!	Bottom drug stress according Weatherly & Martin, 1978
!	(see Blumberg & Mellor, 1983)
!	z0mp5=z0*.5, 
!	where z0=0.01 m, multiplyer .5 indicate the half 
!		height of the bottom box
!-------

	Cdbot1=0.0025
	karman=0.4

	k = 1
	do j = jsouth(myrank), jnorth(myrank)
	do i = iwest (myrank), ieast (myrank)
		if(k <= kv(i,j)) then

			kbot = kv(i,j)
			dh = hzw(kbot+1)
			argument = 50.*dh	!	(1/2)*dh/(1.cm)=0.5*dh/(0.01m)

			Cdbot2 = ( karman/( aLOG(argument) ) )**2
			Cdbot = amax1(Cdbot1, Cdbot2)

			ubot = ul(kbot,i,j)
			vbot = vl(kbot,i,j)

			botcur = sqrt(ubot*ubot+vbot*vbot)
			c = Cdbot*botcur

			cbot(i,j) = c
		end if	
	end do
	end do

!-------
!	lateral friction term
!-------
	select case (mixh_m_mode)	!if mixh_m_mode==2 then just uv_laplace2=uv_laplace2
	case(0)
		u_laplace2 =  u_laplace
		v_laplace2 =  v_laplace
	case(1)
		u_laplace2 =  u_laplace2+u_laplace
		v_laplace2 =  v_laplace2+v_laplace
	end select		

!-------
!	 Compute  right side for momentum equation without sea level gradients
!	-adv(u)-metric-grad(baroclinic pressure)+explicit vertical friction
!	Coriolis term: 	coriolis_alfa*U(l+1) + coriolis_gamma*U(l) + coriolis_betta*U(l-1)
!-------

	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
	do k=1,kv(i,j)
			u_f(k,i,j)=	-flux_u_n (k,i,j)	&
     			-force_x1 (k,i,j)* hzw(k+1)		&
     			+u_laplace2(k,i,j)			&
     			+func_u_b (k,i,j)

     			v_f(k,i,j)=	-flux_u_e (k,i,j)	&
     			-force_x2 (k,i,j)*hzw(k+1)		&
     			+v_laplace2(k,i,j)			&
     			+func_v_b (k,i,j)
	end do
	end do
	end do

!-------
!	add surface b.c.
!-------
	k = 1
	do j = jsouth(myrank), jnorth(myrank)
	do i = iwest (myrank), ieast (myrank)
		if( k <= kv(i,j) ) then
			!Interpolating aice to v-points
			aice_vv = ( (aice(i,j) + aice(i+1,j+1)) + (aice(i+1,j) + aice(i,j+1)) )*0.25

			u_f(k,i,j)=u_f(k,i,j)+(	aice_vv 	* ice_ocn_stress_x_v(i,j)	+		&
						(1.-aice_vv) 	* wind_stress_x_v(i,j)		)/densit

			v_f(k,i,j)=v_f(k,i,j)+(	aice_vv 	* ice_ocn_stress_y_v(i,j)	+		&
						(1.-aice_vv) 	* wind_stress_y_v(i,j)		)/densit
		end if
	end do
	end dO

!-------
!		add bottom friction
!-------
	do j = jsouth(myrank), jnorth(myrank)
	do i = iwest (myrank), ieast (myrank)
		if(kv(i,j) > 0) then
			k = kv(i,j)
			u_f(k,i,j)=u_f(k,i,j)-	cbot(i,j)*ul(k,i,j)
			v_f(k,i,j)=v_f(k,i,j)-	cbot(i,j)*vl(k,i,j)
		end if
	end do
	end do

!-------
!	integral of right side for barotropic equations, explicit scheme
!-------
	rx=0.
	ry=0.

	do j = jsouth(myrank)	,jnorth(myrank)
	do i = iwest (myrank)	,ieast (myrank)
	do k = 1,kv(i,j)
	rx(i,j) = rx(i,j)+dble( u_f(k,i,j) )
	ry(i,j) = ry(i,j)+dble( v_f(k,i,j) )
	end do
	end do
	end do

	
!-------
!	Solving barotropic equations and obtaining new ssl and dx3
!-------
	!barotropic equations
	call o_d_brtr_sweq_sngl

	!updating ssl_f and w_surf_c-----
	if  (scalar_scheme_time==0 .or. (scalar_scheme_time==1.and.mod(time_l-1,lmats)==0) ) then
		call o_d_make_ssl_f(1)
	else
		call o_d_make_ssl_f(2)
	end if
	
#if CMF_VERSION == 2    
    call comp_halo_update_T8(DIM2_t = ssl_f, update_width = 2, change_sign_on_bipolar = 1)
#elif CMF_VERSION == 3
    call HaloUpdateMaster % update(ssl_f, update_width = 2, grid_type = 'T')
#endif	
	!updating dx3's
	call o_d_make_dx3___f
	
!-------
!	Continue baroclinic solver: add terms due to coriolis term
!-------
	do j = jsouth(myrank), jnorth(myrank)
	do i = iwest (myrank), ieast (myrank)
	do k = 1,kv(i,j)
			b = ctime*htsec*coriolis_betta*coriolis(i,j)
			e = ctime*htsec*coriolis_gamma*coriolis(i,j)
			
			u_f(k,i,j) = (ctime*htsec)*u_f(k,i,j) + u0(k,i,j)*dx30(k+1,i,j) + b*v0(k,i,j)* dx30(k+1,i,j) + e*vc(k,i,j)* dx3_v_c(k+1,i,j)
			v_f(k,i,j) = (ctime*htsec)*v_f(k,i,j) + v0(k,i,j)*dx30(k+1,i,j) - b*u0(k,i,j)* dx30(k+1,i,j) - e*uc(k,i,j)* dx3_v_c(k+1,i,j)
        end do
	end do
	end do

!--------------------------------------
!	solve implicit part of momentue quation
!	implicit approximation of vertical mixing terms
!-------------------------------------
	c = ctime*htsec
	ccrank = 0.5*c

	do 10 j = jsouth(myrank), jnorth(myrank)
	do 10 i = iwest (myrank), ieast (myrank)
		if( kv(i,j) <= 0 ) go to 10

!-------
!		Transform the equation in the matrix form according to:
!		-a(k)*T(k-1) + c(k)*T(k) - b(k)*t(k+1) = r(k)
!		compute the matrix terms for the following expresions
!			u-.5*d(k*du/dz)/dz
!			v-.5*d(k*dv/dz)/dz
!-------
		kz1 = 1
		kz2 = kv(i,j)

		k = 1
		arix(k) = 0.0

		do k = 2,kz2
			arix(k) = ccrank*km(k,i,j)*hzv(k)
		end do

		do k = 1,kz2-1
			brix(k) = ccrank*km(k+1,i,j)*hzv(k+1)
		end do
		brix(kz2) = 0.

		do k = 1,kz2
			crix(k) = arix(k)+brix(k)+dx3_v_f(k+1,i,j)
		end do

!-------
!	add the coefficient when the bottom friction is taken implicitly
!		crank-Nicholson scheme
!-------
!	k = kz2-1
!	brix(k) = brix(k)+(c*cbot(i,j)*hzwv(k+1))
!	k = kz2
!	crix(k) = crix(k)+(c*cbot(i,j)*hzwv(k+1))
!	crix(k) = crix(k)+(c*cbot(i,j)*hvbbv(i,j))


		if(coriolis_alfa == 0 .or. abs(x2_v(i,j)).lt.0.5 ) then
!-------
!		Solve two independent implicit equations
!			u-.5*d(k*du/dz)/dz=ru
!			v-.5*d(k*dv/dz)/dz=ru
!-------
!-------
!		Solve the Matrix Equation by factorising Method
!-------
			if(kz2 >= 2) then
				do k = 1,kz2
					rrix(k) = u_f(k,i,j)
				end do

				call factorize(kz1,kz2,arix,brix,crix,rrix,frix,bbrix,rrrix)
		
				do k = 1,kz2
					u_f(k,i,j)=frix(k)
				end do

				do k = 1,kz2
					rrix(k)=v_f(k,i,j)
				end do

				call factorize(kz1,kz2,arix,brix,crix,rrix,frix,bbrix,rrrix)

				do k = 1,kz2
					v_f(k,i,j)=frix(k)
				end do
			end if

		else
!-------
!		Solve implicit system of equations:
!			u-l*v-.5*d(k*du/dz)/dz=ru
!			v+l*u-.5*d(k*dv/dz)/dz=ru
!-------
			factor_a = ctime*htsec*coriolis_alfa*coriolis(i,j)
			
			do k = 1,kz2
				arix(k) = arix(k)/dx3_v_f(k+1,i,j)
				brix(k) = brix(k)/dx3_v_f(k+1,i,j)
				crix(k) = crix(k)/dx3_v_f(k+1,i,j)
			
				ru_impl(k)=u_f(k,i,j)/dx3_v_f(k+1,i,j)
				rv_impl(k)=v_f(k,i,j)/dx3_v_f(k+1,i,j)
			end do

			call o_d_vel_lp(arix(1:kz2),brix(1:kz2),crix(1:kz2),factor_a,ru_impl(1:kz2),rv_impl(1:kz2),u_impl(1:kz2),v_impl(1:kz2),kz2,kz2)

			do k=1,kz2
				u_f(k,i,j)=u_impl(k)
				v_f(k,i,j)=v_impl(k)
			end do
		end if	

10	continue
	end
