!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	ocean dynamic equations.
!	Defining and initializing basic time-mean variables and surface fluxes
!
! authors: 
!	2017.05 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!
! reference:
!	K. V. Ushakov, R. A. Ibraev, Russian Journal of Earth Sciences, 2018.
!--------------------------------------

#include "o_defines.f90"
module o_diagn_basic_module
	use o_diagn_means_module
	save
	type(mean_field) :: u_mean,v_mean,t_mean,s_mean,ssl_mean,aice_mean,hice_mean
	logical,parameter :: integrate=.true.	!integrate means over basin? Warning: several mpi_allreduces, so 'period' is assumed to be big enough
	integer,parameter :: dg_basic_unit=201
	character(300) dg_string
	
	CONTAINS
	subroutine o_diagn_basic_module_init
		call u_mean%init(is3d=.TRUE.)
		call v_mean%init(is3d=.TRUE.)
		call t_mean%init(is3d=.TRUE.)
		call s_mean%init(is3d=.TRUE.)
		call ssl_mean%init(is2d=.TRUE.)
		call aice_mean%init(is2d=.TRUE.)
		call hice_mean%init(is2d=.TRUE.)
		if(myrank==1 .and. integrate) then
			open(unit=dg_basic_unit,file='means_state_'//run_experiment//'.txt',status='replace')
			write(dg_basic_unit,'(a)') 'year	ssl	aice	hice	t3m	t45m	t205m	t475m	s3m	s45m	s205m	s475m'
		end if
	end subroutine o_diagn_basic_module_init
	
	subroutine o_diagn_basic(period)
		use o_basic_module
		use ice_module
		use o_diagn_horizon_module
		implicit none
		integer period	!period of averaging (steps)

		call o_diagn_time_mean(dim3_arr_in=u_c, arr_out=u_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim3_arr_in=v_c, arr_out=v_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim3_arr_in=t_c(:,:,:,1), arr_out=t_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim3_arr_in=t_c(:,:,:,2), arr_out=s_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=real(ssl_c(iwest(myrank)-1:ieast(myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1),main_kind), arr_out=ssl_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=real(aice,main_kind), arr_out=aice_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=real(hice,main_kind), arr_out=hice_mean, finalize_condition=(MOD(time_l,period) == 0))
	
		if(integrate .and. MOD(time_l,period) == 0) then
			write(dg_string,'(12(f8.2))') float(time_l)/minday/365.,	o_diagn_horizon_mean(ssl_mean % collector2d,1,.true.),			&
											o_diagn_horizon_mean(aice_mean % collector2d,1,.true.),			&
											o_diagn_horizon_mean(hice_mean % collector2d,1,.true.),			&
											o_diagn_horizon_mean(t_mean % collector3d(1,:,:),1,.true.),		&
											o_diagn_horizon_mean(t_mean % collector3d(5,:,:),5,.true.),		&	!45 m
											o_diagn_horizon_mean(t_mean % collector3d(17,:,:),17,.true.),		&	!205 m
											o_diagn_horizon_mean(t_mean % collector3d(24,:,:),24,.true.),		&	!475 m
											o_diagn_horizon_mean(s_mean % collector3d(1,:,:),1,.true.),		&
											o_diagn_horizon_mean(s_mean % collector3d(5,:,:),5,.true.),		&
											o_diagn_horizon_mean(s_mean % collector3d(17,:,:),17,.true.),		&
											o_diagn_horizon_mean(s_mean % collector3d(24,:,:),24,.true.)
			if(myrank==1) write(dg_basic_unit,'(a)') trim(dg_string)
		end if
	end subroutine o_diagn_basic
end module o_diagn_basic_module

!===========================================================================================================================================

module o_diagn_surface_module		!time-mean of surface fluxes absorbed by INMIO dynamic core
	use o_diagn_means_module
	save
	type(mean_field) :: ssr_mean,str_mean,sns_mean,ltn_mean,prec_mean,evap_mean
	type(mean_field) :: wind_stress_x_t_mean,wind_stress_y_t_mean
	type(mean_field) :: total_surf_heat_mean
	logical,parameter :: integrate=.true.	!integrate means over basin? Warning: several mpi_allreduces, so 'period' is assumed to be big enough
	integer,parameter :: dg_surf_unit=202
	character(300) dg_string
	
	
	CONTAINS
	subroutine o_diagn_surface_module_init
		call ssr_mean%init(is2d=.TRUE.)
		call str_mean%init(is2d=.TRUE.)
		call sns_mean%init(is2d=.TRUE.)
		call ltn_mean%init(is2d=.TRUE.)
		call prec_mean%init(is2d=.TRUE.)
		call evap_mean%init(is2d=.TRUE.)
		call wind_stress_x_t_mean%init(is2d=.TRUE.)
		call wind_stress_y_t_mean%init(is2d=.TRUE.)
		call total_surf_heat_mean%init(is2d=.TRUE.)
		if(myrank==1 .and. integrate) then
			open(unit=dg_surf_unit,file='means_flux_'//run_experiment//'.txt',status='replace')
			write(dg_surf_unit,'(a)') 'year	ssr	str	sns	ltn	prec	evap	total_heat'
		end if
	end subroutine o_diagn_surface_module_init
	
	subroutine o_diagn_surface(period)
		use o_tf_module
		use ice_module
		use o_td_module
		use o_diagn_horizon_module
		implicit none
		integer period			!period of averaging (steps)
		
		call o_diagn_time_mean(dim2_arr_in=dble(ssr), arr_out=ssr_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=dble(str), arr_out=str_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=dble(sns), arr_out=sns_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=dble(ltn), arr_out=ltn_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=dble(prec), arr_out=prec_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=dble(evap), arr_out=evap_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=dble(wind_stress_x_t), arr_out=wind_stress_x_t_mean, finalize_condition=(MOD(time_l,period) == 0))
		call o_diagn_time_mean(dim2_arr_in=dble(wind_stress_y_t), arr_out=wind_stress_y_t_mean, finalize_condition=(MOD(time_l,period) == 0))
		
		!Total surface heat flux per m2 of grid cell
		!Note, that atm-ocn fluxes are per m2 of open water, while ice-ocn fluxes are per m2 of grid cell
		total_surf_heat=(ssr*(1.-albedo_ocean)+str+sns+ltn)*(1.-aice) + ice_ocn_ssr+ice_ocn_heat
		call o_diagn_time_mean(dim2_arr_in=dble(total_surf_heat), arr_out=total_surf_heat_mean, finalize_condition=(MOD(time_l,period) == 0))
	
		if(integrate .and. MOD(time_l,period) == 0) then
			write(dg_string,'(8(f8.2))') float(time_l)/minday/365.,	o_diagn_horizon_mean(ssr_mean % collector2d,1,.false.),			&
											o_diagn_horizon_mean(str_mean % collector2d,1,.false.),			&
											o_diagn_horizon_mean(sns_mean % collector2d,1,.false.),			&
											o_diagn_horizon_mean(ltn_mean % collector2d,1,.false.),			&
											o_diagn_horizon_mean(prec_mean % collector2d,1,.true.)*24*3600*365*1000,&	![mm/year]
											o_diagn_horizon_mean(evap_mean % collector2d,1,.false.)*24*3600*365,	&	![mm/year]
											o_diagn_horizon_mean(total_surf_heat_mean % collector2d,1,.true.)
			if(myrank==1) write(dg_surf_unit,'(a)') trim(dg_string)
		end if
	end subroutine o_diagn_surface
end module o_diagn_surface_module

!===================================================================================================================================

module o_diagn_a2o_module	!time mean of raw surface fluxes coming from atm model
	use o_diagn_means_module
	save
	type(mean_field) :: sns_from_atm_mean, ltn_from_atm_mean, atm_ocn_windstr_x_mean, atm_ocn_windstr_y_mean, total_surf_heat_from_atm_mean
	
	CONTAINS
	subroutine o_diagn_a2o_module_init
		call sns_from_atm_mean%init(is2d=.TRUE.)
		call ltn_from_atm_mean%init(is2d=.TRUE.)
		call atm_ocn_windstr_x_mean%init(is2d=.TRUE.)
		call atm_ocn_windstr_y_mean%init(is2d=.TRUE.)
		call total_surf_heat_from_atm_mean%init(is2d=.TRUE.)
	end subroutine o_diagn_a2o_module_init
	
	subroutine o_diagn_a2o
		use o_tf_module
		use ice_module
		use o_td_module
		use atm_module
		implicit none
		
		call o_diagn_time_mean(dim2_arr_in=dble(sns_from_atm), arr_out=sns_from_atm_mean, finalize_condition=(MOD(time_l,time_l_save_cp) == 0))
		call o_diagn_time_mean(dim2_arr_in=dble(ltn_from_atm), arr_out=ltn_from_atm_mean, finalize_condition=(MOD(time_l,time_l_save_cp) == 0))
		if(allocated(atm_ocn_windstr_x)) call o_diagn_time_mean(dim2_arr_in=dble(atm_ocn_windstr_x), arr_out=atm_ocn_windstr_x_mean, finalize_condition=(MOD(time_l,time_l_save_cp) == 0))
		if(allocated(atm_ocn_windstr_y)) call o_diagn_time_mean(dim2_arr_in=dble(atm_ocn_windstr_y), arr_out=atm_ocn_windstr_y_mean, finalize_condition=(MOD(time_l,time_l_save_cp) == 0))
		
		!Total surface heat flux per m2 of grid cell
		!Note, that atm-ocn fluxes are per m2 of open water, while ice-ocn fluxes are per m2 of grid cell
		total_surf_heat=(ssr*(1.-albedo_ocean)+str+sns_from_atm+ltn_from_atm)*(1.-aice) + ice_ocn_ssr+ice_ocn_heat
		call o_diagn_time_mean(dim2_arr_in=dble(total_surf_heat), arr_out=total_surf_heat_from_atm_mean, finalize_condition=(MOD(time_l,time_l_save_cp) == 0))
	end subroutine o_diagn_a2o
end module o_diagn_a2o_module


