
!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	diagnostical parameters of solution 
!	Module defining 2D and 3D time-averaging techniques
!	This is a system file.
!
! authors: 
!	2017.09 / Gromov, Igor, IO RAS, Russia, gromov@phystech.edu
!	2020.02 / Ushakov Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!--------------------------------------


#include "o_defines.f90"
  module o_diagn_means_module
    use o_grid_module
    use o_par_module
    implicit none
    
    TYPE, public :: mean_field
      real(kind=collector_kind), allocatable :: collector2d(:,:),collector3d(:,:,:)
      real counter
      logical reinit
      CONTAINS
        procedure :: init => mean_field_init
        procedure :: add => mean_field_add
        procedure :: finalize => mean_field_finalize
    end type mean_field
    
    CONTAINS
      
      subroutine mean_field_init(this,is2d,is3d)
      !allocate 2d or 3d collector array
        implicit none
        class(mean_field) :: this
        logical,optional :: is2d,is3d
        
        if(present(is2d))then
	  allocate (this % collector2d  (iwest(myrank)-1:ieast(myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1))
	  this % collector2d = 0.
        end if
        if(present(is3d))then
	  allocate (this % collector3d  (kl,iwest(myrank)-1:ieast(myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1))
	  this % collector3d = 0.
        end if
	this % counter = 0.
      end subroutine mean_field_init
  
      subroutine mean_field_add(this,arr2d,arr3d) 	
      !this subroutine is used to accumulate data in "collector*d" array. It processes 2D or 3D arrays
        implicit none
        class(mean_field) :: this
        real(kind=main_kind),optional :: arr2d(iwest(myrank)-1:ieast(myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)
        real(kind=main_kind),optional :: arr3d(kl,iwest(myrank)-1:ieast(myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) 
        
        !if on previous step saved DG, must set collector array to zero
        if(this%reinit == .TRUE.)then
	  if(present(arr3d))this % collector3d=0.
	  if(present(arr2d))this % collector2d=0.
	  this%reinit=.FALSE.  
        end if
        
        if(present(arr3d))this % collector3d=this % collector3d+arr3d
        if(present(arr2d))this % collector2d=this % collector2d+arr2d
        this % counter = this % counter + 1.
      end subroutine mean_field_add
  
      subroutine mean_field_finalize(this)
        implicit none
        class(mean_field) :: this
        if(allocated(this%collector2d))this % collector2d = this % collector2d / this % counter
        if(allocated(this%collector3d))this % collector3d = this % collector3d / this % counter
        this%reinit=.TRUE.
        this % counter = 0.
      end subroutine mean_field_finalize
      
  subroutine o_diagn_time_mean(dim3_arr_in, dim2_arr_in, arr_out, finalize_condition)
    !use o_diagn_mean_field_module
    class(mean_field) :: arr_out
    real(kind=main_kind),intent(in),optional :: dim3_arr_in(kl,	iwest(myrank)-1:ieast(myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)
    real(kind=main_kind),intent(in),optional :: dim2_arr_in(	iwest(myrank)-1:ieast(myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)
    logical :: finalize_condition
!process errors
    if(present(dim3_arr_in).and.(.not.allocated(arr_out%collector3d)))call ocn_error("Subroutine o_diagn_time_mean: dim3_arr_in must have 3 dimensions","LCL")
    if(present(dim2_arr_in).and.(.not.allocated(arr_out%collector2d)))call ocn_error("Subroutine o_diagn_time_mean: dim2_arr_in must have 2 dimensions","LCL")
    if(present(dim2_arr_in).and.present(dim3_arr_in))call ocn_error("Subroutine o_diagn_time_mean: dim2_arr_in and dim3_arr_in are presented","LCL")
    if((.not.present(dim2_arr_in)).and.(.not.present(dim3_arr_in)))call ocn_error("Subroutine o_diagn_time_mean: dim2_arr_in and dim3_arr_in are not presented","LCL")
!
    if(present(dim3_arr_in))call arr_out%add(arr3d=dim3_arr_in)
    if(present(dim2_arr_in))call arr_out%add(arr2d=dim2_arr_in)
    if(finalize_condition==.TRUE.)call arr_out%finalize()
  end subroutine o_diagn_time_mean
  
  end module o_diagn_means_module
