#include "o_defines.f90"


!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	grid
!	module for variables
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru 
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!
! reference:
!	R. A. Ibraev, K. V. Ushakov, and R. N. Khabeev, 2012.
!		Eddy-resolving 1/10° model of the World Ocean,
!		Izv., Atmos. Ocean. Phys. 48 (1), 37–46.
!	Ibrayev R.A., 2001: Model of enclosed and semi-enclosed sea hydrodynamics. 
!		Russ. J. Numer. Anal. Math. Modelling, 2001, vol. 16, No 4, p. 291-304.
!--------------------------------------

!===================================
! Contains grid arrays and variables
!===================================

	module o_grid_module
	
	implicit none	
	save
		
	integer ocn_steps_in_day
	logical ocn_read_cp
	logical ocn_das_on 		! ON/OFF data assimilation on OCN (send/recv data to DAS)
	
	real (kind=8) x1_ssc_size
	real (kind=8) x2_ssc_size, x2_ssc_step
	real (kind=8) x1_ssc_west_south
	real (kind=8) x2_ssc_west_south

	integer i, j, k, m

	integer (kind=4) kl, ml, mltke, klp
	integer (kind=4) il, jl
	integer (kind=4) key_lb, key_bio
	integer (kind=4) bsc_flag, il_bsc, jl_bsc, i_bsc,j_bsc
	integer (kind=4) ssc_flag, il_ssc, jl_ssc, jl_ssc_ext 
	integer (kind=4) i_cycle, j_cycle
	integer (kind=4) mixh_m_mode
	integer (kind=4) mixh_t_mode
	integer (kind=4) i_solution_control, j_solution_control, k_solution_control	!point for controlling solution in read/save subr.

	integer (kind=4) sweq_steps
	integer (kind=4) iwe,iea,jso,jno  
!-------
!	space grid and coordinate parameters
!-------

	real (kind=8), allocatable:: x1_v	(:,:)
	real (kind=8), allocatable:: x2_v	(:,:)
	real (kind=8), allocatable:: x1_t	(:,:)
	real (kind=8), allocatable:: x2_t	(:,:)

	real (kind=8), allocatable:: dx1_nw_v	(:,:)
	real (kind=8), allocatable:: dx1_ne_v	(:,:)
	real (kind=8), allocatable:: dx2_en_v	(:,:)
	real (kind=8), allocatable:: dx2_es_v	(:,:)
	real (kind=8), allocatable:: dx1_nw_t	(:,:)
	real (kind=8), allocatable:: dx1_ne_t	(:,:)
	real (kind=8), allocatable:: dx2_en_t	(:,:)
	real (kind=8), allocatable:: dx2_es_t	(:,:)

	real (kind=main_kind), allocatable:: dx3_v_l	(:,:,:)
	real (kind=main_kind), allocatable:: dx3_t_l	(:,:,:)
	real (kind=main_kind), allocatable:: dx3_v_c	(:,:,:)
	real (kind=main_kind), allocatable:: dx3_t_c	(:,:,:)
	real (kind=main_kind), allocatable:: dx3_v_f	(:,:,:)
	real (kind=main_kind), allocatable:: dx3_t_f	(:,:,:)


	real (kind=8), allocatable:: csu(:)
	real (kind=8), allocatable:: cst(:)

	real (kind=4), allocatable:: z		(:)
	real (kind=4), allocatable:: hz	(:)
	real (kind=4), allocatable:: hzv	(:)
	real (kind=4), allocatable:: zw	(:)
	real (kind=4), allocatable:: hzw	(:)
	real (kind=4), allocatable:: hzwv	(:)
	real (kind=4), allocatable:: hkm	(:)
	real (kind=4), allocatable:: hkp	(:)

	real (kind=8), allocatable:: area_vbox	(:,:)
	real (kind=8), allocatable:: area_vbox_v(:,:)
	real (kind=8), allocatable:: area_tbox	(:,:)
	real (kind=8), allocatable:: area_tbox_v(:,:)

	! Time variables

	integer minday,lmats
	real hthour,htsec
	real (kind=4)		tau
	integer	(kind=4)	sweq_pl

!-------
!	botom topography arrays
!-------
	integer (kind=4), allocatable:: kv (:,:)
	integer (kind=4), allocatable:: kt (:,:)
	real (kind=4), allocatable:: hv (:,:)

	real (kind=4), allocatable:: kv_real (:,:)
	real (kind=4), allocatable:: kt_real (:,:)		
	integer (kind=4) pool_min_number
!-------
!	switch off keys
!-------
	integer (kind=4) its, key_ic
	integer (kind=4) mix_unstable_levels
	integer (kind=4) nloop,iprmix !,imix

!-------
!	parameters of region decomposition for mpi
!-------
	integer (kind=4) myrank !, il_belt, jl_belt,size, rank
	integer (kind=4) inp,jnp,decomp

	integer (kind=4), allocatable:: jsouth	(:)
	integer (kind=4), allocatable:: jnorth	(:)
	integer (kind=4), allocatable:: iwest	(:)
	integer (kind=4), allocatable:: ieast	(:)

!------
!	some basic values
!------
	real (kind=4)	gravit, pi, densit, radius, radv, omega
	real (kind=8)	pi180
	real (kind=4)	robert_asselin
	integer atm_model, t_potential_convert
	
	character(50) namelist_path
	
	contains
	
!========================================================================================

    subroutine set_grid_subdomains(rank, iw, ie, js, jn)
        integer, intent(in) :: rank, iw, ie, js, jn
          
        myrank = rank
                    
        allocate (jsouth(myrank:myrank ))
        allocate (jnorth(myrank:myrank ))
        allocate (iwest (myrank:myrank) )
        allocate (ieast (myrank:myrank) )

        iwest(myrank) = iw
        ieast(myrank) = ie
        jsouth(myrank) = js
        jnorth(myrank) = jn     
        
    end subroutine

!========================================================================================
    
	subroutine allocate_o_grid_coord_module
	namelist /OCN_GRID/ i_cycle, j_cycle, ssc_flag, bsc_flag, x1_ssc_west_south, x2_ssc_west_south

	! Coordinates
	allocate (x1_v	(1:il,1:jl))
	allocate (x2_v	(1:il,1:jl))
	allocate (x1_t	(1:il,1:jl))
	allocate (x2_t	(1:il,1:jl))
	x1_v =0.
	x2_v =0.
	x1_t =0.
	x2_t =0.

	! Z-arrays
	allocate (z	(kl+1))
	allocate (hz	(kl+1))
	allocate (hzv	(kl+1))
	allocate (zw	(kl+1))
	allocate (hzw	(kl+1))
	allocate (hzwv	(kl+1))
	allocate (hkm	(kl  ))
	allocate (hkp	(kl  ))

	! Topography
	allocate (kv 		(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (kt 		(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (hv 		(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	kv =0
	kt =0
	hv =0.

	! Temporary for read_ic subroutine
	allocate (kv_real (iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (kt_real (iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	kv_real = 0
	kt_real = 0
	
	end subroutine allocate_o_grid_coord_module

!=======================================================================================================================================================================	

	subroutine allocate_o_grid_step_module

	! Grid steps
	allocate (dx1_nw_v(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx1_ne_v(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx2_en_v(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx2_es_v(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx1_nw_t(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx1_ne_t(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx2_en_t(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx2_es_t(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	dx1_ne_v = 0
	dx1_nw_v = 0
	dx2_en_v = 0
	dx2_es_v = 0
	dx1_ne_t = 0
	dx1_nw_t = 0
	dx2_en_t = 0
	dx2_es_t = 0

	allocate (dx3_v_l(kl+1,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx3_t_l(kl+1,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	dx3_v_l = 0.
	dx3_t_l = 0.
	allocate (dx3_v_c(kl+1,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx3_t_c(kl+1,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	dx3_v_c = 0.
	dx3_t_c = 0.
	allocate (dx3_v_f(kl+1,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (dx3_t_f(kl+1,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1) )
	dx3_v_f = 0.
	dx3_t_f = 0.

	allocate (csu(0:jl+1))
	allocate (cst(0:jl+1))
	csu = 0.
	cst = 0.

	! Box areas
	allocate (area_vbox 	(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (area_vbox_v 	(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (area_tbox	(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	allocate (area_tbox_v	(iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) )
	area_vbox	= 0.
	area_vbox_v	= 0.
	area_tbox	= 0.
	area_tbox_v	= 0.

	end subroutine allocate_o_grid_step_module

!=======================================================================================================================================================================	
	
	end module o_grid_module
