
!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	grid
!	compute grid variables
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru 
!
! reference:
!	Murray R.J. 1996. 
!		Explicit Generation of Orthogonal Grids for Ocean Models,
!		J. Comp. Physics, 126, 251-273.
!	Ibrayev R.A., Khabeev R.N., Ushakov K.V., 2012. 
!		Eddy-resolving 1/10 Model of the World Ocean. 
!		Izvestiya, Atmospheric and Oceanic Physics, 2012, Vol. 48, No. 1, pp. 37–46.
!--------------------------------------



!~~~~~~
	module o_bsc_module
	save

	real (kind=8) x1_pole, x2_pole, x2_0
	real (kind=8) x1_bsc_west_south, x2_bsc_west_south

	real (kind=8), allocatable :: x1_bsc_t (:)
	real (kind=8), allocatable :: x2_bsc_t (:)
	real (kind=8), allocatable :: x1_bsc_v (:)
	real (kind=8), allocatable :: x2_bsc_v (:)

	real (kind=8), allocatable :: x1_bsc_t_radian (:)
	real (kind=8), allocatable :: x2_bsc_t_radian (:)
	real (kind=8), allocatable :: x1_bsc_v_radian (:)
	real (kind=8), allocatable :: x2_bsc_v_radian (:)

	real (kind=8), allocatable:: lame2_bsc_t (:,:)
	real (kind=8), allocatable:: lame1_bsc_t (:,:)
	real (kind=8), allocatable:: lame2_bsc_v (:,:)
	real (kind=8), allocatable:: lame1_bsc_v (:,:)

	real (kind=8), allocatable:: dx1_nw_bsc_v (:,:)
	real (kind=8), allocatable:: dx1_ne_bsc_v (:,:)
	real (kind=8), allocatable:: dx1_sw_bsc_v (:,:)
	real (kind=8), allocatable:: dx1_se_bsc_v (:,:)

	real (kind=8), allocatable:: dx2_en_bsc_v (:,:)
	real (kind=8), allocatable:: dx2_es_bsc_v (:,:)
	real (kind=8), allocatable:: dx2_wn_bsc_v (:,:)
	real (kind=8), allocatable:: dx2_ws_bsc_v (:,:)

	real (kind=8), allocatable:: dx1_ne_bsc_t (:,:)
	real (kind=8), allocatable:: dx1_nw_bsc_t (:,:)
	real (kind=8), allocatable:: dx1_se_bsc_t (:,:)
	real (kind=8), allocatable:: dx1_sw_bsc_t (:,:)

	real (kind=8), allocatable:: dx2_en_bsc_t (:,:)
	real (kind=8), allocatable:: dx2_wn_bsc_t (:,:)
	real (kind=8), allocatable:: dx2_es_bsc_t (:,:)
	real (kind=8), allocatable:: dx2_ws_bsc_t (:,:)

	end module o_bsc_module



!~~~~~~
	subroutine o_ini_set_grid_bsc_coordinate
!~~~~~~

!--------------------------------------
!		define grid in 3-d space and time
!
!ssc - spherical system of coordinates
!bsc - bipolar system of coordinates
!tsc - tripolar system of coordinates
!----------------------------------------------------------------------

	use o_grid_module
	use o_bsc_module

	implicit none

	real (kind=8) b_bsc, cos_0
	real (kind=8) x1, x2
	real (kind=8) cos_x1_t_bsc, cos_x2_t_bsc
	real (kind=8) cos_x1_v_bsc, cos_x2_v_bsc
	real (kind=8) sin_x1_v_bsc

	allocate (x1_bsc_t (il_bsc) )
	allocate (x2_bsc_t (jl_bsc) )
	allocate (x1_bsc_v (il_bsc) )
	allocate (x2_bsc_v (jl_bsc) )

	allocate (x1_bsc_t_radian (il_bsc) )
	allocate (x2_bsc_t_radian (jl_bsc) )
	allocate (x1_bsc_v_radian (il_bsc) )
	allocate (x2_bsc_v_radian (jl_bsc) )
	
!--------------------------------------------
!2.	include bipolar system of coordinates
!--------------------------------------------


	x1_pole = x1_bsc_west_south	!The ssc x1 coordinate of bsc South(!) Pole 
	x2_pole = x2_bsc_west_south	!The ssc x2 coordinate of bsc North and South Poles
	x2_0    = 90.-x2_pole


!-------
!2.1.	x1,x2 coordinates of t-points in bsc (degrees)
!-------
	do i=1,il_bsc
	x1_bsc_t(i) = -90. + (180./real(il_bsc-1)) * real(i-1)
	end do

	do j=1,jl_bsc
	x2_bsc_t(j) = -90. + (180./real(jl_bsc-1)) * real(j-1)
	end do


!-------
!2.2.	x1,x2 coordinates of v-points in bsc (degrees)
!-------
	do i=1,il_bsc-1
	x1_bsc_v(i) = 0.5*(x1_bsc_t(i) + x1_bsc_t(i+1))
	end do

	do j=1,jl_bsc-1
	x2_bsc_v(j) = 0.5*(x2_bsc_t(j) + x2_bsc_t(j+1))
	end do

	return                                                            
	end                                                               





!~~~~~~
	subroutine o_ini_set_grid_bsc_step
!~~~~~~

!--------------------------------------
!		define grid in 3-d space and time
!
!ssc - spherical system of coordinates
!bsc - bipolar system of coordinates
!tsc - tripolar system of coordinates
!----------------------------------------------------------------------

	use o_grid_module
	use o_bsc_module

	implicit none

	integer counter


	real (kind=8) b_bsc, cos_0
	real (kind=8) x1, x2
	real (kind=8) cos_x1_t_bsc, cos_x2_t_bsc
	real (kind=8) cos_x1_v_bsc, cos_x2_v_bsc
	real (kind=8) sin_x1_v_bsc


	
! Allocate arrays

	allocate (lame2_bsc_t	(il_bsc,jl_bsc) )
	allocate (lame1_bsc_t	(il_bsc,jl_bsc) )
	allocate (lame2_bsc_v	(il_bsc,jl_bsc) )
	allocate (lame1_bsc_v	(il_bsc,jl_bsc) )

	allocate (dx1_nw_bsc_v	(il_bsc,jl_bsc) )
	allocate (dx1_ne_bsc_v	(il_bsc,jl_bsc) )
	allocate (dx1_sw_bsc_v	(il_bsc,jl_bsc) )
	allocate (dx1_se_bsc_v	(il_bsc,jl_bsc) )

	allocate (dx2_en_bsc_v	(il_bsc,jl_bsc) )
	allocate (dx2_es_bsc_v	(il_bsc,jl_bsc) )
	allocate (dx2_wn_bsc_v	(il_bsc,jl_bsc) )
	allocate (dx2_ws_bsc_v	(il_bsc,jl_bsc) )

	allocate (dx1_ne_bsc_t	(il_bsc,jl_bsc) )
	allocate (dx1_nw_bsc_t	(il_bsc,jl_bsc) )
	allocate (dx1_se_bsc_t	(il_bsc,jl_bsc) )
	allocate (dx1_sw_bsc_t	(il_bsc,jl_bsc) )

	allocate (dx2_en_bsc_t	(il_bsc,jl_bsc) )
	allocate (dx2_wn_bsc_t	(il_bsc,jl_bsc) )
	allocate (dx2_es_bsc_t	(il_bsc,jl_bsc) )
	allocate (dx2_ws_bsc_t	(il_bsc,jl_bsc) )


!-------
!2.3.	lame coefficients
!-------
	b_bsc = tan(x2_0*pi180/2.)
	cos_0 = cos(x2_0*pi180)

	do j=1,jl_bsc
	do i=1,il_bsc
		x1 = x1_bsc_t(i)*pi180
		x2 = x2_bsc_t(j)*pi180
		cos_x1_t_bsc = cos(x1)
		cos_x2_t_bsc = cos(x2)
		lame2_bsc_t(i,j) = 					&
			2*radius*b_bsc*(b_bsc*b_bsc+1.)*		&
			(cos_x2_t_bsc*cos_x1_t_bsc*cos_0+1.) / 	&
			( ( 1.+cos_x2_t_bsc*cos_x1_t_bsc) + 		&
			b_bsc*b_bsc*					&
			( 1.-cos_x2_t_bsc*cos_x1_t_bsc))**2
		lame1_bsc_t (i,j) = cos_x2_t_bsc * lame2_bsc_t (i,j)	
	end do
	end do


	do j=1,jl_bsc-1
	do i=1,il_bsc-1
		x1 = x1_bsc_v(i)*pi180
		x2 = x2_bsc_v(j)*pi180
		cos_x1_v_bsc = cos(x1)
		cos_x2_v_bsc = cos(x2)
		lame2_bsc_v(i,j) = 					&
			2*radius*b_bsc*(b_bsc*b_bsc+1.)*		&
			(cos_x2_v_bsc*cos_x1_v_bsc*cos_0+1.) / 	&
			( ( 1.+cos_x2_v_bsc*cos_x1_v_bsc) + 		&
			b_bsc*b_bsc*					&
			( 1.-cos_x2_v_bsc*cos_x1_v_bsc))**2
		lame1_bsc_v (i,j) = cos_x2_v_bsc * lame2_bsc_v (i,j)
	end do
	end do

!-------
!2.3.	lengths of lateral boundaries of v-cells in bsc
!-------
	do j=1,jl_bsc-1
	do i=1,il_bsc-1
	dx1_nw_bsc_v(i,j) = lame1_bsc_t(i  ,j+1)*(x1_bsc_v(i  )-x1_bsc_t(i  ))
	dx1_ne_bsc_v(i,j) = lame1_bsc_t(i+1,j+1)*(x1_bsc_t(i+1)-x1_bsc_v(i  ))
	dx1_sw_bsc_v(i,j) = lame1_bsc_t(i  ,j  )*(x1_bsc_v(i  )-x1_bsc_t(i  ))
	dx1_se_bsc_v(i,j) = lame1_bsc_t(i+1,j  )*(x1_bsc_t(i+1)-x1_bsc_v(i  ))

	dx2_en_bsc_v(i,j) = lame2_bsc_t(i+1,j+1)*(x2_bsc_t(j+1)-x2_bsc_v(j  ))
	dx2_es_bsc_v(i,j) = lame2_bsc_t(i+1,j  )*(x2_bsc_v(j  )-x2_bsc_t(j  ))
	dx2_wn_bsc_v(i,j) = lame2_bsc_t(i  ,j+1)*(x2_bsc_t(j+1)-x2_bsc_v(j  ))
	dx2_ws_bsc_v(i,j) = lame2_bsc_t(i  ,j  )*(x2_bsc_v(j  )-x2_bsc_t(j  ))
	end do
	end do


	dx1_nw_bsc_v = dx1_nw_bsc_v * pi180
	dx1_ne_bsc_v = dx1_ne_bsc_v * pi180
	dx1_sw_bsc_v = dx1_sw_bsc_v * pi180
	dx1_se_bsc_v = dx1_se_bsc_v * pi180

	dx2_en_bsc_v = dx2_en_bsc_v * pi180
	dx2_es_bsc_v = dx2_es_bsc_v * pi180
	dx2_wn_bsc_v = dx2_wn_bsc_v * pi180
	dx2_ws_bsc_v = dx2_ws_bsc_v * pi180

!-------
!2.4.	lengths of lateral boundaries of t-cells in bsc
!-------
	do j=1,jl_bsc-1		!ne, en
	do i=1,il_bsc-1
	dx1_ne_bsc_t(i,j) = lame1_bsc_v(i  ,j  )*(x1_bsc_v(i  )-x1_bsc_t(i  ))
	dx2_en_bsc_t(i,j) = lame2_bsc_v(i  ,j  )*(x2_bsc_v(j  )-x2_bsc_t(j  ))
	end do
	end do
	do j=1,jl_bsc-1		!nw, wn
	do i=2,il_bsc
	dx1_nw_bsc_t(i,j) = lame1_bsc_v(i-1,j  )*(x1_bsc_t(i  )-x1_bsc_v(i-1))
	dx2_wn_bsc_t(i,j) = lame2_bsc_v(i-1,j  )*(x2_bsc_v(j  )-x2_bsc_t(j  ))
	end do
	end do
	do j=2,jl_bsc
	do i=1,il_bsc-1
	dx1_se_bsc_t(i,j) = lame1_bsc_v(i  ,j-1)*(x1_bsc_v(i  )-x1_bsc_t(i  ))
	dx2_es_bsc_t(i,j) = lame2_bsc_v(i  ,j-1)*(x2_bsc_t(j  )-x2_bsc_v(j-1))
	end do
	end do
	do j=2,jl_bsc
	do i=2,il_bsc
	dx1_sw_bsc_t(i,j) = lame1_bsc_v(i-1,j-1)*(x1_bsc_t(i  )-x1_bsc_v(i-1))
	dx2_ws_bsc_t(i,j) = lame2_bsc_v(i-1,j-1)*(x2_bsc_t(j  )-x2_bsc_v(j-1))
	end do
	end do


	dx1_ne_bsc_t = dx1_ne_bsc_t * pi180
	dx1_nw_bsc_t = dx1_nw_bsc_t * pi180
	dx1_se_bsc_t = dx1_se_bsc_t * pi180
	dx1_sw_bsc_t = dx1_sw_bsc_t * pi180

	dx2_en_bsc_t = dx2_en_bsc_t * pi180
	dx2_wn_bsc_t = dx2_wn_bsc_t * pi180
	dx2_es_bsc_t = dx2_es_bsc_t * pi180
	dx2_ws_bsc_t = dx2_ws_bsc_t * pi180



!-------
!	free memory
!-------
	deallocate (lame2_bsc_t	 )
	deallocate (lame1_bsc_t	 )
	deallocate (lame2_bsc_v	 )
	deallocate (lame1_bsc_v	 )

	return                                                            
	end                                                               

