
!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	grid
!	compute grid variables
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru 
!
! reference:
!	Murray R.J. 1996. 
!		Explicit Generation of Orthogonal Grids for Ocean Models,
!		J. Comp. Physics, 126, 251-273.
!	Ibrayev R.A., Khabeev R.N., Ushakov K.V., 2012. 
!		Eddy-resolving 1/10 Model of the World Ocean. 
!		Izvestiya, Atmospheric and Oceanic Physics, 2012, Vol. 48, No. 1, pp. 37–46.
!--------------------------------------


!~~~~~~
	module o_ssc_module
!~~~~~~
	save

	real (kind=8), allocatable :: x1_ssc_t	(:)
	real (kind=8), allocatable :: x2_ssc_t	(:)
	real (kind=8), allocatable :: x1_ssc_v	(:)
	real (kind=8), allocatable :: x2_ssc_v	(:)


	real (kind=8), allocatable:: lame2_ssc_t (:,:)
	real (kind=8), allocatable:: lame1_ssc_t (:,:)
	real (kind=8), allocatable:: lame2_ssc_v (:,:)
	real (kind=8), allocatable:: lame1_ssc_v (:,:)

	real (kind=8), allocatable:: dx1_nw_ssc_v (:,:)
	real (kind=8), allocatable:: dx1_ne_ssc_v (:,:)
	real (kind=8), allocatable:: dx1_nw_ssc_t (:,:)
	real (kind=8), allocatable:: dx1_ne_ssc_t (:,:)

	real (kind=8), allocatable:: dx2_en_ssc_v (:,:)
	real (kind=8), allocatable:: dx2_es_ssc_v (:,:)
	real (kind=8), allocatable:: dx2_en_ssc_t (:,:)
	real (kind=8), allocatable:: dx2_es_ssc_t (:,:)


	end module o_ssc_module

!--------------------------------------
!	define grid in sub-region with spherical system of coordinates
!
!ssc - spherical system of coordinates
!bsc - bipolar system of coordinates
!tsc - tripolar system of coordinates
!----------------------------------------------------------------------

	subroutine o_ini_set_grid_ssc_coordinate

	use o_grid_module
	use o_ssc_module

	implicit none

!-------
!	set of (x1,x2,z) coordinates of (u,v) points 
!	(x,y)	in angle degrees
!		Grid with constant increments in x and y directions
!		dx(degree).ne.dy(degree)
!	(z)	meters depth from top to bottom
!	1 degree = 60 minutes = 3600 seconds
!	d(lambda) = (1/18) degree = 3'20'' = 200'' = 6.3 points of initial grid
!	d(fi) =     (1/24) degree = 2'30'' = 150'' = 5 points of initial grid
!-------

	allocate ( x1_ssc_t	(0:il_ssc+1) )
	allocate ( x2_ssc_t	(0:jl_ssc+1) )
	allocate ( x1_ssc_v	(0:il_ssc+1) )
	allocate ( x2_ssc_v	(0:jl_ssc+1) )


!-------
!#1.	compute coordinates of t-points
!-------
	if (i_cycle.eq.0) then
	do i=0,il_ssc+1
	x1_ssc_t(i)=x1_ssc_west_south + (x1_ssc_size/real(il_ssc-3)) * real(i-2)
	end do
	do j=0,jl_ssc+1
	x2_ssc_t(j)=x2_ssc_west_south 	+ (x2_ssc_size/real(jl_ssc-3)) * real(j-2)
	end do
	end if		!if (i_cycle.eq.0) then

	if (i_cycle.eq.1) then
	do i=0,il_ssc+1
	x1_ssc_t(i)=x1_ssc_west_south + (x1_ssc_size/real(il_ssc-2)) * real(i-2)
	end do
	do j=0,jl_ssc+1
	x2_ssc_t(j)=x2_ssc_west_south + (x2_ssc_size/real(jl_ssc-2)) * real(j-2)
	end do
	end if		!if (i_cycle.eq.1) then


!-------
!#2.	compute coordinates of v-points
!-------

	do i=0,il_ssc
	x1_ssc_v(i   )=               0.5*(x1_ssc_t(i )+x1_ssc_t(i+1))
	end do
	x1_ssc_v(il_ssc+1)=x1_ssc_t(il_ssc+1) + 0.5*(x1_ssc_t(il_ssc)-x1_ssc_t(il_ssc-1))


	do j=0,jl_ssc
	x2_ssc_v(j   )=               0.5*(x2_ssc_t(j )+x2_ssc_t(j+1 ))
	end do
	x2_ssc_v(jl_ssc+1)=x2_ssc_t(jl_ssc+1) + 0.5*(x2_ssc_t(jl_ssc)-x2_ssc_t(jl_ssc-1))



	return                                                            
	end                                                               



	subroutine o_ini_set_grid_ssc_step

!--------------------------------------
!	define grid in sub-region with spherical system of coordinates
!
!ssc - spherical system of coordinates
!bsc - bipolar system of coordinates
!tsc - tripolar system of coordinates
!----------------------------------------------------------------------

	use o_grid_module
	use o_ssc_module

	implicit none

!	Allocate arrays
	allocate ( lame1_ssc_v (0:il_ssc+1,0:jl_ssc+1) )
	allocate ( lame1_ssc_t (0:il_ssc+1,0:jl_ssc+1) )

	allocate ( lame2_ssc_v (0:il_ssc+1,0:jl_ssc+1) )
	allocate ( lame2_ssc_t (0:il_ssc+1,0:jl_ssc+1) )

	allocate ( dx1_nw_ssc_v (il_ssc,jl_ssc) )
	allocate ( dx1_ne_ssc_v (il_ssc,jl_ssc) )
	allocate ( dx2_en_ssc_v (il_ssc,jl_ssc) )
	allocate ( dx2_es_ssc_v (il_ssc,jl_ssc) )

	allocate ( dx1_nw_ssc_t (il_ssc,jl_ssc) )
	allocate ( dx1_ne_ssc_t (il_ssc,jl_ssc) )
	allocate ( dx2_en_ssc_t (il_ssc,jl_ssc) )
	allocate ( dx2_es_ssc_t (il_ssc,jl_ssc) )

!-------
!#3.	compute horizontal grid steps and depending parameters of the model
!-------

!-------
!#3.1	grid step for west-east direction
!-------
!-------
!#3.1.2	define lame coefficients for 2d coordinate transformations
!-------
	cst	=0.
	do j=0,jl_ssc+1
	cst(j)=cos(x2_ssc_t(j)*pi180)                                                 
	end do

	csu	=0.
	do j=0,jl_ssc
	csu(j)=0.5*(cst(j)+cst(j+1))
	end do


	do j=0,jl_ssc+1
	do i=0,il_ssc+1
	lame1_ssc_t(i,j)	=radius*cst(j)
	lame1_ssc_v(i,j)	=radius*csu(j)
	end do
	end do


!-------
!#3.1.3	lengths of lateral boundaries of v-cells
!-------
	do j=1 ,jl_ssc
	do i=1, il_ssc
	dx1_nw_ssc_v(i,j) = lame1_ssc_t(i  ,j+1)*(x1_ssc_v(i  )-x1_ssc_t(i  )) * pi180
	dx1_ne_ssc_v(i,j) = lame1_ssc_t(i+1,j+1)*(x1_ssc_t(i+1)-x1_ssc_v(i  )) * pi180
	end do
	end do


!-------
!#3.1.4	lengths of lateral boundaries of t-cells
!-------
	do j=1 ,jl_ssc
	do i=1, il_ssc
	dx1_nw_ssc_t(i,j) = lame1_ssc_v(i-1,j  )*(x1_ssc_t(i  )-x1_ssc_v(i-1)) * pi180
	dx1_ne_ssc_t(i,j) = lame1_ssc_v(i  ,j  )*(x1_ssc_v(i  )-x1_ssc_t(i  )) * pi180
	end do
	end do

!-------
!#3.1.5	free memory
!-------
	deallocate ( lame1_ssc_t )
	deallocate ( lame1_ssc_v )




!-------
!#3.2	grid step for south-north direction
!-------
!-------
!#3.2.2	define lame coefficients for 2d coordinate transformations
!-------

	do j=0,jl_ssc+1
	do i=0,il_ssc+1
	lame2_ssc_t(i,j)	=radius
	lame2_ssc_v(i,j)	=radius
	end do
	end do

!-------
!#3.2.3	lengths of lateral boundaries of v-cells
!-------
	do j=1 ,jl_ssc
	do i=1, il_ssc
	dx2_en_ssc_v(i,j) = lame2_ssc_t(i+1,j+1)*(x2_ssc_t(j+1)-x2_ssc_v(j  )) * pi180
	dx2_es_ssc_v(i,j) = lame2_ssc_t(i+1,j  )*(x2_ssc_v(j  )-x2_ssc_t(j  )) * pi180
	end do
	end do


!-------
!#3.1.4	lengths of lateral boundaries of t-cells
!-------
	do j=1 ,jl_ssc
	do i=1, il_ssc
	dx2_en_ssc_t(i,j) = lame2_ssc_v(i  ,j  )*(x2_ssc_v(j  )-x2_ssc_t(j  )) * pi180
	dx2_es_ssc_t(i,j) = lame2_ssc_v(i  ,j-1)*(x2_ssc_t(j  )-x2_ssc_v(j-1)) * pi180
	end do
	end do
!-------
!#3.1.5	free memory
!-------
	deallocate ( lame2_ssc_t )
	deallocate ( lame2_ssc_v )

	return                                                            
	end                                                               
