#include "o_defines.f90"


!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	ocean thermodynamic equations
!	scheme of central differences for tracer advection (with mean fluxes accumulation for diagnostics)
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru 
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!--------------------------------------

!--------------------------------------
!	integration of advection term of scalar equation 
!	t(l+1)=t(l)+advection(t(l))*dt
!--------------------------------------
	module o_td_adv_cd
	contains
	subroutine o_td_scalar_advection_cd (ctime,uc,vc,tc,mht_u,mht_v)
	use o_basic_module, only: w,ct3
	use o_grid_module
	use o_d_brcl_module
	use o_td_module, only: t_l,t_f,flow_t_n,flow_t_e,flow_t_b,adv_t
	use o_diagn_means_module
	use o_par_module
	implicit none
	
	real (kind=main_kind), intent(in), dimension(kl, iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1) :: uc,vc
	real (kind=main_kind), intent(in), dimension(kl, iwest (myrank)-1:ieast (myrank)+1, jsouth(myrank)-1:jnorth(myrank)+1,ml) :: tc
	type(mean_field),optional :: mht_u,mht_v
	integer, intent(in) ::	ctime		!=1 or 2, for Euler or leap-frog time-stepping respectively
	
	
!-------
! flow through the walls of t-box [m3/sec]
!-------
	!lateral walls [m3/sec]
	do j=jsouth(myrank)-1	,jnorth(myrank)
	do i=iwest (myrank)	,ieast (myrank)
	do k=1, kl
		flow_v_n(k,i,j) =									&
			(vc(k,i-1,j)*dx3_v_c(k+1,i-1,j) + vc(k,i,j)*dx3_v_c(k+1,i,j)   ) * 0.5 *	&
			(dx1_nw_t(i,j) + dx1_ne_t(i,j))
	end do
	end do
	end do

	do j=jsouth(myrank)	,jnorth(myrank)
	do i=iwest (myrank)-1	,ieast (myrank)
	do k=1, kl
		flow_u_e (k,i,j) =									&
			(uc(k,i,j)*dx3_v_c(k+1,i,j) + uc(k,i,j-1)*dx3_v_c(k+1,i,j-1)) * 0.5 *	&
			(dx2_en_t(i,j) + dx2_es_t(i,j))
	end do	
	end do
	end do

	!bottom walls [m3/sec]/area_tbox
	do j = jsouth(myrank)	,jnorth(myrank)
	do i = iwest (myrank)	,ieast (myrank)
		do k = 1, kt(i,j)-1
			flow_w_b(k+1,i,j) = w(k+1,i,j)*ct3(k+1,i,j)	!flow through the bottom of the k-th horizon
		end do
		flow_t_b(kt(i,j),i,j) = 0.	!"ears" for k-axis
		flow_t_b(0      ,i,j) = 0.
	end do
	end do

	do m = 1,ml

		!-------
		!central differences scheme for advection
		!-------
		do j = jsouth(myrank)-1	,jnorth(myrank)
		do i = iwest (myrank)-1	,ieast (myrank)
		do k = 1, kl
			flow_t_n(k,i,j) = flow_v_n(k,i,j) * (tc(k,i,j,m) + tc(k,i,j+1,m)) * 0.5
			flow_t_e(k,i,j) = flow_u_e(k,i,j) * (tc(k,i,j,m) + tc(k,i+1,j,m)) * 0.5
		end do
		end do
		end do

		!Adding heat fluxes to their collectors to make means (e.g. annual mean MHT)
 		if(present(mht_u).and.present(mht_v).and.m==1) then
			func_u_e=0.
			func_v_n=0.
			do j = jsouth(myrank)	,jnorth(myrank)
			do i = iwest (myrank)	,ieast (myrank)
			do k = 1, kt(i,j)
 				func_u_e(k,i,j)= flow_t_e(k,i,j)/( dx2_en_t(i,j)+dx2_es_t(i,j) )
 				func_v_n(k,i,j)= flow_t_n(k,i,j)/( dx1_nw_t(i,j)+dx1_ne_t(i,j) )
			end do
			end do
			end do
			call o_diagn_time_mean(dim3_arr_in=func_u_e, arr_out=mht_u, finalize_condition=(MOD(time_l,time_l_save_cp) == 0))
			call o_diagn_time_mean(dim3_arr_in=func_v_n, arr_out=mht_v, finalize_condition=(MOD(time_l,time_l_save_cp) == 0))
			
			!rotate heat flux arrays if finalized
			if(MOD(time_l,time_l_save_cp) == 0) then
				do k=1,kl
					call tsc_ssc_rotate_v(	mht_u % collector3d(k,:,:), mht_v % collector3d(k,:,:),			&
								mht_u % collector3d(k,:,:), mht_v % collector3d(k,:,:),			&
								iwest(myrank)-1,ieast(myrank)+1,jsouth(myrank)-1,jnorth(myrank)+1)	
				end do
			end if
 		end if
		
		if(ctime==0) return !This allows us to make means without disturbing t_f (e.g. to know heat transported by 3month-mean velocities)
		
		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1, kt(i,j)-1
			flow_t_b(k,i,j) = flow_w_b(k+1,i,j) * (tc(k,i,j,m) + tc(k+1,i,j,m)) * 0.5	!some historical confusion: indices of fluxes on the bottom of the box #k are flow_t_b(k) but flow_w_b(k+1)
		end do
		end do
		end do


		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1, kt(i,j)
			adv_t (k,i,j) = area_tbox_v(i,j) * (				&
					(flow_t_e(k,i  ,j  ) - flow_t_e(k,i-1,j)) +	&
					(flow_t_n(k,i  ,j  ) - flow_t_n(k,i,j-1))  )
		end do
		end do
		end do

		if (ctime.eq.1) then
			do j = jsouth(myrank)	,jnorth(myrank)
			do i = iwest (myrank)	,ieast (myrank)
			do k = 1		,kt(i,j)
				t_f(k,i,j,m) = tc(k,i,j,m)*ct3(k,i,j)*dx3_t_c(k+1,i,j)
			end do
			end do
			end do
		end if

		if (ctime.eq.2) then
			do j = jsouth(myrank)	,jnorth(myrank)
			do i = iwest (myrank)	,ieast (myrank)
			do k = 1		,kt(i,j)
				t_f(k,i,j,m) = t_l(k,i,j,m)*ct3(k,i,j)*dx3_t_l(k+1,i,j)
			end do
			end do
			end do
		end if

		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1		,kt(i,j)
			t_f(k,i,j,m) = t_f(k,i,j,m) - real(ctime,kind=main_kind)*htsec* (adv_t (k,i,j) + flow_t_b(k,i,j)-flow_t_b(k-1,i,j))
		end do
		end do
		end do

	end do			!do m=1,ml
	end subroutine o_td_scalar_advection_cd
	end module o_td_adv_cd	
