#include "o_defines.f90"


!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	ocean thermodynamic equations
!	flux correction scheme for tracers advection (with mean fluxes accumulation for diagnostics)
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru 
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!
! reference:
!	Zalezak, S.T., 1979. 
!		Fully multidimensional flux-corrected transport algorithm for fluids. 
!		J. Comput. Phys., 31, 335-362.
!--------------------------------------


!~~~~~~
	module o_td_adv_fctz
	save
	real (kind=main_kind),allocatable,dimension(:,:,:) :: ta,tb,tmax,tmin,c_t_b_tbox,c_t_e_tbox,c_t_n_tbox,rin,rout
	
	contains
	
	subroutine allocate_o_td_adv_fctz
		use o_grid_module
		allocate(	ta 	(kl,iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)		)
		allocate(	tb 	(kl,iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)		)
		allocate(	tmax	(kl,iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)		)
		allocate(	tmin 	(kl,iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)		)
		allocate(	c_t_b_tbox	(kl,iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)	)
		allocate(	c_t_e_tbox	(kl,iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)	)
		allocate(	c_t_n_tbox	(kl,iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)	)
		allocate(	rin	(kl,iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)		)
		allocate(	rout	(kl,iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)		)
	end subroutine allocate_o_td_adv_fctz
	
	subroutine o_td_scalar_advection_fctz(ctime,tc,tl,dx3_tl,mht_u,mht_v)
!~~~~~~
!--------------------------------------
!	integration of advection term of scalar equation 
!	t(l+1)=t(l)+advection(t(l))*dt using flux correction of Zalesak, J.Comp. Phys, 1979
!--------------------------------------
	use o_basic_module
	use o_grid_module
	use o_d_brcl_module, only: func_u_e,func_u_n,func_u_b,flow_u_e,flow_v_n,flow_w_b,func_v_e,func_v_n
	use o_td_module, only: flow_t_e,flow_t_n,flow_t_b,adv_t,t_f
	use o_par_module
	use o_diagn_means_module

	implicit none
	
	real (kind=main_kind)  tc	(kl,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1,ml)
	real (kind=main_kind)  tl	(kl,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1,ml)
	real (kind=main_kind)  dx3_tl	(kl+1,	iwest (myrank)-1:ieast (myrank)+1,jsouth(myrank)-1:jnorth(myrank)+1)
	
	real(kind=main_kind) taimjz,taipjz,taizjm,taizjp,taijzm,taijzp
	real(kind=main_kind) tbimjz,tbipjz,tbizjm,tbizjp,tbijzm,tbijzp
	real (kind=8) pin,qin,pout,qout
	type(mean_field),optional :: mht_u,mht_v
	integer ctime

!-------
!		flow through the walls of t-box  
!-------
	!lateral walls [m3/sec]
	do j = jsouth(myrank)	,jnorth(myrank)
	do i = iwest (myrank)-1	,ieast (myrank)
		do k = 1,kt(i,j) 
			flow_u_e(k,i,j) =									&
				(u_c(k,i,j)*dx3_v_c(k+1,i,j) + u_c(k,i,j-1)*dx3_v_c(k+1,i,j-1))*0.5*		&
				(dx2_en_t(i,j) + dx2_es_t(i,j))	
		end do	
		flow_t_b(kt(i,j),i,j)=0.	!This initialization is need further for correct calculation of the near-bottom cell
		flow_t_b(0      ,i,j)=0.
	end do
	end do

	do j = jsouth(myrank)-1	,jnorth(myrank)
	do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j) 
			flow_v_n(k,i,j) =									&
				(v_c(k,i-1,j)*dx3_v_c(k+1,i-1,j) + v_c(k,i,j)*dx3_v_c(k+1,i,j))*0.5*		&
				(dx1_nw_t(i,j) + dx1_ne_t(i,j))
		end do	
	end do
	end do

	!top and bottom walls [m3/sec]/area_tbox
	do j = jsouth(myrank)	,jnorth(myrank)
	do i = iwest (myrank)	,ieast (myrank)
	do k = 1,kt(i,j)-1    
		flow_w_b(k+1,i,j)=w(k+1,i,j)*ct3(k+1,i,j)
	end do
	end do
	end do

	do m=1,ml
!-------
!			upwind scheme for horizontal advection, flow_t_*=[deg*m3/sec]
!-------
		do j = jsouth(myrank)-1	,jnorth(myrank)
		do i = iwest (myrank)-1	,ieast (myrank)
		do k = 1,kt(i,j)  
			if(flow_u_e (k,i,j).ge.0.) then
				flow_t_e (k,i,j) = flow_u_e (k,i,j)*tl(k,i,j,m)
			else
				flow_t_e (k,i,j) = flow_u_e (k,i,j)*tl(k,i+1,j,m)
			end if
				
			if(flow_v_n(k,i,j).ge.0.) then
				flow_t_n(k,i,j) = flow_v_n(k,i,j)*tl(k,i, j,m  )
			else
				flow_t_n(k,i,j) = flow_v_n(k,i,j)*tl(k,i,j+1,m)
			end if
		end do
		end do
		end do			
		
!-------
!			upwind scheme for vertical advection. flow_t_b=[deg*m3/sec]/area_tbox=[deg/sec]*ct3*dx3
!-------
		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j)-1
			if(flow_w_b(k+1,i,j).ge.0.) then
				flow_t_b(k,i,j) = flow_w_b(k+1,i,j)*tl(k,i,j,m)	!some historical confusion: indices of fluxes on the bottom of the box #k are flow_t_b(k) but flow_w_b(k+1)
			else
				flow_t_b(k,i,j) = flow_w_b(k+1,i,j)*tl(k+1,i,j,m)
			end if
		end do
		end do
		end do			
	
!-------
!		calculating transported and diffuzed solution, adv_t=[deg*m3/sec]/area_tbox=[deg/sec]*ct3*dx3
!-------
		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j) 
			adv_t(k,i,j) = area_tbox_v(i,j) * (					&
     				(flow_t_e (k,i  ,j  ) - flow_t_e (k,i-1,j  )) +			&
     				(flow_t_n (k,i  ,j  ) - flow_t_n (k,i  ,j-1))  )
			
			!if(m==1) flux_e_dg_adv(i,j)=flux_e_dg_adv(i,j)+flow_t_e(k,i,j)/( dx2_en_t(i,j)+dx2_es_t(i,j) )
			!if(m==1) flux_n_dg_adv(i,j)=flux_n_dg_adv(i,j)+flow_t_n(k,i,j)/( dx1_nw_t(i,j)+dx1_ne_t(i,j) )
		end do
		end do
		end do			
		
		!Saving heat fluxes to make means (e.g. annual mean MHT)
 		!finalization and rotation will be after adding antidiffusive fluxes
 		if(present(mht_u).and.present(mht_v).and.m==1) then
			do j = jsouth(myrank)	,jnorth(myrank)
			do i = iwest (myrank)	,ieast (myrank)
			do k = 1, kt(i,j)
 				func_v_e(k,i,j)= flow_t_e(k,i,j)/( dx2_en_t(i,j)+dx2_es_t(i,j) )
 				func_v_n(k,i,j)= flow_t_n(k,i,j)/( dx1_nw_t(i,j)+dx1_ne_t(i,j) )
			end do
			end do
			end do
 		end if

		!t_f=[deg]*ct3*dx3
		if(ctime.ne.0) then	!ctime=0 means that we only want to calculate fluxes (e.g. MHT) without disturbing t_f
		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j)
			t_f(k,i,j,m) = tl(k,i,j,m)*ct3(k,i,j)*dx3_tl(k+1,i,j) -					&
						htsec*real(ctime)*(adv_t (k,i,j) + flow_t_b(k,i,j)-flow_t_b(k-1,i,j) )
		
		end do
		end do
		end do	
		end if
!-------
!			centered difference scheme for horizontal advection,func_u_*=[deg*m3/sec]
!-------
		do j = jsouth(myrank)-1	,jnorth(myrank)
		do i = iwest (myrank)-1	,ieast (myrank)
		do k = 1,kt(i,j)
			func_u_n(k,i,j) = flow_v_n (k,i,j)*(tc(k,i,j,m)+tc(k,i,j+1,m))*0.5
			func_u_e(k,i,j) = flow_u_e (k,i,j)*(tc(k,i,j,m)+tc(k,i+1,j,m))*0.5
		end do
		end do
		end do			

!-------
!			centered difference scheme for vertical advection,func_u_b=[deg*m3/sec]/area_tbox
!-------
		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j)-1
			func_u_b(k,i,j) = flow_w_b(k+1,i,j)*(tc(k,i,j,m)+tc(k+1,i,j,m))*0.5
		end do
		end do
		end do			

!-------
!		flow_t_* become antidiffuzive fluxes (page 336 of Zalesak, 1979)
!-------
		do j = jsouth(myrank)-1	,jnorth(myrank)
		do i = iwest (myrank)-1	,ieast (myrank)
		do k = 1,kt(i,j)  
			flow_t_e (k,i,j) = func_u_e (k,i,j) -flow_t_e (k,i,j) 
			flow_t_n (k,i,j) = func_u_n (k,i,j) -flow_t_n (k,i,j) 
		end do
		end do
		end do			
	
		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j)-1  
			flow_t_b (k,i,j) = func_u_b (k,i,j) -flow_t_b (k,i,j) 
		end do
		end do
		end do			
	
!-------
!		bounding values (eqs. 17' and 18')
!-------
#if CMF_VERSION == 2    
	call comp_halo_update_T(DIM3_t = t_f(:,:,:,m), update_width = 1, change_sign_on_bipolar = 1)
#elif CMF_VERSION == 3
	call HaloUpdateMaster % update(t_f(:,:,:,m), update_width = 1, grid_type = 'T')
#endif  		
		
		do j = jsouth(myrank)-1	,jnorth(myrank)+1
		do i = iwest (myrank)-1	,ieast (myrank)+1
		do k = 1,kt(i,j)  
			ta(k,i,j) = max(tl(k,i,j,m)*ct3(k,i,j)*dx3_tl(k+1,i,j),t_f(k,i,j,m))
			tb(k,i,j) = min(tl(k,i,j,m)*ct3(k,i,j)*dx3_tl(k+1,i,j),t_f(k,i,j,m))
		end do
		end do
		end do			

		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j)  
								taimjz = ta(k,i-1,j  )
								tbimjz = tb(k,i-1,j  )
			if(cv(k,i-1,j  )+cv(k,i-1,j-1).lt.0.1)	then 
								taimjz = ta(k,i  ,j  )
								tbimjz = tb(k,i  ,j  )
								end if
								
								taipjz = ta(k,i+1,j  )
								tbipjz = tb(k,i+1,j  )
			if(cv(k,i  ,j  )+cv(k,i  ,j-1).lt.0.1)	then 
								taipjz = ta(k,i  ,j  )
								tbipjz = tb(k,i  ,j  )
								end if
								
								taizjp = ta(k,i  ,j+1)
								tbizjp = tb(k,i  ,j+1)
			if(cv(k,i-1,j  )+cv(k,i  ,j  ).lt.0.1)	then 
								taizjp = ta(k,i  ,j  )
								tbizjp = tb(k,i  ,j  )
								end if
								
								taizjm = ta(k,i  ,j-1)
								tbizjm = tb(k,i  ,j-1)
			if(cv(k,i-1,j-1)+cv(k,i  ,j-1).lt.0.1)	then
								taizjm = ta(k,i  ,j  )
								tbizjm = tb(k,i  ,j  )
								end if
			taijzp = ta(min(k+1,kt(i,j)),i,j)
			tbijzp = tb(min(k+1,kt(i,j)),i,j)
			taijzm = ta(max(k-1,1)      ,i,j)
			tbijzm = tb(max(k-1,1)      ,i,j)
			
			tmax(k,i,j) = max(ta(k,i,j),	taimjz,taipjz,taizjm,taizjp,taijzp,taijzm)
			tmin(k,i,j) = min(tb(k,i,j),	tbimjz,tbipjz,tbizjm,tbizjp,tbijzp,tbijzm)
		end do
		end do
		end do			

!-------
!		total inward and outward antidiffuzive fluxes (eqs. 7'-12')
!		Note that the expression calculated is t_c*ct3*dx3+advection*dt, so 
!		all temperatures (t_max, t_f, ...) here are "real"temperatures*ct3*dx3, 
!		horizontal heat fluxes are "real"fluxes [deg*m3/sec]
!-------
		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j)  
			!incoming total heat [grad*m3/s]
			pin = 	max(0.d0,flow_t_e(k,i-1,j))-min(0.d0,flow_t_e(k,i,j)) +		&
     				max(0.d0,flow_t_n(k,i,j-1))-min(0.d0,flow_t_n(k,i,j)) +  	&
     	      		      (	max(0.d0,flow_t_b(k-1,i,j))-min(0.d0,flow_t_b(k,i,j)) )	*area_tbox(i,j)
	
			!maximum allowed incoming total heat [deg*m3/s] during this timestep
			qin = (tmax(k,i,j)-t_f(k,i,j,m))*area_tbox(i,j)/htsec/real(ctime)
		
			if(pin.gt.1.d-10) then
				rin(k,i,j) = min(1.d0, qin/pin) 
			else
				rin(k,i,j) = 0.
			end if

			!outgoing total heat [grad*m3/s]
			pout = 	max(0.d0,flow_t_e(k,i,j))-min(0.d0,flow_t_e(k,i-1,j)) +		&
     				max(0.d0,flow_t_n(k,i,j))-min(0.d0,flow_t_n(k,i,j-1)) +  	&
     	       		       (max(0.d0,flow_t_b(k,i,j))-min(0.d0,flow_t_b(k-1,i,j)) )*area_tbox(i,j)
	
			!maximum allowed outgoing total heat [grad*m3/s] during this timestep
			qout = (t_f(k,i,j,m)-tmin(k,i,j))*area_tbox(i,j)/htsec/real(ctime)
		
			if(pout.gt.1.d-10) then
				rout(k,i,j) = min(1.d0, qout/pout) 
			else
				rout(k,i,j) = 0.
			end if
		end do
		end do
		end do			
	
#if CMF_VERSION == 2    
        call comp_halo_update_T(DIM3_t = rin, update_width = 1, change_sign_on_bipolar = 1)
        call comp_halo_update_T(DIM3_t = rout,update_width = 1, change_sign_on_bipolar = 1)
#elif CMF_VERSION == 3
        call HaloUpdateMaster % update(rin, update_width = 1, grid_type = 'T')
        call HaloUpdateMaster % update(rout, update_width = 1, grid_type = 'T')    
#endif          
	
!-------
!		final restriction of antidiffuzive fluxes
!-------
		do j = jsouth(myrank)-1	,jnorth(myrank)
		do i = iwest (myrank)-1	,ieast (myrank)
		do k = 1,kt(i,j)
			if(flow_t_e(k,i,j).ge.0.) then
				c_t_e_tbox(k,i,j) = min(rin(k,i+1,j),rout(k,i,j))
			else
				c_t_e_tbox(k,i,j) = min(rin(k,i,j),rout(k,i+1,j))
			end if

			if(flow_t_n(k,i,j).ge.0.) then
				c_t_n_tbox(k,i,j) = min(rin(k,i,j+1),rout(k,i,j))
			else
				c_t_n_tbox(k,i,j) = min(rin(k,i,j),rout(k,i,j+1))
			end if
		end do
		end do
		end do			

		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j)-1
			if(flow_t_b(k,i,j).gt.0.) then
				c_t_b_tbox(k,i,j) = min(rin(k+1,i,j),rout(k,i,j))
			else
				c_t_b_tbox(k,i,j) = min(rin(k,i,j),rout(k+1,i,j))
			end if
		end do
		if(kt(i,j)>0) c_t_b_tbox(kt(i,j),i,j)=0.
		end do
		end do			
	
		do j = jsouth(myrank)-1	,jnorth(myrank)
		do i = iwest (myrank)-1	,ieast (myrank)
		do k = 1,kt(i,j)
			flow_t_e(k,i,j) = flow_t_e(k,i,j)*c_t_e_tbox(k,i,j)*0.99  
			flow_t_n(k,i,j) = flow_t_n(k,i,j)*c_t_n_tbox(k,i,j)*0.99 
			flow_t_b(k,i,j) = flow_t_b(k,i,j)*c_t_b_tbox(k,i,j)*0.99 
		end do
		end do
		end do
!-------
!		applying restricted antidiffuzive fluxes
!-------
		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j) 
			adv_t(k,i,j) = area_tbox_v(i,j) * (				&
     				(flow_t_e(k,i  ,j  ) - flow_t_e(k,i-1,j)) +		&
     				(flow_t_n(k,i  ,j  ) - flow_t_n(k,i,j-1))  )
			
			!if(m==1) flux_e_dg_adv(i,j)=flux_e_dg_adv(i,j)+flow_t_e(k,i,j)/( dx2_en_t(i,j)+dx2_es_t(i,j) )
			!if(m==1) flux_n_dg_adv(i,j)=flux_n_dg_adv(i,j)+flow_t_n(k,i,j)/( dx1_nw_t(i,j)+dx1_ne_t(i,j) )
		end do
		end do
		end do			
		
		!Adding heat fluxes to their collectors to make means (e.g. annual mean MHT)
 		if(present(mht_u).and.present(mht_v).and.m==1) then
			do j = jsouth(myrank)	,jnorth(myrank)
			do i = iwest (myrank)	,ieast (myrank)
			do k = 1, kt(i,j)
 				func_v_e(k,i,j)= func_v_e(k,i,j) + flow_t_e(k,i,j)/( dx2_en_t(i,j)+dx2_es_t(i,j) )
 				func_v_n(k,i,j)= func_v_n(k,i,j) + flow_t_n(k,i,j)/( dx1_nw_t(i,j)+dx1_ne_t(i,j) )
			end do
			end do
			end do
			call o_diagn_time_mean(dim3_arr_in=func_v_e, arr_out=mht_u, finalize_condition=(MOD(time_l,time_l_save_cp) == 0))
			call o_diagn_time_mean(dim3_arr_in=func_v_n, arr_out=mht_v, finalize_condition=(MOD(time_l,time_l_save_cp) == 0))
			
			!rotate heat flux arrays if finalized
			if(MOD(time_l,time_l_save_cp) == 0) then
				do k=1,kl
					call tsc_ssc_rotate_v(	mht_u % collector3d(k,:,:), mht_v % collector3d(k,:,:),			&
								mht_u % collector3d(k,:,:), mht_v % collector3d(k,:,:),			&
								iwest(myrank)-1,ieast(myrank)+1,jsouth(myrank)-1,jnorth(myrank)+1)	
				end do
			end if
 		end if
		if(ctime==0) return !This allows us to make means without disturbing t_f (e.g. to know heat transported by 3month-mean velocities)
 		
		do j = jsouth(myrank)	,jnorth(myrank)
		do i = iwest (myrank)	,ieast (myrank)
		do k = 1,kt(i,j)  
			t_f(k,i,j,m) = t_f(k,i,j,m)-htsec*real(ctime)*(adv_t(k,i,j)+flow_t_b(k,i,j)-flow_t_b(k-1,i,j))
		end do
		end do
		end do			
		
	end do	!do m=1,ml
	end subroutine o_td_scalar_advection_fctz
	end module o_td_adv_fctz