#include "o_defines.f90"


!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	ocean thermodynamic equations
!	equation of state
!	Compute 3-d density array and do Hydrostatic Adjustment
! 		with partial bottom boxes (McDougall 2003)
! authors: 
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!
! reference:
!	McDougall T. J., Jackett D. R., Wright D. G., Feistel R., 2003. Accurate and computationally efficient 
!		algorithms for potential temperature and density of seawater. J. Atmos. Oceanic Technol., 20, 730–741.
!--------------------------------------

	subroutine o_td_state_mcdougall2003

	use o_basic_module
	use o_grid_module
	use o_d_brcl_module
	use o_diagn_mld_module
	!use shared_module

	implicit none

	real (kind=main_kind) tmixed (ml)
	real (kind=8) press,rho_from_theta,h,dv1,volmix,dmixed,gd
	integer loop,npoint,k0,k1,kk

	!online diagnostics - initializing the step
	!if(o_gen_diag_flag) conv_adj_effect=conv_adj_effect-t_c
!-------
!		calculation of POTENTIAL density for hydrostatic adjustment
!-------
	d_c=0.

	do j=jsouth(myrank)-1   , jnorth(myrank)+1
	do i=iwest (myrank)-1   , ieast (myrank)+1
	do k=1,kl
		if(k.le.kt(i,j)) d_c(k,i,j)=rho_from_theta(dble(t_c(k,i,j,2)), dble(t_c(k,i,j,1)),0.d0)
	end do
	end do
	end do

	if(mix_unstable_levels.eq.1) then
!-------
!		volume of t-boxes in case of regular height of bottom t-boxes
!		volume computed without multiplying by 0.25
!-------

	    flow_u_n = 0.	!array used for tbox_volume

	    do j=jsouth(myrank)-1   , jnorth(myrank)+1
	    do i=iwest (myrank)-1   , ieast (myrank)+1
	    do k=1,kl
	        if(k.le.kt(i,j)  ) then
			h=hzw(k+1)
			if(k.eq.1) h=hzw(k+1)+real(ssl_c(i,j)-seash)

			dv1=4.*ct3(k,i,j)
			flow_u_n(k,i,j)=h*dv1
	        end if
	    end do
	    end do
	    end do

!-------
!		mix unstable levels
!-------
		do 30 j=jsouth(myrank)-1   , jnorth(myrank)+1
		do 30 i=iwest (myrank)-1   , ieast (myrank)+1
			if(kt(i,j) .le.0) go to 30

			do 10 loop=1,nloop
				npoint=0                
				do k=1,kl-1
					if(k.eq.kt(i,j)  ) go to 9
					if(d_c(k,i,j).gt.d_c(k+1,i,j) ) then
						k0=k
						k1=k
1                       			continue
						k1=k1+1

						do m=1,ml
							tmixed(m)=0.
						end do
						Volmix=0.
						do kk=k0,k1
							do m=1,ml
								tmixed(m) = tmixed(m) + t_c(kk,i,j,m)*flow_u_n(kk,i,j)
							end do
							Volmix = Volmix + flow_u_n(kk,i,j)
						end do

						do m=1,ml
							tmixed(m)=tmixed(m)/Volmix
						end do
						dmixed = rho_from_theta(dble(tmixed(2)),dble(tmixed(1)),0.d0)

						do kk=k0,k1
							do m=1,ml
								t_c(kk,i,j,m)=tmixed(m)
							end do
							d_c(kk,i,j)=dmixed
						end do

						npoint=npoint+1
						if(k1.eq.kt(i,j) ) go to 9
						if(d_c(k1,i,j).gt.d_c(k1+1,i,j) ) go to 1
					end if
				end do
9               		continue
				if(npoint.eq.0) go to 30
10			continue
30	    	continue
	end if	!	if(mix_unstable_levels.eq.1) 
       
!-------
!		calculation of IN SITU density for pressure gradient in motor
!-------
	d_potential_c = d_c		!saving potential density for KPP
	do j=jsouth(myrank)-1   , jnorth(myrank)+1
	do i=iwest (myrank)-1   , ieast (myrank)+1
	do k=1,kl
		press=densit*gravit*z(k)*0.0001	!pressure [db]
		if(k.le.kt(i,j)) d_c(k,i,j)=	rho_from_theta(	dble(t_c(k,i,j,2)),	dble(t_c(k,i,j,1)),press)
	end do
	end do
	end do

	if(abs(rho_from_theta(35.d0-dble(S_ref),25.d0,2000.d0)-1031.654229d0).gt.1) stop 'Density error'
	!print *,'Test1',rho_from_theta(35.d0-dble(S_ref),25.d0,2000.d0),'1031.654229'
	!print *,'Test2',rho_from_theta(20.d0-dble(S_ref),20.d0,1000.d0),'1017.726743'
	!print *,'Test3',rho_from_theta(40.d0-dble(S_ref),12.d0,8000.d0),'1062.928258'
	!stop
	!print *,theta_from_t(35.,20.,2000.,0.)

	gd=gravit/densit
	do j=jsouth(myrank)-1   , jnorth(myrank)+1
	do i=iwest (myrank)-1   , ieast (myrank)+1
	do k=1,kl
		if(k.le.kt(i,j)  ) d_c(k,i,j)=gd*d_c(k,i,j)
	end do
	end do
	end do

	!online diagnostics - finishing the step
	!if(o_gen_diag_flag) conv_adj_effect=conv_adj_effect+t_c
		
	return
    end
	
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	real(kind = 8 ) function rho_from_theta(s1,th,p)
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
!         in-situ density from potential temperature, as in 
!         McDougall, Jackett, Wright and Feistel (2003), JAOT, V20, 730-741
!         s               : salinity (psu)
!         th              : potential temperature (deg C, ITS-90)
!         p               : pressure (dbar)
!         rho_from_theta  : in-situ density (kg/m^3)
!         check value     : rho_from_theta(20,20,1000) = 1017.726743139793
!         DRJ on 31/5/02
    use o_basic_module
    implicit real*8(a-h,o-z)

      s=s1+S_ref
      th2 = th*th; sqrts = sqrt(abs(s));
      anum =          9.99843699d+02 +		& 
                 th*( 7.35212840d+00 +		&
                 th*(-5.45928211d-02 +		&
                 th*  3.98476704d-04)) +	&
                  s*( 2.96938239d+00 -		&
                 th*  7.23268813d-03 +		&
                  s*  2.12382341d-03)      
            
      aden =          1.00000000d+00 +		&
                th*( 7.28606739d-03 +		&
                th*(-4.60835542d-05 +		& 
                th* (3.68390573d-07 +		&
                th*  1.80809186d-10))) +	&
                 s*( 2.14691708d-03 -		&
                th*( 9.27062484d-06 +		&
               th2*  1.78343643d-10) +		&
             sqrts*( 4.76534122d-06 +		&
               th2*  1.63410736d-09))

      if(p.ne.0.d0) then
          pth = p*th
          anum = anum +						&
                 p*( 1.04004591d-02 +		&
               th2*  1.03970529d-07 +		&
                 s*  5.18761880d-06 -		&
                 p*( 3.24041825d-08 +		&	!bug? There is '-' in the article, but control point succeeds with '+'
               th2*  1.23869360d-11))
          aden = aden +						&
                 p*( 5.30848875d-06 -		&
          pth*(th2*  3.03175128d-16 +		&
                 p*  1.27934137d-17))
      end if
      rho_from_theta = (anum/aden)
!	Note:   this function should always be run in double precision
!             (since rho is returned rather than sigma = rho-1.0d3)
      return
      end



!~~~~~~
      function theta_from_t(s,t,p,pr)
!~~~~~~
!         potential temperature from in-situ temperature, as in 
!         McDougall, Jackett, Wright and Feistel (2003), JAOT, V20, 730-741
!         s             : salinity (psu)
!         t             : in-situ temperature (deg C, ITS-90)
!         p             : pressure (dbar)
!         pr            : reference pressure (dbar)
!         theta_from_t  : potential temperature (deg C, ITS-90)
!         calls         : entropy_diff and de_dt_FH
!         check value   : theta_from_t(20,20,1000,0) = 19.82767858849492
!                                                           (with nloops=1)
!         check value   : theta_from_t(20,20,1000,0) = 19.82767858849503
!                                                           (with nloops=2)
!         DRJ on 31/5/02
      implicit real(kind=main_kind)(a-h,o-z)
 
      th0 = t+(p-pr)*( 1.067610d-05 - &
                  s*  1.434297d-06 - &
             (p+pr)*  7.566349d-09 + &
                  t*(-8.535585d-06 + &
                  s*  3.074672d-08 + &
                  t*  1.918639d-08 + &
             (p+pr)*  1.788718d-10))

!----------iterations will be tested later--------------
!      de_dt = 13.9d0
!      nloops = 1
!!        NOTE: nloops = 1 gives theta with a maximum error of 3.4x10^-06
!!              nloops = 2 gives theta with a maximum error of 3.6x10^-14
!      n = 1 
!      do while(n.le.nloops)
!          dentropy = entropy_diff(s,t,p,th0,pr)
!          theta = th0-dentropy/de_dt
!          theta = 0.5d0*(theta+th0)
!          de_dt = de_dt_FH(s,theta,pr)
!          theta = th0-dentropy/de_dt
!          n = n+1; th0 = theta
!      end do
!----------------------------------------------------------
      theta_from_t = th0
      return
      end

