
!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!	ocean thermodynamic equations
!	atmospheric boundary layer model driver
!
! authors: 
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru 
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!
! reference:
!	Launiainen, J. and Vihma, T., 1990.
!	Derivation of turbulent surface fluxes – an iterative flux-profile method allowing arbitrary observing heights, 
!	Environ. Softw., 5(3), 113–124.
!--------------------------------------



!--------------------------------------
!	Routine that arranges calculating of heat fluxes 
!	all units are in SI (J, kg, m)
!	output of convert: [evap] = kg/(m**2*s)
!--------------------------------------

	subroutine o_tf_abl_drive

	use o_tf_module
	use o_basic_module
	use o_grid_module
	use o_td_module
	use o_par_module

	implicit none

	real (kind=4) depth  !z, qh
!z	integer time
	real u_current, v_current, u_a_s, v_a_s
	integer ncount

!-------
!#0	if time is not an integer hour then leave the forcing unchanged
!-------

	!if(MOD(time_l-1, time_steps_day/24) /=0) return

!z	time = htsec*REAL(time_l_in_year)

!	do j = jsouth(myrank), jnorth(myrank)+1
!	do i = iwest (myrank), ieast (myrank)+1
!		if(kt(i,j) > 0) wind_speed_a_s(i,j) = wind_speed_m_t(i,j)
!	end do
!	end do

!-------
!	air speed relative to sea 
!-------

	k = 1
	do j = jsouth(myrank), jnorth(myrank)+1
	do i = iwest (myrank), ieast (myrank)+1
		if(kt(i,j) > 0) then
			u_current =					&
				(  	(u_c(1,i  ,j  )			&
					+u_c(1,i-1,j-1))		&
				+	(u_c(1,i  ,j-1)			&
					+u_c(1,i-1,j  ))  )*0.25 !/(4*ct3(k,i,j))
			v_current =					&
				(	(v_c(1,i  ,j  )			&
					+v_c(1,i-1,j-1))		&
				+	(v_c(1,i  ,j-1)			&
					+v_c(1,i-1,j  ))  )*0.25 !/(4*ct3(k,i,j))

			u_a_s = wind_speed_x_t(i,j) - u_current
			v_a_s = wind_speed_y_t(i,j) - v_current

			wind_speed_a_s(i,j) = SQRT(u_a_s**2+v_a_s**2)

		end if
	end do
	end do
#if CMF_VERSION == 2    
	call comp_halo_update_T4(DIM2_t = wind_speed_a_s, update_width = 1, change_sign_on_bipolar = 1)
#elif CMF_VERSION == 3
	call HaloUpdateMaster % update(wind_speed_a_s, update_width = 1, grid_type = 'T')
#endif
!---------------------------------------------------------------------
!	Heat flux at open sea surface is computed in 
!	atmospheric boundary layer model of t.vihma
!	input: tair02, tdew02, clouds, wind(10m)
!	output: tair10, q10, wind stress, qh_sns, qh_ltn, evap
!---------------------------------------------------------------------

	do j = jsouth(myrank)-1, jnorth(myrank)+1
	do i = iwest (myrank)-1, ieast (myrank)+1

		if(kt(i,j) > 0) then

			select case (atm_model_name)
				case("ATM_NCAR")
					call o_tf_abl_convert_sh_ncar( patm(i,j),	&
						  tair02(i,j),				&
						  shumid(i,j),				&
						  wind_speed_a_s(i,j),			&
						  real(t_c(1,i,j,1),kind=4),		&
						  wind_stress_m_t(i,j),			&
						  sns(i,j),				&
						  ltn(i,j),				&
						  evap(i,j),				&
						  ncount,Tair10(i,j),q10(i,j),rl(i,j)	)
				case("ATM_INTERIM")
					call o_tf_abl_convert( patm(i,j),		&
						  tair02(i,j),				&
						  tdew02(i,j),				&
						  shumid(i,j),				&						  
						  wind_speed_a_s(i,j),			&
						  real(t_c(1,i,j,1),kind=4),		&
						  wind_stress_m_t(i,j),			&
						  sns(i,j),				&
						  ltn(i,j),				&
						  evap(i,j),				&
						  ncount,Tair10(i,j),q10(i,j),rl(i,j)	)
				case("ATM_GFS","ATM_INM","ATM_INMCM4")
					call o_tf_abl_convert_sh( patm(i,j),	&
						  tair02(i,j),				&
						  shumid(i,j),				&
						  wind_speed_a_s(i,j),			&
						  real(t_c(1,i,j,1),kind=4),		&
						  wind_stress_m_t(i,j),			&
						  sns(i,j),				&
						  ltn(i,j),				&
						  evap(i,j),				&
						  ncount,Tair10(i,j),q10(i,j),rl(i,j)	)
				
			end select			  
						  
!#################################
!	remove 'precipitation' from the 'frog'
!_cs_11_f76:
			if(evap(i,j) < 0.) then
				evap(i,j) = 0.
				ltn (i,j) = 0.
			end if
!#################################

		end if
	end do
	end do

!-------
! wind stress, recompute wind_stress_m_t to T then to V-point
!-------

	do j = jsouth(myrank), jnorth(myrank)+1
	do i = iwest (myrank), ieast (myrank)+1
		if(kt(i,j) > 0) wind_stress_x_t(i,j) = wind_stress_m_t(i,j)*wind_speed_x_t(i,j)/wind_speed_m_t(i,j)	! it is just a cos(phi)
	end do
	end do

	do j = jsouth(myrank), jnorth(myrank)+1
	do i = iwest (myrank), ieast (myrank)+1
		if(kt(i,j) > 0) wind_stress_y_t(i,j) =	wind_stress_m_t(i,j)*wind_speed_y_t(i,j)/wind_speed_m_t(i,j)	! it is just a sin(phi)
	end do
	end do

!z	do j=jsouth(myrank), jnorth(myrank)
!z	do i=iwest (myrank), ieast (myrank)
!z		if(kv(i,j) > 0) wind_stress_m_v(i,j) = 0.25*(	(wind_stress_m_t(i+1,j+1)+	&
!z								 wind_stress_m_t(i  ,j  ))+	&
!z								(wind_stress_m_t(i+1,j  )+	&
!z								 wind_stress_m_t(i  ,j+1)) )
!z	end do
!z	end do

	do j = jsouth(myrank), jnorth(myrank)
	do i = iwest (myrank), ieast (myrank)
		if(kv(i,j) > 0) wind_stress_x_v(i,j)=0.25*(	(wind_stress_x_t(i+1,j+1)+	&
								 wind_stress_x_t(i  ,j  ))+	&
								(wind_stress_x_t(i+1,j  )+	&
								 wind_stress_x_t(i  ,j+1)) )
	end do
	end do

	do j = jsouth(myrank), jnorth(myrank)
	do i = iwest (myrank), ieast (myrank)
		if(kv(i,j) > 0)  wind_stress_y_v(i,j)=0.25*(	(wind_stress_y_t(i+1,j+1)+	&
								 wind_stress_y_t(i  ,j  ))+	&
								(wind_stress_y_t(i+1,j  )+	&
								 wind_stress_y_t(i  ,j+1)) )
	end do
	end do

	return
	end

