
!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
! authors: 
!	2015.07 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!	2010.05 / Ibrayev, Rashit, INM RAS, Russia, ibrayev@mail.ru 
!
! reference:
!	K. V. Ushakov, R. A. Ibraev, and V. V. Kalmykov, 2015.
!		Simulation of the world ocean climate with a massively parallel numerical model, 
!		Izv., Atmos. Ocean. Phys. 51 (4), 362–380.
!	R. A. Ibraev, K. V. Ushakov, and R. N. Khabeev, 2012.
!		Eddy-resolving 1/10° model of the World Ocean,
!		Izv., Atmos. Ocean. Phys. 48 (1), 37–46.
!	Ibrayev R.A., 2001: Model of enclosed and semi-enclosed sea hydrodynamics. 
!		Russ. J. Numer. Anal. Math. Modelling, 2001, vol. 16, No 4, p. 291-304.
!--------------------------------------



!~~~~~~~
	subroutine process_levitus_var(var, var_name, mis_value)
!~~~~~~~
    
! Process one levitus var
    
    use o_grid_module
    
    character(*), intent(in) :: var_name
    real(4), intent(in) :: mis_value
    
    real(kind=4), dimension(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank), kl), intent(inout) :: var


    ! Copy levels under 5500., because Levitus is undefined there
    do k = 1, kl
        if(z(k) > 5500.) var(:,:,k) = var(:,:, index_of_elem(5500., z))
    end do
    
    ! If Levitus has 'MISSING_VALUE' in ocean point, put nearest value from this level
    do j = jsouth(myrank), jnorth(myrank)
    do i = iwest(myrank), ieast(myrank)
    do k = 1, kt(i,j)
        if(var(i,j,k) == mis_value) var(i,j,k) = o_ini_search_nearest(i,j,var(:,:,k),mis_value)
    end do
    end do
    end do
    
    ! Set 0 instead of MISSING_VALUE
    var = MERGE(REAL(0), var, var == mis_value)
    
    if(var_name=='t_an') call o_user_redefine_T (var)
    if(var_name=='s_an') call o_user_redefine_S (var)
    
    !make (T,S)=0 at land points
    if(var_name=='t_an' .or. var_name=='s_an') then
        do k = 1, kl
        do j=jsouth(myrank) ,jnorth(myrank)
        do i=iwest (myrank) ,ieast (myrank)
            if(k.gt.kt(i,j)) var(i,j,k)=0.
        end do
        end do
        end do
    end if
   
    CONTAINS
    
!========================================================================================

    ! Replace it with search function of Utils module after full shift on CMF-3.0
    function index_of_elem(elem, vect) result (ind)
        integer :: ind
        real(kind=4) :: elem, vect(:)
        
        ind = 1
        do while (elem /= vect(ind))
            ind = ind + 1
        end do
    end function

!========================================================================================
    
!========================
! Search for nearest cell
!========================
    
    real(kind=4) function o_ini_search_nearest(i_c,j_c,field, mis_value)
    
    ! Input variables
    integer i_c, j_c
    real(kind=4) field(iwest(myrank):ieast(myrank), jsouth(myrank):jnorth(myrank))
    real(kind=4) mis_value
    
    ! Work variables
    integer rad
    integer s,n,w,e
    
    ! Search with increasing radius
    do rad = 1, MAX(il,jl)
        s = MAX(j_c-rad, jsouth(myrank))
        n = MIN(j_c+rad, jnorth(myrank))
        w = MAX(i_c-rad, iwest(myrank))
        e = MIN(i_c+rad, ieast(myrank))
        
        ! Get mean value of search result
        if(ANY(field(w:e, s:n) /= mis_value)) then
            o_ini_search_nearest = SUM(field(w:e, s:n), field(w:e, s:n) /= mis_value)/COUNT(field(w:e, s:n) /= mis_value)
            return
        end if
    end do
    
    ! If we found nothing
    o_ini_search_nearest = 0
    
    end function o_ini_search_nearest
    
!================================================================================================================================================================

    end subroutine process_levitus_var
    
    
!==================================
! Define test rectangle 'hv' relief
!==================================

    subroutine o_ini_def_test_hv
    
    use o_grid_module
      
    real(kind = 4) global_topo(il,jl)
    integer i0, j0, ii
    integer radius2
    integer width_mask_pole
    integer counter

    ! Define test rectangle relief 
    global_topo = 0.
    do j = 1, jl
    do i = 1, il
        if(j >= 2 .AND. j <= jl .AND. i >= 1 .AND. i <= il) global_topo(i,j) = z(kl)
    end do
    end do

    ! Make land around barotropic S & N poles
    if(bsc_flag == 1) then

        width_mask_pole = 20

        do j = 1, jl
        do i = 1, il
            if(j >= jl_ssc-1 .AND. j <= jl .AND. i >= 1 .AND. i <= width_mask_pole)                 global_topo(i,j) = 0.
            if(j >= jl_ssc-1 .AND. j <= jl .AND. i >= il-width_mask_pole+1 .AND. i <= il)           global_topo(i,j) = 0.
            if(j >= jl_ssc-1 .AND. j <= jl .AND. i >= il/2-width_mask_pole+1 .AND. i <= il/2+width_mask_pole)   global_topo(i,j) = 0.
        end do
        end do

    end if

    ! Define local hv from global topo
    do j = jsouth(myrank), jnorth(myrank)
    do i = iwest(myrank), ieast(myrank)
        hv(i,j) = global_topo(i,j)
    end do
    end do

    end subroutine o_ini_def_test_hv
    

!=================Subroutines for printing topography to an ASCII file============================= 
!~~~~~~
!   subroutine out_map (experiment,title,bt,x1,x2,il,jl,type_belt)
    subroutine out_map (experiment,title,bt,x1,x2,il,jl)
!~~~~~~

!--------------------------------------
!   output map consisting of several pieces
!--------------------------------------
    integer il,jl,i,j,i0,j0,nbelt
    real (kind=8)   x1 (il,jl)
    real (kind=8)   x2 (il,jl)
    character (len= 3)  experiment
    character (len= *) title
    character*1 bt(il,jl)
    character*1 xc1(4000),xc2(4000)
    character*7 xx

    xc1=' '
    xc2=' '

    open  (11,file=experiment//title)
    write (11,*) 'title for map=',title,'il,jl=',il,jl
    write (*,*) 'title for map=',title,'il,jl=',il,jl

    do i=1,il
    xc1(i)=' '
    if( mod(x1(i,1),1.).eq.0.) xc1(i)='*'
    xc2(i)=' '
    end do 

    do i=1,il
    if( mod(x1(i,1),1.).eq.0.) then
    write(xx(1:7),'(f7.2)') x1(i,1)
    if(i.gt.3)    xc2(i-3)=xx(1:1)
    if(i.gt.2)    xc2(i-2)=xx(2:2)
    if(i.gt.1)    xc2(i-1)=xx(3:3)
                  xc2(i  )=xx(4:4)
    if(i.lt.il)   xc2(i+1)=xx(5:5)
    if(i.lt.il-1) xc2(i+2)=xx(6:6)
    if(i.lt.il-2) xc2(i+3)=xx(7:7)
    end if
    end do 

    do nbelt=1,(il-1)/100+1
    call type_belt(xc1,xc2,bt,x2,il,jl,nbelt)
    end do
    close (11)
    
    return
    end subroutine out_map
            
!~~~~~~
    subroutine type_belt(xc1,xc2,bt,x2,il,jl,nbelt)
!~~~~~~

    integer il,jl,i,j,i0,j0,nbelt
    real (kind=8)   x2 (il,jl)
!   dimension x2(jl)
    character*1 bt(il,jl)
    character*115 line
    character*1 xc1(4000),xc2(4000)

        i0=100*(nbelt-1)+1
        i1=100* nbelt
    if(i1.ge.il) i1=il

!   write(*,*) il,i0,i1,'jl=',jl

    do i=1,115
    line(i:i)=' '
    end do

    write (11,'(i3,1x,100a1)') jl+1-j,(xc2(i),i=i0,i1)
    write (11,'(i3,1x,100a1)') jl+1-j,(xc1(i),i=i0,i1)
    write (11,'(i3,1x,100i1)') jl+1-j,(mod(i,10),i=i0,i1)
    write (11,*)

    do j=1,jl

    write (line(1:3),'(i3)') jl+1-j
    line(4:4)=' '
    do i=i0,i1
    line(i-i0+5:i-i0+5)=bt(i,jl+1-j)
    end do

    if(i1.eq.il) then
    i2=i1-i0+1
    line(i2+6:i2+6)=' '
    write(line(i2+7:i2+9),'(i3)') jl+1-j
    line(i2+10:i2+10)=' '
    write(line(i2+11:i2+16),'(f6.2)') x2(1,jl+1-j)
    end if

!   write (11,'(a115)') line(1:i1+15)
    write (11,'(a116)') line
    end do

    write (11,*)
    write (*,'(i3,1x,100i1)') jl+1-j,(mod(i,10),i=i0,i1)
    write (11,'(i3,1x,100i1)') jl+1-j,(mod(i,10),i=i0,i1)
    write (11,'(i3,1x,100a1)') jl+1-j,(xc1(i),i=i0,i1)
    write (11,'(i3,1x,100a1)') jl+1-j,(xc2(i),i=i0,i1)
    write (11,'(///)')
    return
    end subroutine type_belt

!~~~~~~
!   subroutine out_map_rivers (experiment,title,bt,x1,x2,il,jl,type_belt)
    subroutine out_map_rivers (experiment,title,bt,x1,x2,il,jl)
!~~~~~~

!--------------------------------------
!   output map consisting of several pieces
!--------------------------------------
    integer il,jl,i,j,i0,j0,nbelt
    real (kind=8)   x1 (il,jl)
    real (kind=8)   x2 (il,jl)
    character (len= 3)  experiment
    character (len= *) title
    character*1 bt(il,jl)
    character*1 xc1(4000),xc2(4000)
    character*7 xx

    xc1=' '
    xc2=' '

!   open  (11,file=experiment//title)


!   write (11,*) 'title for map=',title,'il,jl=',il,jl
    backspace 11
    backspace 11
    backspace 11
    backspace 11
    backspace 11
    backspace 11
    backspace 11
    backspace 11

    do i=1,il
    xc1(i)=' '
    if( mod(x1(i,1),1.).eq.0.) xc1(i)='*'
    xc2(i)=' '
    end do 

    do i=1,il
    if( mod(x1(i,1),1.).eq.0.) then
    write(xx(1:7),'(f7.2)') x1(i,1)
    if(i.gt.3)    xc2(i-3)=xx(1:1)
    if(i.gt.2)    xc2(i-2)=xx(2:2)
    if(i.gt.1)    xc2(i-1)=xx(3:3)
                  xc2(i  )=xx(4:4)
    if(i.lt.il)   xc2(i+1)=xx(5:5)
    if(i.lt.il-1) xc2(i+2)=xx(6:6)
    if(i.lt.il-2) xc2(i+3)=xx(7:7)
    end if
    end do 

    do nbelt=1,(il-1)/100+1
    call type_belt(xc1,xc2,bt,x2,il,jl,nbelt)
    end do
    close(11)
    return
    end subroutine out_map_rivers


!~~~~~~
    subroutine charar(m,bt,il,jl)
!~~~~~~
    integer il,jl,i,j,m
        dimension    m(il,jl)
    character*1 bt(il,jl)
        do j=1,jl
        do i=1,il
             bt(i,j)='.'
        if(m(i,j).lt. 0) bt(i,j)='-'
        if(m(i,j).eq.  1) bt(i,j)='1'
        if(m(i,j).eq.  2) bt(i,j)='2'
        if(m(i,j).eq.  3) bt(i,j)='3'
        if(m(i,j).eq.  4) bt(i,j)='4'
        if(m(i,j).eq.  5) bt(i,j)='5'
        if(m(i,j).eq.  6) bt(i,j)='6'
        if(m(i,j).eq.  7) bt(i,j)='7'
        if(m(i,j).eq.  8) bt(i,j)='8'
        if(m(i,j).eq.  9) bt(i,j)='9'
        if(m(i,j).eq. 10) bt(i,j)='0'
        if(m(i,j).eq. 11) bt(i,j)='a'
        if(m(i,j).eq. 12) bt(i,j)='b'
        if(m(i,j).eq. 13) bt(i,j)='c'
        if(m(i,j).eq. 14) bt(i,j)='d'
        if(m(i,j).eq. 15) bt(i,j)='e'
        if(m(i,j).eq. 16) bt(i,j)='f'
        if(m(i,j).eq. 17) bt(i,j)='g'
        if(m(i,j).eq. 18) bt(i,j)='h'
        if(m(i,j).eq. 19) bt(i,j)='i'
        if(m(i,j).eq. 20) bt(i,j)='j'
        if(m(i,j).eq. 21) bt(i,j)='k'
        if(m(i,j).eq. 22) bt(i,j)='l'
        if(m(i,j).eq. 23) bt(i,j)='m'
        if(m(i,j).eq. 24) bt(i,j)='n'
        if(m(i,j).eq. 25) bt(i,j)='o'
        if(m(i,j).eq. 26) bt(i,j)='p'
        if(m(i,j).eq. 27) bt(i,j)='q'
        if(m(i,j).eq. 28) bt(i,j)='r'
        if(m(i,j).eq. 29) bt(i,j)='s'
        if(m(i,j).eq. 30) bt(i,j)='t'

        if(m(i,j).eq. 31) bt(i,j)='1'
        if(m(i,j).eq. 32) bt(i,j)='2'
        if(m(i,j).eq. 33) bt(i,j)='3'
        if(m(i,j).eq. 34) bt(i,j)='4'
        if(m(i,j).eq. 35) bt(i,j)='5'
        if(m(i,j).eq. 36) bt(i,j)='6'
        if(m(i,j).eq. 37) bt(i,j)='7'
        if(m(i,j).eq. 38) bt(i,j)='8'
        if(m(i,j).eq. 39) bt(i,j)='9'
        if(m(i,j).eq. 40) bt(i,j)='0'
        if(m(i,j).eq. 41) bt(i,j)='a'
        if(m(i,j).eq. 42) bt(i,j)='b'
        if(m(i,j).eq. 43) bt(i,j)='c'
        if(m(i,j).eq. 44) bt(i,j)='d'
        if(m(i,j).eq. 45) bt(i,j)='e'
        if(m(i,j).eq. 46) bt(i,j)='f'
        if(m(i,j).eq. 47) bt(i,j)='g'
        if(m(i,j).eq. 48) bt(i,j)='h'
        if(m(i,j).eq. 49) bt(i,j)='i'
        if(m(i,j).eq. 50) bt(i,j)='j'
        if(m(i,j).eq. 51) bt(i,j)='k'
        if(m(i,j).eq. 52) bt(i,j)='l'
        if(m(i,j).eq. 53) bt(i,j)='m'
        if(m(i,j).eq. 54) bt(i,j)='n'
        if(m(i,j).eq. 55) bt(i,j)='o'
        if(m(i,j).eq. 56) bt(i,j)='p'
        if(m(i,j).eq. 57) bt(i,j)='q'
        if(m(i,j).eq. 58) bt(i,j)='r'
        if(m(i,j).eq. 59) bt(i,j)='s'
        if(m(i,j).eq. 60) bt(i,j)='t'

        if(m(i,j).eq. 61) bt(i,j)='1'
        if(m(i,j).eq. 62) bt(i,j)='2'
        if(m(i,j).eq. 63) bt(i,j)='3'
        if(m(i,j).eq. 64) bt(i,j)='4'
        if(m(i,j).eq. 65) bt(i,j)='5'
        if(m(i,j).eq. 66) bt(i,j)='6'
        if(m(i,j).eq. 67) bt(i,j)='7'
        if(m(i,j).eq. 68) bt(i,j)='8'
        if(m(i,j).eq. 69) bt(i,j)='9'
        if(m(i,j).eq. 70) bt(i,j)='0'
        if(m(i,j).eq. 71) bt(i,j)='a'
        if(m(i,j).eq. 72) bt(i,j)='b'
        if(m(i,j).eq. 73) bt(i,j)='c'
        if(m(i,j).eq. 74) bt(i,j)='d'
        if(m(i,j).eq. 75) bt(i,j)='e'
        if(m(i,j).eq. 76) bt(i,j)='f'
        if(m(i,j).eq. 77) bt(i,j)='g'
        if(m(i,j).eq. 78) bt(i,j)='h'
        if(m(i,j).eq. 79) bt(i,j)='i'
        if(m(i,j).eq. 80) bt(i,j)='j'
        if(m(i,j).eq. 81) bt(i,j)='k'
        if(m(i,j).eq. 82) bt(i,j)='l'
        if(m(i,j).eq. 83) bt(i,j)='m'
        if(m(i,j).eq. 84) bt(i,j)='n'
        if(m(i,j).eq. 85) bt(i,j)='o'
        if(m(i,j).eq. 86) bt(i,j)='p'
        if(m(i,j).eq. 87) bt(i,j)='q'
        if(m(i,j).eq. 88) bt(i,j)='r'
        if(m(i,j).eq. 89) bt(i,j)='s'
        if(m(i,j).eq. 90) bt(i,j)='t'

        if(m(i,j).eq. 91) bt(i,j)='1'
        if(m(i,j).eq. 92) bt(i,j)='2'
        if(m(i,j).eq. 93) bt(i,j)='3'
        if(m(i,j).eq. 94) bt(i,j)='4'
        if(m(i,j).eq. 95) bt(i,j)='5'
        if(m(i,j).eq. 96) bt(i,j)='6'
        if(m(i,j).eq. 97) bt(i,j)='7'
        if(m(i,j).eq. 98) bt(i,j)='8'
        if(m(i,j).eq. 99) bt(i,j)='9'
        if(m(i,j).eq.100) bt(i,j)='0'
        if(m(i,j).eq.101) bt(i,j)='a'
        if(m(i,j).eq.102) bt(i,j)='b'
        if(m(i,j).eq.103) bt(i,j)='c'
        if(m(i,j).eq.104) bt(i,j)='d'
        if(m(i,j).eq.105) bt(i,j)='e'
        if(m(i,j).eq.106) bt(i,j)='f'
        if(m(i,j).eq.107) bt(i,j)='g'
        if(m(i,j).eq.108) bt(i,j)='h'
        if(m(i,j).eq.109) bt(i,j)='i'
        if(m(i,j).eq.110) bt(i,j)='j'
        if(m(i,j).eq.111) bt(i,j)='k'
        if(m(i,j).eq.112) bt(i,j)='l'
        if(m(i,j).eq.113) bt(i,j)='m'
        if(m(i,j).eq.114) bt(i,j)='n'
        if(m(i,j).eq.115) bt(i,j)='o'
        if(m(i,j).eq.116) bt(i,j)='p'
        if(m(i,j).eq.117) bt(i,j)='q'
        if(m(i,j).eq.118) bt(i,j)='r'
        if(m(i,j).eq.119) bt(i,j)='s'
        if(m(i,j).eq.120) bt(i,j)='t'


        if(m(i,j).ge.121) bt(i,j)='+'
        end do
        end do
    return
    end subroutine charar
    