#include "ice_2Ddomain"
!-----------------------------------------
!	*****Subroutines for conversion of ATM_* data to CICE format*****
!		
!	Marchuk Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
! authors: 
!	2015 	/ Ibrayev, Rashit, IO RAS, Russia, ibrayev@mail.ru
!	2019.10	/ Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!
! reference:
!	Fadeev R.Yu., Ushakov K.V., Tolstykh M.A., Ibrayev R.A. (2018). Design and development 
!		of the SLAV-INMIO-CICE coupled model for seasonal prediction and climate research.
!		Russian Journal of Numerical Analysis and Mathematical Modelling, 33(6), pp. 333-340.
!	Kalmykov V.V., Ibrayev R.A., Kaurkin M.N., Ushakov K.V. (2018) Compact Modeling Framework v3.0 
!		for high-resolution global ocean-ice-atmosphere models. Geosci. Model Dev., 11, 3983-3997. 
!		https://doi.org/10.5194/gmd-11-3983-2018
!	Kalmykov V.V., Ibrayev R.A. (2013). 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Certificate of state registration of the computer progam number 2013619320. 
!		Date of state registration in the Registry of the computer programs October 1, 2013
!	Kalmykov V.V., Ibrayev R.A., Ushakov K.V. (2013). 
!		Compact modeling framework CMF3.0 for Earth system modeling on massively-parallel computers. 
!		Certificate of state registration of the computer progam number 2017610785. 
!		Date of state registration in the Registry of the computer programs January 18, 2017
!	Ibraev R.A., Ushakov K.V., Khabeev R.N. (2012).
!		Eddy-resolving 1/10° model of the World Ocean,
!		Izvestiya, Atmos. Ocean. Phys. 48 (1), 37–46.
!--------------------------------------
	
	module atm_general_module
	use ice_par_module
	use non_ga_ice_array_module
	implicit none
	save

	!parameters to be received from atm. See some comments below.
	real (kind=4), allocatable, dimension(:,:) ::		&
		Tair_interlayer,	Qa_interlayer,		&
		uatm_interlayer,	vatm_interlayer,	&
		flw_interlayer,		sw_interlayer,		&
		frain_interlayer,	fsnow_interlayer!,	&
		!uatm_rotated_interlayer,vatm_rotated_interlayer
       
	real (kind=4), allocatable, dimension(:,:) ::		&
		sns_atm_interlayer,	ltn_atm_interlayer,	&
		ustress_atm_interlayer,	vstress_atm_interlayer

	!to be sent to atm
	real (kind=4), allocatable, dimension(:,:) ::		&
		aice_atm_interlayer,	tice_atm_interlayer
		
	integer ao_fluxes_mode		!read from namelist, 1 - ice absorbs fluxes calculated by SLAV, 2 - by CICE
	logical ao_conserv_fluxes	!read from namelist, is the interpolation conservative (true) or smoothing (false)
	real(8) tsfc_prev
contains

	subroutine atm_interlayer_allocate
		allocate(uatm_interlayer(idom),		&
			vatm_interlayer(idom),		&
			Qa_interlayer(idom),		&
			Tair_interlayer(idom),		&
			sw_interlayer(idom),		&
			flw_interlayer(idom),		&
			frain_interlayer(idom),		&
			fsnow_interlayer(idom))
         
		allocate(ustress_atm_interlayer(idom),	&
			vstress_atm_interlayer(idom),	&
			sns_atm_interlayer(idom),	&
			ltn_atm_interlayer(idom))

		allocate(aice_atm_interlayer(idom),	&
			tice_atm_interlayer(idom))
			
		tsfc_prev=0.
	end subroutine atm_interlayer_allocate


!~~~~~~~
	subroutine get_atm_fields
!~~~~~~~
		use ice_flux, only: zlvl,uatm,vatm,wind,qa,rhoa,potT,Tair,swvdr,swvdf,swidr,swidf,flw,frain,fsnow,strax,stray,fsensn_f,flatn_f,fsurfn_f,fcondtopn_f
		!use ice_therm_shared, only: calc_Tsfc
		use ice_blocks, only: nx_block, ny_block, nghost
		use ice_grid, only: anglet
		use ice_domain_size, only: ncat
		use ice_constants, only: kseaice,ksno
		use ice_state, only: aice,aicen,vicen,vsnon,trcr,trcrn,nt_Tsfc
		use ice_calendar, only: istep
		use ice_par_module
		real(8) workx,worky,heff,fsurf,Tsfc
		real :: wind_threshold=25.
		real wind_scaling
		integer i,j,n
    
        
		if(atm_model_name=="ATM_INM".and. ao_fluxes_mode==1) then
    
			call rearrange_CPL_to_CICE(sw_interlayer(idom),    swvdr)      !Total shortwave radiation, W/m2, later will be divided into bands
			call rearrange_CPL_to_CICE(flw_interlayer(idom),   flw)        !Downward longwave
			call rearrange_CPL_to_CICE(frain_interlayer(idom), frain)      !kg/m2/s
			call rearrange_CPL_to_CICE(fsnow_interlayer(idom), fsnow)

			call rearrange_CPL_to_CICE(ustress_atm_interlayer(idom),  strax)
			call rearrange_CPL_to_CICE(vstress_atm_interlayer(idom),  stray)
			
			do n=1,ncat
				call rearrange_CPL_to_CICE(sns_atm_interlayer(idom),  fsensn_f(:,:,n,:))
				call rearrange_CPL_to_CICE(ltn_atm_interlayer(idom),  flatn_f(:,:,n,:))
			end do
    
			!TODO: introduce tunable parameters for minimal aicen (2 issues) and surface temperature
			do n = 1, ncat
			do j = 1, ny_block
			do i = 1, nx_block
				fsurf=fsensn_f(i,j,n,1)+flatn_f(i,j,n,1)+swvdr(i,j,1)+flw(i,j,1)
				fsurfn_f(i,j,n,1)=fsurf
				fcondtopn_f(i,j,n,1)=fsurf
	
				!effective ice column height
				if(aicen(i,j,n,1)>1e-3) then
					heff = vicen(i,j,n,1)/aicen(i,j,n,1) + (kseaice/ksno) * (vsnon(i,j,n,1)/aicen(i,j,n,1))
				else
					!fcondtopn_f(i,j,n,1)=0.
					trcrn(i,j,nt_Tsfc,n,1)=0.
					cycle
				end if
	
				!Calculate T_surface from conductivity relation in ice
				!kseaice = [W/m/K], -1.8 = freezing temperature
				tsfc=fsurf*heff/kseaice-1.8 !+tf(i,j,1)
	
				if(tsfc>0) then
					tsfc=0.
					fcondtopn_f(i,j,n,1)=(0.+1.8)*kseaice/heff
				end if
		
				tsfc=tsfc_prev+0.1*(tsfc-tsfc_prev)
				
				trcrn(i,j,nt_Tsfc,n,1)=max(tsfc,-60.)
			end do
			end do
			end do
    
			do j = 1, ny_block
			do i = 1, nx_block
				if(aice(i,j,1)>1e-3) then
					trcr(i,j,nt_tsfc,1)=sum(trcrn(i,j,nt_tsfc,1:ncat,1)*aicen(i,j,1:ncat,1))/aice(i,j,1)
				else 
					trcr(i,j,nt_tsfc,1)=-1.81
				end if
			end do
			end do
    
			!Ice surface temperature is needed by ATM_INM, but not by CICE
			!It might be spoiled by CICE, so we interlayer it here (before CICE step)
			call rearrange_CICE_to_CPL(trcr(:,:,nt_tsfc,:),	tice_atm_interlayer(idom) )    
        
			fsnow=fsnow*1000. !Converting to kg/m2/s
    
			do j = 1+nghost, ny_block - nghost
			do i = 1+nghost, nx_block - nghost
				!-----------------------------------------------------------------
				! Rotate zonal/meridional vectors to local coordinates.
				! Velocity comes in on T grid, but is oriented geographically ---
				! need to rotate to pop-grid FIRST using ANGLET
				! then interpolate to the U-cell centers  (otherwise we
				! interpolate across the pole).
				! Use ANGLET which is on the T grid !
				! Atmo variables are needed in T cell centers in subroutine 
				! atmo_boundary_layer, and are interpolated to the U grid later as 
				! necessary.
				!-----------------------------------------------------------------
               
				workx      = strax(i,j,1) ! wind stress, N/m2
				worky      = stray(i,j,1)
				strax (i,j,1) = workx*cos(ANGLET(i,j,1)) & 
					+ worky*sin(ANGLET(i,j,1))   
				stray (i,j,1) = worky*cos(ANGLET(i,j,1)) & 
					- workx*sin(ANGLET(i,j,1))
			enddo
			enddo            
			
			strax=min(strax,3.)
			strax=max(strax,-3.)
			stray=min(stray,3.)
			stray=max(stray,-3.)
    
		elseif(atm_model_name=="ATM_NCAR" .or. atm_model_name=="ATM_INTERIM" .or. atm_model_name=="ATM_GFS" .or. atm_model_name=="ATM_INMCM4" .or. (atm_model_name=="ATM_INM" .and. ao_fluxes_mode==2)) then
    
			call rearrange_CPL_to_CICE(uatm_interlayer(idom),  uatm)
			call rearrange_CPL_to_CICE(vatm_interlayer(idom),  vatm)
			call rearrange_CPL_to_CICE(qa_interlayer(idom),    qa)         !Spec humidity, CICE,NCAR,GFS,SLAV,INMCM4 - kg/kg, Interim - dewpoint, K
			call rearrange_CPL_to_CICE(tair_interlayer(idom),  tair)       !CICE,NCAR,INTERIM,INMCM4,SLAV-K, GFS-Celsius
			call rearrange_CPL_to_CICE(sw_interlayer(idom),    swvdr)      !Total shortwave radiation, W/m2, later will be divided into bands
			call rearrange_CPL_to_CICE(flw_interlayer(idom),   flw)        !Downward longwave, W/m2
			call rearrange_CPL_to_CICE(frain_interlayer(idom), frain)      !CICE,NCAR,GFS,INMCM4-kg/m2/s=mm/s, Interim - m (accumulated),SLAV-kg/m2 (accumulated)
			call rearrange_CPL_to_CICE(fsnow_interlayer(idom), fsnow)
			call rearrange_CICE_to_CPL(trcr(:,:,nt_tsfc,:), tice_atm_interlayer(idom) )	!tice is only for diagnostics with these atmosphere models 
													!it is interlayered here for uniformity with ATM_INM
    
			if(atm_model_name=="ATM_INTERIM") then
	
				!TEMPRARILY: manual handling of scale_factor and add_offset
				!------------------for MHI 2010.01.01 - 2018.06.30 ------------------------------
				uatm=uatm*0.00424590398992683+34.2944468722238
				vatm=vatm*0.00277015278912139+3.53934826100778
				tair=tair*0.00203636576556611+258.458622320047
				qa=qa*0.00195197522964933+243.274139979182
				fsnow=(fsnow*6.54784668362868e-06+0.214546744338633)/12./3600.  !prec and radiation are accumulated for 12 hours
				frain=0.							!For Interim total prec above ice we assume as snow
				flw=(flw*403.806889277158+12768598.6590554)/12./3600.
				swvdr=(swvdr*559.588115911068+18335464.205942)/12./3600.
				
				!------------------for 1991-2000------------------------------
				!uatm=uatm*0.00130511042549054-3.433330777869
				!vatm=vatm*0.00136506543093925+0.203335106444687
				!tair=tair*0.00205943568179629+257.899238836847
				!qa=qa*0.00196060199989891+243.918209152125
				!fsnow=(fsnow*7.97644449370106e-06+0.26135618018896)/12./3600.
				!frain=0.
				!flw=(flw*425.210931896907+12021815.3945341)/12./3600.
				!swvdr=(swvdr*567.510353562327+18595044.2448232)/12./3600.
				
				
				!------------------for 2001-2013------------------------------
				!uatm=uatm*0.0015557333523461-3.61003079636367
				!vatm=vatm*0.00140508130430828+4.18342526696503
				!tair=tair*0.0020387826582838+258.199297991483
				!qa=qa*0.0019801136353476+242.352128839667
				!fsnow=(fsnow*6.5447033258556e-06+0.21444374909044)/12./3600.  !total prec above ice is treated as snow
				!frain=0.
				!flw=(flw*422.760120855142+12007954.1199396)/12./3600.
				!swvdr=(swvdr*569.291685105214+18653411.3541574)/12./3600.
    
				!------------------for 1979-1990------------------------------
				!uatm=uatm*0.00136057089528129+0.515249890333609
				!vatm=vatm*0.00123852503111803+1.8194872438321
				!tair=tair*0.00205283393235266+257.467410777858
				!qa=qa*0.00195595285761622+243.165701405895
				!fsnow=(fsnow*6.68417526622767e-06+0.219013686660271)/12./3600.  !total prec above ice is treated as snow
				!frain=0.
				!flw=(flw*433.385927700548+12567900.3070362)/12./3600.
				!swvdr=(swvdr*564.853981963286+18508005.573009)/12./3600.
				
				!converting from dewpoint to specific humidity, making data quality control
				do j = 1+nghost, ny_block - nghost
				do i = 1+nghost, nx_block - nghost
					qa(i,j,1)=specific_humidity(sngl(qa(i,j,1)),100000.)
				end do
				end do
				
				fsnow=fsnow*1000. !Converting to kg/m2/s
				
				!In Interim, the first accumulated data come at 12:00 of the first day. Before 12:00 there is "rubbish".
				if(istep <= ice_steps_in_day/2) then 
					swvdr = 0.
					fsnow = 0.
					flw = 0.
				end if
			elseif(atm_model_name=="ATM_GFS") then
				Tair=Tair+273.15	!CICE wants air temperature to be in Kelvins
			elseif(atm_model_name=="ATM_INM") then
				frain=0.
			elseif(atm_model_name=="ATM_INMCM4") then
				frain=0.
			end if
    
			do j = 1+nghost, ny_block - nghost
			do i = 1+nghost, nx_block - nghost
				wind_scaling=sqrt(uatm(i,j,1)*uatm(i,j,1)+vatm(i,j,1)*vatm(i,j,1))/wind_threshold
				
				if( wind_scaling > 1. ) then
					uatm(i,j,1)=uatm(i,j,1)/wind_scaling
					vatm(i,j,1)=vatm(i,j,1)/wind_scaling
					!print *,'Forcing warning in ice: wind speed,i,j = ',wind_scaling*wind_threshold
				end if
			end do
			end do
			
			
			zlvl=10.			!standard reference level(?)
			wind=sqrt(uatm*uatm+vatm*vatm)
			rhoa=1.34			!air density at -10C
			potT=Tair			!it is a simplification
			swvdr=swvdr*0.25		!short wave separation into 4 bands. Must be revised.
			swvdf=swvdr
			swidr=swvdr
			swidf=swvdr
	
			do j = 1+nghost, ny_block - nghost
			do i = 1+nghost, nx_block - nghost

			!-----------------------------------------------------------------
			! Rotate zonal/meridional vectors to local coordinates.
			! Velocity comes in on T grid, but is oriented geographically ---
			! need to rotate to pop-grid FIRST using ANGLET
			! then interpolate to the U-cell centers  (otherwise we
			! interpolate across the pole).
			! Use ANGLET which is on the T grid !
			! Atmo variables are needed in T cell centers in subroutine 
			! atmo_boundary_layer, and are interpolated to the U grid later as 
			! necessary.
			!-----------------------------------------------------------------
				workx      = uatm(i,j,1) ! wind velocity, m/s
				worky      = vatm(i,j,1)
				uatm (i,j,1) = workx*cos(ANGLET(i,j,1)) & 	! convert to POP grid
					+ worky*sin(ANGLET(i,j,1))   		! note uatm, vatm, wind
				vatm (i,j,1) = worky*cos(ANGLET(i,j,1)) & 	! are on the T-grid here
					- workx*sin(ANGLET(i,j,1))
			enddo
			enddo            

		else
			print *,"Atmosphere model not recognized: ",atm_model_name
			stop
		end if
	end subroutine get_atm_fields
  
	!~~~~~~~
	subroutine put_atm_fields
	!~~~~~~~
		use ice_state, only: aice
		call rearrange_CICE_to_CPL(aice, aice_atm_interlayer(idom))
	end subroutine put_atm_fields
  
  
	pure function specific_humidity (Tdew, Patmo)
	!--------------------------------------
	! reference:
	!     Launiainen, J. and Vihma, T., 1990.
	!     Derivation of turbulent surface fluxes – an iterative flux-profile method allowing arbitrary observing heights, 
	!     Environ. Softw., 5(3), 113–124.
	!--------------------------------------
		real (kind=4) specific_humidity
		real (kind=4),intent(in) :: Tdew               ! dew point temperature, K
		real (kind=4),intent(in) :: Patmo              ! atmospheric pressure, Pa
		
		real (kind=4) rsp                              ! reverse saturation pressure, 1/mb 
		
		if(Tdew.lt.273.15) rsp=exp(6141./Tdew -24.3)
		if(Tdew.ge.273.15) rsp=exp(6763.6/Tdew+4.9283*log(Tdew)-54.23)
		rsp = rsp/100                                  ! convert from 1/mb to 1/Pa
		specific_humidity=0.622/(Patmo*rsp  -0.378)    ! kg/kg
	end function specific_humidity
  
end module atm_general_module

