#include "ice_2Ddomain"
module non_ga_ice_array_module

  use ice_par_module

  implicit none
  save

  private

!=================
! Public interface
!=================
  public :: &
       cpl_ice_initialize, &
       rearrange_CICE_to_CPL, &
       rearrange_CPL_to_CICE, &
       check_parameters, &
       ice_cp_process

  interface cpl_ice_initialize
     module procedure check_decomposition
  end interface cpl_ice_initialize

  interface rearrange_CICE_to_CPL
    module procedure copy_CICE8_to_CPL
  end interface

  interface rearrange_CPL_to_CICE
    module procedure copy_CPL_to_CICE8
  end interface

!================
! Private members
!================
 CONTAINS

!========================================================================================
! no new arrays allocation, but checking whether we are in mode with
! the same decomposition

subroutine check_decomposition()
	use ice_distribution, only : distrb
	use ice_domain, only : distrb_info, nblocks
	use ice_blocks, only: get_block_parameter, nx_block, ny_block, nghost

	integer, parameter :: block_size_x = nx_block - 2 * nghost, block_size_y = ny_block - 2 * nghost

	integer iblock0, jblock0
	integer cice_east, cice_south, cice_west, cice_north

	if (nblocks .gt. 1) then
		call write_critical_error()
		write(*, *) "CICE decomposition is incorrect for non-global array coupler mode"
		write(*, *) "Number of blocks of this proccess is greater than 1"
		write(*, *) "nblocks:", nblocks
		write(*, *) "Check comp_ice script and ice_in input namelist in cice"
		write(*, *) "Stopping..."
		stop
	end if
	
	!call MPI_BARRIER(comm_local, ierr)
	if (nblocks == 1) then
		call get_block_parameter(distrb_info % blockGlobalID (1), iblock = iblock0, jblock = jblock0)

		cice_west = 	1 + (iblock0 - 1) * block_size_x
		cice_south = 	1 + (jblock0 - 1) * block_size_y
		cice_east = 	iblock0 * block_size_x
		cice_north = 	jblock0 * block_size_y

		if (cice_east > ice_i_size) &
			cice_east = ice_i_size

		if (cice_north > ice_j_size) &
			cice_north = ice_j_size

		if ( 	cice_east .ne. ice_cpl_east .or.		&
			cice_west .ne. ice_cpl_west .or. 		&
			cice_south .ne. ice_cpl_south .or.	&
			cice_north .ne. ice_cpl_north) then

			call write_critical_error()
			
			write(*, *) "Geopgrahical indices of coupler decomp:"
			write(*, *) ice_i_size
			write(*, *) "west =", ice_cpl_west
			write(*, *) "east =", ice_cpl_east
			write(*, *) "south =", ice_cpl_south
			write(*, *) "north =", ice_cpl_north
			write(*, *) ""
			write(*, *) "Geographical indices of cice decomp:"
			write(*, *) "west =", cice_west
			write(*, *) "east =", cice_east
			write(*, *) "south =", cice_south
			write(*, *) "north =", cice_north
			write(*, *) ""
			write(*, *) "Stopping..."
			stop
		endif ! east = east & ...
	endif ! nblocks == 1

	!call MPI_BARRIER(comm_local, ierr)
	!if (rank_local == 0) write(*, *) "Succesfully checked decomposition equality of cice and coupler",nblocks
	write(*,'(a,14(i,1x))') "Decomposition parameters:",nblocks,ice_rank_local,ice_cpl_west,ice_cpl_east,ice_cpl_south,ice_cpl_north,cice_west,cice_east,cice_south,cice_north,iblock0,jblock0,block_size_x,block_size_y
end subroutine check_decomposition

!========================================================================================
subroutine check_parameters(ao_fluxes_mode)
	use ice_therm_shared, only: calc_Tsfc
	use ice_restoring, only: restore_ice
	use ice_domain, only: nblocks
	use ice_atmo, only: formdrag,calc_strair
	use ice_state, only: tr_aero,tr_FY,tr_iage,tr_pond_cesm,tr_pond_lvl,tr_pond_topo
	use ice_forcing, only: precip_units
	use ice_domain_size, only: max_ntrcr,nilyr
	use ice_transport_driver, only: advection
	use ice_flux, only: update_ocn_f
	use ice_therm_shared, only: ktherm
	!use ice_par_module, only: atm_model_name
	use ice_dyn_shared, only: kdyn,ndte
	use ice_shortwave, only: shortwave
	use ice_restart_shared, only: restart
	
	integer ao_fluxes_mode
	
	
	if (ice_rank_local == 0) write(*,*) '===Checking CICE parameners (see cice_cpl_inter_module.f90)==='
	
	if (atm_model_name == "ATM_NCAR" .or. atm_model_name == "ATM_INTERIM" .or. atm_model_name == "ATM_GFS".or. atm_model_name == "ATM_INMCM4".or. (atm_model_name=="ATM_INM" .and. ao_fluxes_mode==2)) then
		if (.not. calc_Tsfc) 	Stop 'VITIM ERROR: If running ATM_NCAR, ATM_INTERIM, ATM_INMCM4 or ATM_GFS or ATM_INM (CICE fluxes) then turn CICE ABL submodel ON (calc_Tsfc must be .true.)'
		if (.not. calc_strair) 	Stop 'VITIM ERROR: If running ATM_NCAR, ATM_INTERIM, ATM_INMCM4 or ATM_GFS or ATM_INM (CICE fluxes) then turn CICE windstress calculation ON (calc_strair must be .true.)'
		if(ktherm == 0 .and. ice_rank_local == 0) write(*,'(A,i)') 'VITIM WARNING: With ATM_NCAR, ATM_INTERIM, ATM_INMCM4 or ATM_GFS ktherm recommended value is 1 or 2, not ',ktherm
	elseif (atm_model_name == "ATM_INM".and. ao_fluxes_mode==1) then
		if(calc_Tsfc) then 
			if (ice_rank_local == 0) write(*,*)	'VITIM WARNING: running ATM_INM (SLAV fluxes), so CICE ABL submodel is turned OFF (calc_Tsfc set to .false.)'
			calc_Tsfc=.false.
		end if
		if(calc_strair) then 
			if (ice_rank_local == 0) write(*,*)	'VITIM WARNING: running ATM_INM (SLAV fluxes), so CICE windstress calculation is turned OFF (calc_strair set to .false.)'
			calc_strair=.false.
		end if
		if(ktherm > 0 .and. ice_rank_local == 0) write(*,'(A,i)') 'VITIM WARNING: With ATM_INM (SLAV fluxes) ktherm recommended value is 0, not ',ktherm
	else
					Stop 'VITIM ERROR: check parameters: atmosphere model not recognized'
	end if
	
	if(restore_ice) 					Stop 'VITIM ERROR: Please turn off ice restoring on lateral boundaries of the domain (restore_ice must be .false.)'
	!if(nblocks.ne.1)					Stop 'VITIM ERROR: Something gone wrong: nblocks .ne. 1'
	if(formdrag)						Stop 'VITIM ERROR: Please turn formdrag off'
	if(tr_aero)						Stop 'VITIM ERROR: Please turn aerosols off (tr_aero=false)'
	if(advection.ne.'upwind')					Stop 'VITIM ERROR: Please set upwind advection'
	!if(.not.tr_pond_cesm .or. tr_pond_topo .or. tr_pond_lvl) 	Stop 'VITIM ERROR: Please use CESM pond formulation only'
	if(precip_units.ne.'mks' .and. precip_units.ne.'mm_per_sec') 	Stop 'VITIM ERROR: Precipitation units must be kg/m2/s'
	!if(.not. tr_iage .or. .not. tr_FY) 				Stop 'VITIM ERROR: Please turn on ice age and first year tracers (tr_iage=tr_fy=true)'
	if(.not. update_ocn_f)						Stop 'VITIM ERROR: update_ocn_f must be  .true. to include water and salt (not heat!) fluxes due to frazil ice formation'
	if(ktherm>0 .and. nilyr<2) 					Stop 'VITIM ERROR: Please increase number of ice layers (recommended nicelyr=7)'
	
	!kitd=1		!linear remapping in itd redistribution
	!kcatbound=0	!standard formula for category boundaries
	!ndtd=1		!number of dynamics subcycles

	if (ice_rank_local == 0) write(*,*) '===Checking complete succesfully==='
	
	if (ice_rank_local == 1) then
		write(*,*) '===Running with these CICE parameners==='
		write(*,'(A,L)') 'Check: calc_strair = '	,calc_strair
		write(*,'(A,L)') 'Check: calc_Tsfc = '		,calc_Tsfc
		write(*,'(A,2i)')'Check: ktherm, nilyr = '	,ktherm, nilyr
		write(*,'(A,A)') 'Check: precip_units = '	,trim(precip_units)
		write(*,'(A,L)') 'Check: update_ocn_f = '	,update_ocn_f
		write(*,'(A,2i)')'Check: kdyn, ndte = '		,kdyn,ndte
		write(*,'(A,A)') 'Check: shortwave = '		,trim(shortwave)
		write(*,'(A,L)') 'Check: ice_read_cp_flag = '	,ice_read_cp_flag
		write(*,'(A,L)') 'Check: restart = '		,restart
	end if
	
#ifdef CICE_IN_NEMO
	Stop 'VITIM ERROR: Please turn off CICE_IN_NEMO functionality'
#endif
	
end subroutine check_parameters
!========================================================================================

  subroutine write_critical_error
    write(*, *) "#######################################"
    write(*, *) "#######################################"
    write(*, *) "Critical error in cice coupler driver"
    write(*, *) "#######################################"
    write(*, *) "My Rank :", ice_rank_local
    write(*, *) ""
  end subroutine write_critical_error

! Doing nothing
  subroutine copy_CICE8_to_CPL(arr_cice, arr_cpl)
    use ice_blocks, only: nx_block, ny_block, nghost
    use ice_domain, only: nblocks

    real(kind=8), intent(in) :: arr_cice(:,:,:)
    real(kind=4), intent(inout) :: arr_cpl(:,:)

    if (nblocks == 1) &
       arr_cpl = arr_cice(1 + nghost: nx_block - nghost, 1 + nghost: ny_block - nghost, 1)

  end subroutine copy_CICE8_to_CPL

  subroutine copy_CPL_to_CICE8(arr_cpl, arr_cice)
    use ice_blocks, only: nx_block, ny_block, nghost
    use ice_domain, only: nblocks

    real(kind=4), intent(in) :: arr_cpl(:,:)
    real(kind=8), intent(inout) :: arr_cice(:,:,:)

    if (nblocks == 1) &
       arr_cice(1 + nghost:nx_block - nghost, 1 + nghost:ny_block - nghost, 1) = arr_cpl

  end subroutine copy_CPL_to_CICE8
!============================================================================================
  
  subroutine ice_cp_process(aice_intrl)
	use ice_state, only: aice,aicen,vicen,vsnon
	use ice_domain_size, only: ncat
	use ice_blocks, only: nx_block, ny_block
	integer i,j,n
	
	real(4) aice_intrl(idom)
	
	call rearrange_CPL_to_CICE(aice_intrl(idom),	aice)
	
	aicen=0.
	vicen=0.
	vsnon=0.
	do n = 1, ncat
	do j = 1, ny_block
	do i = 1, nx_block
		aicen(i,j,n,1)=aice(i,j,1)/float(ncat)
		if(aicen(i,j,n,1)>0.9) then
			vicen(i,j,n,1)=2.
			vsnon(i,j,n,1)=0.1
		elseif(aicen(i,j,n,1)>0.5) then
			vicen(i,j,n,1)=0.5
			vsnon(i,j,n,1)=0.05
		elseif(aicen(i,j,n,1)>0.001) then
			vicen(i,j,n,1)=0.2
		end if
	end do
	end do
	end do
	
	if(ice_rank_local == 0) write(*,*) '===ice CP transferred to CICE internal arrays==='
	
  end subroutine ice_cp_process
  
end module non_ga_ice_array_module
