
!--------------------------------------
!		
!	Institute of Numerical Mathematics, Russian Academy of Sciences (INM RAS)
!	Shirshov Institute of Oceanology, Russian Academy of Sciences (IO RAS) 
!		Ocean General Circulation Model	
!			INMIO
!
!
! authors: 
!	2015.07 / Ibrayev, Rashit, IO RAS, Russia, ibrayev@mail.ru
!	2018.02 / Ushakov, Konstantin, IO RAS, Russia, ushakovkv@mail.ru
!
! reference:
!	K. V. Ushakov, R. A. Ibraev, and V. V. Kalmykov, 2015.
!		Simulation of the world ocean climate with a massively parallel numerical model, 
!		Izv., Atmos. Ocean. Phys. 51 (4), 362–380.
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Vychisl. Metody Programm., 2013, v. 14, iss. 3, 88-95. 
!	Kalmykov V.V., Ibrayev R.A., 2013. 
!		A framework for the ocean-ice-atmosphere-land coupled modeling on massively-parallel architectures. 
!		Certificate of state registration of the computer progam number 2013619320. 
!		Date of state registration in the Registry of the computer programs October 1, 2013
!	R. A. Ibraev, K. V. Ushakov, and R. N. Khabeev, 2012.
!		Eddy-resolving 1/10° model of the World Ocean,
!		Izv., Atmos. Ocean. Phys. 48 (1), 37–46.
!--------------------------------------

module ice_cice_driver_module
	use ice_par_module
	use non_ga_ice_array_module
	use cice_module
	use atm_general_module
	use ocn_general_module
	implicit none

	integer atm_ice_mapping, ocn_ice_mapping, ice_ocn_mapping, ice_diag_period
	
	CONTAINS
	
!================================================================================================================================================================
	subroutine ice_cice_driver_ini_params
		use o_par_module
		
		logical file_exists
		namelist /ICE_MAIN/ 	ice_i_size, ice_j_size, ice_steps_in_day,			&
					ice_read_cp_flag, ice_np, ice_tripole_flag, ice_i_cycle_flag,	&
					ice_resolution_mode, ice_das_on, ice_das_sens
		namelist /INMIO_SLAV/ 	ao_fluxes_mode, ao_conserv_fluxes

		inquire(file = "./configure/ice_list.in", exist = file_exists)
		if (file_exists) then
			open(91, file= "./configure/ice_list.in")
			read(91, nml = ICE_MAIN)
			close(91)
		else
			stop 'ERROR: ice_list.in not found'
		end if
		
		inquire(file = "./configure/coupling_list.in", exist = file_exists)
		if (file_exists) then
			open(92, file= "./configure/coupling_list.in")
			read(92, nml = INMIO_SLAV)
			close(92)
		else
			stop 'ERROR: coupling_list.in not found'
		end if
		
		ao_fluxes_mode=1
		
		atm_ice_mapping = 60*6
		ocn_ice_mapping = 2*1440/ice_steps_in_day !24 !in minutes
		ice_ocn_mapping = 2*1440/ice_steps_in_day !24
		ice_diag_period = 43800
		
		!------
		!	periods for data exchange in [min]
		!					60*1     min =    60 min ~> 1h, 
		!					60*2     min =   120 min ~> 2h, 
		!					60*6     min =   360 min ~> 6h, 
		!					60*24    min =  1440 min ~> 1d, 
		!					60*24*5  min =  7200 min ~> 5d, 
		!					60*24*10 min = 14400 min ~> 10d, 
		!					60*24*30 min = 43200 min ~> 30d
		!------
	
		!write(*,'(a,4i)') 'ice mapping parameters: ',ocn_ice_mapping,atm_ice_mapping,ice_ocn_mapping,ice_remapping_mode
		
	end subroutine ice_cice_driver_ini_params

!================================================================================================================================================================

	subroutine ice_cice_driver_ini_main
		use ice_timers, only: ice_timer_start, timer_step
		
		!---CICE internal initializations, including timers (drivers/inmio/cice_initmod.f90)---
		call CICE_init(ice_comm_local)
		
		!---test decomposition (non_ga_array_module.f90)---
		call cpl_ice_initialize
		
		!---Checking whether all parameters are suitable to CORE-like experiments (non_ga_array_module.f90)----
		call check_parameters(ao_fluxes_mode)
		
		! start timing entire run
		call ice_timer_start(timer_step)   

		! transfer ice CP to CICE internal arrays
		if(ice_read_cp_flag) call ice_cp_process(aice_interlayer)
		
	end subroutine ice_cice_driver_ini_main
				
!================================================================================================================================================================

	subroutine ice_cice_driver_main_step
		
		use ice_calendar, only: istep, istep1, time, dt, calendar,idate,nyr,month,mday,hour
		use ice_flux, only: init_flux_atm, init_flux_ocn
		use ice_domain, only : nblocks
		real start, finish

		
		call shared_timer
		if(ice_rank_local == 0 .and. mod(istep,ice_steps_in_day/24)==0) then
			!write(*,'(a,1x,6i4)') "ICE STEP NO", istep, nblocks, cmp % west, cmp % east, cmp % south, cmp % north
			call cpu_time(start)
		endif
		
		if (ice_model_name .ne. "ICE_CICE" ) stop 'ice_model_name not initialized'
		if (time_l_in_run==10 .or. time_l_in_run==1000) call check_parameters(ao_fluxes_mode)
		
		if ( ice_das_on .and. istep /= 1 .and. mod(istep-1,ice_steps_in_day)==0 ) then
			print *, "### ICE DAS CALL ### ",istep, ice_steps_in_day, maxval(aice_a_interlayer)
			call das_to_cice_aicen_correction
		endif
		call get_atm_fields	
		call get_ocn_fields

		call ice_step

		call put_ocn_fields	
		call put_atm_fields
		
		istep = istep + 1
		istep1 = istep1 + 1
		time = time + dt       ! determine the time and date
		
		call calendar(time)    ! at the end of the timestep

		call init_flux_atm     ! initialize atmosphere fluxes sent to coupler
		call init_flux_ocn     ! initialize ocean fluxes sent to coupler
		
		call cice_mask_land_for_dg (missval)

		if (ice_rank_local == 0 .and. mod(istep-1,ice_steps_in_day/24)==0) then
			call cpu_time(finish)
			print *, "### ICE step walltime ### ", finish-start
		endif

	!-------
	!	print CPL model time and for control add CICE model time
	!-------
		if (ice_rank_local == 0) then
		write(*,'(a3,a4,i7, i5,i2.2,i2.2,  a1,i2.2,a1,i2.2,  a7,i8,a1,3i2.2,a1,i6,a1)')			& 
			run_experiment,										&
			' ice',											&
			istep,										&
			time_year,time_month,time_day,'/',time_hour,':',time_min,				&
			'  cice ',idate,' ',nyr,month,mday,'/',hour,'h'
		endif

	end subroutine
		
end module ice_cice_driver_module
